/*
 *  GTRModel.cpp
 *  PLEX
 *
 *  A.P. Jason de Koning, jason.de.koning@gmail.com
 *  http://jasondk.org
 *
 *  David Pollock lab, University of Colorado School of Medicine
 *  http://www.evolutionarygenomics.com
 *  All rights reserved.
 
 */

#include "GTRModel.h"
#include "PLEXApp.h"

#pragma mark -
//---------------------------------------------------------------------------
GTRModel::GTRModel(
                   const DoubleMatrix3D& inAlphaMatrix3D,
                   const DoubleMatrix& inFrequencyMatrix,
                   const IntMatrix& inCountMatrix)
:	MixtureModel()
{
	num = &Numerics::getGlobalNumerics();
    
	mAlphaMatrix3D = inAlphaMatrix3D;
	mFrequencyMatrix = inFrequencyMatrix;
	mCountMatrix = inCountMatrix;
	
}		

//---------------------------------------------------------------------------
GTRModel::GTRModel(
                   const GTRModel&	inModel)
:	MixtureModel(inModel)
{
	num = &Numerics::getGlobalNumerics();
    
	mAlphaMatrix3D = inModel.mAlphaMatrix3D;
	mFrequencyMatrix = inModel.mFrequencyMatrix;
	mCountMatrix = inModel.mCountMatrix;
}

//---------------------------------------------------------------------------
GTRModel::~GTRModel(void)
{
}


//---------------------------------------------------------------------------
void GTRModel::swap(
                    GTRModel&	inModel)
{
	MixtureModel::swap(inModel);
	mAlphaMatrix3D.swap(inModel.mAlphaMatrix3D);
	std::swap(mFrequencyMatrix, inModel.mFrequencyMatrix);
	mCountMatrix.swap(inModel.mCountMatrix);
}

//---------------------------------------------------------------------------
void GTRModel::SetCountMatrix(const IntMatrix& inCountMatrix)
{
	mCountMatrix = inCountMatrix;
}


//---------------------------------------------------------------------------
void GTRModel::InitializeFrequencyMatrix(
                                         const IntMatrix& inCountMatrix)
{
	if (mNumStates == 0)
		mNumStates = inCountMatrix.rows();
    
	if (mFrequencyMatrix.empty())
		mFrequencyMatrix.resize(mNumModels, mNumStates, 0);
	
	double	total=0.0;
	
	DoubleVector sum(mNumStates, 0);
    
	for (unsigned int i=0; i<mNumStates; i++)
	{
		for (unsigned int j=0; j < mNumStates; j++)
			sum[i]+= inCountMatrix.mRows.mDataArray[i].mDataArray[j];
        
		total += sum[i];
	}
	
    // INITIALIZE FREQUENCIES TO THEIR OBSERVED VALUES IN THE DATA...
    for (unsigned int k = 0; k < mNumModels; k++)
		for (unsigned int i=0; i < mNumStates; i++)
			mFrequencyMatrix.mRows.mDataArray[k].mDataArray[i] = sum[i]/total;
}

//---------------------------------------------------------------------------
void GTRModel::InitializeAlphaMatrix3D(
                                       const IntMatrix& inCountMatrix)
{
	if (mNumStates == 0)
		mNumStates = inCountMatrix.rows();
    
	if (mFrequencyMatrix.empty())
		InitializeFrequencyMatrix(inCountMatrix);
    
	if (mAlphaMatrix3D.empty())
		mAlphaMatrix3D.resize(mNumModels, mNumStates, mNumStates, 0.0);
	
	DoubleVector sum(mNumStates, 0);
	
	DoubleMatrix Rates(mNumStates, mNumStates, 0);
    
	for (unsigned int i=0; i<mNumStates; i++)
		for (unsigned int j=0; j < mNumStates; j++)
			sum[i]+= inCountMatrix.mRows.mDataArray[i].mDataArray[j];
    
	// Rates initialized here...
	for (unsigned int i=0; i<mNumStates; i++)
		for (unsigned int j=0; j<mNumStates; j++)
			Rates.mRows.mDataArray[i].mDataArray[j] = inCountMatrix.mRows.mDataArray[i].mDataArray[j] / sum[i];
    
	IntMatrix found(mNumStates, mNumStates, 1);
    
	for (unsigned int k = 0; k < mNumModels; k++)
	{
		for (unsigned int i=0; i < mNumStates; i++)
			for (unsigned int j=0; j < mNumStates; j++)
				if (i == j)
					found.mRows.mDataArray[i].mDataArray[j] = 0;
				else
					found.mRows.mDataArray[i].mDataArray[j] = 1;
        
		for (unsigned int i=0; i < mNumStates; i++) {
			for (unsigned int j=0; j < mNumStates; j++) {
				if (found.mRows.mDataArray[i].mDataArray[j]) {
					mAlphaMatrix3D[k].mRows.mDataArray[i].mDataArray[j] = ((Rates.mRows.mDataArray[i].mDataArray[j] / mFrequencyMatrix.mRows.mDataArray[k].mDataArray[j]) + (Rates.mRows.mDataArray[j].mDataArray[i] / mFrequencyMatrix.mRows.mDataArray[k].mDataArray[i])) / 2.0;
				}
			}
		}
	}		
}

void GTRModel::PrintRateMatrix3D(
                                 ostream&		outStream, 
                                 ObservedData& theData) {
	
	m_oPLEXApp->PrintMatrix3D(outStream, mAlphaMatrix3D, theData); 
}

//---------------------------------------------------------------------------
void GTRModel::GTRModelParameterstoRates(void)
{
	if (mNumStates == 0)
		mNumStates = mCountMatrix.rows();
    
	if (mFrequencyMatrix.empty())
		InitializeFrequencyMatrix(mCountMatrix);
    
	if (mAlphaMatrix3D.empty())
		InitializeAlphaMatrix3D(mCountMatrix);
    
	static IntMatrix found(mNumStates, mNumStates, 1);
	found.fill(1);
	
	DoubleMatrix3D *Rates3D = returnRateMatrix3D();
    
	if (Rates3D->empty()) {
		Rates3D->resize(mNumModels, mNumStates, mNumStates, 0);
	} 
	
	for (unsigned int k = 0; k < mNumModels; k++)
	{
        // Yang's formulation...
		mAlphaMatrix3D.mRows.mDataArray[k].mRows.mDataArray[0].mDataArray[2] = 1.0;
		mAlphaMatrix3D.mRows.mDataArray[k].mRows.mDataArray[2].mDataArray[0] = 1.0;
        
		for (unsigned int i=0; i < mNumStates; i++)
			for (unsigned int j=0; j < mNumStates; j++)
				if (i == j)
					found.mRows.mDataArray[i].mDataArray[j] = 0;
				else
					found.mRows.mDataArray[i].mDataArray[j] = 1;
        
		for (unsigned int i=0; i<mNumStates; i++)
		{
			for (unsigned int j=0; j<mNumStates; j++)
				if (found.mRows.mDataArray[i].mDataArray[j]) {
					(*Rates3D)[k].mRows.mDataArray[i].mDataArray[j] = mAlphaMatrix3D[k].mRows.mDataArray[i].mDataArray[j] * mFrequencyMatrix.mRows.mDataArray[k].mDataArray[j];
				}
		}
        
		for (unsigned int i=0; i<mNumStates; i++)
		{
			double sum = 0;
            
			for (unsigned int j=0; j<mNumStates; j++)
			{
				if (i != j)
					sum += (*Rates3D)[k].mRows.mDataArray[i].mDataArray[j];
			}
            
			(*Rates3D)[k].mRows.mDataArray[i].mDataArray[i] = -1 * sum;
		}
        
		double sum = 0.0;
		for (int i=0; i<mNumStates; i++) {
			for (int j=0; j<mNumStates; j++) {
				if (j != i) { 
					sum += (*Rates3D)[k].mRows.mDataArray[i].mDataArray[j] * mFrequencyMatrix.mRows.mDataArray[k].mDataArray[i];
				}
			}
		}
        
		for (int i=0; i<mNumStates; i++) {
			for (int j=0; j<mNumStates; j++) {
				(*Rates3D)[k].mRows.mDataArray[i].mDataArray[j] /= sum;
			}
		}
		
        /*		std::cout << "Normalized Q:"<<endl;
         for (int i=0; i<mNumStates; i++) {
         for (int j=0; j<mNumStates; j++) {
         
         //				std::cout << (mAlphaMatrix3D)[k].mRows.mDataArray[i].mDataArray[j] << "\t";
         std::cout << (*Rates3D)[k].mRows.mDataArray[i].mDataArray[j] << "\t";
         }
         std::cout << endl;
         }
         std::cout << endl;
         */	
	}
}

//---------------------------------------------------------------------------
void GTRModel::InitializeParameter(
                                   const IntMatrix& inCountMatrix)
{
	SetCountMatrix(inCountMatrix);
	InitializeFrequencyMatrix(inCountMatrix);
    
    // TUNED FOR TEST DATASET
	/*
     if (options.datatype == 1) {
		mFrequencyMatrix[0][0] = 0.29497;	// A
		mFrequencyMatrix[0][1] = 0.27374;	// T
		mFrequencyMatrix[0][2] = 0.13194;	// G
		mFrequencyMatrix[0][3] = 0.29935;	// C
	} */
    
	InitializeAlphaMatrix3D(inCountMatrix);
	GTRModelParameterstoRates();
	InitializeFreqModels();
	SetModelNumNucleotides(1);
}

//---------------------------------------------------------------------------
void GTRModel::PrintAlphaMatrix3D(
                                  ostream&	outStream, ObservedData& theData)
{
	m_oPLEXApp->PrintMatrix3D(outStream, mAlphaMatrix3D, theData);
}

//---------------------------------------------------------------------------
void GTRModel::PrintCountMatrix(
                                ostream& outStream, ObservedData& theData)
{
	m_oPLEXApp->PrintMatrix(outStream, mCountMatrix, theData);
}

//---------------------------------------------------------------------------
void GTRModel::PrintFrequencyMatrix(
                                    ostream& outStream, ObservedData& theData)
{
    
	for (unsigned int k = 0; k< mNumModels; k++){
		
		m_oPLEXApp->PrintVector(outStream, theData, mFrequencyMatrix[k]);
	}
}

//---------------------------------------------------------------------------
void GTRModel::UpdateFrequencyParameter(
                                        double& inParameter)
{
	double rannum = Numerics::getGlobalNumerics().Random();
	
	double changeamount = (rannum - 0.5) * mWindowsize;
	
	double proposal = inParameter + changeamount;
    
	if ((proposal <= 0.0 ) || (proposal >= 1.0))
		proposal = inParameter - changeamount;
    
	inParameter = proposal;
}

//---------------------------------------------------------------------------
void GTRModel::UpdateAlphaParameter(
                                    double& inParameter)
{
	static int rejectNum = 0;
	double changeamount, proposal, var;
    
	int notDone = 1;
	while (notDone) {
	
// OVERRIDE THE BELOW COMMENTS FOR EMPIRICALLY TUNED UPDATERS (test dataset)        
 if ( Numerics::getGlobalNumerics().Random() < 0.05 ) {
			changeamount = ( Numerics::getGlobalNumerics().Random() - 0.5) * mWindowsize * inParameter;
			proposal = inParameter + changeamount;
			
			if (proposal <= 0.0 )
				proposal = inParameter - changeamount;
// *
    } else {
        // EMPIRICALLY TUNED FOR TEST DATASET
			var = 0.15;
			
			if (inParameter <= 0.68) {
				var = 0.028;
			} else if (inParameter <= 0.35) { 
				var = 0.009;
			} else if (inParameter <= 0.11) { 
				var = 0.012;
			} else if (inParameter <= 0.06) {
				var = 0.003; 
			}
			var *= 0.5;
            
			changeamount = num->box_muller( 0.0, var ); // (rannum - 0.5) * mWindowsize;
			
			while ( (inParameter + changeamount) <= 0.0 ) {
				changeamount *= -1;
				if ( (inParameter + changeamount) < 0.0 ) {
					changeamount = num->box_muller( 0.0, var ); // (rannum - 0.5) * mWindowsize;
				}					
			}
			proposal = inParameter + changeamount;
		}
// *      
        if (proposal > 10.0) {
            //std::cout << "WHOA, proposal is " << proposal << ", in parm was " << inParameter << endl;
            notDone = 1;
            rejectNum++;
        } else {
            notDone = 0;
        }
		
	}
	inParameter = proposal;
	
	if (rejectNum > 10000) {
		std::cout << "	[WARNING! Proposals are hitting the ceiling a lot on rate parameters! Ceiling hit #" << rejectNum << "]" << endl;
	}
}

//---------------------------------------------------------------------------
int GTRModel::UpdateAlphaParameter( int theModel, int i, int j, int reversibleEvaluation) {
	int rejectNum = 0;
	double changeamount, proposal, var;
	
	// THIS IS NOT VERY EFFICIENT, BUT WILL GUARANTEE VALID PROPOSALS UNDER OPTIONS.CAPUNIFORMIZATION...
	
	double newWaitingTime1, sumCorrection;
    
	double inParameter = mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[i].mDataArray[j];
	double oldRate1     = inParameter * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[j];
	double oldRate2     = inParameter * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[i];
	
	double rateSum = 0.0;
	for (int ii=0; ii < mNumStates; ii++) {
		for (int j=0; j < mNumStates; j++) {
			if (j != ii) rateSum += mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[ii].mDataArray[j] * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[j] * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[ii];
		}
	}
    
	int notDone = 1;
	while (notDone) {
        // OVERRIDE THE BELOW COMMENTS FOR EMPIRICALLY TUNED UPDATERS (test dataset)        
        if ( Numerics::getGlobalNumerics().Random() < 0.05 ) {
			changeamount = ( Numerics::getGlobalNumerics().Random() - 0.5) * mWindowsize * inParameter; 
			proposal = inParameter + changeamount;
			
			if (proposal <= 0.0 )
				proposal = inParameter - changeamount;
// *
		} else {
            // EMPIRICALLY TUNED FOR TEST DATASET
			var = 0.18; 
			
			if (inParameter <= 0.68) { 
				var = 0.114; 
			} else if (inParameter <= 0.35) { 
				var = 0.094;
			} else if (inParameter <= 0.11) { 
				var = 0.078;;
			} else if (inParameter <= 0.06) {
				var = 0.068;  
			}
            
			var *= 0.1;
			
			changeamount = num->box_muller( 0.0, var ); // (rannum - 0.5) * mWindowsize;
			
			while ( (inParameter + changeamount) <= 0.0 ) {
				changeamount *= -1;
				if ( (inParameter + changeamount) < 0.0 ) {
					changeamount = num->box_muller( 0.0, var ); // (rannum - 0.5) * mWindowsize;
				}					
			}
			proposal = inParameter + changeamount;
		}
// *		
		if (proposal > 10.0) {
			//std::cout << "WHOA, proposal is " << proposal << ", in parm was " << inParameter << endl;
			notDone = 1;
			rejectNum++;
		} else {
			if (!options.capUniformizationRate) notDone = 0;
		}
		
		if (options.capUniformizationRate) {
			if (rejectNum > 3) {
				if (num->Random() > 0.5) {
					proposal = inParameter - changeamount;
					if (proposal < 0.0) {
						proposal = inParameter + changeamount;
					}
				}
			}
			
			if (rejectNum > 4) {
				return 0;
			}
			
			double rowSum;
			if ( reversibleEvaluation ) {
				sumCorrection =   (rateSum - oldRate1* mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[i] - oldRate2* mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[j] + (2.0 * proposal * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[j]*mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[i]));
				notDone = 0;
				for (int kk=0; kk<mNumStates; kk++) {
					rowSum = 0.0;
					for (int k=0; k < mNumStates; k++) {
						if (k != kk) rowSum += mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[kk].mDataArray[k] * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[k];
					}
					double term2;
					if (kk == i) {
						term2 = -oldRate1 + (proposal * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[j]);
					} else if (kk == j) {
						term2 = -oldRate2 + (proposal * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[i]);
					} else {
						term2 = 0.0;
					}
					
					newWaitingTime1 = (rowSum + term2) / sumCorrection;
					//newWaitingTime2 = (rowSum2 - oldRate2 + (proposal * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[i])) / sumCorrection;
					
					if ( ( (newWaitingTime1 * options.__largestRate) > uniformizationConstants.mDataArray[theModel] ) ) {
						notDone = 1;
					}
				}
                
				if (notDone == 1) {
					rejectNum++;
				} 			
			} else {
				sumCorrection =  rateSum / (rateSum - oldRate1  + (proposal * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[j]) );
				newWaitingTime1 = -mRateMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[i].mDataArray[i] - oldRate1 + (proposal * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[j]) * sumCorrection;
				if ( (newWaitingTime1 * options.__largestRate) > uniformizationConstants.mDataArray[theModel] ) {
					notDone = 1;
					rejectNum++;
				} else {
					notDone = 0;
				}
			}
			
			
		}
	}

	mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[i].mDataArray[j] = proposal;
	if (reversibleEvaluation) mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[j].mDataArray[i] = proposal;
	
	return 1;
}


//---------------------------------------------------------------------------
void GTRModel::UpdateFrequency(DoubleVector& mFrequency)
{
	bool m_bUpdate = false;
	
	DoubleVector tempFrequency;
    
	while (!m_bUpdate)
	{
		double sum = 0;
        
		tempFrequency = mFrequency;
        
		IntVector found(mNumStates, 0);
        
		for (int i = 0; i < mNumStates; i++)
		{
			if ( i == 0)
				found[i] = 0;
			else
				found[i] = 1;
		}
        
		for (int i = 0; i < mNumStates; i++)
		{
			if (found[i])
			{
				UpdateFrequencyParameter(tempFrequency[i]);
                
				sum += tempFrequency[i];
			}
		}
        
		tempFrequency[0] = 1 - sum;
		
		if ((tempFrequency[0] >= 1.0) || (tempFrequency[0] <= 0.0))
			m_bUpdate = false;
		else
			m_bUpdate = true;
	}
	
	if (m_bUpdate)
		mFrequency = tempFrequency;
}

//---------------------------------------------------------------------------
void GTRModel::UpdateFrequencyMatrix(void)
{
	for (unsigned int k = 0; k< mNumModels; k++)
		UpdateFrequency(mFrequencyMatrix[k]);
}

// new version
//---------------------------------------------------------------------------
void GTRModel::UpdateAlphaMatrix(int theModel, const int& m_nMethod)
{
	int i,j;
	bool m_bUpdate = false;
    
    
	int numUpdated = 0;
	//	tempAlphaMatrix = mAlphaMatrix;
	
	int updateLimit =(int)(Numerics::getGlobalNumerics().Random()*4) + 1;
    
	int tester;
	i=0;j=0;
	while (numUpdated <= updateLimit ) {
		i = (int)(Numerics::getGlobalNumerics().Random() * mNumStates);
		j = (int)(Numerics::getGlobalNumerics().Random() * mNumStates);
		while (j == i) j = (int)(Numerics::getGlobalNumerics().Random() * mNumStates);
		if ( (i==0) && (j==2) ) continue;
		if ( (i==2) && (j==0) ) continue;
		
		tester = UpdateAlphaParameter( theModel, i, j, 1);
		if (tester == 0) continue;
		numUpdated++;
	}
    
	// Following Yang's formulation...
	mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[0].mDataArray[2] = 1.0;
	mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[2].mDataArray[0] = 1.0;
	
	m_bUpdate = true;
}


// older version
//---------------------------------------------------------------------------
void GTRModel::UpdateAlphaMatrix(DoubleMatrix& mAlphaMatrix, const DoubleVector& mFrequency, const int& m_nMethod)
{
	int i,j;
	bool m_bUpdate = false;
	static DoubleMatrix tempAlphaMatrix;
	
	if (options.m_nMethod == 0)
	{
		while (!m_bUpdate)
		{
			tempAlphaMatrix = mAlphaMatrix;
            
			IntMatrix found(mNumStates, mNumStates, 0);
            
			for (i = 0; i < mNumStates; i++)
			{
				for (j = 0; j < mNumStates; j++)
				{
					if ( i == j)
						found.mRows.mDataArray[i].mDataArray[j] = 0;
					else
						found.mRows.mDataArray[i].mDataArray[j] = 1;
				}
			}
            
			int updateCount = 0;
			for (i = 0; i < mNumStates; i++)
			{
				for (j = 0; j < mNumStates; j++)
				{
					if (found.mRows.mDataArray[i].mDataArray[j])
					{
						if (Numerics::getGlobalNumerics().Random() <= ( 0.25 )) {
							UpdateAlphaParameter(tempAlphaMatrix.mRows.mDataArray[i].mDataArray[j]);
							updateCount++;
						}
						//Non-rev!
                        tempAlphaMatrix.mRows.mDataArray[j].mDataArray[i] = tempAlphaMatrix.mRows.mDataArray[i].mDataArray[j];
                        found.mRows.mDataArray[i].mDataArray[j] = 0;
                        //Non-rev!
                        found.mRows.mDataArray[j].mDataArray[i] = 0;
					}
				}
                
				// Following Yang's specification...
				tempAlphaMatrix.mRows.mDataArray[0].mDataArray[2] = 1.0;
				tempAlphaMatrix.mRows.mDataArray[2].mDataArray[0] = 1.0;
				
				double sum = 0;
                
				double tempRate;
                
				for (j = 0; j < mNumStates; j++)
				{
					if (i != j)
					{
						tempRate = tempAlphaMatrix.mRows.mDataArray[i].mDataArray[j] * mFrequency.mDataArray[j];
						sum += tempRate;
					}
				}
                
				tempRate = 1 - sum;
                
				if ((tempRate <= 0.0) || (tempRate >= 1.0))
				{
					m_bUpdate = false;
					break;
				}
				else
				{
					m_bUpdate = true;
					tempAlphaMatrix.mRows.mDataArray[i].mDataArray[i] = 0;
				}
			}
		}
	}
	else
	{
		int numUpdated = 0;
        //	tempAlphaMatrix = mAlphaMatrix;
		
		int updateLimit =(int)(Numerics::getGlobalNumerics().Random()*4) + 1;
		//std::cout << "Going to update " << updateLimit << " rates." << endl;
		
		int i,j;
		i=0;j=0;
		while (numUpdated <= updateLimit ) {
			i = (int)(Numerics::getGlobalNumerics().Random() * mNumStates);
			j = (int)(Numerics::getGlobalNumerics().Random() * mNumStates);
			while (j == i) j = (int)(Numerics::getGlobalNumerics().Random() * mNumStates);
			
			// UpdateAlphaParameter( int theModel, int i, int j, int reversibleEvaluation);
			UpdateAlphaParameter(mAlphaMatrix.mRows.mDataArray[i].mDataArray[j]);
			mAlphaMatrix.mRows.mDataArray[j].mDataArray[i] = mAlphaMatrix.mRows.mDataArray[i].mDataArray[j];
            
			numUpdated++;
		}
        // Following Yang's formulation...
		mAlphaMatrix.mRows.mDataArray[0].mDataArray[2] = 1.0;
		mAlphaMatrix.mRows.mDataArray[2].mDataArray[0] = 1.0;
        
		m_bUpdate = true;
	}
}

//---------------------------------------------------------------------------
void GTRModel::UpdateAlphaMatrix3D(const int& m_nMethod)
{
	DoubleVector tempFrequency(mNumStates, 0);
	
	for (unsigned int k = 0; k< mNumModels; k++)
	{
		tempFrequency = mFrequencyMatrix[k];
		UpdateAlphaMatrix( k, options.m_nMethod );
		//UpdateAlphaMatrix(mAlphaMatrix3D[k], tempFrequency, options.m_nMethod);
	}
}

//---------------------------------------------------------------------------
void GTRModel::UpdateParameter(const int& m_nMethod)
{
	if (options.freeFreqs == 1) UpdateFrequencyMatrix();
	UpdateAlphaMatrix3D(options.m_nMethod);
	GTRModelParameterstoRates();
	UpdateFreqModels();
}

