/*
 *  MCMC.cpp
 *
 *  A.P. Jason de Koning, jason.de.koning@gmail.com
 *  http://jasondk.org
 *
 *  David Pollock lab, University of Colorado School of Medicine
 *  http://www.evolutionarygenomics.com
 *  All rights reserved.

 */

// NOTE:  THIS FILE INCLUDES A VARIETY OF EXPERIMENTAL CODE USED IN DEVELOPMENT.

#include "MCMC.h"
#include <fstream>

#include "RandomGenerator.h"
#include "SiteCategory.h"

using std::cerr;
using std::endl;

//-----------------------------------------------------------------------
class EmptySequenceDetector		//this class is used to detect if the input sequence is empty or not
{
public:
    bool	operator()(
                       const TreeNodeData_&	inData) const;
};

//-----------------------------------------------------------------------
bool
EmptySequenceDetector::operator()(
                                  const TreeNodeData_&	inData) const
{
    return inData.sequence.empty();
}


//---------------------------------------------------------------------------

MCMC::MCMC(void) {
	// Set a pointer to the global numerics object
	num = &Numerics::getGlobalNumerics();
}


// ************************************************************************//
// Functions for tree manipulation
// ************************************************************************//

void MCMC::InitializeTree( Tree& mTree, ObservedData& theData ) {
	
	// Initialize the tree with the sequences, sequence names
	// mTree.for_each(SequenceAssigner(m_vecSequences, m_vecSeqname));

    mTree.assignSequencesToLeaves( theData );
	mTree.RemoveFreeLeaves(EmptySequenceDetector());
	mTree.setinternalnodes();
	
	mTree.root()->propogateRootLink(mTree.root());
}

void MCMC::InitializeTreeAlternate(Tree& mTree, ObservedData& theData) {
	
	// Initialize the tree with the sequences, sequence names
	mTree.root()->propogateRootLink(mTree.root());
}


// ************************************************************************//
// Change in likelihood wrappers
// ************************************************************************//

double MCMC::CalculateCompleteDataLogLikelihood_Delta_B1(
														 Tree& mTree,
                                                         MixtureModel& mModel,
														 IntVector& m_vecSiteAssignment,
														 int theModel, int x) {		
	
	// Calculate the contribution of row x to the complete-data log-likelihood under standard B1
	double logLikelihood;
	mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels(), 1, theModel, x);
	
	mTree.calculateCompleteDataLogLdeltab1( mModel, m_vecSiteAssignment, logLikelihood, theModel, x);	
	return(logLikelihood);
}

double MCMC::CalculateCompleteDataLogLikelihood_Delta_B1u(
														  Tree& mTree,
														  MixtureModel& mModel,
														  ObservedData& theData,
														  IntVector& m_vecSiteAssignment,
														  int theModel, int x, int y) {		
	
	// Calculate the contribution of row x, col y to the complete-data log-likelihood under B1u
	double logLikelihood = mTree.calculateCompleteDataLogLdeltab1uni( mModel, m_vecSiteAssignment, logLikelihood, theModel, x, y);	
   
	// ********** WARNING **********
	// If capUniformizationRate is not being used, then a call to this function can be allowed to increase the uni constant
	if (options.capUniformizationRate == 0) {
		std::cout << "	** WARNING **  You've called a function that can alter the uniformization constant!  CalculateCompleteDataLogLikelihood_Delta_B1u()" << endl;
	}
	
	return(logLikelihood);
}

// ************************************************************************//
// Initialization
// ************************************************************************//

void MCMC::InitializeMixtureModel(ObservedData& theData,
								  MixtureModel& m_Model,
								  const IntMatrix& m_ChangeMatrix) {
	
	// this function to initialize the parameter of each model in the mixture model
	m_Model.SetNumStates( theData.getNumStates() );
//	m_Model.SetStates(m_vecState);
//	m_Model.SetMinState(m_nMinstate);
//	m_Model.SetStateIndex(m_vecStateIndex);
	m_Model.SetNumModels(options.m_nNumModels);
	m_Model.SetFixedModels(options.m_bFixedFreq);
	m_Model.SetUpdateWindowsize(options.m_fUpdatewindow);
	m_Model.InitializeParameter(m_ChangeMatrix);
	
	if (!options.capUniformizationRate) m_Model.updateUniformization();
}

// ************************************************************************//
// Utility functions
// ************************************************************************//

void MCMC::NormalizeMatrix(
						   DoubleMatrix& m_probMatrix) {
	for (int i = 0; i < m_probMatrix.rows(); i++)
	{
		double sum = 0;
		
		for (int j = 0; j< m_probMatrix.cols(); j++)
			sum += m_probMatrix.mRows.mDataArray[i].mDataArray[j];
		
		for (int j = 0; j< m_probMatrix.cols(); j++)
			m_probMatrix.mRows.mDataArray[i].mDataArray[j] = m_probMatrix.mRows.mDataArray[i].mDataArray[j]/sum;
	}
	
}	

int MCMC::sampleFromCumulative( double *theVector, int length ) {
	// This function assumes:
	//    1) LOG likelihoods are NOT HEATED
	//    2) Assumed theVector contains log-likelihood values
	
	// WARNING: this function will destroy the incoming vector...
	
	// Sampling from:  exp(value - max) / sum_i(exp(val_i - max))
	
	int i, theMax;
	double sum = 0.0, max = -std::numeric_limits<double>::max( );
	
	for (i=0; i<length; i++) {
		// Heat the relative log-likelihoods
		theVector[i] *= options.__globalLikelihoodTemp;
		if (theVector[i] > max) { max = theVector[i]; theMax = i; }
	}
	
	for (i=0; i<length; i++) {
		theVector[i] = exp(theVector[i] - max);
		sum += theVector[i];
	}
	
	for (i=0; i<length; i++) {
		theVector[i] /= sum;
		if (i>0) theVector[i] += theVector[i-1];
	}
	
	double rannum = num->Random();
	
	for (i=0; i<length;i++) {
		if (rannum <= theVector[i]) return i;
	}

    return i;
}


// ************************************************************************//
// Model update wrappers
// ************************************************************************//
void MCMC::UpdateModel(
					   MixtureModel& m_Model,
                       const double& m_fWindowsize) {
	
	m_Model.SetUpdateWindowsize(m_fWindowsize );
	m_Model.UpdateParameter(options.m_nMethod);
}

void MCMC::UpdateModel(
					   MixtureModel& m_Model) {
	
	m_Model.UpdateParameter(options.m_nMethod);
}

void MCMC::UpdateBranchLengths(
							   Tree& mTree, 
							   int startNode,
							   GTRModel& mModel, 
							   ObservedData& theData,
							   IntVector& m_vecSiteAssignment,
							   int version, double variance) {
	
	// Versions: 0 = standard branch-wise deltaL, 1 = conjugate gibbs (b1u), 2 = simple proposal evaluation (b1u)
	
	int dirty = mTree.KEEP_updateBranchLength( startNode, variance, mModel, theData, m_vecSiteAssignment, version );

}

void MCMC::UpdateBranchLengths(
							   Tree& mTree, 
							   int startNode,
							   ProteinModel& mModel, 
							   ObservedData& theData,
							   IntVector& m_vecSiteAssignment,
							   int version, double variance) {
	
    // versions: 0 = standard branch-wise deltaL, 1 = conjugate gibbs (b1u), 2 = simple proposal evaluation (b1u)
	int dirty = mTree.KEEP_updateBranchLength( startNode, variance, mModel, theData, m_vecSiteAssignment, version );
	
}

//---------------------------------------------------------------------------------------------------------
void MCMC::ResampleSiteAssignments_MixtureModel(
												vector< vector<int> > &m_vecSequences,
												IntVector& m_vecSiteAssignment,
												IntMatrix3D& count3D,
												MixtureModel& mModel,
                                                ObservedData& theData ) {

	// About to use gibbs sampling to update the assignment of each model to sequences
	
	DoubleMatrix3D ProbMatrix3D;
	std::cout << "DOING THE EXPENSIVE ResampleSiteAssignments_MixtureModel()..." << endl;
	
	unsigned int numModels = mModel.GetNumModels();
	unsigned int numModelStates = count3D.cols();
	unsigned int numCells = count3D.rows();		//number of sequences of sites
	
	if (m_vecSiteAssignment.empty())
		m_vecSiteAssignment.resize(numCells, -1);
	else
		m_vecSiteAssignment.fill(-1);
	
	for (unsigned int k = 0; k < numCells; k++) {
		IntMatrix mCount = count3D[k];
		DoubleVector logLikelihoodModel(numModels, 0);
		double logLikelihoodModelMax = 0;
		
		for (unsigned int m = 0; m < numModels; m++) {
			for (unsigned int i = 0; i < numModelStates; i++)
				for (unsigned int j = 0; j < numModelStates; j++)
					logLikelihoodModel[m] += mCount[i][j] * log(ProbMatrix3D[m][i][j]);
			
			if (m== 0)
				logLikelihoodModelMax = logLikelihoodModel[m];
			else if (logLikelihoodModel[m] > logLikelihoodModelMax)
				logLikelihoodModelMax = logLikelihoodModel[m];
		}
		
		DoubleVector m_fLikelihoodRatio(numModels, 0);
		
		double sum = 0;
		
		for (unsigned int m = 0; m < numModels; m++)
		{
			if (logLikelihoodModel[m] - logLikelihoodModelMax >= -22)
				m_fLikelihoodRatio[m] = exp(logLikelihoodModel[m] - logLikelihoodModelMax);
			else
				m_fLikelihoodRatio[m] = 0;
			
			sum += m_fLikelihoodRatio[m];
		}
		
		DoubleVector m_fCummulativeRatio(numModels, 0);
		
		for (unsigned int m = 0; m < numModels; m++)
		{
			if (m == 0)
				m_fCummulativeRatio[m] = m_fLikelihoodRatio[m] / sum;
			else
				m_fCummulativeRatio[m] = m_fCummulativeRatio[m-1] + m_fLikelihoodRatio[m] / sum;
		}
		
		double rannum = num->Random();
		
		for (unsigned int m = 0; m < numModels; m++)
		{
			if (rannum <= m_fCummulativeRatio[m])
			{
		//		if (!m_bSite)
		//			m_vecSequences[2*k+1] = m;
		//		else
				m_vecSiteAssignment[k] = m;
				
				break;
			}
		}
	}
}

//---------------------------------------------------------------------------------------------------------

int MCMC::ResampleSiteAssignments_MixtureModel(
											   Tree& mTree,
											   IntVector& m_vecSiteAssignment,
											   MixtureModel& mModel,
                                               ObservedData& theData ) {
	return mTree.ResampleSiteAssignments_MixtureModel( mModel, theData, m_vecSiteAssignment );	
}

void MCMC::ResampleSiteAssignments_ASRV_rates(
											  Tree& mTree,
											  MixtureModel& mModel,
											  ObservedData& theData,
                                              IntVector& m_vecSiteAssignment)

{
	
	mTree.KEEP_ResampleSiteAssignments_ASRV_rates( mModel, theData, m_vecSiteAssignment );	
}

void MCMC::InitializeAssignment(
								Tree& mTree,
								IntVector& m_vecSiteAssignment,
								MixtureModel& mModel,
                                ObservedData& theData ) {
	
	ResampleSiteAssignments_MixtureModel(mTree, m_vecSiteAssignment, mModel, theData );	
}

//---------------------------------------------------------------------------

double MCMC::CalculateCompleteDataLogLikelihood(
												Tree& mTree,
												MixtureModel& mModel2,
												IntVector& m_vecSiteAssignment) {	
	
 	return CalculateCompleteDataLogLikelihood( mTree, mModel2, m_vecSiteAssignment, 0);
}

double MCMC::CalculateCompleteDataLogLikelihood(
												Tree& mTree,
												MixtureModel& mModel2,
												IntVector& m_vecSiteAssignment,
												int version) {		
	
	double logLikelihood;
	mTree.setprobmatrixalongbranch( mModel2, mModel2.GetFreqModels() );
	mTree.calculateCompleteDataLogL( mModel2 , m_vecSiteAssignment, logLikelihood, version );	
	
	return(logLikelihood);
}

double MCMC::CalculateGibbsSamplingLnLikelihoodRatio(
													 Tree& mTree,
													 MixtureModel& mModel1,
													 MixtureModel& mModel2,
													 IntVector& m_vecSiteAssignment) {
	
	// calculate the log likelihood ratio of two different models fits the data
	int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
	mTree.root()->setCalculateFullProbabilityDistribution(1);
	mTree.setprobmatrixalongbranch( mModel1, mModel1.GetFreqModels());
	double logLikelihood1 = CalculateCompleteDataLogLikelihood(mTree, mModel1, m_vecSiteAssignment);
	mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
	
	tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
	mTree.root()->setCalculateFullProbabilityDistribution(1);
	mTree.setprobmatrixalongbranch(mModel2, mModel2.GetFreqModels());
	double logLikelihood2 = CalculateCompleteDataLogLikelihood(mTree, mModel2, m_vecSiteAssignment);
	mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
	
	return( (logLikelihood2 - logLikelihood1) );
}

//---------------------------------------------------------------------------------------------------------

int MCMC::B1u_ConjugateGibbsUpdate_rateMatrix_nonRev(
													 Tree& mTree, 
													 ProteinModel& mModel, 
													 IntVector& m_vecSiteAssignment,
													 bool& m_bUpdate) {

	// *** WARNING *** Support for ASRV is 'in progress'
	
	// if (options.allowASRV != 0) { std::cout << "	ERROR! CONJUGATE GIBBS SAMPLING DOESN'T WORK WITH AMONG SITE RATE VARIATION YET..." << endl; exit(-1); }
	
	static DoubleVector maxWaitingTimes;
	static ProteinModel tempModel1;
	tempModel1 = mModel;
	DoubleVector *mu = mModel.returnUniformizationConstants();
	
	// This function is for ProteinModel models only...
	int numStates  = mModel.GetNumStates();
	int numModels  = mModel.GetNumModels();
	DoubleMatrix* stateF = &mModel.returnStateFrequencies();
	
	// Pick a single rate to update...
	int whichRateFrom;
	int whichRateTo;
	int whichModel = -1;
	int modelStart, modelEnd;
	
	DoubleMatrix3D    *CXY = mTree.root()->getChangeMatrix3D( mTree.root() ); //&GetData().countModelchange3D;
	DoubleMatrix      *C1  = mTree.root()->getAugmentedCountMatrix( mTree.root() );
	
	int count=0;
	for (int numUpdates = 0; numUpdates <  (numStates * (numStates-1)); numUpdates++) {
		whichRateFrom  = (int)(num->Random() * (numStates));
		whichRateTo    = (int)(num->Random() * (numStates));
		
		while (whichRateTo == whichRateFrom) {
			whichRateTo = (int)(num->Random() * (numStates));
		}
		
		// Got a rate to do...
		if (whichModel == -1) {
			modelStart = 0; modelEnd = numModels;
		} else {
			modelStart = whichModel; modelEnd = modelStart+1;
		}
		
		for (int theModel=modelStart; theModel < modelEnd; theModel++) {
			count++;
			double oldRate1 = mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo];
			double cxy_1 = (*CXY).mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo];
			double c1_1  = (*C1).mRows.mDataArray[ theModel ].mDataArray[whichRateFrom];

			if (c1_1 < 0) { std::cout << " theModel is " << theModel << ", whichrateFrom is " << whichRateFrom << endl; exit(-1); }
			
			double LambdaX = -mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom];
			double Delta = (*mu).mDataArray[theModel] - LambdaX + oldRate1;
			
			if (c1_1 < 0) { std::cout << "What the heck!  c1_1 is < 0!! " << c1_1 << endl; }
			
			double var1 = num->betaVariate( (cxy_1 + 1), (c1_1 + 1 )) * ( Delta );
			double stateFreq = stateF->mRows.mDataArray[theModel].mDataArray[whichRateTo];
			
			// Override with MLE:
			//double var1a = num->MLE_B1u_instRate( (LambdaX-oldRate1), c1_1, cxy_1, (*mu).mDataArray[theModel]);
			//if ( (var1a > 0) && (var1a == var1a) ) {
			//	var1 = var1a;
			//}
			
			// Validate based on capuniformation rules:
			if (options.allowASRV) {
				if ( ( -(mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom] + oldRate1 - var1 ) * options.__largestRate ) > (*mu).mDataArray[theModel] ) {
					// Don't accept it!
				} else {
					// Accept
					if (options.scaleByFrequencies) {
						mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] = var1 / stateFreq;
					} else {
						mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] = var1 ;
					}
					mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] = var1;
					
					mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom] += oldRate1 - var1;
				}
			} else {
				// ACCEPT without validation...	
				
				if (options.scaleByFrequencies) {
					mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] = var1 / stateFreq;
				} else {
					mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] = var1 ;
				}
				mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] = var1;
				
				mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom] += oldRate1 - var1;
			}			
		}
	}
	return count;
}

int MCMC::UpdateASRV_shape_MC(
							  Tree& mTree, 
							  GTRModel& mModel, 
							  ObservedData& theData,
							  IntVector& m_vecSiteAssignment,
							  bool& m_bUpdate, double var) {
	
	// For uniformized B1, we'd like three versions: BRANCH, RATE, ALL [these will return either the full (ALL) likleihood, or just the parts that depend on rates or branches]
//	const int BOTH    = 0;
	const int RATE    = 1;

	double logLikelihood1a, logLikelihood2;
	double logLR;
	
	static int    unibackupAlloc = 0;
	static double *backupUni;
	if (unibackupAlloc == 0) {
		backupUni = new double[mModel.GetNumModels()];
		unibackupAlloc = 1;
	}
	
	int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
	mTree.root()->setCalculateFullProbabilityDistribution(0);
	
	
	// NOTE: unless the uniformization constant was changed for evaluation, we shouldn't need to recalculate the likelihood here for mModel (tempModel1)
	const int numStates = mModel.GetNumStates();
	const int alignLength = m_vecSiteAssignment.size();
	
	double proposal;
	double startShape = mTree.root()->GetData().siteCats->getShapeParm();
	proposal = -startShape - 5;
	while ( (startShape + proposal) <= 0.0 ) {
		proposal   = num->box_muller(0, var);
	}
	
	int uniChanged = 0;
	
	if (options.m_nMethod == 6) {
		uniChanged = 0;

		// Backup uniformization constants...
		for (int ii=0; ii<mModel.GetNumModels(); ii++) {
			backupUni[ii] = mModel.getUniformization(ii);
		}
		
		mTree.root()->GetData().siteCats->setGammaDistribution( (startShape + proposal) );
		mModel.forceUpdateUniformization();
		mTree.root()->GetData().siteCats->setGammaDistribution( (startShape ) );
		
		// Check uniformization constants...
		for (int ii=0; ii<mModel.GetNumModels(); ii++) {
			if ( (mModel.returnUniformizationConstants()->mDataArray[ii] ) > backupUni[ii] ) {
				uniChanged = 1;
				break;
			}
		}
		
		if ( ( (uniChanged == 1) && (options.capUniformizationRate) ) || (uniChanged == 0) )  {
			for (int ii=0; ii<mModel.GetNumModels(); ii++) {
				mModel.setUniformization(ii, backupUni[ii]);
			}
			if (options.capUniformizationRate && (uniChanged == 1) ) {
				//	std::cout << "REJECTING GAMMA PROPOSAL BECAUSE IT EXCEEDS MU..." << endl;
				return -1;
			}
		} else {
			//		std::cout << "	[gamma proposal caused uni to change from " << backupUni[0] << " to " << mModel.returnUniformizationConstants()->mDataArray[0] << endl;
		}
	}
	
	// ***************************************************************************************************
	// Model 1 likelihood... (only the parts that depend on RATE parameters - when m_nMethod allows)
	logLikelihood1a = CalculateCompleteDataLogLikelihood(mTree, mModel, m_vecSiteAssignment, RATE);
	
	int resampleAssignments = 0;
	
	mTree.root()->GetData().siteCats->setGammaDistribution( (startShape + proposal) );
	if ( num->Random() < 0.1 ) {
		resampleAssignments = 1;
		
		mTree.root()->GetData().siteCats->backupSiteAssignments();
		ResampleSiteAssignments_ASRV_rates(mTree, mModel, theData, m_vecSiteAssignment );	
	}
	
	// Model 2 likelihood... (only the parts that depend on RATE parameters - when m_nMethod allows)
	logLikelihood2 = CalculateCompleteDataLogLikelihood(mTree, mModel, m_vecSiteAssignment, RATE);
	mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
	
	// Got the log-likelihood ratio (assuming the larger of the two uniformization constants)...
	logLR = (logLikelihood2 - logLikelihood1a);
	
	//std::cout << "		[Gamma shape proposal: LR is " << logLR << "; from " << startShape << " -> " << (startShape + proposal) << " ]" << endl;
	
	
	// Accept / reject?

	m_bUpdate = false;
	if (logLR >= 0.0 )
	{
		m_bUpdate = true;
	} else {
		if (num->evaluateProposal_logL( logLR ) ) {
			m_bUpdate = true;
		} else {
			m_bUpdate = false;
		}
	}
	// ***************************************************************************************************

	if (m_bUpdate == false) {
		mTree.root()->GetData().siteCats->setGammaDistribution( startShape );
		if ( (options.m_nMethod == 6)  ) {
			for (int ii=0; ii<mModel.GetNumModels(); ii++) {
				mModel.setUniformization(ii, backupUni[ii]);
			}
		}
		
		if (resampleAssignments == 1) {
			if (options.m_nMethod == 6) {
				for (int ii=0; ii<m_vecSiteAssignment.size(); ii++) {
					mTree.root()->reassignSiteCountsBetweenRateCategories_Uni(ii, mTree.root()->GetData().siteCats->getAssignment(ii), mTree.root()->GetData().siteCats->getBackupAssignment(ii), theData, m_vecSiteAssignment );
				}
			}			
			mTree.root()->GetData().siteCats->restoreSiteAssignments();
		}
		return 0;
	} else {
		return 1;
	}
	
}

int MCMC::UpdateASRV_shape_MC(
							  Tree& mTree, 
							  ProteinModel& mModel, 
							  ObservedData& theData,
							  IntVector& m_vecSiteAssignment,
							  bool& m_bUpdate, double var) {
	
	// For uniformized B1, we'd like three versions: BRANCH, RATE, ALL [these will return either the full (ALL) likleihood, or just the parts that depend on rates or branches]
//	const int BOTH    = 0;
	const int RATE    = 1;
//	const int BRANCH  = 2;
	
	double logLikelihood1a, logLikelihood2;
	double logLR;
	
	static int    unibackupAlloc = 0;
	static double *backupUni;
	if (unibackupAlloc == 0) {
		backupUni = new double[mModel.GetNumModels()];
		unibackupAlloc = 1;
	}
	
	// ***************************************************************************************************
	// Get the likelihoods of the start and end
	// ***************************************************************************************************
	int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
	mTree.root()->setCalculateFullProbabilityDistribution(0);
	
	
	// NOTE: unless the uniformization constant was changed for evaluation, we shouldn't need to recalculate the likelihood here for mModel (tempModel1)
	const int numStates = mModel.GetNumStates();
	const int alignLength = m_vecSiteAssignment.size();
	
	double proposal;
	double startShape = mTree.root()->GetData().siteCats->getShapeParm();
	
	proposal = -startShape - 5;
	while ( (startShape + proposal) <= 0.0 ) {
		proposal   = num->box_muller(0, var);
	}
	
	int uniChanged = 0;
	
	if (options.m_nMethod == 6) {
		uniChanged = 0;
		// Backup uniformization constants...
		for (int ii=0; ii<mModel.GetNumModels(); ii++) {
			backupUni[ii] = mModel.getUniformization(ii);
		}
		
		mTree.root()->GetData().siteCats->setGammaDistribution( (startShape + proposal) );
		mModel.forceUpdateUniformization();
		mTree.root()->GetData().siteCats->setGammaDistribution( (startShape ) );
		
		// Check uniformization constants...
		for (int ii=0; ii<mModel.GetNumModels(); ii++) {
			if ( (mModel.returnUniformizationConstants()->mDataArray[ii] ) > backupUni[ii] ) {
				uniChanged = 1;
				break;
			}
		}
		
		if ( ( (uniChanged == 1) && (options.capUniformizationRate) ) || (uniChanged == 0) )  {
			for (int ii=0; ii<mModel.GetNumModels(); ii++) {
				mModel.setUniformization(ii, backupUni[ii]);
			}
			if (options.capUniformizationRate && (uniChanged == 1) ) {
				return -1;
			}
		} else {
			//		std::cout << "	[gamma proposal caused uni to change from " << backupUni[0] << " to " << mModel.returnUniformizationConstants()->mDataArray[0] << endl;
		}
	}
	
	// Model 1 likelihood... (only the parts that depend on RATE parameters - when m_nMethod allows)
	logLikelihood1a = CalculateCompleteDataLogLikelihood(mTree, mModel, m_vecSiteAssignment, RATE);
	
	int resampleAssignments = 0;

	mTree.root()->GetData().siteCats->setGammaDistribution( (startShape + proposal) );
	if ( num->Random() < 0.667 ) { //.1 ) {
		resampleAssignments = 1;
		
		mTree.root()->GetData().siteCats->backupSiteAssignments();
		ResampleSiteAssignments_ASRV_rates(mTree, mModel, theData, m_vecSiteAssignment );	
	}

	// Model 2 likelihood... (only the parts that depend on RATE parameters - when m_nMethod allows)
	logLikelihood2 = CalculateCompleteDataLogLikelihood(mTree, mModel, m_vecSiteAssignment, RATE);
	mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
	
	// Got the log-likelihood ratio (assuming the larger of the two uniformization constants)...
	logLR = (logLikelihood2 - logLikelihood1a);
	
	//std::cout << "		[Gamma shape proposal: LR is " << logLR << "; from " << startShape << " -> " << (startShape + proposal) << " ]" << endl;
	// ***************************************************************************************************
	
	// Accept / reject?
	m_bUpdate = false;
	if (logLR >= 0.0 )
	{
		m_bUpdate = true;
	} else {
		
		if (num->evaluateProposal_logL( logLR ) ) {
			m_bUpdate = true;
		} else {
			m_bUpdate = false;
		}
	}

	if (m_bUpdate == false) {
		mTree.root()->GetData().siteCats->setGammaDistribution( startShape );
		if ( (options.m_nMethod == 6)  ) {
			for (int ii=0; ii<mModel.GetNumModels(); ii++) {
				mModel.setUniformization(ii, backupUni[ii]);
			}
		}
		
		if (resampleAssignments == 1) {
			if (options.m_nMethod == 6) {
				for (int ii=0; ii<m_vecSiteAssignment.size(); ii++) {
					mTree.root()->reassignSiteCountsBetweenRateCategories_Uni(ii, mTree.root()->GetData().siteCats->getAssignment(ii), mTree.root()->GetData().siteCats->getBackupAssignment(ii), theData, m_vecSiteAssignment );
				}
			}
			mTree.root()->GetData().siteCats->restoreSiteAssignments();
			
		}
		return 0;
	} else {
		//	std::cout << "	[Updated gamma shape from " << startShape << " to " << (startShape + proposal) << "]" << endl;
		return 1;
	}
	
}

void MCMC::UpdateRateModelParameters_MC(
										Tree& mTree, 
										GTRModel& mModel, 
										ObservedData& theData,
										IntVector& m_vecSiteAssignment,
										bool& m_bUpdate)
{
	// For uniformized B1, we'd like three versions: BRANCH, RATE, ALL [these will return either the full (ALL) likleihood, or just the parts that depend on rates or branches]
//	const int BOTH    = 0;
	const int RATE    = 1;
//	const int BRANCH  = 2;
	
	static int firstCall = 1;
	static GTRModel tempModel1(mModel);
	static GTRModel tempModel2(mModel);
	if (firstCall == 1) {
		firstCall = 0;
	} else {
		tempModel1 = mModel;
		tempModel2 = mModel;
		firstCall = 0;
	}
	
	double logLikelihood1a, logLikelihood2;
	double logLR;
	
	int numModels = mModel.GetRateMatrix3D().rows();
	
	for (int whichModel = 0; whichModel < numModels; whichModel++) {
		double startMu; //, newMu;
		
		Vector<double> *theFreqs1, *theFreqs2;
		theFreqs1 = &tempModel1.returnStateFrequencies()[whichModel];
		theFreqs2 = &tempModel2.returnStateFrequencies()[whichModel];
		double windowsize = mModel.GetUpdateWindowsize();
		
		if (options.capUniformizationRate) startMu =  mModel.getUniformization(whichModel);
		
		// THIS IS INEFFICIENT: SHOULD CHECK IF THE PROPOSAL EXCEEDS MU IN THE
		// UPDATEMODEL() FXN AND REFLECT BACK IF CAPUNI IS ON...
		
		// Multidimensional update...
		UpdateModel(tempModel2, windowsize);
	
		// ***************************************************************************************************
		// This version for enforcing a capped uniformization constant...
		// ***************************************************************************************************
		/*		if (options.capUniformizationRate) {
		 tempModel2.forceUpdateUniformization(whichModel);
		 newMu = tempModel2.getUniformization(whichModel); /// / 1.04;
		 
		 int rejectCount = 0;
		 while (newMu > startMu) {
		 tempModel2.setUniformization(whichModel, startMu);
		 for (int ii=0; ii<theFreqs2->size(); ii++) {
		 theFreqs2->mDataArray[ii] = theFreqs1->mDataArray[ii];
		 }
		 
		 UpdateModel(tempModel2, windowsize, options.m_nMethod);
		 tempModel2.forceUpdateUniformization(whichModel);
		 newMu = tempModel2.getUniformization(whichModel); /// / 1.04;
		 rejectCount++;
		 //		std::cout << "	startmu was " << startMu << "; newmu is " << newMu << endl;
		 if (rejectCount > 10) { 
		 std::cout << "*" << endl; 
		 if (windowReduce < 5) { 
		 mModel.SetUpdateWindowsize( (windowsize * 0.9) ); 
		 std::cout << "	[reduced proposal window size to " << mModel.GetUpdateWindowsize() << "]" << endl; 
		 windowReduce++;
		 continue;
		 }
		 return;
		 }
		 }
		 //	std::cout << " *** OKAY *** "<< endl;
		 
		 tempModel2.setUniformization(whichModel, (startMu) );
		 } */
		// ***************************************************************************************************
		
		int uniDirtyFlag = 0;
		// ***************************************************************************************************
		// This version has to check if the implied uniformization constant from the proposal needs to be
		// used for evaluating the delta log-L
		// ***************************************************************************************************
		if (!options.capUniformizationRate) {
			tempModel1.updateUniformization(whichModel);
			tempModel2.forceUpdateUniformization(whichModel);
			
			double theMax = 0.0;
			DoubleVector *mu1 = tempModel1.returnUniformizationConstants();
			DoubleVector *mu2 = tempModel2.returnUniformizationConstants();
			if ( (*mu1)[whichModel] > theMax) { theMax = (*mu1)[whichModel]; }
			if ((*mu2)[whichModel] > theMax) { theMax = (*mu2)[whichModel]; }
			
			uniDirtyFlag = 0;
			if ( theMax >= (*mu1)[whichModel] ) {
				// Set both models' uniformization constant to be the same...
				tempModel2.setUniformization(whichModel, theMax ); // * 1.01
				tempModel1.setUniformization(whichModel, theMax );
				uniDirtyFlag = 1;
			} else {
				tempModel2.setUniformization(whichModel, (*mu1)[whichModel] );
			}
		}	
		
		//		tempModel2.forceUpdateUniformization();
		
		// ***************************************************************************************************
		
		// ***************************************************************************************************
		// Get the likelihoods of the start and end
		// ***************************************************************************************************
		int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
		mTree.root()->setCalculateFullProbabilityDistribution(0);
		
		// NOTE: unless the uniformization constant was changed for evaluation, we shouldn't need to recalculate the likelihood here for mModel (tempModel1)
		
		// Model 1 likelihood... (only the parts that depend on RATE parameters - when m_nMethod allows)
		logLikelihood1a = CalculateCompleteDataLogLikelihood(mTree, tempModel1, m_vecSiteAssignment, RATE);
		
		// Model 2 likelihood... (only the parts that depend on RATE parameters - when m_nMethod allows)
		logLikelihood2 = CalculateCompleteDataLogLikelihood(mTree, tempModel2, m_vecSiteAssignment, RATE);
		
		mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
		
		// Got the log-likelihood ratio (assuming the larger of the two uniformization constants)...
		logLR = (logLikelihood2 - logLikelihood1a);
		
		//		std::cout << "LR is " << logLR << endl;
		// ***************************************************************************************************
		
		
		// ***************************************************************************************************
		// Accept / reject?
		// ***************************************************************************************************
		m_bUpdate = false;
		if (logLR >= 0.0 )
		{
			mModel = tempModel2;
			m_bUpdate = true;
		} else {
			if (num->evaluateProposal_logL( logLR ) ) {
				mModel = tempModel2;
				m_bUpdate = true;
			} else {
				m_bUpdate = false;
			}
		}
		// ***************************************************************************************************
		
		
		
		if (m_bUpdate == true) {
			if (!options.capUniformizationRate) mModel.forceUpdateUniformization(whichModel);
			if (options.m_nMethod == 6) mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
		} else {
			// Update the prob-matrix at the root...
			if (options.m_nMethod == 6) mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
	
		}
	}	
}


int MCMC::UpdateRateModelParameters_MC(
									   double heat,
									   Tree& mTree, 
									   ProteinModel& mModel, 
									   ObservedData& theData,
									   IntVector& m_vecSiteAssignment,
									   bool& m_bUpdate)
{
	if (options.m_nMethod == 6) return UpdateModelParameters_nonRAR_MC( heat, mTree, mModel, theData, m_vecSiteAssignment,  m_bUpdate);
	
	ProteinModel tempModel1(mModel);
	ProteinModel tempModel2(mModel);
	double logLR;
	
	double windowsize = mModel.GetUpdateWindowsize();
	
	UpdateModel(tempModel2, windowsize);
	
	logLR = CalculateGibbsSamplingLnLikelihoodRatio(mTree, tempModel1, tempModel2, m_vecSiteAssignment);

	if ( num->evaluateProposal_logL( 1.0, (1.0 - logLR), 1.0, (1.0/(1.0+heat)) ) == 1 ) {
		mModel = tempModel2;
		m_bUpdate = true;
	} else {
		m_bUpdate = false;
	}
	
	if (m_bUpdate) return 1;
	else return 0;
}

int MCMC::UpdateModelParameters_nonRAR_MC(
										  double heat,
										  Tree& mTree, 
										  ProteinModel& mModel, 
										  ObservedData& theData,
										  IntVector& m_vecSiteAssignment,
										  bool& m_bUpdate) {
	// For uniformized B1, we'd like three versions: BRANCH, RATE, ALL [these will return either the full (ALL) likleihood, or just the parts that depend on rates or branches]
	//const int BOTH    = 0;
	const int RATE    = 1;
	//const int BRANCH  = 2;
	
	ProteinModel tempModel1(mModel);
	ProteinModel tempModel2(mModel);
	double logLR;
	int numModels = mModel.GetRateMatrix3D().rows();
	
	for (int whichModel = 0; whichModel < numModels; whichModel++) {	
		double logLikelihood1a, logLikelihood2;
		int numUpdated = 0;
		int tempUpdate = 0;
		
		Vector<double> *theFreqs1, *theFreqs2;
		theFreqs1 = &tempModel1.returnStateFrequencies()[whichModel];
		theFreqs2 = &tempModel2.returnStateFrequencies()[whichModel];
		
		double startMu, newMu;
		double windowsize = 0.01; // mModel.GetUpdateWindowsize();
		
		if (options.capUniformizationRate) startMu =  mModel.getUniformization(whichModel);
		UpdateModel(tempModel2, windowsize);
		//[blk]		tempModel2.rewriteAlphaMatrix3D(); // write the updated rate matrix
		//[blk]		tempModel2.ProteinModelParameterstoRates(); // write the updated rate matrix scaled by freqs
		
		if (options.capUniformizationRate) {
			tempModel2.forceUpdateUniformization(whichModel);
			newMu = tempModel2.getUniformization(whichModel) / 1.04;
			
			int rejectCount = 0;
			while (newMu > startMu) {
				tempModel2.setUniformization(whichModel, startMu);
				for (int ii=0; ii<theFreqs2->size(); ii++) {
					theFreqs2->mDataArray[ii] = theFreqs1->mDataArray[ii];
				}
				
				UpdateModel(tempModel2, windowsize);
				//[blk]			tempModel2.rewriteAlphaMatrix3D(); // write the updated rate matrix
				//[blk]			tempModel2.ProteinModelParameterstoRates(); // write the updated rate matrix scaled by freqs
				
				tempModel2.forceUpdateUniformization(whichModel);
				newMu = tempModel2.getUniformization(whichModel) / 1.04;
				rejectCount++;
				
				if (rejectCount > 10) { return 0;  std::cout << "*" << endl; }
			}		
			
			tempModel2.setUniformization(whichModel, (startMu) );
		}
		
		if (whichModel != -1) {
			numUpdated++;
		} else {
			numUpdated += mModel.GetNumModels();
		}
		
		int uniDirtyFlag = 0;
		if (!options.capUniformizationRate) {
			// Get the larger uniformization constant over tempModel1 and tempModel2...
			tempModel2.forceUpdateUniformization(whichModel);
			
			double theMax = 0.0;
			DoubleVector *mu1 = tempModel1.returnUniformizationConstants();
			DoubleVector *mu2 = tempModel2.returnUniformizationConstants();
			if ( (*mu1)[whichModel] > theMax) { theMax = (*mu1)[whichModel]; }
			if ((*mu2)[whichModel] > theMax) { theMax = (*mu2)[whichModel]; }
			
			uniDirtyFlag = 0;
			if ( theMax > (*mu1)[whichModel] ) {
				// Set both models' uniformization constant to be the same...
				tempModel2.setUniformization(whichModel, theMax * 1.04);
				tempModel1.setUniformization(whichModel, theMax * 1.04);
				uniDirtyFlag = 1;
			}
			//tempModel2.forceUpdateUniformization(whichModel);
			//tempModel1.forceUpdateUniformization(whichModel);
		}	
		
		int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
		mTree.root()->setCalculateFullProbabilityDistribution(0);
		
		// Model 1 likelihood... (only the parts that depend on RATE parameters)
		mTree.setprobmatrixalongbranch(tempModel1, tempModel1.GetFreqModels());
		if (uniDirtyFlag == 1) mTree.KEEP_CountFullChangesOverTree( tempModel1, theData, m_vecSiteAssignment );	
		
		logLikelihood1a = CalculateCompleteDataLogLikelihood(mTree, tempModel1, m_vecSiteAssignment, RATE);
		
		// Model 2 likelihood... (only the parts that depend on RATE parameters)
		mTree.setprobmatrixalongbranch(tempModel2, tempModel2.GetFreqModels());
		if (uniDirtyFlag == 1) mTree.KEEP_CountFullChangesOverTree( tempModel2, theData, m_vecSiteAssignment );	
		
		logLikelihood2 = CalculateCompleteDataLogLikelihood(mTree, tempModel2, m_vecSiteAssignment, RATE);
		
		mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
		
		// Got the likelihood ratio (assuming the larger of the two uniformization constants)...
		logLR = (logLikelihood2 - logLikelihood1a);		
		//	std::cout << "LR is " << logLR << endl;
		
		m_bUpdate = false;
		if (logLR >= 0.0 )
		{
			mModel = tempModel2;
			numUpdated += tempUpdate;
			m_bUpdate = true;
		} else {
			
			if (num->evaluateProposal_logL( logLR ) ) {
				mModel = tempModel2;
				numUpdated += tempUpdate;
				m_bUpdate = true;
			} else {
				m_bUpdate = false;
				numUpdated = 0;
			}
		}
		
		if (m_bUpdate == true) {
			//	std::cout << "UPDATE!" << endl;
			if (!options.capUniformizationRate) mModel.forceUpdateUniformization(whichModel);
			mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
		} else {
			// Need to reset the counts
			// [test] 	if (uniDirtyFlag == 1) mTree.KEEP_CountFullChangesOverTree( mModel, theData, m_vecSiteAssignment );	
		}
	}	
	
	
	if (m_bUpdate) return 1;
	else return 0;
}

//---------------------------------------------------------------------------------------------------------
int MCMC::ResampleParameterAssignments_RAR(
										   Tree& mTree, 
										   ProteinModel& mModel, 
										   ObservedData& theData,
										   IntVector& m_vecSiteAssignment,
										   bool& m_bUpdate) {
	// This function is for discrete-distributed rates only (for now)!
	
	// Unless otherwise specified, do update of 100% of parameters...
	return ResampleParameterAssignments_RAR( mTree, mModel, theData, m_vecSiteAssignment, m_bUpdate, 1.0);
	
}

int MCMC::UpdateRARDistributionParameters(
										  double heat,
										  Tree& mTree, 
										  ProteinModel& mModel, 
										  ObservedData& theData,
										  IntVector& m_vecSiteAssignment,
										  bool& m_bUpdate, double updateFraction, double *proposalVariances) {
	
	return UpdateRARDistributionParameters( heat, mTree, mModel, theData, m_vecSiteAssignment,  m_bUpdate, updateFraction, proposalVariances, 0 );
}

int MCMC::UpdateRARDistributionParameters(
										  double heat,
										  Tree& mTree, 
										  ProteinModel& mModel, 
										  ObservedData& theData,
										  IntVector& m_vecSiteAssignment,
										  bool& m_bUpdate, double updateFraction, double *proposalVariances, int whichModel) {
	
	if (options.m_nMethod == 6) return UpdateRARDistributionParameters_Uni_MC( heat, mTree, mModel, theData, m_vecSiteAssignment,  m_bUpdate, updateFraction, proposalVariances, whichModel);
	
	// This does a proposal affecting all models in the mixture.  Maybe not a good idea...
	
	double logLR;

	int numUpdated = 0;
	int tempUpdate = 0;
	
	if (!mModel.getRARdistributed()) { std::cout << "GammaShapeWithSweep update called when the model doesn't have gamma-stuff set!" << endl; exit(-1); }
	//std::cout << ":::::::::: GAMMA SHAPE PARM IS " << mModel.getGammaShape(0) << "::::::::::::::::::" << endl << endl;
	ProteinModel tempModel1(mModel);
	ProteinModel tempModel2(mModel);
	
	// Propose an update to the gamma shape parameter...
	if (whichModel != -1) {
		tempModel2.UpdateGammaShape(whichModel, proposalVariances[whichModel]);
		numUpdated++;
	} else {
		tempModel2.UpdateGammaShape3D(proposalVariances);
		numUpdated += mModel.GetNumModels();
	}

	// Do a parameter sweep for a fraction of rates...

	// We are going to do the parameter sweep over all models...
	if (tempModel2.getDependentRates() && (mModel.GetNumModels() > 1) ) {
		tempUpdate += ResampleParameterAssignments_RAR_dependentAssignment(mTree, tempModel2, theData, m_vecSiteAssignment, m_bUpdate, updateFraction, -1);
		tempModel2.UpdateAlphaMatrix3D(options.m_nMethod); // block re-assignment of rates update
	} else {
		for (int i=0; i<mModel.GetNumModels(); i++) {
			tempUpdate += ResampleParameterAssignments_RAR( mTree, tempModel2, theData, m_vecSiteAssignment, m_bUpdate, updateFraction, i );

			if (whichModel != -1) {
				tempModel2.UpdateAlphaMatrix3D(options.m_nMethod, i); // block re-assignment of rates update
			} else {
				tempModel2.UpdateAlphaMatrix3D(options.m_nMethod); // block re-assignment of rates update
			}
		}
	}
	tempModel2.ProteinModelParameterstoRates();
	
	// Evaluate the full proposal LR...
	logLR = CalculateGibbsSamplingLnLikelihoodRatio(mTree, tempModel1, tempModel2, m_vecSiteAssignment);

	m_bUpdate = false;
	
	if (logLR >= 0.0 ) {
		mModel = tempModel2;
		numUpdated += tempUpdate;
		//if need to output the sampled parameters
		m_bUpdate = true;
	} else {
		if (num->evaluateProposal_logL( logLR ) ) {
			mModel = tempModel2;
			numUpdated += tempUpdate;
			//if need to output the sampled parameters
			m_bUpdate = true;
		} else {
			m_bUpdate = false;
			numUpdated = 0;
		}
	}
	
	return numUpdated;
}

// --- B1 uni version here -----
int MCMC::RunMCMCGammaShapeWithSweepUni(
										double heat,
										Tree& mTree, 
										ProteinModel& mModel, 
										ObservedData& theData,
                                        IntVector& m_vecSiteAssignment,
										bool& m_bUpdate, double updateFraction, double *proposalVariances, int whichModel) {
	
	// TEMPORARILY OVERRIDE...
	return UpdateRARDistributionParameters_Uni_MC( heat, mTree, mModel, theData, m_vecSiteAssignment,  m_bUpdate, updateFraction, proposalVariances, whichModel); 
	double logLR;
//	double logLikelihood1, logLikelihood2;
	int numUpdated = 0;
	int tempUpdate = 0;
	if (!mModel.getRARdistributed()) { std::cout << "GammaShapeWithSweep update called when the model doesn't have gamma-stuff set!" << endl; exit(-1); }
	
	ProteinModel tempModel1(mModel);
	ProteinModel tempModel2(mModel);
	
	// Propose an update to the gamma shape parameter...
	if (whichModel != -1) {
		tempModel2.UpdateGammaShape(whichModel, proposalVariances[whichModel]);
		numUpdated++;
	} else {
		tempModel2.UpdateGammaShape3D(proposalVariances);
		numUpdated += mModel.GetNumModels();
	}
	
	// This may not be safe - should check how the enlargement is taking place
	tempModel2.UpdateAlphaMatrix3D(options.m_nMethod); // block re-assignment of rates update
	tempModel2.ProteinModelParameterstoRates();
	tempModel2.enlargeUniformization();
	
	// We are going to do the parameter sweep over all models...
	if (tempModel2.getDependentRates() && (mModel.GetNumModels() > 1) ) {
		tempModel2.UpdateAlphaMatrix3D(options.m_nMethod); // block re-assignment of rates update
		tempUpdate += ResampleParameterAssignments_RAR_dependentAssignment(mTree, tempModel2, theData, m_vecSiteAssignment, m_bUpdate, 1.0, -1);
	} else {
		for (int i=0; i<mModel.GetNumModels(); i++) {
			tempUpdate += ResampleParameterAssignments_RAR_B1u( mTree, tempModel2, theData, m_vecSiteAssignment, m_bUpdate, 1.0, i );
		}
	}
	
	tempModel2.UpdateAlphaMatrix3D(options.m_nMethod); // block re-assignment of rates update
	tempModel2.ProteinModelParameterstoRates();
	tempModel2.updateUniformization();
	
	// Evaluate the full proposal LR...
	logLR = CalculateGibbsSamplingLnLikelihoodRatio(mTree, tempModel1, tempModel2, m_vecSiteAssignment);
	
	m_bUpdate = false;
	
	if (logLR >= 0.0 ){
		mModel = tempModel2;
		mModel.setUniformization( tempModel2.returnUniformizationConstants() );
		
		numUpdated += tempUpdate;
		
		//if need to output the sampled parameters
		m_bUpdate = true;
	} else {
		if (num->evaluateProposal_logL( logLR ) ) {
			//			std::cout << "CHANGE ACCEPTED - OK" << endl;
			mModel = tempModel2;
			mModel.setUniformization( tempModel2.returnUniformizationConstants() );
			
			numUpdated += tempUpdate;

			m_bUpdate = true;
		} else {
			m_bUpdate = false;
			numUpdated = 0;
		}
	}
	
	mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
	return numUpdated;
}


// --- B1 uni version here -----
int MCMC::UpdateRARDistributionParameters_Uni_MC(
												 double heat,
												 Tree& mTree, 
												 ProteinModel& mModel, 
												 ObservedData& theData,
												 IntVector& m_vecSiteAssignment,
												 bool& m_bUpdate, double updateFraction, double *proposalVariances, int whichModel) {
	
	// ************************************************************************************************************************************************
	// Update the RAR parameters using a Metropolis-Hastings step (for RAR models with B1u partial sampling)
	// ************************************************************************************************************************************************
	
	// THIS VERSION UPDATES THE RATES PER CATEGORY WITH AN MC STEP...
	// std::cout << "	[Starting RAR distribution update...]" << endl;
	//return;
	
	//const int BOTH    = 0;
	const int RATE    = 1;
	//const int BRANCH  = 2;
	
	// This does a proposal affecting all models in the mixture.  Maybe not a good idea...
	//	return 0;
	
	double logLR;
	double logLikelihood1a, logLikelihood2;
	int numUpdated = 0;
	int tempUpdate = 0;
	if (!mModel.getRARdistributed()) { std::cout << "GammaShapeWithSweep update called when the model doesn't have gamma-stuff set!" << endl; exit(-1); }
	
	//	DoubleVector *mu = mModel.returnUniformizationConstants();
	ProteinModel tempModel1(mModel);
	ProteinModel tempModel2(mModel);
	
	std::vector<double> *theRates1, *theRates2;
	theRates1 = tempModel1.returnDiscreteRates();
	theRates2 = tempModel2.returnDiscreteRates();
	
	//std::cout << "MODEL 2: " << endl;
	if (whichModel == -1) { std::cout << "ERROR UPDATE CALLED FOR WHICHMODEL -1" << endl; exit(-1); }
	
	double startMu, newMu;
	if (options.capUniformizationRate) startMu =  mModel.getUniformization(whichModel);
	
	tempModel2.UpdateRateDistributionRAR_orderedDiscrete_MC(whichModel, 0.1); // last was 0.005 //proposalVariances[whichModel]);
	tempModel2.rewriteAlphaMatrix3D(); // write the updated rate matrix
	tempModel2.ProteinModelParameterstoRates(); // write the updated rate matrix scaled by freqs
	
	if (options.capUniformizationRate) {
		newMu = tempModel2.returnOptimalUniformization(whichModel);
		
		int rejectCount = 0;
		while (newMu > startMu) {
			for (int ii=0; ii<theRates2->size(); ii++) {
				theRates2->at(ii) = theRates1->at(ii);
			}
			
			tempModel2.UpdateRateDistributionRAR_orderedDiscrete_MC(whichModel, 1.0); //0.001); // last was 0.005 //proposalVariances[whichModel]);			
			tempModel2.rewriteAlphaMatrix3D(); // write the updated rate matrix
			tempModel2.ProteinModelParameterstoRates(); // write the updated rate matrix scaled by freqs
			
			newMu = tempModel2.returnOptimalUniformization(whichModel);
			rejectCount++;
			if (rejectCount > 10) { std::cout << "*" << endl; return 0; } // std::cout << "*" << endl;
		}		
	}
	
	if (whichModel != -1) {
		numUpdated++;
	} else {
		numUpdated += mModel.GetNumModels();
	}
	
	
	// Don't do this all the time!
	if (num->Random() <= 0.15) {
		// Something in here is failing to update uni ****
	
		if (tempModel2.getDependentRates() && (mModel.GetNumModels() > 1) ) {
			tempUpdate += ResampleParameterAssignments_RAR_dependentAssignment(mTree, tempModel2, theData, m_vecSiteAssignment, m_bUpdate, 1.0, -1);
		} else {
			for (int i=0; i<mModel.GetNumModels(); i++) {
				tempUpdate += ResampleParameterAssignments_RAR_B1u( mTree, tempModel2,  theData, m_vecSiteAssignment, m_bUpdate, 1.0, i );
			}
		}
		
		// Rebuild rate matricies based on parameter assignments; reset uniformization rate...
		tempModel2.rewriteAlphaMatrix3D();
		tempModel2.ProteinModelParameterstoRates();
	}
	
	int uniDirtyFlag = 0;
	if (!options.capUniformizationRate) {
		// Get the larger uniformization constant over tempModel1 and tempModel2...
		tempModel2.forceUpdateUniformization(whichModel);
		
		double theMax = 0.0;
		DoubleVector *mu1 = tempModel1.returnUniformizationConstants();
		DoubleVector *mu2 = tempModel2.returnUniformizationConstants();
		if ( (*mu1)[whichModel] > theMax) { theMax = (*mu1)[whichModel]; }
		if ( (*mu2)[whichModel] > theMax) { theMax = (*mu2)[whichModel]; }
		
		uniDirtyFlag = 0;
		if ( theMax > (*mu1)[whichModel] ) {
			// Set both models' uniformization constant to be the same...
			tempModel2.setUniformization(whichModel, theMax * 1.01);
			tempModel1.setUniformization(whichModel, theMax * 1.01);
			uniDirtyFlag = 1;
		}
	}
	
	int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
	mTree.root()->setCalculateFullProbabilityDistribution(0);
	
	// Model 1 likelihood... (only the parts that depend on RATE parameters)
	mTree.setprobmatrixalongbranch(tempModel1, tempModel1.GetFreqModels());
	if (uniDirtyFlag == 1) mTree.KEEP_CountFullChangesOverTree( tempModel1, theData, m_vecSiteAssignment );	
	
	logLikelihood1a = CalculateCompleteDataLogLikelihood(mTree, tempModel1, m_vecSiteAssignment, RATE);
	
	// Model 2 likelihood... (only the parts that depend on RATE parameters)
	mTree.setprobmatrixalongbranch(tempModel2, tempModel2.GetFreqModels());
	if (uniDirtyFlag == 1) mTree.KEEP_CountFullChangesOverTree( tempModel2, theData, m_vecSiteAssignment );	
	logLikelihood2 = CalculateCompleteDataLogLikelihood(mTree, tempModel2, m_vecSiteAssignment, RATE);
	
	mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
	
	// Got the likelihood ratio (assuming the larger of the two uniformization constants)...
	logLR = (logLikelihood2 - logLikelihood1a);
	
	m_bUpdate = false;
	if (logLR >= 0.0 )
	{
		mModel = tempModel2;
		numUpdated += tempUpdate;
		m_bUpdate = true;
	} else {
		
		if (num->evaluateProposal_logL( logLR ) ) {
			mModel = tempModel2;
			numUpdated += tempUpdate;
			m_bUpdate = true;
		} else {
			m_bUpdate = false;
			numUpdated = 0;
		}
	}
	
	if (m_bUpdate == true) {
		if (!options.capUniformizationRate) mModel.forceUpdateUniformization(whichModel);
		mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
	} else {
		// Need to reset the counts
		// [test] 	if (uniDirtyFlag == 1) mTree.KEEP_CountFullChangesOverTree( mModel, theData, m_vecSiteAssignment );	
	}
	
	return numUpdated;
}


#include <vector>

int MCMC::ResampleParameterAssignments_RAR(
										   Tree& mTree, 
										   ProteinModel& mModel, 
										   ObservedData& theData,
                                           IntVector& m_vecSiteAssignment,
										   bool& m_bUpdate, double updateFraction)
{
	return ResampleParameterAssignments_RAR( mTree, mModel, theData, m_vecSiteAssignment, m_bUpdate, updateFraction, 0 );
}

int MCMC::ResampleParameterAssignments_RAR(
										   Tree& mTree, 
										   ProteinModel& mModel, 
										   ObservedData& theData,
										   IntVector& m_vecSiteAssignment,
										   bool& m_bUpdate, double updateFraction, int whichModel) {
	
	if (options.m_nMethod == 6) { return ResampleParameterAssignments_RAR_B1u( mTree, mModel, theData, m_vecSiteAssignment, m_bUpdate, updateFraction, whichModel); }
	
	if ( (options.m_nMethod != 1) && (options.m_nMethod != 3)) { std::cout << "ERROR! ParameterSweep only implemented for B1-type methods at present" << endl; exit(-1); }
	
	// This function is for discrete-distributed rates only (for now)!
	// [updating this to use the updateFraction as an upper limit]
	
	ProteinModel tempModel1(mModel);
	ProteinModel tempModel2(mModel);
	
	std::vector<double> *theRates;
	std::vector<int> *rateAssignments;
	int theModel, i,j,k, startSite;
	int startingState;
	int newState;
	double theFraction;
	double logLikelihood1, logLikelihood2;
	
	int numStates = mModel.GetRateMatrix3D().cols();
	int numModels = mModel.GetRateMatrix3D().rows();
	int numRates = mModel.returnNumRates();
	
	double LRsum;
	double marginalConditional[(int)numRates];
	double rannum;
	
	theRates = mModel.returnDiscreteRates();
	if (updateFraction < 1.0) { theFraction = num->Random() * updateFraction; } else { theFraction = updateFraction; }
	
	int modelStart, modelEnd;
	if (whichModel == -1) {
		modelStart = 0; modelEnd = numModels;
	} else {
		modelStart = whichModel; modelEnd = modelStart+1;
	}
	
	int jj;
	int numUpdated = 0;
	
	for (theModel=modelStart; theModel < modelEnd; theModel++) {
		for (i=0; i<numStates; i++) {
			if (mModel.getReversible()) { startSite = i; } else { startSite = 0; }
			
			logLikelihood1 = CalculateCompleteDataLogLikelihood_Delta_B1(mTree, tempModel1, m_vecSiteAssignment, theModel, i);
			//logLikelihood1a = CalculateCompleteDataLogLikelihood(mTree, tempModel1, m_vecSiteAssignment, options.m_nMethod);
			///			std::cout << "model " << theModel << ", startSite " << i << "; initial conditional lnL = " << logLikelihood1 << endl;
			
			for (jj=startSite; jj<numStates; jj++) {
				j = int(num->Random() * (numStates-startSite)) + startSite;
				
				if ( (j != i) ) {
					if (theFraction != 1.0) {
						rannum = num->Random(); // [JdK] Numerics::getGlobalNumerics().Random();
						if (rannum > theFraction) continue;
					}
					numUpdated++;
					
					rateAssignments = tempModel1.returnRateAssignments();
					startingState = (*rateAssignments)[ ((theModel * numStates * numStates) + (i * numStates) + j) ];
					LRsum = 0.0;
					for (k=0;k<numRates; k++) {
						if (k != startingState) {
							tempModel2 = tempModel1;
							tempModel2.UpdateRateAssignmentB1(theModel,i,j,k);
							//					logLikelihood2a = CalculateCompleteDataLogLikelihood(mTree, tempModel2, m_vecSiteAssignment, options.m_nMethod);
							logLikelihood2 = CalculateCompleteDataLogLikelihood_Delta_B1(mTree, tempModel2, m_vecSiteAssignment, theModel, i);
							
							LRsum += (logLikelihood2 - logLikelihood1);
							marginalConditional[k] = logLikelihood2;
						} else {
							//			std::cout << "State transition " << i << " -> " << j << "; current state  " << k << ": LR for relevant part = " << 0 << " [" << logLikelihood1 << "]" << endl;
							marginalConditional[k] = logLikelihood1;
						}
					}
					newState = sampleFromCumulative(&marginalConditional[0], numRates);
					//	std::cout << i << "->" << j << ": newState chosen = " << newState << endl;
					
					// Now need to set everything...
					tempModel1.UpdateRateAssignmentB1(theModel,i,j,newState);
					logLikelihood1 = CalculateCompleteDataLogLikelihood_Delta_B1(mTree, tempModel1, m_vecSiteAssignment, theModel, i);
				}
			}
		}
	}
	//	std::cout << "FINISHED FULL PARAMETER SWEEP..." << endl;
	
	mModel = tempModel1;
	///	logLikelihood1 = CalculateCompleteDataLogLikelihood(mTree, mModel, m_vecSiteAssignment, options.m_nMethod);
	///	std::cout << "New log-likelihood is " << logLikelihood1 << endl;
	m_bUpdate = true;
	return numUpdated;
	
	
}

// --- B1 uniformized version ------
int MCMC::ResampleParameterAssignments_RAR_B1u(
											   Tree& mTree, 
											   ProteinModel& mModel, 
											   ObservedData& theData,
											   IntVector& m_vecSiteAssignment,
											   bool& m_bUpdate, double updateFraction, int whichModel) {
	// ************************************************************************************************************************************************
	// Resample rate parameter assignments to state exchanges under RAR models and B1u partial sampling
	// ************************************************************************************************************************************************
	if ( (options.m_nMethod != 6)) { std::cout << "ERROR! ParameterSweep only implemented for B1-type UNIFORMIZED methods at present" << endl; exit(-1); }
	
	// This function is for discrete-distributed rates only (for now)!
	// [updating this to use the updateFraction as an upper limit]
	
	// Apr 2010: re-writing to never exceed the current uniformization constant
	
	std::vector<double> *theRates;
	std::vector<int> *rateAssignments;
	
	int theModel, i,j,k, startSite;
	int startingState;
	int newState;
	double theFraction;
	double logLikelihood2;
	
	int numStates = mModel.GetRateMatrix3D().cols();
	int numModels = mModel.GetRateMatrix3D().rows();
	int numRates = mModel.returnNumRates();
	int allStatesInvalid = 0;
	double marginalConditional[(int)numRates];
	double rannum;
	
	theRates = mModel.returnDiscreteRates();
	
	if (updateFraction < 1.0) { theFraction = num->Random() * updateFraction; } else { theFraction = updateFraction; }
	
	int modelStart, modelEnd;
	if (whichModel == -1) {
		modelStart = 0; modelEnd = numModels;
	} else {
		modelStart = whichModel; modelEnd = modelStart+1;
	}
	
	int jj;
	int numUpdated = 0;
	
	int accept, reject;
	accept = reject = 0;
	
	for (theModel=modelStart; theModel < modelEnd; theModel++) {
		for (int ii=0; ii<numStates*1.1; ii++) {
			if (mModel.getReversible()) { startSite = i; } else { startSite = 0; }
			
			// Randomly select row, i
			i = int(num->Random() * numStates + startSite);
			
			// Update more than every state transition
			for (jj=startSite; jj<(int)(numStates*1.4); jj++) {
				// Randomly select a j
				j = int(num->Random() * (numStates-startSite)) + startSite;
				
				if ( (j != i) ) {
					if (theFraction != 1.0) {
						rannum = num->Random(); // [JdK] Numerics::getGlobalNumerics().Random();
						if (rannum > theFraction) continue;
					}
					numUpdated++;
					
					// WE COULD ADD SOMETHING USEFUL HERE -- IF THERE ARE ZERO COUNTS THEN RANDOMLY SELECT A RATE ASSIGNMENT!  DONT HAVE TO DO THE MARGINALS
					rateAssignments = mModel.returnRateAssignments();
					
					startingState = (*rateAssignments)[ ((theModel * numStates * numStates) + (i * numStates) + j) ];
					DoubleVector *mu = mModel.returnUniformizationConstants();
					
					if (  -mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[i].mDataArray[i] > (*mu).mDataArray[theModel]) { 
						std::cout << "ERROR!  Before resampling, LambdaX for state " << i << " is " << (-mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[i].mDataArray[i]) << ", while mu is " << (*mu).mDataArray[theModel] << endl;
					}
					
					double remainderOfLambdaX = -mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[i].mDataArray[i] - (*theRates)[ (theModel * mModel.returnNumRates() ) + startingState ];
					
					allStatesInvalid = 1;
					// Evaluate change in likelihood by reassigning state transition (i,j) to all discrete rate parameters, k
					for (k=0;k<numRates; k++) {
						if ( ( remainderOfLambdaX + (*theRates)[ (theModel * mModel.returnNumRates() ) + k ] ) > (*mu).mDataArray[theModel] ) {
							// Skip this state
							marginalConditional[k] = -std::numeric_limits<double>::max( );
							reject++;
						//	cout << "	Rejecting state " << k << "; lambda remainder is " << remainderOfLambdaX << " while rate is " << (*theRates)[k] << endl;
						} else {
							allStatesInvalid = 0;
							
							mModel.UpdateRateAssignmentB1(theModel,i,j,k);
							logLikelihood2 = CalculateCompleteDataLogLikelihood_Delta_B1u(mTree, mModel, theData, m_vecSiteAssignment,  theModel, i, j);
							
							marginalConditional[k] = logLikelihood2;
							//cout << "Exchange " << i << ", " << j << "; assingment to " << k << " gives " << marginalConditional[k] << endl;
							accept++;
						}
					}
					//exit(-1);
					
					// Sometimes the choices at one row state assignment can invalidate all choices at the current state...
					if ( allStatesInvalid == 1) {
						jj -= 1;
						std::cout << "	[WARNING: all states for exchange " << i << ", " << j << " are now invalid.  Trying elsewhere in the row.]" << endl;
						mModel.UpdateRateAssignmentB1(theModel,i,j,startingState);
						continue;
					}
					
					newState = sampleFromCumulative(&marginalConditional[0], numRates);
					while ( marginalConditional[ newState ] == -std::numeric_limits<double>::max( ) ) {
						std::cout << "	[ Rare error: we chose a restricted state!  Lazy programmer! ]" << endl;
						newState = sampleFromCumulative(&marginalConditional[0], numRates);
					}						
					mModel.UpdateRateAssignmentB1(theModel,i,j,newState);
				}
			}
		}
	}
	
	m_bUpdate = true;
//	if ( ( accept/(accept+reject) ) <= 0.5 ) {
//		std::cout << "		[ Possible problem in RAR resampler; ratio of in-bounds and out-bounds proposals (due to uniformization constraint): " << ( accept/(accept+reject) ) << " ]" << endl;
//	}
	return numUpdated;
}


// GTR version:
int MCMC::UpdateGTRrates_unidimensional_b1u(
											Tree& mTree, 
											GTRModel& mModel, 
											ObservedData& theData,
											IntVector& m_vecSiteAssignment,
											double updateFraction, int whichModel) {

	// ********************************************************************************************************************************
	// This version was used for the B1u paper; de Koning et al (2009) MBE
	// ********************************************************************************************************************************
	// NOTE!  THIS ONLY WORKS WHEN USING UNNORMALIZED RATE MATRICES
	
    if ( (options.m_nMethod != 6)) { std::cout << "ERROR! ParameterSweep only implemented for B1-type UNIFORMIZED methods at present" << endl; exit(-1); }
	DoubleVector *mu = mModel.returnUniformizationConstants();
	
	double logger1, logger2, deltaL;
	DoubleMatrix3D    *realChangeCount3D = mTree.root()->getChangeMatrix3D( mTree.root() ); //&GetData().countModelchange3D;
//	DoubleMatrix3D *ChangeCount3D     = mTree.root()->getChangeMatrix3Duni( mTree.root() ); //&GetData().countModelchange3D;
	DoubleMatrix    *augmentedCount      = mTree.root()->getAugmentedCountMatrix( mTree.root() );
	
	double oldRate, oldRate_1, proposal, proposal_1, oldRate_2, proposal_2, changeamount, var;
	//int numUpdated = 0;
	
	// This function is for GTR models only...
	int numStates  = mModel.GetRateMatrix3D().cols();
	int numModels  = mModel.GetRateMatrix3D().rows();
	DoubleMatrix* stateF = &mModel.returnStateFrequencies();
	
	// Pick a single rate to update...
	int whichRateFrom;
	int whichRateTo;
	
	updateFraction = 1.5;
	
	int updateLimit = (updateFraction * (numStates * (numStates-1) ) );
	int rejectFlag = 0;
	int rejectCount = 0;
	
	for (int numUpdates = 0; numUpdates < updateLimit; numUpdates++) {
		
		// Put a limit on number of attempts we make to sample when hitting the mu boundary under capuniformization...
		if (rejectCount > 10) {
			rejectFlag = 0; rejectCount = 0;
			updateLimit++;
			continue;
		}
		
		if (rejectFlag == 0) {
			// Don't resample the states if we just rejected a move due to capuniformization...
			whichRateFrom  = (int)(num->Random() * (numStates));
			whichRateTo = (int)(num->Random() * (numStates));
			
			while (whichRateTo == whichRateFrom) {
				whichRateTo = (int)(num->Random() * (numStates));
			}
		}
		
		rejectFlag = 0;
		// Got a rate to do...
		int modelStart, modelEnd;
		if (whichModel == -1) {
			modelStart = 0; modelEnd = numModels;
		} else {
			modelStart = whichModel; modelEnd = modelStart+1;
		}
		
		for (int theModel=modelStart; theModel < modelEnd; theModel++) {
			double startMu = mu->mDataArray[theModel];
			oldRate = mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] / stateF->mRows.mDataArray[theModel].mDataArray[whichRateTo]; //mModel.
			
			if ( Numerics::getGlobalNumerics().Random() < 0.05 ) {
				changeamount = ( Numerics::getGlobalNumerics().Random() - 0.5) * 0.001 * oldRate; //trying to scale change amount by parm size to equalize mixing...
				proposal = oldRate + changeamount;
				
				if (proposal <= 0.0 )
					proposal = oldRate - changeamount;
			} else {
				var = 0.15;
				
				if (oldRate <= 0.68) {
					var = 0.028;
				} else if (oldRate <= 0.35) { 
					var = 0.009;
				} else if (oldRate <= 0.11) {
					var = 0.012; 
				} else if (oldRate <= 0.06) {
					var = 0.003;  
				}
				var *= 0.5;
				
				changeamount = num->box_muller( 0.0, var ); // (rannum - 0.5) * mWindowsize;
				proposal = oldRate + changeamount;
				
				while ( proposal <= 0.0 ) {
					changeamount *= -1;
					proposal = oldRate + changeamount;
				}
			}
			proposal_1 = proposal * stateF->mRows.mDataArray[theModel].mDataArray[whichRateTo];
			proposal_2 = proposal * stateF->mRows.mDataArray[theModel].mDataArray[whichRateFrom];
			oldRate_1  = oldRate * stateF->mRows.mDataArray[theModel].mDataArray[whichRateTo];
			oldRate_2  = oldRate * stateF->mRows.mDataArray[theModel].mDataArray[whichRateFrom];
			
			// **** Standard B1 likelihood evaluation *****
			// logLikelihood1  = CalculateCompleteDataLogLikelihood_Delta_B1(mTree, tempModel1, m_vecSiteAssignment, options.m_nMethod, theModel, i);
			
			// ***************************************************************************************************
			// Need to see if mu requires updating...
			// ***************************************************************************************************
			double newLambda;
			// Row 1...
			newLambda= -mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom] - oldRate_1 + proposal_1;
			newLambda /= 1.04;
			if (newLambda > mu->mDataArray[theModel]) {
				if (!options.capUniformizationRate) {
					mu->mDataArray[theModel] = newLambda;
				} else {
					// proposal causes newLambda to exceed our fixed mu...
					rejectFlag = 1; rejectCount++; updateLimit++;
					std::cout << "* ";
					continue;
				}
			}
			
			// Row 2...
			newLambda = -mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateTo] - oldRate_2 + proposal_2;
			newLambda /= 1.04;
			if (newLambda > mu->mDataArray[theModel]) {
				if (!options.capUniformizationRate) {
					mu->mDataArray[theModel] = newLambda;
				} else {
					// proposal causes newLambda to exceed our fixed mu...
					rejectFlag = 1; rejectCount++; updateLimit++;
					std::cout << "* ";
					continue;
				}
			}
			
			// Now do change in likelihood...
			deltaL = (*realChangeCount3D).mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] * num->logarithm( (oldRate_1 / proposal_1) );
			deltaL += (*realChangeCount3D).mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateFrom] * num->logarithm( (oldRate_2 / proposal_2) );
			
			// Now do the virtual substitutions (may be different across rates classes)...
			int rateClassLimit = options.numCatASRV;
			if ( options.allowASRV != 1 ) rateClassLimit = 1;
			for (int rateclass= 0; rateclass < rateClassLimit; rateclass++ ) {
				logger1 = mu->mDataArray[theModel] + (mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom] * mTree.root()->GetData().siteCats->getRate( rateclass ) );
				logger2 = mu->mDataArray[theModel] + (mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom] + oldRate_1 - proposal_1)* mTree.root()->GetData().siteCats->getRate( rateclass );
				if (logger1 <= 0) logger1  = 1e-70;
				if (logger2 <= 0) logger2  = 1e-70;
				
				deltaL += (*augmentedCount).mRows.mDataArray[ num->returnModelIndex( theModel, rateclass, 0) ].mDataArray[whichRateFrom] * ( num->logarithm( ( logger1 )) - num->logarithm( logger2 ) );
			}
			
			// Now do the virtual substitutions (may be different across rates classes)...
			rateClassLimit = options.numCatASRV;
			if ( options.allowASRV != 1 ) rateClassLimit = 1;
			for (int rateclass= 0; rateclass < rateClassLimit; rateclass++ ) {
				logger1 = mu->mDataArray[theModel] + (mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateTo] * mTree.root()->GetData().siteCats->getRate( rateclass) );
				logger2 = mu->mDataArray[theModel] + (mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateTo] + oldRate_2 - proposal_2) * mTree.root()->GetData().siteCats->getRate( rateclass);
				if (logger1 <= 0) logger1 = 1e-70;
				if (logger2 <= 0) logger2 = 1e-70;
				
				deltaL += (*augmentedCount).mRows.mDataArray[ num->returnModelIndex( theModel, rateclass, 0) ].mDataArray[whichRateTo] * ( num->logarithm( ( logger1 )) - num->logarithm( logger2 ) );
			}

			deltaL *= -1.0;
			
			if ( num->evaluateProposal_logL( 1.0, (1.0 - deltaL) ) == 1 ) {
				// ACCEPT
				mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] = proposal;
				mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateFrom] = proposal;
				
				mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] = proposal_1;
				mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateFrom] = proposal_2;
				mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom] += oldRate_1 - proposal_1;
				mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateTo]     += oldRate_2 - proposal_2;
				
				// To be consistent with the forceUpdateUniformization() in the multidimensional proposals...
				if ( mu->mDataArray[theModel] != startMu ) mu->mDataArray[theModel] *= 1.01;
				
			} else {
				mu->mDataArray[theModel] = startMu;
			}
		}
	}
	return 1;
}

// GTR version:
int MCMC::UpdateGTRrates_unidimensional(
										Tree& mTree, 
										GTRModel& mModel, 
										ObservedData& theData,
										IntVector& m_vecSiteAssignment,
										double updateFraction, int whichModel) {

	// ********************************************************************************************************************************
	// This version was used for the B1u paper; de Koning et al (2009) MBE
	// ********************************************************************************************************************************

//	double likelihood1, likelihood2;
	
	double oldRate, oldRate_1, proposal, proposal_1, oldRate_2, proposal_2, changeamount, var;
	
	// This function is for GTR models only...
	
	//	double logLikelihood1, logLikelihood2;
	int numStates  = mModel.GetRateMatrix3D().cols();
	int numModels  = mModel.GetRateMatrix3D().rows();
	DoubleMatrix* stateF = &mModel.returnStateFrequencies();
	
	// Pick a single rate to update...
	int whichRateFrom;
	int whichRateTo;

	updateFraction = 1.5; 
	
	int updateLimit = (updateFraction * (numStates * (numStates-1) ) );
//	int rejectFlag = 0;
//	int rejectCount = 0;
	
	for (int numUpdates = 0; numUpdates < updateLimit; numUpdates++) {
		whichRateFrom  = (int)(num->Random() * (numStates));
		whichRateTo = (int)(num->Random() * (numStates));
		
		while (whichRateTo == whichRateFrom) {
			whichRateTo = (int)(num->Random() * (numStates));
		}
		
		// Got a rate to do...
		int modelStart, modelEnd;
		if (whichModel == -1) {
			modelStart = 0; modelEnd = numModels;
		} else {
			modelStart = whichModel; modelEnd = modelStart+1;
		}
		
		for (int theModel=modelStart; theModel < modelEnd; theModel++) {
			//likelihood1 = CalculateCompleteDataLogLikelihood_Delta_B1(mTree, mModel,m_vecSiteAssignment, theModel, whichRateFrom);
			double full1 = CalculateCompleteDataLogLikelihood(mTree, mModel, m_vecSiteAssignment, theModel);
			
			oldRate = mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] / stateF->mRows.mDataArray[theModel].mDataArray[whichRateTo]; //mModel.
			
			if ( Numerics::getGlobalNumerics().Random() < 0.05 ) {
				changeamount = ( Numerics::getGlobalNumerics().Random() - 0.5) * 0.001 * oldRate; //trying to scale change amount by parm size to equalize mixing...
				proposal = oldRate + changeamount;
				
				if (proposal <= 0.0 )
					proposal = oldRate - changeamount;
			} else {
				var = 0.15;
				
				// Approximate posterior variances determined from output
				// (for analysis of 224 taxon cyt-b data set in de Koning et al, 2009)
				if (oldRate <= 0.68) { 
					var = 0.028;  
				} else if (oldRate <= 0.35) { 
					var = 0.009;
				} else if (oldRate <= 0.11) { 
					var = 0.012; 
				} else if (oldRate <= 0.06) {
					var = 0.003; 
				}
				var *= 0.5;
				
				changeamount = num->box_muller( 0.0, var ); // (rannum - 0.5) * mWindowsize;
				proposal = oldRate + changeamount;
				
				while ( proposal <= 0.0 ) {
					changeamount *= -1;
					proposal = oldRate + changeamount;
				}
			}
			proposal_1 = proposal * stateF->mRows.mDataArray[theModel].mDataArray[whichRateTo];
			proposal_2 = proposal * stateF->mRows.mDataArray[theModel].mDataArray[whichRateFrom];
			oldRate_1  = oldRate * stateF->mRows.mDataArray[theModel].mDataArray[whichRateTo];
			oldRate_2  = oldRate * stateF->mRows.mDataArray[theModel].mDataArray[whichRateFrom];
			
			// Set all the parms for likelihood evaluation:
			mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] = proposal;
			mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateFrom] = proposal;
			
			mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] = proposal_1;
			mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateFrom] = proposal_2;
			mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom] += oldRate_1 - proposal_1;
			mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateTo]     += oldRate_2 - proposal_2;
			
			// **** Standard B1 likelihood evaluation *****
			double full2 = CalculateCompleteDataLogLikelihood(mTree, mModel, m_vecSiteAssignment, theModel);
			
			if ( num->evaluateProposal_logL( full2, full1 ) == 1 ) {
				// ACCEPT
			} else {
				// REJECT
				mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] = oldRate;
				mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateFrom] = oldRate;
				
				mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] = oldRate_1;
				mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateFrom] = oldRate_2;
				mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom] += -oldRate_1 + proposal_1;
				mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateTo]     += -oldRate_2 + proposal_2;
			}
		}
	}
	return 1;
}

// ********************************************************************************************************************************

int MCMC::ResampleParameterAssignments_RAR_dependentAssignment(
															   Tree& mTree, 
															   ProteinModel& mModel, 
															   ObservedData& theData,
															   IntVector& m_vecSiteAssignment,
															   bool& m_bUpdate, double updateFraction, int whichModel) {

	// *** WARNING *** This is the old version.  A few function calls need to be checked/optimized for B1u
		
	// Need to adapt to B1u and put in likleihood heating via options.__globalLikleihoodHeat
		std::cout << "NOT ADAPTED YET!" << endl; return 0;


	static ProteinModel tempModel1;
	static ProteinModel tempModel2;
	
	tempModel1 = mModel;
	tempModel2 = mModel;
	int numUpdated = 0;
	
//	double logLR;
	if (whichModel != -1) { std::cout << "ERROR!  _dependentassignment only works for whichModel == -1 for now" << endl; exit(-1); }	
	if ( (options.m_nMethod != 1) && (options.m_nMethod != 3)) { std::cout << "ERROR! ParameterSweep only implemented for B1 at present" << endl; exit(-1); }
	
	int theModel, i,j,k, startSite;
	int startingState;
	std::vector<double> *theRates;
	std::vector<int> *rateAssignments;
//	double windowsize = mModel.GetUpdateWindowsize();
	double logLikelihood1, logLikelihood2;
	
	int newState;
	int numStates = mModel.GetRateMatrix3D().cols();
	int numModels = mModel.GetRateMatrix3D().rows();
	
	int numRates = mModel.returnNumRates();
	
	theRates = mModel.returnDiscreteRates();
	
	double theFraction;
	
	if (updateFraction < 1.0) { theFraction = num->Random() * updateFraction; } else { theFraction = updateFraction; }

	double LRsum;
	int theDim = (int)numRates;
	
	double rannum;
	
	int modelStart, modelEnd;
	if (whichModel == -1) {
		modelStart = 0; modelEnd = numModels; 
	} else {
		modelStart = whichModel; modelEnd = modelStart+1;
	}
	
	double marginalConditional[(int)numRates*numModels];
	double marginalConditional2[(int)numRates];
	double marginalConditional3[(int)numRates];
	double marginalConditional4[(int)numRates];
	
	if (whichModel == -1) { theDim = numRates; } else { theDim = 1; }
	
	//int numUpdates = 0;
	// Iterate over all relevant i->j's (taking care to not be redundant on symmetric models)
	for (i=0; i<numStates; i++) {
		if (mModel.getReversible()) { startSite = i; } else { startSite = 0; }
		for (j=startSite; j<numStates; j++) {
			// Don't do all of them, if theFraction is < 1
			if (theFraction != 1.0) {
				rannum = num->Random(); 
				if (rannum > theFraction) continue;
			}
			//numUpdates++;
			
			// Now iterate over all models and calculate the deltaLnL...
			for (theModel=modelStart; theModel < modelEnd; theModel++) {
				//if ( (j==0) ) {
				numUpdated++;
				
				logLikelihood1  = CalculateCompleteDataLogLikelihood_Delta_B1(mTree, tempModel1, m_vecSiteAssignment, theModel, i);

				if ( j != i ) {
					// Current state (just evaluated and stored as logLikelihood1)
					rateAssignments = tempModel2.returnRateAssignments();
					startingState = (*rateAssignments)[ ((theModel * numStates * numStates) + (i * numStates) + j) ];

					// Now do the rest of them...
					LRsum = 0.0;
					for (k=0;k<numRates; k++) {
						if (k != startingState) {
							tempModel2 = tempModel1;
							
							tempModel2.UpdateRateAssignmentB1(theModel,i,j,k);
							// [full eval] logLikelihood2a = CalculateCompleteDataLogLikelihood(mTree, tempModel2, m_vecSiteAssignment,);
							logLikelihood2 = CalculateCompleteDataLogLikelihood_Delta_B1(mTree, tempModel2, m_vecSiteAssignment, theModel, i);
							marginalConditional[(theModel-modelStart)*theDim + k] = logLikelihood2;
						} else {
							marginalConditional[(theModel-modelStart)*theDim + k] = logLikelihood1;
						}
					}
				} // j!=i
			} // for models..
			
			// These are fixed things we need whether we have dependencies or not.  Can reuse!
			for (theModel=0; theModel < numModels; theModel++) {
				// Now for each model, go through and collect the dependent rates stuff for #0, then: 1) assign; 2) sample...
				// [DO #0, and shared assignments LAST]
				if ( j != i ) {
					if ( ( theModel == 0 ) && ( mModel.getDependentRates() ) ) {
						// If we have dependent rates, go through all the non-zero models to collect the contribution of 'non-zero' mixture components to this state transitions marginal rate distribution...
						for (k=0; k<numRates; k++) {
							marginalConditional2[k] = marginalConditional[k];
						}
						
						for (int j1=1; j1 < numModels; j1++) {
							if (tempModel1.dependencyIndicators[j1][i][j] == 1) {
								for (k=0; k<numRates; k++) {
									marginalConditional2[k] += marginalConditional[(j1-modelStart)*theDim + k];
								}
							} // if dependency
						} // for models
					} // if model==0 and dependent rates
					
					// NOW, set marginalConditional3 for the current theModel...
					double maxL = -std::numeric_limits<double>::max( );
					for (k=0;k<numRates; k++) {
						marginalConditional3[k] = marginalConditional[(theModel-modelStart)*theDim + k];
						if (marginalConditional3[k] > maxL) { maxL = marginalConditional3[k]; }
					}
					
					// Get LRs...
					LRsum=0.0;
					double tempVal;
					for (k=0;k<numRates; k++) {
						tempVal = exp((marginalConditional3[k] - maxL));
						marginalConditional3[k] = tempVal;
						LRsum += marginalConditional3[k];
					}
					
					// Finish transformation...
					for (k=0;k<numRates; k++) {
						marginalConditional3[k] /= LRsum;
					}
					
					// MarginalConditional3 now contains the posterior prob's for rate assignments at model 'theModel' (assuming no dependency)
					
					//  Now set marginalConditional4 based on all the OTHER dependent rates (excluding me, if i'm dependent)...
					if ( (theModel > 0) && (tempModel1.dependencyIndicators[theModel][i][j] == 1) ) {
						for (k=0; k<numRates; k++) {
							marginalConditional4[k] = marginalConditional2[k] - marginalConditional[(theModel-modelStart)*theDim + k];
						}
					} // if dependency
					else if (theModel >0)  {
						for (k=0; k<numRates; k++) {
							marginalConditional4[k] = marginalConditional2[k]; // - marginalConditional[(theModel-modelStart)*theDim + k];
						}
					}
					
					if (theModel > 0) {
						// Transform marginalconditional4 to probabilities...
						tempVal = 0.0;
						maxL = -std::numeric_limits<double>::max( );
						for (k=0;k<numRates; k++) {
							if (marginalConditional4[k] > maxL) { maxL = marginalConditional4[k]; }
						}
						
						// Get LRs...
						LRsum = 0.0;
						for (k=0;k<numRates; k++) {
							tempVal = exp((marginalConditional4[k] - maxL));
							marginalConditional4[k] = tempVal;
							LRsum += marginalConditional4[k];
						}
						
						// Finish transformation...
						for (k=0;k<numRates; k++) {
							marginalConditional4[k] /= LRsum;
						}
						
						// MarginalConditional4 now contains the posterior prob's for rate assignments across all the other dependent categories
					}
					
					double pSame, pDiff, probDiff;
					pSame = pDiff = 0.0;
					probDiff = 0.0;

					double differencePrior = mModel.getIndependentFractionPrior();
					
					// Calculate the probability of having the same rate assignment as model #0...
					if (theModel > 0) {
						for (int i1=0; i1<numRates; i1++) {
							for (int j1=0; j1<numRates; j1++) {
								if (j1==i1) { pSame += ( marginalConditional4[i1] ) * (marginalConditional3[j1]); }
								else { pDiff += ( marginalConditional4[i1] * marginalConditional3[j1] ) ;  }
							}
						}
						
						probDiff = ( pDiff * differencePrior ) / ( (pDiff * differencePrior) + (pSame * (1.0-differencePrior) ) );
						//		std::cout << "model " << theModel << ", for " << i << " -> " << j << ", the probability of being different is " << probDiff << "; pdiff is " << pDiff << ", pSame is " << pSame << endl;
						
						double u = num->Random();
						if (u <= probDiff) {
							tempModel1.dependencyIndicators[theModel][i][j] = 0;
							
							double rannum = num->Random();; // [JdK] Numerics::getGlobalNumerics().Random();
							
							newState = 0;
							//double thebest = 0.0; int best = 0;
							
							for (int ii=0; ii<numRates;ii++) {
								if (ii>0) { marginalConditional3[ii] += marginalConditional[ii-1]; }
							}
							
							for (int ii=0; ii<numRates;ii++) {
								if (rannum <= marginalConditional3[ii]) {newState=ii; break;}
							}
							
							// Update stuff...
							tempModel1.UpdateRateAssignmentB1(theModel,i,j,newState);
							//[Don't need to do this?] logLikelihood1 = CalculateCompleteDataLogLikelihood_Delta_B1(mTree, tempModel1, m_vecSiteAssignment, theModel, i);
							
						} else {
							tempModel1.dependencyIndicators[theModel][i][j] = 1;
						}
					}
				} // j!=i
			} // for themodel
			
			// NOW sample the dependent guys...
			int numDependent = 0;
			if ( j != i ) {
				for (int j1=1; j1 < numModels; j1++) {
					if (tempModel1.dependencyIndicators[j1][i][j] == 1) {
						numDependent++;
					}
				}
				
				// special case: model==0 and NO dependent rates...
				if (numDependent == 0) {
					newState = sampleFromCumulative(&marginalConditional[0],numRates);
					tempModel1.UpdateRateAssignmentB1(0,i,j,newState);
					//tempModel1.ProteinModelParameterstoRates();
				} else {
					for (k=0; k<numRates; k++) {
						marginalConditional2[k] = marginalConditional[k];
					}
					
					for (int j1=1; j1 < numModels; j1++) {
						if (tempModel1.dependencyIndicators[j1][i][j] == 1) {
							for (k=0; k<numRates; k++) {
								marginalConditional2[k] += marginalConditional[(j1-modelStart)*theDim + k];
							}
						} // if dependency
					} // for models
					
					// NOTE: samplefromcumulative now assumes LOG liks
					newState = sampleFromCumulative(&marginalConditional2[0],numRates);
					tempModel1.UpdateRateAssignmentB1(0,i,j,newState);
					
					for (int j1=1; j1 < numModels; j1++) {
						if (tempModel1.dependencyIndicators[j1][i][j] == 1) {
							tempModel1.UpdateRateAssignmentB1(j1,i,j,newState);
						} // if dependency
					} // for models
					//					tempModel1.ProteinModelParameterstoRates();
					
					
				}
			} // j!=i
			
			// STEP 2.  IF (THEMODEL > 0) NOW NEED TO DRAW u TO DETERMINE IF THIS MODEL COMPONENT WILL BE SHARED OR NOT.  SET INDICATOR APPROPRIATELY
			
			// STEP 2A.  IF SHARED, RECALCULATE THE POSTERIOR DISTRIBUTION BY SUMMING THE APPROPRIATE LOG-L VALUES
			
			// STEP 3.  IF (INDICATOR SAYS DEPENDENT || MODELCOMPONENT == 0) SAMPLE FROM POSTERIOR AND SET NEW STATE
		} // j
	} // i

	//	std::cout << "FINISHED FULL PARAMETER SWEEP..." << endl;
	
	mModel = tempModel1;
	m_bUpdate = true;

	return numUpdated;
}


//---------------------------------------------------------------------------------------------------------
int MCMC::setAllAncestralAndTransientStates_local(
												  Tree& mTree,
												  int& rannode,
                                                  MixtureModel& theModel,
												  ObservedData& theData,
												  IntVector& m_vecSiteAssignment,
                                                  DoubleMatrix& stateFreqMatrix) {
    
    // Adding stateFreqMatrix here because sometimes we want to call with different frequencies at the root of the tree...
//	return mTree.ResampleAncestralAndTransientStates_local(rannode, theModel, theData, m_vecSiteAssignment, stateFreqMatrix );

	return mTree.ResampleAncestralAndTransientStates_local( rannode, theModel, theData, m_vecSiteAssignment, stateFreqMatrix );

}


// ------------------------------------------------------------------------------------------------------
// New utility functions added by Jason to help with code clean-up and reorg / Dec 2009
// ------------------------------------------------------------------------------------------------------
void MCMC::setAllProbMatricesOnTree( Tree& mTree, MixtureModel& mModel ) {
	// Force a recalculation of all branches' probability matrices using whatever method is set
	// (Note: this will force their calculation even for B1u, which usually will not require it)
	
	int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
	
	mTree.root()->setCalculateFullProbabilityDistribution(1);
	mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
	mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
}

/*
void MCMC::_setDataPackPointers( DataPack& theDataPack,
								IntVector& m_vecSiteAssignment,
								Tree& theTree,
                                ObservedData& theData) {
	
	theDataPack.m_vecSiteAssignment = m_vecSiteAssignment;
	theDataPack.theTree             = theTree;
	theDataPack.theData             = theData;
}

 */

int MCMC::setAllAncestralAndTransientStates_full( Tree& mTree, MixtureModel& mModel, ObservedData& theData, IntVector& m_vecSiteAssignment) {
	// Do the full ancestral/transient state updater...
	
	int numUpdated = 0;
	// Not sure if the forceUpdateStationaryFreqs() call is always required...
	mModel.forceUpdateStationaryFreqs();
	
	if (  (mModel.getPredefinedMatrix() > -1) && mModel.getReversible() ) {
		numUpdated += mTree.AssignAncestralStatesByMarginalProbability( mModel, theData, m_vecSiteAssignment, mModel.returnStateFrequencies(), 1);
	} else {
		numUpdated += mTree.AssignAncestralStatesByMarginalProbability( mModel, theData, m_vecSiteAssignment, mModel.returnStateFrequencies(), 1);
	} 
	
	return numUpdated;
}


void MCMC::countChangesAlongTree( Tree& mTree, MixtureModel& mModel, ObservedData& theData, IntVector& m_vecSiteAssignment) {
	mTree.KEEP_CountFullChangesOverTree( mModel, theData, m_vecSiteAssignment );
}

//---------------------------------------------------------------------------------------------------------
void MCMC::RunMarkovChain(
						  Tree& mTree,
						  GTRModel& mModel,
						  IntVector& m_vecSiteAssignment,
						  ObservedData& theData,
						  bool& m_bModelUpdate,
						  bool& m_bStateUpdate,
						  bool& m_bAssignment) {
	
    // MESSY UPDATER FUNCTION FOR GTR ANALYSIS; approximately what was used in de Koning et al (2009)
    
	static int modelCount = 0;
	static int assignCount = 0;
	static int shapeCount = 0;
	static int ancstateCount = 0;
	static int blCount = 0;
	
	static int updateCount = 0;
	double rannum1 = num->Random();
	int numModels = mModel.GetNumModels();
    
	if (updateCount == 0) {
		// std::cout << "Performing initial marginal reconstruction (sample)..." << endl;
		ancstateCount++;
		
		if (options.m_nMethod == 6) {
			mModel.updateUniformization();
			int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
			mTree.root()->setCalculateFullProbabilityDistribution(1);
			mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
			mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
		} else {
			mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
		}
		
		mTree.initSiteLikelihoods( m_vecSiteAssignment, theData.getNumStates() );
		mTree.AssignAncestralStatesByMarginalProbability( mModel, theData, m_vecSiteAssignment, mModel.returnStateFrequencies(), 1 );
		
		// Now reset uniformization constant to something bigger...
		if (options.m_nMethod == 6) mModel.enlargeUniformization();
		
        if (options.m_nMethod == 6) {
            mModel.setUniformization(0, 1.0);
        }
        
		if ( options.allowASRV && (options.numCatASRV > 1) ) {
			if (options.m_nMethod != 6) {
				mTree.KEEP_CountFullChangesOverTree( mModel, theData, m_vecSiteAssignment );
			} else {
				mTree.KEEP_CountFullChangesOverTree( mModel, theData, m_vecSiteAssignment );	
			}
			
			mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
			ResampleSiteAssignments_ASRV_rates(mTree, mModel, theData, m_vecSiteAssignment);	
		}
		
		if (options.m_nMethod != 6) {
			mTree.KEEP_CountFullChangesOverTree( mModel, theData, m_vecSiteAssignment );
		} else {
			mTree.KEEP_CountFullChangesOverTree( mModel, theData, m_vecSiteAssignment );	
		}
		
		
	}
	else if ( ((updateCount % options.outputIncompleteLogL) == 0) && (options.outputIncompleteLogL > 0) ) {
		ancstateCount++;
        
//        cout << "UNI CONSTANT IS: " << mModel.returnUniformizationConstants()->mDataArray[0] << endl;
        
		int tempCalcProbDist2 = mTree.root()->getCalculateFullProbabilityDistribution();
		mTree.root()->setCalculateFullProbabilityDistribution(1);
		mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
		mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist2);
       
        // [DIAGNOSTIC CHECK:] cout << "Alternative lnL says: " << mTree.root()->incompleteDataLikelihood( mTree.root(), mModel.returnStateFrequencies(), mTree.GetModelFrequencyVector(), m_vecSiteAssignment, theData ) << endl;
        
		// Force a full ancestral state update every so many gens...
		mModel.updateUniformization();
		int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
		mTree.root()->setCalculateFullProbabilityDistribution(1);
		mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
		mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
		
		mTree.initSiteLikelihoods( m_vecSiteAssignment, theData.getNumStates() );
		mTree.AssignAncestralStatesByMarginalProbability( mModel, theData, m_vecSiteAssignment, mModel.returnStateFrequencies(), 1 );
		
		// Now reset uniformization constant to something bigger...
		if (options.m_nMethod == 6) mModel.enlargeUniformization();
		
		// Don't need to do this every time, do we?
		mTree.KEEP_CountFullChangesOverTree( mModel, theData, m_vecSiteAssignment );
	}
	
	updateCount++;
	
	double ancstateProb;
	ancstateProb = options.ancStateUpdateFreq;
	if (updateCount < 2000) { 
		ancstateProb = 0.99; 
	}
	
	// ********************************************************************************************************************************
	// Output new parameter file 'posteriorRateMatrix'
	// ********************************************************************************************************************************
	if ( (updateCount % options.outputFrequency == 0) ) {
		static int fileOpen = 0;
		static ofstream posteriorMeanMatrix;
		if (fileOpen == 0) {
			posteriorMeanMatrix.open( "posteriorRateMatrix", fstream::out);
			fileOpen = 1;
		}
		
		// output the posterior mean rates
		
		// output SUM over: rate * siteRate * siteRateFraction
		static DoubleVector siteRateFraction;
		if (siteRateFraction.empty()) { siteRateFraction.resize( options.numCatASRV, 0 ); }
		else siteRateFraction.fill(0.0);
		
		for (int ii=0; ii<m_vecSiteAssignment.size(); ii++) {
			siteRateFraction.mDataArray[ mTree.root()->GetData().siteCats->getAssignment( ii ) ]++;
		}
		
		for (int jj=0; jj<options.numCatASRV; jj++) {
			siteRateFraction.mDataArray[jj] /= m_vecSiteAssignment.size();
		}
		
		static int firstOutput = 1;
		if (firstOutput == 1) {
			posteriorMeanMatrix << "TC	TA	TG	CA	CG";
			if (options.allowASRV) {
				posteriorMeanMatrix << "	Shape";
			} 
			if (options.updatebls) {
				posteriorMeanMatrix << "	TreeLength" << endl;
			}
			std::cout << endl;
			firstOutput = 0;
		}
		
		static double treeLen;
		if ( mTree.root()->GetData().treeLength != 0 ) {
			treeLen =  mTree.root()->GetData().treeLength;
		}
		
		for (int mm=0; mm<mModel.GetNumModels(); mm++) {
			if (mModel.GetNumModels() > 1) posteriorMeanMatrix << "Model class " << mm << ":" << endl;
			int iPairs[6] = { 1, 0, 1, 0, 2};
			int jPairs[6] = { 3, 1, 2, 3, 3}; 
			for (int ii=0; ii<5; ii++) {
				double outputRate = 0.0;
				//for (int kk=0; kk<options.numCatASRV; kk++) {
				//	outputRate += mModel.returnAlphaMatrix3D()->mRows.mDataArray[mm].mRows.mDataArray[iPairs[ii]].mDataArray[jPairs[ii]] * mTree.root()->GetData().siteCats->getRate( kk ) * siteRateFraction.mDataArray[kk];
				//}
				outputRate = mModel.returnAlphaMatrix3D()->mRows.mDataArray[mm].mRows.mDataArray[iPairs[ii]].mDataArray[jPairs[ii]] / mModel.returnAlphaMatrix3D()->mRows.mDataArray[mm].mRows.mDataArray[0].mDataArray[2];
				posteriorMeanMatrix << outputRate;
				if (ii < 4) posteriorMeanMatrix << "\t";
			} 
			if (options.allowASRV) {
				posteriorMeanMatrix << "\t" << mTree.root()->GetData().siteCats->getShapeParm();
			}
			if (options.updatebls) {
				posteriorMeanMatrix << "\t" << treeLen;
			}
			posteriorMeanMatrix << endl;
		} 
	}
	// ********************************************************************************************************************************
	
	
	if (rannum1 < ancstateProb)	{ 
		double rannum2 = num->Random();
		
		// Branch-length updater
		if ( (options.updatebls == 1) && (num->Random() < options.brlenupdatefreq)) {
			blCount++;
			bool m_bUpdate;
			int rannode = (int)( num->Random() * options.m_nNumBranches);
			
			// Fully calculate all substitution prob matrices...
			int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
			
			// Version 0: full branch-specific likelihood
			// Version 1: conjugate gibbs on bl's
			// Version 2: simple change in lik (b1u)
			// Version 3: MLE for bl (b1u)
			
			int version = 0;
			double var = 0.2;
			if (updateCount < 5000) { var = 1.0; }
			
			if (options.m_nMethod == 6) {
				// Conjugate Gibbs sampler seems to be working very well...
				if (updateCount < 5000) { version = 1; }
				else { version = 1; }
			}
			
			if ( version == 0 ) {
				mTree.root()->setCalculateFullProbabilityDistribution(1);
				mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
			}
			
			UpdateBranchLengths( mTree, rannode, mModel, theData, m_vecSiteAssignment, version, var );
			
			// Now reset the flag...
			mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
			
		} else if ( (rannum2 < 0.99) || ( (numModels <=1) && (options.numCatASRV <=1) ) )		//update model
		{
			m_bStateUpdate = false;
			m_bAssignment = false;
			
			static int capped = 0;
			if (options.m_nMethod == 6) {
				if ( (updateCount >= 5000) && (capped == 0) ) {
					std::cout << "	[ CAPPING THE UNIFORMIZATION CONSTANT and the NUMBER OF TREE NODES ]" << endl;
					options.capUniformizationRate = 1;
				
                    // This should be averaged across model classes...
                    double theRateFactor = mModel.getRateSum(); // / mModel.GetNumModels();
                    
                   // std::cout << "Got rate sum of " << theRateFactor << endl;
                    
                   // mModel.scaleDiscreteRates( theRateFactor ) ;
                    
                    double mu = mModel.returnUniformizationConstants()->mDataArray[0] / theRateFactor;
                    
                    // Change uniformization and ratescale for Qs
                    mModel.setUniformization(0, mu ); //(mu / theRateFactor) );
                    
                    mModel.scaleRateMatrices( theRateFactor );
                    
                    // Transform bls
                    mTree.root()->TransformBranchLength_B1u( mTree.root(), ( 1.0 / theRateFactor), mu );

                    
					// Need to re-subdivide the tree...
					mTree.subdivideLongBranches( mTree.root(), options.maxBL, mModel.returnUniformizationConstants(), 1 );
					
					options.capNumTreeNodes = true;
					std::cout << "		[ --- THE UNI CONSTANT IS " << (*mModel.returnUniformizationConstants())[0] << " --- ]" << endl;
					capped = 1;
				}
			}
			
			// SHOULD NOT DO UNIDIMENSIONAL PROPOSALS IF WE'RE NORMALIZING THE RATE MATRIX...
			if ( (updateCount > 5000) && (num->Random() <= 0.95) && ( options.m_nMethod == 6) && (1==2) ) { // ( m_nMethod != 4)
				// The unidimensional mechanism for B1u (updates many rate parameters one-by-one)...
				
				if (options.m_nMethod == 6) {
					UpdateGTRrates_unidimensional_b1u( mTree, mModel, theData, m_vecSiteAssignment, 1.0, -1 );
					mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels(), 1);			
				} else {
					// This version should work with any method...
					UpdateGTRrates_unidimensional( mTree, mModel, theData, m_vecSiteAssignment, 1.0, -1 );
					mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels(), 1);			
				}
				
			} else {
				// Testing it out...
				//	B1u_ConjugateGibbsUpdate_rateMatrix_rev(mTree, mModel, m_vecSiteAssignment, options.m_nMethod, m_bModelUpdate);
				//	B1u_ConjugateGibbsUpdate_rateMatrix_nonRev(mTree, mModel, m_vecSiteAssignment, options.m_nMethod, m_bModelUpdate);
				
				UpdateRateModelParameters_MC(mTree, mModel, theData, m_vecSiteAssignment, m_bModelUpdate);
			}
			modelCount++;
			
			// Call the gamma shape updater for ASRV if appropriate...
			// 1% of 95% of parameter updates
			if ( options.allowASRV && (options.numCatASRV > 1) && options.updateASRVparms && (num->Random() < 0.05) ) {
				shapeCount++;
				double var;
				var = 0.01;
				UpdateASRV_shape_MC(mTree, mModel, theData, m_vecSiteAssignment, m_bModelUpdate, var);
				
			}
			else		// update assignment
			{
				// 1% of parameter updates
				m_bModelUpdate = true;
				m_bAssignment = true;
				m_bStateUpdate = false;
				
				
				if ( options.allowASRV && (options.numCatASRV > 1) ) {
					assignCount++;
					if (options.m_nMethod != 6) {
						int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
						mTree.root()->setCalculateFullProbabilityDistribution(1);
						mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
						mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
					}			
					
					ResampleSiteAssignments_ASRV_rates(mTree, mModel, theData, m_vecSiteAssignment);	
				}
				
				if ( numModels >= 1) {
					ResampleSiteAssignments_MixtureModel(mTree, m_vecSiteAssignment, mModel, theData );	
				}
			}
		}
    } else			// update internal nodes
		{
			ancstateCount++;
			m_bModelUpdate = true;
			m_bAssignment = false;
			m_bStateUpdate = true;
			
			//---------------------------------
			// About to update internal states 
			//---------------------------------
			
			// 'Full' marginal ancestral state sampler...
			if (  (options.m_nMethod == 3) || (options.m_nMethod == 5) || (options.m_nMethod == 4) || ( options.forceFullMarginalUpdate == 1 ) ) { //==6 && (updateCount<2000) ) ) ) { //&& (updateCount < 2000) ) ) { // || ( ( m_nMethod == 6) && (num->Random() < 0.25) ) ) { // || (options.m_nMethod == 6)
				if (options.m_nMethod == 6) mModel.updateUniformization();
				
				int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
				mTree.root()->setCalculateFullProbabilityDistribution(1);
				mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
				mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
				
				//		mModel.updateStationaryFreqs();
				mTree.initSiteLikelihoods( m_vecSiteAssignment, theData.getNumStates() );
				//	mTree.AssignAncestralStatesByMarginalProbability(options.m_nMethod, m_vecStateIndex, numModels, mModel.returnStationaryFrequencies(), m_vecSiteAssignment, m_vecStateIndex, m_vecState, m_nMinstate, m_nNumStates, 1);
				mTree.AssignAncestralStatesByMarginalProbability( mModel, theData, m_vecSiteAssignment, mModel.returnStateFrequencies(), 1 );
				
			} else {
              //  cout << "** DOING LOCAL UPDATER **" << endl;
				// Original 'local' updater...
			
                if (options.m_nMethod == 6) mModel.updateUniformization();
				//[regular method]
				int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
				mTree.root()->setCalculateFullProbabilityDistribution(1);
				mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
				mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
				                
				int rannode = (int)( num->Random() * mTree.numRegularNodes);
				//			std::cout << "Doing ancestral state sample form node " << rannode << "..." << endl;
				//[regular]
				setAllAncestralAndTransientStates_local(  mTree, rannode, mModel, theData, m_vecSiteAssignment, mModel.returnStateFrequencies() );
				//[N4 method]
				// RunMCMCGibbsInternalStatesByN4(mTree, rannode, mModel.returnStateFrequencies(), m_vecSiteAssignment, m_vecStateIndex, m_vecState, m_nMinstate, m_nNumStates, mModel.returnRateMatrix3D(), mModel.returnUniformizationConstants());
				
				
			}
			
			// Now reset uniformization constant to something bigger...
			if (options.m_nMethod == 6) mModel.enlargeUniformization();
			
			// Don't need to do this every time, do we?
			mTree.KEEP_CountFullChangesOverTree( mModel, theData, m_vecSiteAssignment );
		}
		
		
		double logLikelihood1 = CalculateCompleteDataLogLikelihood(mTree, mModel, m_vecSiteAssignment);
		//[this one]
		//	std::cout << " ***** Complete data log-likelihood:  	" << logLikelihood1 << endl;
		mTree.completeDataLnL = logLikelihood1;
    
	m_bModelUpdate = true;
	if ( (updateCount % 10000) == 0) std::cout << "	*** Ancstate updates: " << ancstateCount << ", model updates: " << modelCount << ", bl counts: " << blCount << ", assign counts: " << assignCount << ", shape update counts: " << shapeCount << endl;
}


int MCMC::RunMarkovChain(
						 Tree& mTree,
						 ProteinModel& mModel,
                         IntVector& m_vecSiteAssignment,
                         ObservedData &theData,
						 bool& m_bModelUpdate,
						 bool& m_bStateUpdate,
						 bool& m_bAssignment,
						 std::ofstream& posteriorPredictiveOut) {
	
    int numModels = mModel.GetNumModels();
    
	static int realNumGens = 0;
	int numUpdated = 0;
	bool success = false;
	double likelihood;
	
	// **************************************************************************************************
	// Static variables to be saved across multiple calls...
	// **************************************************************************************************
	static double	*proposalVariances;
	static int		updateCount = 1;
	static int		lastUpdateType = -1;
	static int		proposalInit = 0;
	static int blCount = 0;
	
	// Initialize proposal variances...
	if (proposalInit == 0) {
		proposalVariances = new double [numModels];
		
		for (int i=0; i<numModels; i++) {
			proposalVariances[i] = 0.001;
		}
		proposalInit = 1;
	}
	
	// **************************************************************************************************
	// Set some of the move type probabilities...
	// **************************************************************************************************
	double ancestralStateProb = 0.99;
	double mixtureAssignmentProb = 0.85;
	
	if (updateCount < 500) { ancestralStateProb = 0.9; } else if (updateCount < 1000) { ancestralStateProb = 0.9; } else { ancestralStateProb = 0.999; } //mixtureAssignmentProb used to be 0.8 constant
	if (updateCount > 200) ancestralStateProb = options.ancStateUpdateFreq;
	if (!mModel.getRARdistributed()) ancestralStateProb = options.ancStateUpdateFreq;
	
	
	// **************************************************************************************************
	// INITIALIZATIONS FOR FIRST GENERATION...
	// **************************************************************************************************
	if (updateCount == 1) {
		// Some branch-specific initializers...
		// for COX1             mTree.MarkNodesBelow(272, m_vecSiteAssignment, m_vecStateIndex, m_vecState, m_nMinstate, m_nNumStates);		
		// for whole genome		mTree.MarkNodesBelow(270, m_vecSiteAssignment, m_vecStateIndex, m_vecState, m_nMinstate, m_nNumStates);		
		
		// Force a full ancestral state update
		std::cout << "		Performing initial ancestral state sample..." << endl;
		if (options.m_nMethod == 6) mModel.updateUniformization();
		
		//if ( (options.doRAR == 0) && (mModel.getPredefinedMatrix() == -1) ) {
		for (int ii=0; ii<numModels; ii++) {
			mModel.setUniformization( ii, 1.0 ); // 0.8			}
		} 
		//}
		if (mModel.getPredefinedMatrix()) {
			if (options.m_nMethod == 6) mModel.updateUniformization();
		}
		
		setAllProbMatricesOnTree( mTree, mModel );
		
		// Must be called manually, first...
		mTree.initSiteLikelihoods( m_vecSiteAssignment, theData.getNumStates() );
		numUpdated += setAllAncestralAndTransientStates_full( mTree, mModel, theData, m_vecSiteAssignment );
		
		// [blocking now Mar 22] if (options.m_nMethod == 6) mModel.enlargeUniformization();
		
		if ( options.allowASRV && (options.numCatASRV > 1) ) {
			countChangesAlongTree( mTree, mModel, theData, m_vecSiteAssignment );
			
			mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
			if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount << " [ Resampling site assignments to ASRV rate distribution ]" << endl; 
			
			ResampleSiteAssignments_ASRV_rates(mTree, mModel, theData, m_vecSiteAssignment);	
		}
		
		countChangesAlongTree( mTree, mModel, theData, m_vecSiteAssignment );
		
	}
	
	if ( (updateCount > 1) && ((updateCount % options.outputIncompleteLogL) == 0) && (options.outputIncompleteLogL > 0) ) {
		
		// Update uniformization only works before the cap is set at the end of burnin
		if (options.m_nMethod == 6) mModel.updateUniformization();
		
		setAllProbMatricesOnTree( mTree, mModel );
		numUpdated += setAllAncestralAndTransientStates_full(  mTree, mModel, theData, m_vecSiteAssignment );
		countChangesAlongTree( mTree, mModel, theData, m_vecSiteAssignment );
		
		// **************************************************************************************************
		// PP coevo
		// **************************************************************************************************
		if (options.coevol) {
			if (numModels > 1) {
				std::cout << "	ERROR, NO SUPPORT FOR MIXTURES WITH SELETED OPTIONS YET." << endl;
				exit(-1);
			}
			
			static ofstream posteriorCoevolution;
			static int posteriorCoevolutionOpen = 0;
			if (posteriorCoevolutionOpen == 0) {
				posteriorCoevolution.open( "posteriorCoevolution", fstream::out);
				posteriorCoevolutionOpen = 1;
			}
			
			posteriorCoevolution << "Gen " << updateCount << "\t";
			
			// Dwell-time based MI for coevolution:
			DoubleVector* treeDepth;
			
			// Doing site ii versus jj...
			for (int ii=0; ii<m_vecSiteAssignment.size(); ii++) {
				posteriorCoevolution << "Site " << ii << "\t";
				
				// Just output differences in dwell versus observed state times:
				if (mTree.root()->GetData().observedFreqs.empty()) {
					mTree.root()->GetData().observedFreqs.resize( 1, mModel.GetNumStates(), 0.0); 
				} else {
					//	if (ii==0) {
					mTree.root()->GetData().observedFreqs.fill(0.0);
					//	}
				}
				
				double siteRate1 = 1.0;
				double siteRate2 = 1.0;
				if (options.allowASRV) {
					// set the site rates here
					siteRate1 = mTree.root()->GetData().siteCats->getRate( mTree.root()->GetData().siteCats->getAssignment( ii ) );
				}
				
				// Display the site's dwell time stuff (if comparing to frequencies)...
				/*
				 DoubleMatrix* marginalDwell = mTree.getRoot()->getMarginalDwell(mTree.getRoot());
				 mTree.calculateDwellTimeOverBranches(ii, siteRate1, mModel, theData, m_vecSiteAssignment);	
				 DoubleVector* treeDepth = &mTree.getRoot()->GetData().tempTreeDepth;
				 
				 // Force stationary freqs update...
				 mModel.updateStationaryFreqs(1);
				 
				 double tempSum = 0.0;
				 for (int i=0; i<mModel.GetNumStates(); i++) {
				 std::cout << "Site\t" << ii << "\tState\t" << i << "\t" << ( marginalDwell->mRows.mDataArray[ 0 ].mDataArray[ i ] ) << "\t" << (mTree.root()->GetData().observedFreqs.mRows.mDataArray[0].mDataArray[i] / mTree.leaves()) << "\t" << ( mModel.returnStationaryFrequencies().mRows.mDataArray[0].mDataArray[i] ) << "\t" << endl;
				 //		std::cout << "site " << ii << ", state " << i << ", margDwell: " << tempMarginal.mRows.mDataArray[0].mDataArray[i] << endl;
				 tempSum += (marginalDwell->mRows.mDataArray[ 0 ].mDataArray[ i ] );
				 }
				 //std::cout << "	sum over dwell times at site is " << tempSum << endl;
				 */
				
				// comment out below if just getting the dwell versus freq times...			
				DoubleMatrix* marginalDwell = mTree.getRoot()->getMarginalDwell(mTree.getRoot());
				DoubleMatrix* jointDwell    = mTree.getRoot()->getJointDwell( mTree.getRoot() );
				
				// *******************************************************************************************************
				// Site ii -> 'tempMarginal'
				// *******************************************************************************************************
				double numSubs1 = mTree.calculateDwellTimeOverBranches( ii, siteRate1, mModel, theData, m_vecSiteAssignment );	
				
				if (numSubs1 < options.coevol_minNumSubstitutions) {
					// Just output a row of 0.0's for the MI statistic and continue...
					for (int jj=0; jj < m_vecSiteAssignment.size(); jj++) {
						if (jj >= ii) {
							posteriorCoevolution << "0.0" << "\t";
						} else { 
							posteriorCoevolution << "--\t";
						}
					} posteriorCoevolution << endl;
					continue;
				}
				
				// Will copy the results into 'tempMarginal' so we can combine stats on sites ii and jj later...
				static DoubleMatrix tempMarginal;
				if (tempMarginal.empty()) tempMarginal.resize( numModels, mModel.GetNumStates(), 0.0 );
				else tempMarginal.fill( 0.0 );
				
				double sum = 0.0;
				//	for (int k=0; k<numModels; k++) {
				for (int i=0; i<mModel.GetNumStates(); i++) {
					sum += tempMarginal.mRows.mDataArray[0].mDataArray[i] = marginalDwell->mRows.mDataArray[ 0 ].mDataArray[ i ];
					//		std::cout << "site " << ii << ", state " << i << ", margDwell: " << tempMarginal.mRows.mDataArray[0].mDataArray[i] << endl;
				}
				//	}
				
				for (int jj=0; jj<m_vecSiteAssignment.size(); jj++) {
					
					if (jj >= ii) {
						// *******************************************************************************************************
						// Site jj -> marginalDwell
						// *******************************************************************************************************
						if (options.allowASRV) {
							// set the site rates here
							siteRate2 = mTree.root()->GetData().siteCats->getRate( mTree.root()->GetData().siteCats->getAssignment( jj ) );
						}
						
						sum = 0.0;
						double numSubs2 = mTree.calculateDwellTimeOverBranches( jj, siteRate2, mModel, theData, m_vecSiteAssignment );	
						//for (int k=0; k<numModels; k++) {
						for (int i=0; i<mModel.GetNumStates(); i++) {
							sum += marginalDwell->mRows.mDataArray[0].mDataArray[i];
						}
						//}
						//					std::cout << "site " << ii << " marginal sum is " << sum << endl;
						
						if (numSubs2 < options.coevol_minNumSubstitutions) {
							if (jj >= ii) {
								posteriorCoevolution << "0.0" << "\t";
							} else { 
								posteriorCoevolution << "--\t";
							}
							continue;
						}						
						// *******************************************************************************************************
						// Now joint dwell times for ii x jj...
						// *******************************************************************************************************
						treeDepth = &mTree.getRoot()->GetData().tempTreeDepth;
						//	std::cout << "DIAG OUTPUT: TREE DEPTH IS " << (*treeDepth)[0] << endl;
						
						if (ii == 0) {
							std::cout << "	** Posterior rate for site " << jj << ": " << (numSubs2) << endl; // / (*treeDepth)[0]) << endl;
						}
						
						mTree.calculateJointDwellTimeOverBranches( ii,jj, siteRate1, siteRate2, mModel, theData, m_vecSiteAssignment );	
						
						// *******************************************************************************************************
						// Calculate phylogenetic MI statistic...
						// *******************************************************************************************************
						double sitePairMI = 0.0;
						
						double total = 0.0;
						double a,b,c;
						double aTot, bTot, cTot;
						aTot = bTot = cTot = 0.0;
						
						// We can test the effectiveness of scaling by the joint entropy when everything is done properly...
						// [make this an option, maybe] double jointEntropy = 0.0;
						
						for (int i=0; i<( mModel.GetNumStates() ); i++) {
							for (int j=0; j<( mModel.GetNumStates() ); j++) {
								total += jointDwell->mRows.mDataArray[0].mDataArray[ (i*mModel.GetNumStates()) + j] / (*treeDepth)[0];
								a = jointDwell->mRows.mDataArray[0].mDataArray[ (i*mModel.GetNumStates()) + j] / (*treeDepth)[0];
								
								// Marginal dwell times are already scaled by the tree depth in the Tree::calculateDwellTimeOverBranches() function...
								b = tempMarginal.mRows.mDataArray[0].mDataArray[i];
								c = marginalDwell->mRows.mDataArray[0].mDataArray[j];
								
								double temp = 0.0;
								if ( (a>0) && ( (b * c) > 0 ) ) {
									//	jointEntropy -= a * log2( a );
									temp = ( a ) * ( log2( a ) - (log2( b * c) ) );
									
									if ( (temp == temp ) ) {
										sitePairMI += temp;
									}
									//	std::cout << "state " << i << " x " << j << "; " << temp << ": " << a << ", " << b << ", " << c << endl;
								}
							}
						}
						
						//	sitePairMI /= jointEntropy;
						
						if ( (sitePairMI > 0.00001) ) {
							posteriorCoevolution << setprecision(5) << setw(10) << sitePairMI << "\t";
						} else {
							posteriorCoevolution << "0.0" << "\t";
						}
						//							std::cout << "pMI for site pair:\t" << ii << "\t" << jj << "\t" << setprecision(5) << setw(10) << sitePairMI;
						//						if ( (sitePairMI > 0.1) || (sitePairMI < -0.1) ) std::cout << "\t" << "*";
						
					} else { posteriorCoevolution << "--\t"; }
				} posteriorCoevolution << endl;
				
				
				if (options.coevol_doPosteriorPredictiveNull == 1) {
					// Site comparison for ii vs jj is done...
					std::cout << "	*** Testing posterior predictive.." << endl;
					
					double scaleFactor = 0.0;
					for (int aa=0; aa< mModel.GetNumStates(); aa++) {
						scaleFactor -= mModel.returnRateMatrix3D()->mRows.mDataArray[0].mRows.mDataArray[aa].mDataArray[aa];
					}
					std::cout << "The rate matrix scaling factor is currently " << scaleFactor << ", tree depth is " << (*treeDepth)[0] << endl;
					
					// Site ii (tempMarginal)...
					mModel.ProteinModelParameterstoRates( &tempMarginal );
					double scaleFactorPrime = 0.0;
					for (int aa=0; aa< mModel.GetNumStates(); aa++) {
						scaleFactorPrime -= mModel.returnRateMatrix3D()->mRows.mDataArray[0].mRows.mDataArray[aa].mDataArray[aa];
					}
					
					std::cout << "Using the marginal dwell time estimates, the scaling factor is " << scaleFactorPrime << endl;
					std::cout << "Going to use the transformation scalar of " << ( scaleFactor * numSubs1/(*treeDepth)[0] ) / scaleFactorPrime << " to attain a rate of " << numSubs1/(*treeDepth)[0] << endl;
					
					//	std::cout << mTree.simulatePosteriorPredictiveTestStatistic_multinomialLikelihood(mModel.returnStateFrequencies(), m_vecSiteAssignment, m_vecStateIndex, m_vecState, m_nMinstate, m_nNumStates) << endl;
					//	std::cout << "did it!" << endl;
				}				
			} posteriorCoevolution << endl;
			
			// [don't exit! -- this is for the frequency output stuff] 
			// exit(-1);
		} // coevolution stuff
		
	}
	
	// HEATING NO LONGER APPARENTLY NECESSARY
	double heat = 0.0;
	
	// **************************************************************************************************
	// MAIN LOOP...
	// **************************************************************************************************
	while (!success) { 
		
		// **************************************************************************************************
		// Branch-length updater
		// **************************************************************************************************
		// Don't start updating branchlengths until 1500 gens in...
		if ( (options.updatebls == 1) && (num->Random() < options.brlenupdatefreq) ) { // && ( updateCount > 3500) ) {
			success = true;
			blCount++;
			
			bool m_bUpdate;
			int rannode = (int)( num->Random() * options.m_nNumBranches);
			
			// Fully calculate all substitution prob matrices...
			//	int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
			
			// Version 0: full branch-specific likelihood
			// Version 1: conjugate gibbs on bl's
			// Version 2: simple change in lik (b1u)
			// Version 3: MLE for bl (b1u)
			
			int version = 0;
			double var = 0.2;
			if (updateCount < 5000) { var = 1.0; }
			
			if (options.m_nMethod == 6) {
				// Conjugate Gibbs sampler seems to be working very well...
				if (updateCount < 5000) { version = 1; }
				else { version = 1; }
			}
			
			if ( version == 0 ) {
				setAllProbMatricesOnTree( mTree, mModel );
			}
			
			UpdateBranchLengths( mTree, rannode, mModel, theData, m_vecSiteAssignment, version, var );
			if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount << " [ Branch-length updates complete.  Tree-length: " << mTree.root()->GetData().treeLength << "; numNodes: " << ( mTree.root()->CountLeaves() + mTree.root()->CountInternalNodes() ) << " ]" << endl;
			
		} else {
			
			// **************************************************************************************************
			// Cap the uniformization constant after burn-in ends
			// **************************************************************************************************
			// Cap the uniformization constant and number of nodes after a burnin period
			// This prevents the branchlength resampler from adding many, many nodes.
			static int capped = 0;			
			//		static int uniCheckCount = 0;
			
			if (options.m_nMethod == 6) {
				if ( (updateCount >= 5000) && (capped == 0) ) {
					std::cout << "	[ CAPPING THE UNIFORMIZATION CONSTANT and the NUMBER OF TREE NODES ]" << endl;
					std::cout << "		" << updateCount << " [ Starting tree-length: " << mTree.root()->GetData().treeLength << "; numNodes: " << ( mTree.root()->CountLeaves() + mTree.root()->CountInternalNodes() ) << " ]" << endl;// mModel.updateUniformization();
					
					// Should re-scale the rate matrix and branchlengths here...
					
					options.capUniformizationRate = 1;
						// This should be averaged across model classes...
                        double theRateFactor = mModel.getRateSum(); // / mModel.GetNumModels();
						
						std::cout << "Got rate sum of " << theRateFactor << endl;
						
						mModel.scaleDiscreteRates( theRateFactor ) ;
						
						double mu = mModel.returnUniformizationConstants()->mDataArray[0] / theRateFactor;
                    
                        // Change uniformization and ratescale for Qs
                        mModel.setUniformization(0, mu ); //(mu / theRateFactor) );
 
                        mModel.scaleRateMatrices( theRateFactor );

                        // Transform bls
                        mTree.root()->TransformBranchLength_B1u( mTree.root(), ( 1.0 / theRateFactor), mu );

					if (options.updatebls) {		
                        // Test: transform
						mTree.subdivideLongBranches( mTree.root(), options.maxBL, mModel.returnUniformizationConstants(), 1 );

						
						
						theRateFactor = mModel.getRateSum() / mModel.GetNumModels();
						
						std::cout << "Got rate sum of " << theRateFactor << endl;
						
						setAllProbMatricesOnTree( mTree, mModel );
						countChangesAlongTree(mTree, mModel, theData, m_vecSiteAssignment);
						numUpdated += setAllAncestralAndTransientStates_full(  mTree, mModel, theData, m_vecSiteAssignment );
						countChangesAlongTree( mTree, mModel, theData, m_vecSiteAssignment);
						
						
						// Ran the bl updater...
						int rannode;
						double treeLengthDelta = 0.6;
						int goodtimes=0;
						double lastTreeLength = mTree.root()->GetData().treeLength;
						
						//options.capNumTreeNodes = true;
						while ( (goodtimes < 3) ) {
							rannode = (int)( num->Random() * options.m_nNumBranches);
							
							UpdateBranchLengths( mTree, rannode, mModel, theData, m_vecSiteAssignment, 1, 0.2 );
							std::cout << "		" << updateCount << " [ Branch-length updates complete.  Tree-length: " << mTree.root()->GetData().treeLength << "; numNodes: " << ( mTree.root()->CountLeaves() + mTree.root()->CountInternalNodes() ) << " ]" << endl;
							
							// Need to re-subdivide the tree...
							// mTree.subdivideLongBranches( mTree.root(), options.maxBL, mModel.returnUniformizationConstants(), 1 );
							
							setAllProbMatricesOnTree( mTree, mModel );
							countChangesAlongTree(mTree, mModel, theData, m_vecSiteAssignment);
							numUpdated += setAllAncestralAndTransientStates_full(  mTree, mModel, theData, m_vecSiteAssignment );
							countChangesAlongTree( mTree, mModel, theData, m_vecSiteAssignment);
							
							treeLengthDelta = abs( mTree.root()->GetData().treeLength - lastTreeLength );
							cout << "		[ TREELENGTH CHANGE WAS " << treeLengthDelta << " ]" << endl;
							if (treeLengthDelta <= 3.0) { goodtimes++; }
							else { goodtimes = 0; }
							
							lastTreeLength = mTree.root()->GetData().treeLength;
							
						}	
						//options.capNumTreeNodes = false;
						
						
						// Need to re-subdivide the tree...
						mTree.subdivideLongBranches( mTree.root(), options.maxBL, mModel.returnUniformizationConstants(), 1 );
						
						options.capNumTreeNodes = true;
					}
					
					std::cout << "		[ --- THE UNI CONSTANT IS " << (*mModel.returnUniformizationConstants())[0] << " --- ]" << endl;
					capped = 1;
					
					
					//std::cout << "Finished.  lets bail! " << endl;
					
					setAllProbMatricesOnTree( mTree, mModel );
					countChangesAlongTree( mTree, mModel, theData, m_vecSiteAssignment );
					numUpdated += setAllAncestralAndTransientStates_full(  mTree, mModel, theData, m_vecSiteAssignment );
					countChangesAlongTree(mTree, mModel, theData, m_vecSiteAssignment );
					
					//exit(11);
					
				}
				
			}
			
			double whichUpdate = num->Random();
			if ( (whichUpdate < ancestralStateProb)  ) {
				double rannum2 = num->Random(); //[jdk] Numerics::getGlobalNumerics().Random();
				
				//	mixtureAssignmentProb = 0.0000001;
				
				// [DEBUG]: USE THIS LINE TO PREVENT THE MODEL UPDATER FROM RUNNING TWICE IN A ROW:
				// if (lastUpdateType == 0) { success = false; continue; }
				
				// UPDATE MODEL...
				if ( (rannum2 < mixtureAssignmentProb) || (numModels <=1)) {
					m_bStateUpdate = false;
					m_bAssignment = false;
					success = true;
					
					//----------------------------
					// About to update Model rates
					//----------------------------
					int whichModel = 0;
					
					// CONJUGATE GIBBS PHASE FOR RATE PARAMETERS ( initial burnin with unrestricted non-reversible model )...
					if ( ( (options.doRAR == 0) && (options.fixedRates == -1)  && (options.m_nMethod == 6) )  || ( ( updateCount < 5000 ) && (options.m_nMethod == 6) &&  (options.fixedRates == -1) ) ) {
						
						mModel.forceUpdateStationaryFreqs();
						
						if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount << " [Resampling full non-rev model by Conjugate Gibbs / RunMCMCFullParameterSweep_dependentAssignment 100%]" << endl;
						numUpdated += B1u_ConjugateGibbsUpdate_rateMatrix_nonRev(mTree, mModel, m_vecSiteAssignment, m_bModelUpdate);
						//	numUpdated += UpdateRateModelParameters_MC(heat, mTree, mModel, m_vecSiteAssignment, m_vecStateIndex, m_nMinstate, options.m_nMethod, m_bModelUpdate);
						
						if ( (num->Random() <= 0.2) ) { // was 0.05
							if ( options.allowASRV && (options.numCatASRV > 1) ) {
								//		assignCount++;
								if (options.m_nMethod != 6) {
									setAllProbMatricesOnTree( mTree, mModel );
								}			
								
								if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount << " [ Resampling site assignments to ASRV rate distribution ]" << endl; 
								ResampleSiteAssignments_ASRV_rates(mTree, mModel, theData, m_vecSiteAssignment);
                            }
							
							if ( ( mModel.GetNumModels() > 1) ) { 
								if ( (updateCount % options.outputFrequency) == 0) std::cout << "		[ Resampling site assignments ]" << endl; 
								numUpdated += ResampleSiteAssignments_MixtureModel(mTree, m_vecSiteAssignment, mModel, theData);	
							}
						}					
						m_bModelUpdate = true;
					} else {
						// if NOT doing conjugate Gibbs any longer...
						
						static int resetAfterConjugatePhase = 0;
						// GOING TO SET THE RAR RATE DISTRIBUTION BASED ON THE LAST CONJUGATE GIBBS RATE DISTRIBUTION...
						
						if ( (resetAfterConjugatePhase == 0) && (mModel.getRARdistributed()) ) { 
							std::cout << "		[ Setting initial RAR rate distribution ]" << endl; 
							
							mModel.updateUniformization();
							likelihood = CalculateCompleteDataLogLikelihood(mTree, mModel, m_vecSiteAssignment);
							
							static int *oldAssignments;
							static int assignedAlready = 0;
							if (assignedAlready == 0) {
								oldAssignments = new int[ mModel.GetNumModels() * mModel.GetNumStates() * mModel.GetNumStates() ];
								assignedAlready = 1;
							} 
							
							double distParm1, distParm2;
							distParm1 = mModel.getGammaScale(0);
							distParm2 = mModel.getGammaShape(0);
							
							mModel.forceUpdateStationaryFreqs();
							
							mModel.backupRateAssignments(oldAssignments);
							mModel.setDistributionParmsFromCounts( mTree );
							
							mModel.forceUpdateStationaryFreqs();
							
							int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
							mTree.root()->setCalculateFullProbabilityDistribution(0);
							mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
							mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
							
							// Set the uniformization constant to the maximum possible row-sum
							mModel.setMaximumUniformization_RAR();
							
							options.capUniformizationRate = 1;
							resetAfterConjugatePhase = 1;
							
							// RAR DISTRIBUTION INITIALIZATION, CAPPING OF TREENODES, ETC...
							if (mModel.getRARdistributed() && (updateCount > 5500) ) {
								
								if ((updateCount==1) || ( (updateCount%30)==0)) {
									// Do a RAR parameter assignment sweep...
									if (mModel.getDependentRates() && (mModel.GetNumModels() > 1)) {	
										if ( (updateCount % options.outputFrequency) == 0)  std::cout << "		" << updateCount << " [Resampling rate parameter values with re-assignment sweep / RunMCMCFullParameterSweep_dependentAssignment 100%]" << endl;
										numUpdated += ResampleParameterAssignments_RAR_dependentAssignment(mTree, mModel, theData, m_vecSiteAssignment, m_bModelUpdate, 1.0, -1);
									} else {
										whichModel = (num->Random() * numModels);
										
										if ( (updateCount % options.outputFrequency) == 0)  std::cout << "		" << updateCount << " [Resampling rate parameter assignments / (a) RunMCMCFullParameterSweep 100%]" << endl;
										numUpdated += ResampleParameterAssignments_RAR(mTree, mModel, theData, m_vecSiteAssignment, m_bModelUpdate, 1.0, -1);
									}
								} else {
									if (mModel.getDependentRates()) {	
										whichModel = -1;
									} else {
										whichModel = int(num->Random() * numModels); //[jdk] Numerics::getGlobalNumerics().Random();
									}	
									
									if ( num->Random() <= 0.3) { // 0.3 ) { // was last 0.3 // BLOCKED July 2009
										// RAR parameter sweep...
										if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount << " [Resampling rate parameter assignments / (b) RunMCMCFullParameterSweep / FULL]" << endl;
										
										if (mModel.getDependentRates()) {	
											numUpdated += ResampleParameterAssignments_RAR_dependentAssignment(mTree, mModel, theData, m_vecSiteAssignment,  m_bModelUpdate, 0.99, whichModel);
										} else {
											numUpdated += ResampleParameterAssignments_RAR(mTree, mModel, theData, m_vecSiteAssignment, m_bModelUpdate, 1.0, whichModel);
										}
										
									} else {
										if ( (num->Random() <= 0.5) || (mModel.getFixedFreqs() == 1) ) {
											// Update the RAR parameter distribution...
											if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount << " [Resampling rate parameter values with re-assignment sweep / RunMCMCDistributionParametersWithSweep <= 100%]" << endl;
											numUpdated += UpdateRARDistributionParameters(heat, mTree, mModel, theData, m_vecSiteAssignment, m_bModelUpdate, 1.0, proposalVariances, whichModel);
										} else {
											if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount << " Resampling other model parameters (non-RAR) / RunMCMCGibbsSamplingModel" << endl;
											numUpdated += UpdateRateModelParameters_MC(heat, mTree, mModel, theData, m_vecSiteAssignment,  m_bModelUpdate);
										}
									}
								}
							} else {
								// if NOT doing getRARdistributed():
								if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount << " Resampling model parameters (non-RAR) / RunMCMCGibbsSamplingModel" << endl;
								numUpdated += UpdateRateModelParameters_MC(heat, mTree, mModel, theData, m_vecSiteAssignment,  m_bModelUpdate);
								
                                if (m_bModelUpdate) mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels(),1);
							}
							
							if (options.updatebls) {
								double theRateFactor = mModel.getRateSum() / mModel.GetNumModels();
								
								std::cout << "Got rate sum of " << theRateFactor << endl;
								
								mModel.scaleDiscreteRates( theRateFactor ) ;
						
								double mu;
								for (int mm=0; mm < mModel.GetNumModels(); mm++) {
									mModel.scaleRateMatrices( mm, theRateFactor );
									
									mu = mModel.getUniformization(mm);
									mModel.setUniformization( mm , 1.0); // mu / theRateFactor );
								}
								
								theRateFactor = mModel.getRateSum() / mModel.GetNumModels();

								// Ran the bl updater...
								int rannode;
								double treeLengthDelta = 0.6;
								int goodtimes=0;
								double lastTreeLength = mTree.root()->GetData().treeLength;
								
								//options.capNumTreeNodes = true;
								while ( (goodtimes < 3) ) {
									rannode = (int)( num->Random() * options.m_nNumBranches);
									
									UpdateBranchLengths( mTree, rannode, mModel, theData, m_vecSiteAssignment, 1, 0.2 );
									std::cout << "		" << updateCount << " [ Branch-length updates complete.  Tree-length: " << mTree.root()->GetData().treeLength << "; numNodes: " << ( mTree.root()->CountLeaves() + mTree.root()->CountInternalNodes() ) << " ]" << endl;
									
									// Need to re-subdivide the tree...
									// mTree.subdivideLongBranches( mTree.root(), options.maxBL, mModel.returnUniformizationConstants(), 1 );
									
									setAllProbMatricesOnTree( mTree, mModel );
									countChangesAlongTree(mTree, mModel, theData, m_vecSiteAssignment);
									numUpdated += setAllAncestralAndTransientStates_full(  mTree, mModel, theData, m_vecSiteAssignment );
									countChangesAlongTree( mTree, mModel, theData, m_vecSiteAssignment);
									
									treeLengthDelta = abs( mTree.root()->GetData().treeLength - lastTreeLength );
									cout << "		[ TREELENGTH CHANGE WAS " << treeLengthDelta << " ]" << endl;
									if (treeLengthDelta <= 3.0) { goodtimes++; }
									else { goodtimes = 0; }
									
									lastTreeLength = mTree.root()->GetData().treeLength;
									
								}	
								//options.capNumTreeNodes = false;
								
								// Need to re-subdivide the tree...
								mTree.subdivideLongBranches( mTree.root(), options.maxBL, mModel.returnUniformizationConstants(), 1 );
								
							}
							options.capNumTreeNodes = true;
							std::cout << "		[ --- THE UNI CONSTANT IS " << (*mModel.returnUniformizationConstants())[0] << " --- ]" << endl;
							capped = 1;
							
							
							//std::cout << "Finished.  lets bail! " << endl;
							
							setAllProbMatricesOnTree( mTree, mModel );
							countChangesAlongTree( mTree, mModel, theData, m_vecSiteAssignment );
							numUpdated += setAllAncestralAndTransientStates_full( mTree, mModel, theData, m_vecSiteAssignment );
							countChangesAlongTree( mTree, mModel, theData, m_vecSiteAssignment );
							
						}
						
						// Now updates rates, assignments, etc.
						if (mModel.getRARdistributed() ) { //&& (updateCount > 5500) ) {
							
							if ((updateCount==1) || ( (updateCount%30)==0)) {
								// Do a RAR parameter assignment sweep...
								
								if (mModel.getDependentRates() && (mModel.GetNumModels() > 1)) {	
									if ( (updateCount % options.outputFrequency) == 0)  std::cout << "		" << updateCount << " [Resampling rate parameter values with re-assignment sweep / RunMCMCFullParameterSweep_dependentAssignment 100%]" << endl;
									numUpdated += ResampleParameterAssignments_RAR_dependentAssignment(mTree, mModel, theData, m_vecSiteAssignment, m_bModelUpdate, 1.0, -1);
								} else {
									whichModel = (num->Random() * numModels);
									
									if ( (updateCount % options.outputFrequency) == 0)  std::cout << "		" << updateCount << " [Resampling rate parameter assignments / (a) RunMCMCFullParameterSweep 100%]" << endl;
									numUpdated += ResampleParameterAssignments_RAR(mTree, mModel, theData, m_vecSiteAssignment, m_bModelUpdate, 1.0, -1);
								}
								
							} else {
								if (mModel.getDependentRates()) {	
									whichModel = -1;
								} else {
									whichModel = int(num->Random() * numModels); //[jdk] Numerics::getGlobalNumerics().Random();
								}	
								
								if ( num->Random() <= 0.3) { // 0.3 ) { // was last 0.3 // BLOCKED July 2009
									// RAR parameter sweep...
								
									if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount << " [Resampling rate parameter assignments / (b) RunMCMCFullParameterSweep / FULL]" << endl;
									
									if (mModel.getDependentRates()) {	
										numUpdated += ResampleParameterAssignments_RAR_dependentAssignment(mTree, mModel, theData, m_vecSiteAssignment,  m_bModelUpdate, 0.99, whichModel);
									} else {
										numUpdated += ResampleParameterAssignments_RAR(mTree, mModel, theData, m_vecSiteAssignment, m_bModelUpdate, 1.0, whichModel);
									}
								
								} else {
									// Update the RAR rate parameter distribution...
									
									if ( (num->Random() <= 0.5) || (mModel.getFixedFreqs() == 1) ) {
										// Update the RAR parameter distribution...
										if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount << " [Resampling rate parameter values with re-assignment sweep / RunMCMCDistributionParametersWithSweep <= 100%]" << endl;
										numUpdated += UpdateRARDistributionParameters(heat, mTree, mModel, theData, m_vecSiteAssignment, m_bModelUpdate, 1.0, proposalVariances, whichModel);
									} else {
										if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount << " Resampling other model parameters (non-RAR) / RunMCMCGibbsSamplingModel" << endl;
										numUpdated += UpdateRateModelParameters_MC(heat, mTree, mModel, theData, m_vecSiteAssignment,  m_bModelUpdate);
									}
								
								 }
							}
						} else {
							// if NOT doing getRARdistributed():
							if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount << " Resampling model parameters (non-RAR) / RunMCMCGibbsSamplingModel" << endl;
							numUpdated += UpdateRateModelParameters_MC(heat, mTree, mModel, theData, m_vecSiteAssignment,  m_bModelUpdate);
							
							if (m_bModelUpdate) mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels(),1);
						}
						
						
						//[debug]
						likelihood = CalculateCompleteDataLogLikelihood(mTree, mModel, m_vecSiteAssignment);
					}
					lastUpdateType = 0;
					
					// Call the gamma shape updater for ASRV if appropriate...
					// 1% of 95% of parameter updates
					if ( options.allowASRV && (options.numCatASRV > 1) && options.updateASRVparms && (num->Random() < 0.05) ) {
						//	shapeCount++;
						double var;
						//				if (options.m_nMethod == 6) var = 0.001;
						//				else var = 0.2;
						var = 0.01;
						
						//if (updateCount > 8000) { var = 0.02; }
						if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount << "[ Resampling ASRV distribution parameters ]" << endl; 
						
						UpdateASRV_shape_MC( mTree, mModel, theData, m_vecSiteAssignment, m_bModelUpdate, var );
						// Don't need to do this every time, do we?
						// countChangesAlongTree( mTree, mModel, theData, m_vecSiteAssignment );
					}
					
				} else {
					// UPDATE SITE ASSIGNMENT UNDER MIXTURE MODELS...
					if (  (lastUpdateType == 1))  { success = false; continue; }
					
					m_bModelUpdate = false;
					m_bAssignment = true;
					m_bStateUpdate = false;
					success = true;
					
					if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount << " [Resampling site-class memberships / ResampleSiteAssignments_MixtureModel]" << endl;			
					if ( options.allowASRV && (options.numCatASRV > 1) ) {
						//		assignCount++;
						if (options.m_nMethod != 6) {
							setAllProbMatricesOnTree( mTree, mModel );	
						}			
						
						ResampleSiteAssignments_ASRV_rates(mTree, mModel, theData, m_vecSiteAssignment);	
					}
					
					if ( numModels >= 1) {
						numUpdated += ResampleSiteAssignments_MixtureModel( mTree, m_vecSiteAssignment, mModel, theData );	
					}
					
					// numUpdated += ResampleSiteAssignments_MixtureModel(mTree, m_vecSiteAssignment, mModel, m_nAlignLength, theData, m_nNumStates);	
					lastUpdateType = 1;
					
					// not necessary for b1xm
					// mTree.CountChangesAlongEveryBranch(options.m_nMethod, theData, m_nNumStates);
					
					/*
					 // update the model changes of each branch
					 countChangesAlongTree( mTree, mModel, theData, m_vecSiteAssignment );
					 */
					
				}
			} else {
				// UPDATE ANCESTRAL STATES / TRANSIENT STATES...
				if ( (updateCount > 200) && (lastUpdateType == 2) ) { success = false; continue; }
				
				m_bModelUpdate = false;
				m_bAssignment = false;
				m_bStateUpdate = true;
				lastUpdateType = 2;
				success = true;
				//---------------------------------
				// About to update internal states 
				//---------------------------------
				int rannode;
				//			double fullUpdateFreq = 0.3;
				
				int which;
				
				which = 1;
				if ( options.forceFullMarginalUpdate == 1 ) { which = 2; }
				if ( updateCount < 200 ) { which = 2; }
				
				// Branchlength sampler for B1xM...
				//	if (options.m_nMethod == 3) {
				//		std::cout << "		[B1xM: doing branchlength sampler...]" << endl;
				//		mTree.setprobmatrixalongbranch(options.m_nMethod, mModel.GetNumStates(), mModel.GetRateMatrix3D(), mModel.returnUniformizationConstants(),  mModel.GetFreqModels());
				//		mModel.updateStationaryFreqs();
				//		numUpdated += RunMCMCGibbsSampling_B1xM_branchlengths(mTree, rannode, mModel.returnStateFrequencies(), m_vecSiteAssignment, m_vecStateIndex, m_vecState, m_nMinstate, m_nNumStates);
				//	}
				
				if ( (which == 1) ) {
					//if ( (updateCount % options.outputFrequency) == 0) 
					if ( (updateCount % options.outputFrequency) == 0) std::cout << "		[RESAMPLING ANCESTRAL STATES WITH LOCAL UPDATER..]" << endl;
					rannode = (int)( num->Random() * mTree.numRegularNodes); // options.m_nNumBranches); //mTree.leaves() ); // options.m_nNumBranches);

					setAllProbMatricesOnTree( mTree, mModel );
					
					mModel.forceUpdateStationaryFreqs();
                    
                    // Differentiate between root frequencies: parameters for predefined & reversible model.  O/w use stationary freqs.
					if (  (mModel.getPredefinedMatrix() > -1) && mModel.getReversible() ) {
						numUpdated += setAllAncestralAndTransientStates_local( mTree, rannode, mModel, theData, m_vecSiteAssignment, mModel.returnStateFrequencies() );
					} else {
						numUpdated += setAllAncestralAndTransientStates_local( mTree, rannode, mModel, theData, m_vecSiteAssignment, mModel.returnStationaryFrequencies() );
					}
					
					if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount << " [setAllAncestralAndTransientStates_local from node " << rannode << "]" << endl;			
					
				} else  {
					if ( (updateCount % options.outputFrequency) == 0) std::cout << "		[RESAMPLING ANCESTRAL STATES WITH FULL MARGINAL METHOD...]" << endl;
					
					setAllProbMatricesOnTree( mTree, mModel );
					numUpdated += setAllAncestralAndTransientStates_full(  mTree, mModel, theData, m_vecSiteAssignment );
					
					if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount << " [Complete ancestral state assignment (marginal sample)]" << endl;			
					
				}
				
				//----------------------------------------------------------------------		
				// Re-counting changes along the branch after states have been updated
				//----------------------------------------------------------------------		
				if (!options.capUniformizationRate) mModel.forceUpdateUniformization();
				mTree.CountChangesAlongEveryBranch( theData);
				
				// update the model changes of each branch
				countChangesAlongTree(mTree, mModel, theData, m_vecSiteAssignment );
				
				likelihood = CalculateCompleteDataLogLikelihood(mTree, mModel, m_vecSiteAssignment);
				mTree.completeDataLnL = likelihood;
				
				// BLOCK
				/*	// PP
				 static int hasBeenDone = 0;
				 if (hasBeenDone == 0) {
				 hasBeenDone = 1;
				 //		std::cout << "Calculating the observed T(X)..." << endl;
				 posteriorPredictiveOut << "Observed:" << "\t" << mTree.calculateObservedSequenceTestStatistic(mModel.returnStateFrequencies(), m_vecSiteAssignment, m_vecStateIndex, m_vecState, m_nMinstate, m_nNumStates) << endl;
				 //		std::cout << endl;
				 }
				 
				 if (updateCount > 1000) {
				 //	if ( num->Random()	<= 0.02 ) {
				 mModel.updateStationaryFreqs();
				 
				 std::cout << endl << "	[Calculating the multinomial likelihood, T(X), of a posterior predictive sample...]" << endl << endl;
				 posteriorPredictiveOut << "Predicted:" << "\t" << mTree.simulatePosteriorPredictiveTestStatistic_multinomialLikelihood(mModel.returnStateFrequencies(), m_vecSiteAssignment, m_vecStateIndex, m_vecState, m_nMinstate, m_nNumStates) << endl;
				 //std::cout << endl;
				 //	}
				 }
				 */
			}	
		}		
		updateCount++;
		realNumGens++;
	} // while !success
	
	return numUpdated;
}

