/*
 *  MCMC.h
 *  PLEX
 *
 *  A.P. Jason de Koning, jason.de.koning@gmail.com
 *  http://jasondk.org
 *
 *  David Pollock lab, University of Colorado School of Medicine
 *  http://www.evolutionarygenomics.com
 *  All rights reserved.
 
 */


#ifndef MCMC_h__
#define MCMC_h__

#include "Numerics.h"

#include "GTRModel.h"
#include "ProteinModel.h"
#include "Tree.h"
#include "Options.h"
#include "ObservedData.h"

extern Options options;

class MCMC
{
public:
	MCMC(void);
	
	void B1u_ConjugateGibbsUpdate_rateMatrix_nonRev(
													Tree& mTree, 
													GTRModel& mModel, 
													IntVector& m_vecSiteAssignment,
													bool& m_bUpdate);
	
	int B1u_ConjugateGibbsUpdate_rateMatrix_nonRev(
												   Tree& mTree, 
												   ProteinModel& mModel, 
												   IntVector& m_vecSiteAssignment,
												   bool& m_bUpdate);
	
	void B1u_ConjugateGibbsUpdate_rateMatrix_rev(
												 Tree& mTree, 
												 GTRModel& mModel, 
												 IntVector& m_vecSiteAssignment,
												 bool& m_bUpdate);
	
	void InitializeTree( Tree& mTree, ObservedData& theData);		//assign node sequence, augment internal sequence
	
	void InitializeTreeAlternate( Tree& mTree, ObservedData& theData);
    
    void InitializeMixtureModel(ObservedData& theData,
                                MixtureModel& m_Model,
                                const IntMatrix& m_ChangeMatrix);
        
	
	void GetModelUpdateWindowsize(
								  Tree& mTree, 
								  GTRModel& mModel, 
								  const int& AcceptCutoff,
								  const int& m_nMaxWindow,
								  const int& m_nTotalWindow,
								  const int& m_nSimulationGeneration);		//run the MCMC to get the windowsize in parameter update with phylogeny and mixture GTR model
	
	void GetModelUpdateWindowsize(
										Tree& mTree, 
										ProteinModel& mModel, 
										const int& AcceptCutoff);	// added by Jason
	
	void RunMarkovChain(
						Tree& mTree,
						GTRModel& mModel,
						bool& m_bModelUpdate,
						bool& m_bStateUpdate);				//run markov chain in one step with phylogeny, GTR model and mixture with weight scheme
	
	int RunMCMCGibbsSampling_B1xM_branchlengths(
												Tree& mTree,
												const int& rannode,
												const DoubleMatrix& stateFreqMatrix,
												IntVector& m_vecSiteAssignment,
												ObservedData& theData);
	
	void GetModelUpdateWindowsize(
								  const IntMatrix3D& count,
								  GTRModel& mModel, 
								  const int& AcceptCutoff);		//test the update windowsize of repetitive elements, GTR model and fixed weight scheme
	
	void RunMarkovChain(
						const IntMatrix3D& count, 
						GTRModel& mModel,
						bool& m_bUpdate);				//run one step in markov chain with repetitive elements, GTR model, mixture with weight scheme
	
	void RunMarkovChain(
						ObservedData& theData,
						const IntMatrix3D& countEach3D,	//changes of each site or sequences as the input
						IntMatrix3D& countModel3D,		//changes of each model after assignment
						GTRModel& mModel,
						IntVector& m_vecSiteAssignment,
						bool& m_bModel,
						bool& m_bAssignment);				//run one step in markov chain with repetitive elements, GTR model, mixture with Gibbs sampling
	
	
		void InitializeAssignment(
							  //initialize the assignment of model to site/lineage with phylogeny input
							  Tree& mTree,
							  IntVector& m_vecSiteAssignment,
							  MixtureModel& mModel,
							  ObservedData& theData );
	
	void RunMarkovChain(
						//run markov chain of model update, model assignment, internal state update with phylogeny
						Tree& mTree,
						GTRModel& mModel,
						IntVector& m_vecSiteAssignment,
						ObservedData& theData,
						bool& m_bModelUpdate,
						bool& m_bStateUpdate,
						bool& m_bAssignment);
		
    
     int RunMarkovChain(
                       Tree& mTree,
                       ProteinModel& mModel,
                       IntVector& m_vecSiteAssignment,
                       ObservedData &theData,
                       bool& m_bModelUpdate,
                       bool& m_bStateUpdate,
                       bool& m_bAssignment,
                       std::ofstream& posteriorPredictiveOut);
    
	void RunMarkovChain(
							  ObservedData& theData,
							  const IntMatrix3D& countEach3D,	//changes of each site or sequences as the input
							  IntMatrix3D& countModel3D,		//changes of each model after assignment
							  ProteinModel& mModel,
							  IntVector& m_vecSiteAssignment,
							  bool& m_bModel,
							  bool& m_bAssignment); 
	
	void RunMarkovChain(
							  Tree& mTree,
							  ProteinModel& mModel,
							  const IntVector& m_vecStateIndex, 
							  ObservedData& theData,
							  bool& m_bModelUpdate,
							  bool& m_bStateUpdate); //added by Jason
	
	double CalculateCompleteDataLogLikelihood(	//calculate the tree log likelihood when use gibbs sampling for mixture
											  Tree& mTree,
											  MixtureModel& mModel,
											  IntVector& m_vecSiteAssignment);
	
	double CalculateCompleteDataLogLikelihood(
											  Tree& mTree,
											  MixtureModel& mModel,
											  IntVector& m_vecSiteAssignment,
											  int version);
	
	double CalculateCompleteDataLogLikelihood_Delta_B1u(
														Tree& mTree,
														MixtureModel& mModel,
														ObservedData& theData,
														IntVector& m_vecSiteAssignment,
														int theModel, int x, int y);
	int ResampleParameterAssignments_RAR_B1u(
											 Tree& mTree, 
											 ProteinModel& mModel, 
											 ObservedData& theData,
                                             IntVector& m_vecSiteAssignment,
											 bool& m_bUpdate, double updateFraction, int whichModel);
	
	int UpdateGTRrates_unidimensional_b1u(
										  Tree& mTree, 
										  GTRModel& mModel, 
										  ObservedData& theData,
										  IntVector& m_vecSiteAssignment,
										  double updateFraction, int whichModel);
	
	int UpdateGTRrates_unidimensional(
									  Tree& mTree, 
									  GTRModel& mModel, 
									  ObservedData& theData,
									  IntVector& m_vecSiteAssignment,
									  double updateFraction, int whichModel);
	
	Numerics *num;
	
	int setAllAncestralAndTransientStates_full( Tree& mTree, MixtureModel& mModel, ObservedData& theData, IntVector& m_vecSiteAssignment );
	void countChangesAlongTree( Tree& mTree, MixtureModel& mModel, ObservedData& theData, IntVector& m_vecSiteAssignment );

protected:
	
private:
	void NormalizeMatrix(
						DoubleMatrix& m_probMatrix);
	
	int sampleFromCumulative( double *theVector, int length );

	void UpdateBranchLengths(Tree& mTree, 
							 int startNode,
							 GTRModel& mModel, 
							 ObservedData& theData,
							 IntVector& m_vecSiteAssignment,
							 int version, double variance
		);
	
		void UpdateBranchLengths(Tree& mTree, 
								 int startNode,
								 ProteinModel& mModel, 
								 ObservedData& theData,
								 IntVector& m_vecSiteAssignment,
								 int version, double variance
		);
	
	void RunMCMCChain(
					  const IntMatrix3D& count,
					  GTRModel& mModel1,
					  GTRModel& mModel2,
					  const double& windowsize,
					  const int& generations,
					  int& accept_counter);
	
	
	void ResampleSiteAssignments_MixtureModel(
											  vector< vector<int> > &m_vecSequences,
											  IntVector& m_vecSiteAssignment,
											  IntMatrix3D& count3D,			//the changes of each sequence/sites, sequence/site * numStates * numStates
											  MixtureModel& mModel,
                                              ObservedData& theData );
    
	int ResampleSiteAssignments_MixtureModel(
											 Tree& mTree,
											 IntVector& m_vecSiteAssignment,
											 MixtureModel& mModel,
                                             ObservedData& theData );
    
	void ResampleSiteAssignments_ASRV_rates(
											Tree& mTree,
											MixtureModel& mModel,
											ObservedData& theData,
                                            IntVector& m_vecSiteAssignment);
	
	int UpdateASRV_shape_MC(
							Tree& mTree, 
							GTRModel& mModel, 
							ObservedData& theData,
							IntVector& m_vecSiteAssignment,
							bool& m_bUpdate, double proposalVariance);
	
	int UpdateASRV_shape_MC(
							Tree& mTree, 
							ProteinModel& mModel, 
							ObservedData& theData,
							IntVector& m_vecSiteAssignment,
							bool& m_bUpdate, double proposalVariance);
	
	void RunMCMCModel(
					  const IntMatrix3D& count3D, 
					  GTRModel& mModel,
					  bool& m_bUpdate);
	
	void RunMCMCModel(
							const IntMatrix3D& count3D, 
							ProteinModel& mModel,
							bool& m_bUpdate);
	
	void UpdateModel(
					 MixtureModel& m_Model,
					 const double& m_fWindowsize);				//Update GTR model with a specific window size
	
	void UpdateModel(
					 MixtureModel& m_Model);				//Update GTR model once the window size has been set
	

	
	void UpdateRateModelParameters_MC(Tree& mTree, 
									  GTRModel& mModel, 
									  ObservedData& theData,
									  IntVector& m_vecSiteAssignment,
									  bool& m_bUpdate);
		
	int UpdateModelParameters_nonRAR_MC(
										double heat,
										Tree& mTree, 
										ProteinModel& mModel,
										ObservedData& theData,
										IntVector& m_vecSiteAssignment,
										bool& m_bUpdate);
	
	
	int UpdateRateModelParameters_MC(
									 double heat,
									 Tree& mTree, 
									 ProteinModel& mModel, 
									 ObservedData& theData,
									 IntVector& m_vecSiteAssignment,
									 bool& m_bUpdate);
	
	int ResampleParameterAssignments_RAR(
										 Tree& mTree, 
										 ProteinModel& mModel, 
										 ObservedData& theData,
                                         IntVector& m_vecSiteAssignment,
										 bool& m_bUpdate);
	
	int ResampleParameterAssignments_RAR(
										 Tree& mTree, 
										 ProteinModel& mModel, 
										 ObservedData& theData,
										 IntVector& m_vecSiteAssignment,
										 bool& m_bUpdate, double updateFraction, int whichModel);
	
	int ResampleParameterAssignments_RAR(
										 Tree& mTree, 
										 ProteinModel& mModel, 
										 ObservedData& theData,
										 IntVector& m_vecSiteAssignment,
										 bool& m_bUpdate, double updateFraction);
	
	int ResampleParameterAssignments_RAR_dependentAssignment(
															 Tree& mTree, 
															 ProteinModel& mModel, 
															 ObservedData& theData,
															 IntVector& m_vecSiteAssignment,
															 bool& m_bUpdate, double updateFraction, int whichModel);
	
	int UpdateRARDistributionParameters(
										double heat,
										Tree& mTree, 
										ProteinModel& mModel, 
										ObservedData& theData,
										IntVector& m_vecSiteAssignment,
										bool& m_bUpdate, double updateFraction, double *proposalVariances);
	
	int UpdateRARDistributionParameters(
										double heat,
										Tree& mTree, 
										ProteinModel& mModel, 
										ObservedData& theData,
										IntVector& m_vecSiteAssignment,
										bool& m_bUpdate, double updateFraction, double *proposalVariances, int whichModel);
	
	int RunMCMCGammaShapeWithSweepUni(
									  double heat,
									  Tree& mTree, 
									  ProteinModel& mModel, 
									  ObservedData& theData,
									  IntVector& m_vecSiteAssignment,
									  bool& m_bUpdate, double updateFraction, double *proposalVariances, int whichModel);
	
	int UpdateRARDistributionParameters_Uni_MC(
											   double heat,
											   Tree& mTree, 
											   ProteinModel& mModel, 
											   ObservedData& theData,
											   IntVector& m_vecSiteAssignment,
											   bool& m_bUpdate, double updateFraction, double *proposalVariances, int whichModel);
	
	int setAllAncestralAndTransientStates_local(
												Tree& mTree,
												int& rannode,
												MixtureModel& theModel,
												ObservedData& theData,
												IntVector& m_vecSiteAssignment,
												DoubleMatrix& stateFreqMatrix);
	
	double CalculateGibbsSamplingLnLikelihoodRatio(
												   Tree& mTree,
												   MixtureModel& mModel1,
												   MixtureModel& mModel2,
												   IntVector& m_vecSiteAssignment);
	
	double CalculateCompleteDataLogLikelihood_Delta_B1(
													 Tree& mTree,
													 MixtureModel& mModel,
													 IntVector& m_vecSiteAssignment,
													 int theModel, int x);
		
	void setAllProbMatricesOnTree( Tree& mTree, MixtureModel& mModel );
	
};



#endif // MCMC_h__
