/*
 *  MixtureModel.cpp
 *  PLEX
 *
 *  A.P. Jason de Koning, jason.de.koning@gmail.com
 *  http://jasondk.org
 *
 *  David Pollock lab, University of Colorado School of Medicine
 *  http://www.evolutionarygenomics.com
 *  All rights reserved.
 
 */


#include "MixtureModel.h"
#include "PLEXApp.h"

// pragma mark -
//---------------------------------------------------------------------------
MixtureModel::	MixtureModel(
		const DoubleMatrix3D& inRateMatrix3D,
		const int& inNumStates,
		const IntVector& inStateIndex,
		const CharVector& inStates,
		const int& inMinState,
		const int& inNumModels,
		const DoubleVector& inFreqModels,
		const bool& inFixedFreq,
		const double& inWindowsize,
		const int& inNucleotides)
{
	m_oPLEXApp = new(PLEXApp);
	num = &Numerics::getGlobalNumerics();

	if (uniformizationConstants.empty()) { uniformizationConstants.resize( inNumModels, 2.0 ); }
//	uniformizationConstants = new double[inNumModels];
//	for (int i=0; i<inNumModels; i++) { uniformizationConstants[i] = 1.0; }
	
	mRateMatrix3D = inRateMatrix3D;
	mNumStates = inNumStates;
	mStateIndex = inStateIndex;
	mStates = inStates;
	mMinState = inMinState;
	mNumModels = inNumModels;
	mFreqModels = inFreqModels;
	mFixedFreq = inFixedFreq;
	mWindowsize = inWindowsize;
	mNucleotides = inNucleotides;
	
	num = &Numerics::getGlobalNumerics();

}		

//---------------------------------------------------------------------------
MixtureModel::MixtureModel(
		const MixtureModel&	inMixtureModel)
{
//	m_oPLEXApp = new(PLEXApp);
	num = &Numerics::getGlobalNumerics();

//	uniformizationConstants = new double[inMixtureModel.mNumModels];
	if (uniformizationConstants.empty()) { uniformizationConstants.resize( inMixtureModel.mNumModels, 2.0 ); }
	for (int i=0; i<inMixtureModel.mNumModels; i++) { uniformizationConstants[i] = inMixtureModel.uniformizationConstants[i]; }

	mNumStates = inMixtureModel.mNumStates;
	mStateIndex = inMixtureModel.mStateIndex;
	mStates = inMixtureModel.mStates;
	mMinState = inMixtureModel.mMinState;
	mNumModels = inMixtureModel.mNumModels;
	mFreqModels = inMixtureModel.mFreqModels;
	mFixedFreq = inMixtureModel.mFixedFreq;
	mWindowsize = inMixtureModel.mWindowsize;
	mNucleotides = inMixtureModel.mNucleotides;
	
	mRateMatrix3D = inMixtureModel.mRateMatrix3D;
	m_oPLEXApp = inMixtureModel.m_oPLEXApp;
	num = &Numerics::getGlobalNumerics();
}

//---------------------------------------------------------------------------
MixtureModel::~MixtureModel(void)
{
//	delete uniformizationConstants;
//	delete m_oPLEXApp;
}
					
										
DoubleMatrix3D* MixtureModel::returnRateMatrix3D(void) {
	return &mRateMatrix3D;
}

//---------------------------------------------------------------------------
/*const MixtureModel& MixtureModel::operator=(
				const MixtureModel&	inMixtureModel)
{
	MixtureModel temp(inMixtureModel);
	swap(temp);
	return *this;
} */
											
//---------------------------------------------------------------------------
void MixtureModel::swap(
			MixtureModel&	inMixtureModel)
{
	mRateMatrix3D.swap(inMixtureModel.mRateMatrix3D);
	std::swap(uniformizationConstants, inMixtureModel.uniformizationConstants);
	std::swap(mNumStates, inMixtureModel.mNumStates);
	std::swap(mStateIndex, inMixtureModel.mStateIndex);
	std::swap(mStates, inMixtureModel.mStates);
	std::swap(mMinState, inMixtureModel.mMinState);
	std::swap(mNumModels, inMixtureModel.mNumModels);
	std::swap(mFreqModels, inMixtureModel.mFreqModels);
	std::swap(mFixedFreq, inMixtureModel.mFixedFreq);
	std::swap(mWindowsize, inMixtureModel.mWindowsize);
	std::swap(mNucleotides, inMixtureModel.mNucleotides);
}

int MixtureModel::updateUniformization(void) {
	return updateUniformization(-1);
}

void MixtureModel::setUniformization(DoubleVector* vals) {
	if (uniformizationConstants.empty()) { uniformizationConstants.resize( mNumModels, 0.0 ); }

	for (int i=0; i<mNumModels; i++) {
		uniformizationConstants.mDataArray[i] = vals->mDataArray[i];
	//	std::cout << "		SetUniformization() -> setting to " << vals->mDataArray[i] << endl;
	}
}

void MixtureModel::setUniformization(int which, double val) {
	if (uniformizationConstants.empty()) { uniformizationConstants.resize( mNumModels, 0.0 ); }
	
	uniformizationConstants.mDataArray[which] = val;
}

double MixtureModel::getRateSum( void ) {
	return getRateSum( -1 );
}

double MixtureModel::getRateSum( int which ) {
	double theSum = 0.0;
	int startModel, endModel;
	
	if (which == -1) {
		startModel = 0;
		endModel   = mNumModels;
	} else {
		startModel = which;
		endModel   = which + 1;
	}
	
	for (int i=startModel; i<endModel; i++) {
		for (int j=0; j< mNumStates; j++) {
			// theSum += -mRateMatrix3D.mRows.mDataArray[i].mRows.mDataArray[j].mDataArray[j];
			theSum += mStationaryFrequencyMatrix.mRows.mDataArray[i].mDataArray[j] * (-mRateMatrix3D.mRows.mDataArray[i].mRows.mDataArray[j].mDataArray[j]);
		}	
	}
	
	return theSum;
}

double MixtureModel::scaleRateMatrices( double scaleFactor ) {
	return scaleRateMatrices( -1, scaleFactor );
}

double MixtureModel::scaleRateMatrices( int which, double scaleFactor ) {
	double theSum = 0.0;
	int startModel, endModel;
	
	if (which == -1) {
		startModel = 0;
		endModel   = mNumModels;
	} else {
		startModel = which;
		endModel   = which + 1;
	}
	
	for (int i=startModel; i<endModel; i++) {
		for (int j=0; j< mNumStates; j++) {
			for (int k=0; k< mNumStates; k++) {
				mAlphaMatrix3D.mRows.mDataArray[i].mRows.mDataArray[j].mDataArray[k] /= scaleFactor;
				mRateMatrix3D.mRows.mDataArray[i].mRows.mDataArray[j].mDataArray[k] /= scaleFactor;
			}
		}
	}
	
	return theSum;
}


int MixtureModel::updateUniformization(int theModel) { 
	// Update the uniformization constant for the requested model
	// - will always update if a rate matrix row exceeds the current value
	// - will also occassionally reset completely to make sure we don't stay too large
	// * returns 1 if update was made
	//   (models will by sync'd without reporting an update)
	
	int i, j, startModel, endModel;
	int updated = 0;

	// If capUniformization flag is set, only do this the first time its been called...
	static int init = 0;
	if (init == 0) {
		init = 1;
	} else { 
		if ( options.capUniformizationRate ) return -1;
	}
	
	// If no model specified, do them all...
	if (theModel == -1) {
		startModel=0;
		endModel=mNumModels;
	} else {
		startModel = theModel;
		endModel=theModel+1;
	}
	
	// Initialize stuff...
	if (uniformizationConstants.empty()) { uniformizationConstants.resize( mNumModels, 0.0 ); }
	
	// Every once and a while, completely reset
	if (num->Random() < 0.2) {
		for (i=startModel; i<endModel; i++) {
			uniformizationConstants[i] = 0.0;
		}
	}
	
	double largestRate = 1.0;
	if (options.allowASRV) largestRate = options.__largestRate;
 
	int anyupdated = 0;
	for (i=startModel; i<endModel; i++) {
		updated = 0;
		for (j=0; j<mNumStates; j++) {
			if (uniformizationConstants[i] < (-mRateMatrix3D.mRows.mDataArray[i].mRows.mDataArray[j].mDataArray[j] * largestRate) ) {
				uniformizationConstants[i] = (-mRateMatrix3D.mRows.mDataArray[i].mRows.mDataArray[j].mDataArray[j] * largestRate);
				updated = 1; anyupdated = 1;
			}
		}

		// Enlarge it a bit...
		if (updated == 1) uniformizationConstants[i] *= 1.01;
	}

	// Now set the unif constants to be the same between models [Feb 25, 2009]
	// ( this is to simplify the site specific likelihood calculations, and the branch function calculations )
	double biggest = 0;
	for (i=startModel; i<endModel; i++) {
		if ( uniformizationConstants[i] > biggest ) {
			biggest = uniformizationConstants[i];
		}
	}
	for (i=startModel; i<endModel; i++) {
		uniformizationConstants[i] = biggest;
	}

	// Debug output...
	/* for (i=startModel; i<endModel; i++) {
		if (anyupdated != 0) { std::cout << "		[ (std) UPDATED UNIFORMIZATION " << i << " CONSTANT to " << uniformizationConstants[i] << " ]" << endl; }
		std::cout << "		[ DIFF BETWEEN BIGGEST LAMBDA_X and MU is: " << (uniformizationConstants[i] - biggest) << " ]" << endl;
	} */
	 
	return updated;
}

int MixtureModel::forceUpdateUniformization() { 
	// Call forceUpdate() for all models
	
	for (int i=0; i<mNumModels; i++) {
		forceUpdateUniformization(i);
	}
	
	return 1;
}
	
int MixtureModel::forceUpdateUniformization(int theModel) { 
	// Mandatory update on uniformization constants
	
	int i, j, startModel, endModel;
	int updated = 0;
	
	if ( options.capUniformizationRate ) {
	//	std::cout << " ** forceUpdateUniformization() was called after capUni was set.  Overriding. **" << endl;
		return 0;
	}
	
	// If no model requested, do all...
	if (theModel == -1) {
		startModel=0;
		endModel=mNumModels;
	} else {
		startModel = theModel;
		endModel=theModel+1;
	}
	
	if (uniformizationConstants.empty()) { uniformizationConstants.resize( mNumModels, 0.0 ); }
	
	double largestRate = 1.0;
	if (options.allowASRV) largestRate = options.__largestRate;

	int anyupdated = 0;
	for (i=startModel; i<endModel; i++) {
		updated = 0;
		for (j=0; j<mNumStates; j++) {
			if (uniformizationConstants[i] < (-mRateMatrix3D.mRows.mDataArray[i].mRows.mDataArray[j].mDataArray[j] * largestRate) ) {
				uniformizationConstants[i] = (-mRateMatrix3D.mRows.mDataArray[i].mRows.mDataArray[j].mDataArray[j] * largestRate);
				updated = 1; anyupdated = 1;
			}
			
		}
		if (updated == 1) uniformizationConstants[i] *= 1.01;
	}
	
	// Now set the unif constants to be the same between models [Feb 25, 2009]
	// ( this is to simplify the site specific likelihood calculations, and the branch function calculations )
	double biggest = 0;
	for (i=startModel; i<endModel; i++) {
		if ( uniformizationConstants[i] > biggest ) {
			biggest = uniformizationConstants[i];
		}
	}
	for (i=startModel; i<endModel; i++) {
		uniformizationConstants[i] = biggest;
	}
	
	// Debug output:
	/* for (i=startModel; i<endModel; i++) {
		if (anyupdated == 1) std::cout << "		[ (force) UPDATED UNIFORMIZATION " << i << " CONSTANT to " << uniformizationConstants[i] << " ]" << endl;
			std::cout << "test" << endl;
			std::cout << "		[ DIFF BETWEEN BIGGEST LAMBDA_X and MU is: " << (uniformizationConstants[i] - biggest) << " ]" << endl;
	} */
	
	return anyupdated;
}

double MixtureModel::returnOptimalUniformization(int theModel) { 
	// Check the request model and update the uniformization constant...
	// [this does not set anything, just checks]
	
	int i, j, startModel, endModel;
	int updated = 0;
	
	if (theModel == -1) {
		startModel=0;
		endModel=mNumModels;
	} else {
		startModel = theModel;
		endModel=theModel+1;
	}
	
	if (uniformizationConstants.empty()) { uniformizationConstants.resize( mNumModels, 0.0 ); }

	double bestUni = -std::numeric_limits<double>::max( );
	
	for (i=startModel; i<endModel; i++) {
		updated = 0;
		for (j=0; j<mNumStates; j++) {
			if (bestUni < (-mRateMatrix3D.mRows.mDataArray[i].mRows.mDataArray[j].mDataArray[j]) ) {
				bestUni = (-mRateMatrix3D.mRows.mDataArray[i].mRows.mDataArray[j].mDataArray[j]);
			}
			
		}
	}
	
	return bestUni;
}

int MixtureModel::enlargeUniformization(void) {
	return enlargeUniformization(-1);
}

int MixtureModel::enlargeUniformization(int theModel) { 
	// Check the request model and update the uniformization constant...
	// [block - working on GTR] return updateUniformization(theModel);

	int i, j, startModel, endModel;
	int updated = 0;
	double factor = 1.01;

	if (theModel == -1) {
		startModel=0;
		endModel=mNumModels;
	} else {
		startModel = theModel;
		endModel=theModel+1;
	}
	
	if (uniformizationConstants.empty()) { uniformizationConstants.resize( mNumModels, 0.0 ); }
	
	double largestRate = 1.0;
	if (options.allowASRV) largestRate = options.__largestRate;
	//std::cout << "Largest rate is " << largestRate << endl;

	for (i=startModel; i<endModel; i++) {
		for (j=0; j<mNumStates; j++) {
			if (uniformizationConstants[i] < (-mRateMatrix3D.mRows.mDataArray[i].mRows.mDataArray[j].mDataArray[j] * largestRate) ) {
				uniformizationConstants[i] = (-mRateMatrix3D.mRows.mDataArray[i].mRows.mDataArray[j].mDataArray[j] * largestRate);
				updated = 1;
			}
		}
		if (updated == 1) uniformizationConstants[i] *= factor;
	}
	
	for (i=startModel; i<endModel; i++) {
		if (updated == 1) std::cout << "		[ ENLARGED UNIFORMIZATION " << i << " CONSTANT to " << uniformizationConstants[i] << " ]" << endl;
	}

//	if (updated == 1) {
//		std::cout << "	[ ENLARGED UNIFORMIZATION CONSTANT ]" << endl;
//	}
	return updated;
}

void MixtureModel::initRateScale(double val) {
	rateScaleFactor = new double[GetNumModels()];
	
	//	std::cout << "INITRATESCALE CALLED WITH VAL " << val << endl;
	//	std::cout << "num models returns " << GetNumModels() << endl;
	
	for (int i=0; i< GetNumModels(); i++) {
		setRateScale(i, val);
		//		std::cout << "	SETTIN RATE " << i << " to " << val << endl;
	}
}

void MixtureModel::initRateScale(int num, double val) {
	rateScaleFactor = new double[num];
	
	//	std::cout << "INITRATESCALE CALLED WITH VAL " << val << endl;
	
	for (int i=0; i< num; i++) {
		setRateScale(i, val);
		//		std::cout << "	SETTIN RATE " << i << endl;
	}
}

void MixtureModel::setReversible(int val) {
	reversibleModel = val;
}

int MixtureModel::getReversible(void) {
	return reversibleModel;
}



void MixtureModel::updateStationaryFreqs() {
	updateStationaryFreqs(0);
}

void MixtureModel::forceUpdateStationaryFreqs() {
	updateStationaryFreqs(1);
}

void MixtureModel::updateStationaryFreqs(int force) {
	// We need to get the equilibrium (stationary) distribution...
	static int init = 0;
	static double *theQmat, *trans;
	static int *indexVector;
	static double *pi;
	
	//[block]		
	//	return;
	if ( (getPredefinedMatrix() > -1) && getReversible() ) return;
	if ( (options.freeFreqs == 1) && ( force != 1) ) return;
	
	// std::cout << "[ RECALCULATING STATIONARY FREQS ]" <<endl;// return;
	
	double d;
	int i,j,k;
	
	if (init == 0) {
		trans = new double[ (mNumStates+1) * (mNumStates+1) ];
		theQmat = new double[ (mNumStates+1) * (mNumStates+1) ];
		indexVector = new int[ mNumStates +1 ];
		pi = new double[ mNumStates +1 ];
		
		if (mStationaryFrequencyMatrix.empty() ) {
			mStationaryFrequencyMatrix.resize( mNumModels, mNumStates);
		}
		init = 1;
	}
	
	for (i=0; i<mNumModels; i++) {
		
		theQmat[0] = 0.0;
		for (j=0; j<mNumStates; j++) {
			for (k=0; k<mNumStates; k++) {
				trans[(k*mNumStates)+j] = mRateMatrix3D.mRows.mDataArray[i].mRows.mDataArray[j].mDataArray[k];
			//	std::cout << "mat at " << j << ", " << k << ": " << trans[(k*mNumStates)+j] << endl;
			}
		}
		
		for (j=0; j<mNumStates-1; j++) {
			pi[(j+1)] = -trans[(j*mNumStates) + (mNumStates-1)];
			for (k=0; k<mNumStates-1; k++) {
				theQmat[ (j+1)*(mNumStates-1) + (k+1) ] = trans[(j*mNumStates) + k] + pi[(j+1)];
			}
		}
		
		
		// Permorm the decomposition
		num->NR_ludcmp(theQmat, (mNumStates-1), indexVector, &d);
		
		/*	std::cout << "d is " << d << endl;
		 
		 for (j=1; j<=mNumStates-1; j++) {
		 std:: cout << "		index " << indexVector[j] << endl;
		 }
		 
		 for (j=0; j<mNumStates-1; j++) {
		 for (k=0; k<mNumStates-1; k++) {
		 std::cout << theQmat[(j)*mNumStates+(k+1)] << "\t";
		 } std::cout << endl;
		 } std::cout << endl;
		 
		 std::cout << "Calling NR fxn 2..." << endl;
		 */
		// Backsubstitution
		num->NR_lubksb(theQmat, (mNumStates-1), indexVector, pi);
		
		double sum = 0.0;
		//		std::cout << "Got stationary dist..." << endl;
		for (int theState=1; theState<=mNumStates-1; theState++) {
			// Previous method just updated the 'frequency parameters'
			// mFrequencyMatrix.mRows.mDataArray[i].mDataArray[theState-1] = pi[theState];
			mStationaryFrequencyMatrix.mRows.mDataArray[i].mDataArray[theState-1] = pi[theState];
			sum += pi[theState];
		}
		// mFrequencyMatrix.mRows.mDataArray[i].mDataArray[(mNumStates-1)] = 1.0 - sum;
		mStationaryFrequencyMatrix.mRows.mDataArray[i].mDataArray[(mNumStates-1)] = 1.0 - sum;
		
	} // all models (i)
}

//---------------------------------------------------------------------------
void MixtureModel::InitializeParameter(
			const IntMatrix& m_ChangeMatrix)
{
}

//---------------------------------------------------------------------------
void MixtureModel::InitializeFreqModels(void)
{
	if (mNumModels == 0)
	{	
		cerr << "number of models in mixture model could not be zero" << endl;
		
		exit(-1);
	}
		
	if (mFreqModels.empty())
		mFreqModels.resize(mNumModels,(double) 1/ (double)mNumModels);
}

//---------------------------------------------------------------------------
void MixtureModel::UpdateParameter(
		const int& m_nMethod)
{
}

//---------------------------------------------------------------------------
void MixtureModel::UpdateFreqParameter(
		double& inParameter)
{
// [jdk]	PLEXApp* m_oPLEXApp = new(PLEXApp);

	// This loop only is executed if the passed-in value is out of bounds...	
	do
	{
		double rannum =Numerics::getGlobalNumerics().Random();
	
		double changeamount = (rannum - 0.5) * mWindowsize;
	
		double proposal = inParameter + changeamount;
							
		if ((proposal < 0.0 ) || (proposal > 1.0))
			proposal = inParameter - changeamount;
		
		inParameter = proposal;
	}
	while ((inParameter > 1.0) || (inParameter < 0.0));
							
// [jdk]	delete(m_oPLEXApp);
}

//---------------------------------------------------------------------------
void MixtureModel::UpdateFreqModels(void)
{
	// This is for model component weights...
	
	if (!mFixedFreq)
	{
		DoubleVector tempFreqModels = mFreqModels;
	
		double sum;
		
		do
		{
			sum = 0;
			
			unsigned int k;
			
			for (k = 0; k < mNumModels -1; k++)
			{
				UpdateFreqParameter(tempFreqModels[k]);
				
				sum += tempFreqModels[k];
			}
			
			tempFreqModels[k] = 1 - sum;
		}
		while (sum > 1);
		
		SetFreqModels(tempFreqModels);
	}
}

//---------------------------------------------------------------------------
CharVector& MixtureModel::GetStates(void)
{
	return(mStates);
}
								
//---------------------------------------------------------------------------
int& MixtureModel::GetMinState(void)
{
	return(mMinState);
}
			
//---------------------------------------------------------------------------
IntVector& MixtureModel::GetStateIndex(void)
{
	return(mStateIndex);
}

//---------------------------------------------------------------------------
int& MixtureModel::GetNumStates(void)
{
	return(mNumStates);
}

//---------------------------------------------------------------------------
DoubleMatrix3D& MixtureModel::GetRateMatrix3D(void)
{
	return(mRateMatrix3D);
}

//---------------------------------------------------------------------------
int& MixtureModel::GetNumModels(void) 
{
	return(mNumModels);
}
			
//---------------------------------------------------------------------------
DoubleVector& MixtureModel::GetFreqModels(void)
{
	return(mFreqModels);
}
			
//---------------------------------------------------------------------------
double& MixtureModel::GetUpdateWindowsize(void)
{
	return(mWindowsize);
}

//---------------------------------------------------------------------------
int& MixtureModel::GetModelNumNucleotides(void)
{
	return(mNucleotides);
}

//---------------------------------------------------------------------------
const CharVector& MixtureModel::GetStates(void) const
{
	return(mStates);
}
								
//---------------------------------------------------------------------------
const int& MixtureModel::GetMinState(void) const
{
	return(mMinState);
}
			
//---------------------------------------------------------------------------
const IntVector& MixtureModel::GetStateIndex(void) const
{
	return(mStateIndex);
}

//---------------------------------------------------------------------------
const int& MixtureModel::GetNumStates(void) const
{
	return(mNumStates);
}

//---------------------------------------------------------------------------
const DoubleMatrix3D& MixtureModel::GetRateMatrix3D(void) const
{
	return(mRateMatrix3D);
}

//---------------------------------------------------------------------------
const int& MixtureModel::GetNumModels(void) const
{
	return(mNumModels);
}
			
//---------------------------------------------------------------------------
const DoubleVector& MixtureModel::GetFreqModels(void) const
{
	return(mFreqModels);
}
			
//---------------------------------------------------------------------------
const double& MixtureModel::GetUpdateWindowsize(void) const
{
	return(mWindowsize);
}

//---------------------------------------------------------------------------
const int& MixtureModel::GetModelNumNucleotides(void) const
{
	return(mNucleotides);
}

//---------------------------------------------------------------------------
void MixtureModel::SetStates(
			const CharVector& inStates)
{
	mStates = inStates;
}
								
//---------------------------------------------------------------------------
void MixtureModel::SetMinState(
			const int& inMinState)
{
	mMinState = inMinState;
}
			
//---------------------------------------------------------------------------
void MixtureModel::SetStateIndex(
			const IntVector& inStateIndex)
{
	mStateIndex = inStateIndex;
}
			
//---------------------------------------------------------------------------
void MixtureModel::SetNumStates(
			const int& inNumStates)
{
	mNumStates = inNumStates;
}

//---------------------------------------------------------------------------
void MixtureModel::SetNumModels(
			const int& inNumModels)
{
	mNumModels = inNumModels;
}

//---------------------------------------------------------------------------
void MixtureModel::SetRateMatrix3D(
			const DoubleMatrix3D& inRateMatrix3D)
{
	mRateMatrix3D = inRateMatrix3D;
}

//---------------------------------------------------------------------------
void MixtureModel::SetFreqModels(
			const DoubleVector& inFreqModels)
{
	mFreqModels = inFreqModels;
}

//---------------------------------------------------------------------------
void MixtureModel::SetFixedModels(
			const bool& inFixedFreq)
{
	mFixedFreq = inFixedFreq;
}

//---------------------------------------------------------------------------
void MixtureModel::SetUpdateWindowsize(const double& inWindowsize)
{
	mWindowsize = inWindowsize;
}

//---------------------------------------------------------------------------
void MixtureModel::SetModelNumNucleotides(const int& inNucleotides)
{
	mNucleotides = inNucleotides;
}

//---------------------------------------------------------------------------
void MixtureModel::PrintRateMatrix3D(ostream& outStream, ObservedData& theData)
{
	m_oPLEXApp->PrintMatrix3D(outStream, mRateMatrix3D, theData);
}

//---------------------------------------------------------------------------
/* void MixtureModel::PrintDinucleotideModelRateMatrix3D(
			ostream&		outStream, ObservedData& theData)
{
// [jdk]	PLEXApp* m_oPLEXApp = new(PLEXApp);

	StringVector tempStates(mNumStates);
	
	int k = 0;
	
	for (int i = 0; i< mStates.size(); i++)
	{
		for (int j = 0; j < mStates.size(); j++)
		{
			tempStates[k] = mStates[i];
			
			std::string tempstring(1, mStates[j]);
			
			tempStates[k] += tempstring;
			
			k++;
		}
	}

	m_oPLEXApp->PrintMatrix3D(outStream,  mRateMatrix3D, theData);
	
// [jdk]	delete(m_oPLEXApp);
}
*/
//---------------------------------------------------------------------------
void MixtureModel::PrintRateMatrix3DLine(
			ostream&		outStream)
{
// [jdk]	PLEXApp* m_oPLEXApp = new(PLEXApp);

	m_oPLEXApp->PrintMatrix3DLine(outStream, mStates, mRateMatrix3D);
	
// [jdk]	delete(m_oPLEXApp);
}

//---------------------------------------------------------------------------
void MixtureModel::PrintModelAssignments(ostream& outStream, IntVector& m_vecSiteAssignment) {
//[jdk]	PLEXApp* m_oPLEXApp = new(PLEXApp);
	if (outStream)
	{
		outStream << "MIXTURE ASSIGNMENTS:\t";	
	}
	m_oPLEXApp->PrintIntVector(outStream, m_vecSiteAssignment);
	
//[jdk]	delete(m_oPLEXApp);
}


//---------------------------------------------------------------------------
void MixtureModel::PrintFreqVector(
			ostream& 		outStream)
{
// [jdk]	PLEXApp* m_oPLEXApp = new(PLEXApp);

	m_oPLEXApp->PrintFreqVector(outStream, mFreqModels);
	
// [jdk]	delete(m_oPLEXApp);
}


//---------------------------------------------------------------------------
void MixtureModel::PrintRateParameters(ostream&	outStream) {
	// not defined for the general case; override in subclasses [jdk]
}

