//
//  NewickTreeParser.cpp


#include "NewickTreeParser.h"
//#include "TreeParser.cpp"

#pragma mark -
#pragma mark Delimiters
//---------------------------------------------------------------------------
const char*	NewickTreeParser::cDelimiters					= "(,);";
const char*	NewickTreeParser::cQuotes						= "'";
const char*	NewickTreeParser::cTreeTerminator				= ";";
const char*	NewickTreeParser::cNodeSeparator				= ",";
const char*	NewickTreeParser::cOpenParentNodeDelimiter	= "(";
const char*	NewickTreeParser::cCloseParentNodeDelimiter	= ")";
const char*	NewickTreeParser::cWhitespaceChars			= " \t\n\r";

#pragma mark Error messages
const char*	NewickTreeParser::cTerminatorNotFound			= "';' expected to terminate the tree, but not found";
const char*	NewickTreeParser::cMissingCommaErr			= "',' expected to separate nodes, but not found";
const char*	NewickTreeParser::cMultiNodesAtTopErr			= "Unexpected ',' encountered: Only one node allowed at top level";

#pragma mark -
//---------------------------------------------------------------------------
NewickTreeParser::NewickTreeParser(void) {
	this->SkipWhitespace(true);
	this->SetWhitespaceChars(cWhitespaceChars);
	this->SetDelimiters(cDelimiters);
	this->SetQuoteChars(cQuotes);
	this->RemoveQuotes(true);
}

//---------------------------------------------------------------------------

NewickTreeParser::~NewickTreeParser(void)
{
}

//---------------------------------------------------------------------------

bool
NewickTreeParser::Read(
                       std::istream&	inTreeStream,
                       tree_type&		outTree)
{
	node_list	nodes;
	bool success = ReadNodeList(inTreeStream, true, nodes);
    
	if (success)
	{
		std::string token;
		success = this->ReadToken(inTreeStream, token);
		if (success)
		{
			if (token == cTreeTerminator)
			{
				if (nodes.size())
					outTree.SetRoot(nodes[0]);
			}
			else
			{
				success = false;
				this->ReportError(cTerminatorNotFound);
			}
		}
	}
    
	return success;
}

//---------------------------------------------------------------------------

bool
NewickTreeParser::Write(
                        std::ostream&		outTreeStream,
                        const tree_type&	outTree)
{
	return false;
}

//---------------------------------------------------------------------------

bool
NewickTreeParser::ReadNode(
                           std::istream&	inTreeStream,
                           node_type*&		outNode)
{
	node_type*	node = 0;
	
	std::string		token;
	bool success = this->ReadToken(inTreeStream, token);
	if (success)
	{
		if (!this->IsDelimiter(token)) {
			node = this->MakeNode(token);
           // cout << "MAKING NODE WITH " << token << endl;
        }
		else
		{
			if (token == cOpenParentNodeDelimiter)
			{
				node_list children;
				success = ReadNodeList(inTreeStream, false, children);
				if (success)
				{
					node = new node_type;
					if (node) {
						node->AssignChildren(children);
					}
					else success = false;
                    
					success = this->ReadToken(inTreeStream, token);
					if (success)
						success = (token == cCloseParentNodeDelimiter);
                    
					if (success)
						success = this->ReadToken(inTreeStream, token);
                    
					if (success)
					{
						if (this->IsDelimiter(token))
							this->PutBackToken(token);
						else { MakeNode(token, node); }
					}
				}
			}
		}
        
		success = node;
	}
	
	outNode = node;
    
   // exit(-1);
	return success;
}

//---------------------------------------------------------------------------
// A node list is a comma separated list of nodes.

bool
NewickTreeParser::ReadNodeList(
                               std::istream&	inTreeStream,
                               bool			inLimitToOne,
                               node_list&		outNodes)
{
	bool success				= true;
	bool lastTokenWasSeparator	= true;
	
	node_list	nodes;
	
	std::string	token;
	do
	{
		success = this->ReadToken(inTreeStream, token);
		if (success)
		{
			node_type*	node = 0;
			
			if (this->IsDelimiter(token))
			{
				if (token == cOpenParentNodeDelimiter)
				{
					if (lastTokenWasSeparator)
					{
						this->PutBackToken(token);
						success = ReadNode(inTreeStream, node);
						lastTokenWasSeparator = false;
					}
					else
					{
						success = false;
						this->ReportError(cMissingCommaErr);
						break;
					}
				}
				else if (token == cNodeSeparator)
				{
					if (inLimitToOne)
					{
						success = false;
						this->ReportError(cMultiNodesAtTopErr);
						break;
					}
					
					if (lastTokenWasSeparator)
						node = new node_type;
                    
					lastTokenWasSeparator = true;
				}
				else
				{
					this->PutBackToken(token);
					break;
				}
			}
			else
			{
				if (lastTokenWasSeparator)
				{
					this->PutBackToken(token);
					success = ReadNode(inTreeStream, node);
					lastTokenWasSeparator = false;
				}
				else
				{
					success = false;
					this->ReportError(cMissingCommaErr);
					break;
				}
			}
			
			if (success && node)
				nodes.push_back(node);
		}
	} while(success);
	
	if (success)
		outNodes.swap(nodes);
	else
	{
		while(!outNodes.empty())
		{
			delete outNodes.back();
			outNodes.pop_back();
		}
	}
	
	return success;
}
