/*
 *  Numerics.cpp
 *  PLEX
 *
 *  A.P. Jason de Koning, jason.de.koning@gmail.com
 *  http://jasondk.org
 *
 *  David Pollock lab, University of Colorado School of Medicine
 *  http://www.evolutionarygenomics.com
 *  All rights reserved.
 
 */


#include "Numerics.h"
#include "PLEXApp.h"
#include <iostream>

// For SSE2
// #define	MEXP	2281
//#include "dSFMT.c"

#include "dSFMT.h"
#include "dSFMT.c"


//  #include "SFMT.h"
//  #include "SFMT.c"

#define TINY 1.0e-20;

//---------------------------------------------------------------------------
Numerics& Numerics::getGlobalNumerics(void)
{
	static Numerics	globalGenerator;
	return globalGenerator;
}

double Numerics::Random(void) {
//	return globalGenerator.Random();

// Old version
/*	static int blocked = 0;
 static int counter=0;
	if (counter > blockSize) {
		std::cout << "	[Getting new block of " << blockSize << " random numbers]" << endl;
		fill_array64(randArray, blockSize);
		counter=0;
	}
	
	return to_res53(randArray[counter++]);
*/
/*
 static int counter=0;

// New version
	if (counter > blockSize) {
		while (blocked != 0);
		if (blocked == 0) {
			blocked = 1;
			std::cout << "	[Getting new block of " << blockSize << " random numbers]" << endl;
			dsfmt_fill_array_close_open(&dsfmt, randArray, blockSize);
			counter=0;
		}
		blocked = 0;
	}	
	
	// Will this work??
	while (blocked == 1);
	return randArray[counter++]; //to_res53(randArray[counter++]);
*/

	return dsfmt_genrand_close_open(&dsfmt);

}

double Numerics::returnVariateByARSP( int numDists, double *CXY, double *C1, double *Deltas, double mu, double *workingSpace ) {
 	// AUTOMATED REJECTION SAMPLING FROM PRODUCT OF DISTRIBUTIONS (ARSP) UNDER B1u
	// This requires workingSpace be pre-allocated to 2 x numDists...

	int i, reject;
	double *maxLnLs, *MLEs;
	double variate;
	double acceptProb = 0;
	double u;

	maxLnLs = &workingSpace[0];
	MLEs    = &workingSpace[numDists];
	
	for (i=0; i<numDists; i++) {
		// Override
//		Deltas[i] = Deltas[0];		
		MLEs[i]    = MLE_B1u_instRate( (mu - Deltas[i]), C1[i], CXY[i], mu );
		maxLnLs[i] = normalizedLikelihood_B1u_instRate( MLEs[i], (mu - Deltas[i]), C1[i], CXY[i], mu) ;	
	//	printf("Component %d, got MLE of %f (lnL at max is: %f); %f %f %f %f\n", i, MLEs[i], maxLnLs[i], (mu - Deltas[i]), C1[i], CXY[i], mu );
	}
	
	int fn0 = 0;
	double theMax = -std::numeric_limits<double>::max( );
	for (i=0; i<numDists; i++) {
		if ( maxLnLs[i] > theMax ) { fn0 = i; theMax = maxLnLs[i]; }
	}
//	printf ("Best component is %d\n", fn0);
	
	//printf("Sampling loop:\n");
	reject=1;
	
	variate=0;
	acceptProb = 0;
	while (reject > 0) {
		if ( (reject % 100) == 0 ) { std::cout << "		[ WARNING: 100 REJECTIONS FROM AUTOMATED REJECTION SAMPLER ]" << endl; }
		
		variate = betaVariate( (CXY[ fn0 ] + 1), (C1[ fn0 ] +1) ) * Deltas[ fn0 ];
//		printf("getting beta var from fn0: %f\n", variate); 
		
//		acceptProb = 1;
		acceptProb = 0;
		for (i=0; i<numDists; i++) {
			if ( (i != fn0) && ( variate < Deltas[i] ) ) {
				// acceptProb *= ( normalizedLikelihood_B1u_instRate( variate, (mu - Deltas[i]), C1[i], CXY[i], mu ) / maxLnLs[i]);
//				printf("%d: Adding to aceptProb %f, max for this component is %f\n", i, normalizedLikelihood_B1u_instRate( variate, (mu - Deltas[i]), C1[i], CXY[i], mu ), maxLnLs[i] );
				acceptProb += ( normalizedLikelihood_B1u_instRate( variate, (mu - Deltas[i]), C1[i], CXY[i], mu ) - maxLnLs[i]);
			} else if ( variate >= Deltas[i] ) {
				printf ("	PROPOSED VARIATE %f EXCEEDS DELTA (%f)! mu needs to be increased...\n", variate, Deltas[i]);
			}
		}
//		printf( "Proposed %f; acceptance prob is %f (exp is %f)\n", variate, acceptProb, exp(acceptProb));
		u = Random();
		
		if (u <= exp(acceptProb)) {
			reject = 0;
			printf("ACCEPT: %f\n", variate);
		} else {
			reject++;
		//	printf("\tREJECT\n");
		}
	}

	return variate;
}

double Numerics::gammln(double xx) {
	double x,tmp,ser;
	static double cof[6]={76.18009173,-86.50532033,24.01409822,-1.231739516,0.120858003e-2,-0.536382e-5};
	int j;
	
	x=xx-1.0;
	tmp=x+5.5;
	tmp -= (x+0.5)*log(tmp);
	ser=1.0;
	for (j=0;j<=5;j++) {
		x += 1.0;
		ser += cof[j]/x;
	}
	return -tmp+log(2.50662827465*ser);
}

double Numerics::factln(double n) {
//	static double a[101];
	
	if (n < 0){ printf("Err: negative factorial in routine FACTLN"); exit(-1); }
	if (n <= 1) return 0.0;
//	if (n <= 100) return a[n] ? a[n] : (a[n]=gammln(n+1.0));
	else return gammln(n+1.0);
}

double Numerics::LnGamma (double alpha) {
	/* COURTESY OF ZIHENG YANG
	 returns ln(gamma(alpha)) for alpha>0, accurate to 10 decimal places.  
	 Stirling's formula is used for the central polynomial part of the procedure.
	 Pike MC & Hill ID (1966) Algorithm 291: Logarithm of the gamma function.
	 Communications of the Association for Computing Machinery, 9:684
	 */
	double x=alpha, f=0, z;
	
	if (x<7) {
		f=1;  z=x-1;
		while (++z<7)  f*=z;
		x=z;   f=-log(f);
	}
	z = 1/(x*x);
	return  f + (x-0.5)*log(x) - x + .918938533204673 
	+ (((-.000595238095238*z+.000793650793651)*z-.002777777777778)*z
	   +.083333333333333)/x;  
}


double Numerics::betaVariate( double a, double b ) {
	// Modified from Radford Neal
	/* BETA GENERATOR. Generates a real number, r, in (0,1), with density
	 proportional to r^(a-1) * (1-r)^(b-1).  Things are fiddled to avoid
	 the end-points, and to make the procedure symmetric between a and b. */
	
	double x, y, r;
	
	do
	{ 
		x = gammaVariate(a);
		y = gammaVariate(b);
		
		r = 1.0 + x/(x+y);
		r = r - 1.0;
	
		if (r<=0 || r>=1) { std::cout << "got a beta out of bounds! " << r << " (given a,b = " << a << ", " << b << "; x,y = " << x << ", " << y << "; " << (x/(x+y)) << endl; }
	} while (r<=0.0 || r>=1.0);
	
	return r;
}


double Numerics::gammaVariate( double aa ) {
	
	// Modified from Radford Neal
	/* GAMMA GENERATOR.  Generates a positive real number, r, with density
	 proportional to r^(a-1) * exp(-r).  See Devroye, p. 410 and p. 420.
	 Things are fiddled to avoid ever returning a value that is very near
	 zero. */
	
	double b, c, X, Y, Z, U, V, W;
	
	if (aa<0.00001) {
		X = aa;
	}
	
	else if (aa<=1)
	{
		U = Random();
		X = gammaVariate(1+aa) * pow(U,1/aa);
	}
	
	else if (aa<1.00001)
	{ X = expVariate();
	}
	
	else
	{
		b = aa-1;
		c = 3*aa - 0.75;
		
		for (;;)
		{
			U = Random();
			V = Random();
	
			W = U*(1-U);
			Y = sqrt(c/W) * (U-0.5);
			X = b+Y;
			if (X>=0)
			{
				Z = 64*W*W*W*V*V;
				
				if (Z <= 1 - 2*Y*Y/X || log(Z) <= 2 * (b*log(X/b) - Y)) break;
			}
		}
	}
	
	return X<1e-30 && X<aa ? (aa<1e-30 ? aa : 1e-30) : X;
}

double Numerics::expVariate (void)
// Modified from Radford Neal:
// EXPONENTIAL GENERATOR.  See Devroye, p. 29.  Written so as to never return exactly zero.
{
	return -log(Random());
}

double Numerics::factorial(int x) {
	return exp(factln(x));
	// skipping the obvious thing
	int i;
	int fac = 1;
	for (i=2; i<=x; i++) fac *= i;
	return fac;
}

/* double Numerics::normalizedLikelihood_B1u_instRate( double lambdaXY, double LambdaXsp, double C1, double CXY, double mu) {
	double returnVal = pow( (1 - lambdaXY/(mu-LambdaXsp)), C1) * pow( (lambdaXY/(mu - LambdaXsp)), CXY);
	double normConst = ( factorial( (CXY+C1) ) / ( factorial( CXY) * factorial(C1) )) * (1 + CXY + C1) /  (mu - LambdaXsp);
	returnVal *= normConst;

	return returnVal;
} */

double Numerics::normalizedLikelihood_B1u_instRate( double lambdaXY, double LambdaXsp, double C1, double CXY, double mu) {
	// GammaLn version...
	double returnVal = C1 * log((1 - lambdaXY/(mu-LambdaXsp))) + CXY * log( (lambdaXY/(mu - LambdaXsp)) );
	double normConst = ( factln( (CXY+C1) ) - ( factln( CXY) + factln(C1) )) + log((1 + CXY + C1) /  (mu - LambdaXsp) );
	returnVal += normConst;
	
	return returnVal;
}

double Numerics::MLE_B1u_instRate_double( double LambdaXsp1, double LambdaXsp2, double C11, double C12, double CXY1, double CXY2, double mu) {
	double sum0;
	double a,n;
	double DeltaX1, DeltaX2;
	DeltaX1 = (mu - LambdaXsp1);
	DeltaX2 = (mu - LambdaXsp2);
	
	n    = CXY1+CXY2+C11+C12;
	sum0 = C11 * DeltaX1 + C12 * DeltaX2;
	a    = sum0 - DeltaX1*n - DeltaX2*n;
	
	return ( -1/(2*n) * ( a + sqrt( a*a - 4.0 * (CXY1 + CXY2) * DeltaX1 * DeltaX2 * n )	) );
}

void Numerics::simpleSort(double *list, int arraySize) {
	for (int i = 0; i < arraySize ; i++) {
		int potentialSmallest = i; //assume this is smallest
		
		//look over remaining elements to find smallest.
		for (int j = i + 1; j < arraySize; j++) {
			if (list[j] < list[potentialSmallest]) {
				//remember index for latter swap.
				potentialSmallest = j;
			}
		}
		
		//swap smallest remaining element
		double temp = list[i];
		list[i] = list[potentialSmallest];
		list[potentialSmallest] = temp;
	}
}

double Numerics::gsl_stats_quantile_from_sorted_data(const double sorted_data[],const size_t stride,const size_t n,const double f) {
	// From Gnu GSL...
	
	const double index = f * (n - 1) ;
	const size_t lhs = (int)index ;
	const double delta = index - lhs ;
	double result;
	
	if (n == 0)
		return 0.0 ;
	
	if (lhs == n - 1)
    {
		result = sorted_data[lhs * stride] ;
    }
	else
    {
		result = (1 - delta) * sorted_data[lhs * stride] + delta * sorted_data[(lhs + 1) * stride] ;
    }
	
	return result ;
}

double Numerics::MLE_B1u_branchlength( double C1, double C0, double mu) {
    // C1 is the # of Cxx(virtual) + Cxy changes along the branch; C0 is the number of Cxx (no sub) events
	return ( C1 / (C0 * mu) );
}

double Numerics::MLE_B1u_instRate( double LambdaXsp, double C1, double CXY, double mu) {
	return ( (CXY * (mu - LambdaXsp)) / (CXY + C1) );
}

double Numerics::Variance_B1u_instRate( double LambdaXsp, double C1, double CXY, double mu) {
	// variance at the MLE...
	return ( sqrt( ( ( C1 * pow( (mu - LambdaXsp), 2)) / ( pow( (C1+CXY),3)))));
}

double Numerics::Variance_B1u_GivenLambda( double lambdaXY, double LambdaX, double C1, double CXY, double mu) {
	// asymptotic variance is:
	return sqrt( 1.0 / ( (CXY / pow(lambdaXY,2)) + ( C1 / pow( (mu - LambdaX - lambdaXY), 2 ) ) ) );
}

int Numerics::evaluateProposal_logL( double logLikRatio ) {
	return evaluateProposal_logL( 1.0, (1.0 - logLikRatio), 1.0, 1.0, 1.0, 1.0 );
}

int Numerics::evaluateProposal_logL( double newLik, double lastLik ) {
	return evaluateProposal_logL( newLik, lastLik, 1.0, 1.0, 1.0, 1.0 );
}

int Numerics::evaluateProposal_logL( double newLik, double lastLik, double likHeat, double posteriorHeat ) {
	return evaluateProposal_logL( newLik, lastLik, 1.0, 1.0, likHeat, posteriorHeat );
}

int Numerics::evaluateProposal_logL( double newLik, double lastLik, double proposalRatio, double priorRatio, double likHeat, double posteriorHeat ) {
	// Expecting the likelihoods to be input as log-likelihoods
	double acceptanceCutoff;
	
	if (options.__globalLikelihoodTemp != 1.0) {
		likHeat = options.__globalLikelihoodTemp;
	}
	
	if (likHeat != 1.0) {
		acceptanceCutoff = exponential( ( likHeat * (newLik - lastLik) ) );
//		acceptanceCutoff = pow( exponential( (newLik - lastLik) ), likHeat);
	} else {
		acceptanceCutoff = exponential( (newLik - lastLik) );
	}	
	
	acceptanceCutoff *= priorRatio * proposalRatio;

	if (posteriorHeat != 1.0) {
		double tempval = acceptanceCutoff;
		acceptanceCutoff = pow( tempval, posteriorHeat);
	}
	
	if ( acceptanceCutoff >= 1 ) {
		return 1;
	} else {
		double u = Random();

		if (u < acceptanceCutoff ) {
			return 1;
		} else {
			return 0;
		}	
	}
}



// --- Feb 2009: adding stuff for conjugate gibbs sampling -----------------------

int Numerics::NR_ludcmp(double *a, int n, int *indx, double *d) {
	// Adapted from Numerical Recipes
	int i,imax,j,k;
	double big,dum,sum,temp;
	double *vv;
	
	vv = new double[n+1];

//	vv=vector(1,n);
	*d=1.0;
	for (i=1;i<=n;i++) {
		big=0.0;
		for (j=1;j<=n;j++) {
		//	std::cout << "fabs of a[" << i << ", " << j << "] = " << fabs(a[i*n+j]) << endl;
			if ((temp=fabs(a[i*n+j])) > big) big=temp;
		}
		if (big == 0.0) { std::cout << "Singular matrix in routine ludcmp: error doing LU decomposition" << endl; return -1;} //exit(-1); }
		vv[i]=1.0/big;
	}
	
	for (j=1;j<=n;j++) {
		for (i=1;i<j;i++) {
			sum=a[i*n+j];
			for (k=1;k<i;k++) sum -= a[i*n+k]*a[k*n+j];
			a[i*n+j]=sum;
		}
		big=0.0;
		for (i=j;i<=n;i++) {
			sum=a[i*n+j];
			for (k=1;k<j;k++)
				sum -= a[i*n+k]*a[k*n+j];
			a[i*n+j]=sum;
			if ( (dum=vv[i]*fabs(sum)) >= big) {
				big=dum;
				imax=i;
			}
		}

		if (j != imax) {
			for (k=1;k<=n;k++) {
				dum=a[imax*n+k];
				a[imax*n+k]=a[j*n+k];
				a[j*n+k]=dum;
			}
			*d = -(*d);
			vv[imax]=vv[j];
		}
	
		indx[j]=imax;
		if (a[j*n+j] == 0.0) a[j*n+j]=TINY;
		if (j != n) {
			dum=1.0/(a[j*n+j]);
			for (i=j+1;i<=n;i++) a[i*n+j] *= dum;
		}
	}
	
	delete [] vv;
    return 1;
//	free_vector(vv,1,n);
}

void Numerics::NR_lubksb(double *a, int n, int *indx, double *b) {
	// Adapted from Numerical Recipes
	int i,ii=1,ip,j;
	double sum;

	for (i=1;i<=n;i++) {
		ip=indx[i];
		sum=b[ip];
		b[ip]=b[i];
		if (ii)
			for (j=ii;j<=i-1;j++) sum -= a[i*n+j]*b[j];
		else if (sum) ii=i;
		b[i]=sum;
	}
	for (i=n;i>=1;i--) {
		sum=b[i];
		for (j=i+1;j<=n;j++) sum -= a[i*n+j]*b[j];
		b[i]=sum/a[i*n+i];
	}
}

void Numerics::setMode(int val) {
	theMode = val;
}

double Numerics::box_muller(double m, double s)	{
	// Normal random variate generator: mean m, std.dev s

	double x1, x2, w, y1;
	static double y2;
	static int use_last = 0;

	if (use_last) {
		y1 = y2;
		use_last = 0;
	} else {
		do {
			x1 = 2.0 * Numerics::getGlobalNumerics().Random() - 1.0;
			x2 = 2.0 * Numerics::getGlobalNumerics().Random() - 1.0;
			w = x1 * x1 + x2 * x2;
		} while ( w >= 1.0 );

		w = sqrt( (-2.0 * log( w ) ) / w );
		y1 = x1 * w;
		y2 = x2 * w;
		use_last = 1;
	}

	return( m + y1 * s );
}

int Numerics::returnModelIndex( int modelassignment, int rateCatAssignment, int discreteBranchLengthIndex ) {
	/* if ( options.numCatASRV < 1 ) {
	 std::cout << “ERROR! numCatASRV must be >= 1 always” << endl;
	 exit(-1);
	 }
	 
	 if ( options.numDiscreteBLs < 1 ) {
	 std::cout << “ERROR! numCatASRV must be >= 1 always” << endl;
	 exit(-1);
	 } */
	
	return ( modelassignment * ( options.numCatASRV * options.numDiscreteBLs ) + (discreteBranchLengthIndex * options.numCatASRV) + rateCatAssignment );
}
		
			
double Numerics::calculateB1uprob( const DoubleMatrix &theMatrix, double t, int x, int y) {

	// This version gets the optimal mu...
	double mu = 0;
	for (int i=0; i<theMatrix.rows(); i++) { if ( (-1.0 * theMatrix.mRows.mDataArray[i].mDataArray[i]) > mu ) { mu = -1.0 * theMatrix.mRows.mDataArray[i].mDataArray[i]; } }// std::cout << "mu is " << mu << endl;

	// TEMPORARY FIX FOR ZERO RATES IN MTMAM...
	//if (  theMatrix.mRows.mDataArray[x].mDataArray[y] < 0.00001 ) {  theMatrix.mRows.mDataArray[x].mDataArray[y] = 0.00001; }

	double returnVal;
	if ( x != y ) {
		returnVal = theMatrix.mRows.mDataArray[x].mDataArray[y] * t / ( 1 + mu * t );
	} else {
		returnVal = (1 + ( mu + theMatrix.mRows.mDataArray[x].mDataArray[x]) * t) / ( 1 + mu * t );
//		returnVal = 1.0 + ( theMatrix.mRows.mDataArray[x].mDataArray[x] * t ) / (1 + mu * t );
	}
	
	if (returnVal <= 0) { returnVal = 1e-70; }
	
//	std::cout << "Val: " << returnVal << endl;
	return returnVal;
}	


double Numerics::calculateB1uprob( const DoubleMatrix &theMatrix, double t, int x, int y, double mu ) {
	return calculateB1uprob(theMatrix, t, x, y, mu, 1.0);
}
	
double Numerics::calculateB1uprob( const DoubleMatrix &theMatrix, double t, int x, int y, double mu, double rateScaleFactor ) {
	double returnVal;
	
	// TEMPORARY FIX FOR ZERO RATES IN MTMAM...
	//if (  theMatrix.mRows.mDataArray[x].mDataArray[y] < 0.00001 ) {  theMatrix.mRows.mDataArray[x].mDataArray[y] = 0.00001; }
	
	if ( x != y ) {
		returnVal = (theMatrix.mRows.mDataArray[x].mDataArray[y] * rateScaleFactor) * t / ( 1 + mu * t );
	} else {
		returnVal = (1 + ( mu + (theMatrix.mRows.mDataArray[x].mDataArray[x] * rateScaleFactor)) * t) / ( 1 + mu * t );
	}
	
	if (returnVal <= 0) { returnVal = 1e-70; }

//	std::cout << "Val: " << returnVal << endl;
	return returnVal;
}	

double Numerics::calculateB1prob( const DoubleMatrix &theMatrix, double t, int x, int y ) {
	double returnVal;
	
	// Override and use UNIprob...
	//double mu = 0;
	//for (int i=0; i<theMatrix.rows(); i++) { if ( (-1.0 * theMatrix.mRows.mDataArray[i].mDataArray[i]) > mu ) { mu = -1.0 * theMatrix.mRows.mDataArray[i].mDataArray[i]; } }// std::cout << "mu is " << mu << endl;
	//return calculateB1uprob( theMatrix, t, x, y, mu );
	
	if (x != y) {
		if (theMatrix.mRows.mDataArray[x].mDataArray[x] - theMatrix.mRows.mDataArray[y].mDataArray[y] > 0.0001 || theMatrix.mRows.mDataArray[y].mDataArray[y] - theMatrix.mRows.mDataArray[x].mDataArray[x] > 0.0001) {
			returnVal = ( theMatrix.mRows.mDataArray[x].mDataArray[y] * (exponential(theMatrix.mRows.mDataArray[x].mDataArray[x] * t) - exponential(theMatrix.mRows.mDataArray[y].mDataArray[y] * t)) / (theMatrix.mRows.mDataArray[x].mDataArray[x] - theMatrix.mRows.mDataArray[y].mDataArray[y]) );
		} else {
			returnVal = ( theMatrix.mRows.mDataArray[x].mDataArray[y] * (exponential(theMatrix.mRows.mDataArray[x].mDataArray[x] * t) * t ) );
		}
	} else {
		returnVal = ( exponential(theMatrix.mRows.mDataArray[x].mDataArray[x] * t) );
	}
	
	if (returnVal <= 0) { returnVal = 1e-70; }
	return returnVal;
}	

double Numerics::calculateB2prob( const DoubleMatrix &theMatrix, double t, int x, int y, int z ) {
	double returnVal;
//	// Test
//	return 1.0; //calculateB1prob( theMatrix, t, x, y ); 
	
	if (x == y || y == z) {	// only one substitution or no substitution
		return 0.0;
	} else {
		if (theMatrix.mRows.mDataArray[y].mDataArray[y]- theMatrix.mRows.mDataArray[z].mDataArray[z] < 0.0001 || theMatrix.mRows.mDataArray[z].mDataArray[z] - theMatrix.mRows.mDataArray[y].mDataArray[y] < 0.0001)
		// when rate[y] == rate[z]
		{
			if (theMatrix.mRows.mDataArray[x].mDataArray[x]- theMatrix.mRows.mDataArray[z].mDataArray[z] < 0.0001 || theMatrix.mRows.mDataArray[z].mDataArray[z] - theMatrix.mRows.mDataArray[x].mDataArray[x] < 0.0001)
			// when rate[x] = rate[z]
				returnVal = ( theMatrix.mRows.mDataArray[x].mDataArray[y] * theMatrix.mRows.mDataArray[y].mDataArray[z] * exponential(theMatrix.mRows.mDataArray[z].mDataArray[z] * t) * t * t * 0.5 );
			else
			// when rate[x] != rate[z]
				returnVal = (theMatrix.mRows.mDataArray[x].mDataArray[y] * theMatrix.mRows.mDataArray[y].mDataArray[z] * exponential(theMatrix.mRows.mDataArray[z].mDataArray[z] * t) * (exponential((theMatrix.mRows.mDataArray[x].mDataArray[x] - theMatrix.mRows.mDataArray[z].mDataArray[z]) * t) - 1 - t * (theMatrix.mRows.mDataArray[x].mDataArray[x] - theMatrix.mRows.mDataArray[z].mDataArray[z])) / pow(theMatrix.mRows.mDataArray[x].mDataArray[x] - theMatrix.mRows.mDataArray[z].mDataArray[z], 2) );
		}
		else
		{
			// when rate[y] != rate[z]
			if (theMatrix.mRows.mDataArray[x].mDataArray[x]- theMatrix.mRows.mDataArray[y].mDataArray[y] < 0.0001 || theMatrix.mRows.mDataArray[y].mDataArray[y] - theMatrix.mRows.mDataArray[x].mDataArray[x] < 0.0001)
				// when rate[x] == rate[y]
				returnVal = (theMatrix.mRows.mDataArray[x].mDataArray[y] * theMatrix.mRows.mDataArray[y].mDataArray[z] * t * exponential(theMatrix.mRows.mDataArray[y].mDataArray[y] * t) / (theMatrix.mRows.mDataArray[y].mDataArray[y] - theMatrix.mRows.mDataArray[z].mDataArray[z]) -theMatrix.mRows.mDataArray[x].mDataArray[y] * theMatrix.mRows.mDataArray[y].mDataArray[z] * (exponential(theMatrix.mRows.mDataArray[x].mDataArray[x] * t)-exponential(theMatrix.mRows.mDataArray[z].mDataArray[z] * t))/((theMatrix.mRows.mDataArray[y].mDataArray[y] - theMatrix.mRows.mDataArray[z].mDataArray[z])*(theMatrix.mRows.mDataArray[x].mDataArray[x] - theMatrix.mRows.mDataArray[z].mDataArray[z])));
			else
			{
				// rate[x] != rate[y]
				if ((theMatrix.mRows.mDataArray[x].mDataArray[x]- theMatrix.mRows.mDataArray[z].mDataArray[z] < 0.0001 || theMatrix.mRows.mDataArray[z].mDataArray[z] - theMatrix.mRows.mDataArray[x].mDataArray[x] < 0.0001))
					// rate[x] = rate[z]
					returnVal = (theMatrix.mRows.mDataArray[x].mDataArray[y] * theMatrix.mRows.mDataArray[y].mDataArray[z] * (exponential(theMatrix.mRows.mDataArray[x].mDataArray[x] * t)-exponential(theMatrix.mRows.mDataArray[y].mDataArray[y] * t))/((theMatrix.mRows.mDataArray[y].mDataArray[y] - theMatrix.mRows.mDataArray[z].mDataArray[z])*(theMatrix.mRows.mDataArray[x].mDataArray[x] - theMatrix.mRows.mDataArray[y].mDataArray[y])) - theMatrix.mRows.mDataArray[x].mDataArray[y] * theMatrix.mRows.mDataArray[y].mDataArray[z] * t * exponential(theMatrix.mRows.mDataArray[z].mDataArray[z] * t) / (theMatrix.mRows.mDataArray[y].mDataArray[y] -theMatrix.mRows.mDataArray[z].mDataArray[z]));
				else
					// rate[x] != rate[z]
					returnVal = (theMatrix.mRows.mDataArray[x].mDataArray[y] * theMatrix.mRows.mDataArray[y].mDataArray[z] * ((theMatrix.mRows.mDataArray[y].mDataArray[y] - theMatrix.mRows.mDataArray[z].mDataArray[z]) * exponential(theMatrix.mRows.mDataArray[x].mDataArray[x] * t) + (theMatrix.mRows.mDataArray[z].mDataArray[z] - theMatrix.mRows.mDataArray[x].mDataArray[x]) * exponential(theMatrix.mRows.mDataArray[y].mDataArray[y] * t) + (theMatrix.mRows.mDataArray[x].mDataArray[x] - theMatrix.mRows.mDataArray[y].mDataArray[y]) * exponential(theMatrix.mRows.mDataArray[z].mDataArray[z] * t))/((theMatrix.mRows.mDataArray[y].mDataArray[y] - theMatrix.mRows.mDataArray[z].mDataArray[z]) * (theMatrix.mRows.mDataArray[x].mDataArray[x] - theMatrix.mRows.mDataArray[y].mDataArray[y]) * (theMatrix.mRows.mDataArray[x].mDataArray[x] - theMatrix.mRows.mDataArray[z].mDataArray[z])));
			}
		}
	}
	
	if (returnVal <= 0) { returnVal = 1e-70; }
	return returnVal;
}	

void Numerics::initLookup(int theN) {
	n = theN;
	logFloor = log(1e-70);
	logLookupTable = new double[n];
	expLookupTable = new double[n];
	precomputeLogTable(logLookupTable);
	precomputeExpTable(expLookupTable);
	theMode = 1;
}
#include <stdio.h>
Numerics::Numerics(void) {
	initLookup(1000000);
	printf("Test.  num->exponential(1) = %f, expo(1) = %f\n", exp(1.0), exponential(1.0));

	// Allocate mem for SFMT...
	/* size = get_min_array_size64();
    if (size < R_SIZE) {
	size = R_SIZE;
    } */


	// Ignore the above, use my blocksize...
	#if defined(__APPLE__) || \
		(defined(__FreeBSD__) && __FreeBSD__ >= 3 && __FreeBSD__ <= 6)
		printf("malloc used\n");
		randArray = (double *)malloc(sizeof(double) * blockSize);
		if (randArray == NULL) {
		printf("can't allocate memory.\n");
		exit(-1);
		}
	#elif defined(_POSIX_C_SOURCE)
		printf("posix_memalign used\n");
		if (posix_memalign((void **)&randArray, 16, sizeof(double) * blockSize) != 0) {
		printf("can't allocate memory.\n");
		exit(-1);
		}
	#elif defined(__GNUC__) && (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 3))
		printf("memalign used\n");
		randArray = memalign(16, sizeof(double) * blockSize);
		if (randArray == NULL) {
		printf("can't allocate memory.\n");
		exit(-1);
		}
	#else // in this case, gcc doesn't support SSE2
		printf("malloc used\n");
		randArray = malloc(sizeof(double) * blockSize);
		if (randArray == NULL) {
		printf("can't allocate memory.\n");
		exit(-1);
		}
	#endif
	std::cout << "	Getting initial block of random numbers (dSFMT) using seed of " << options.randseed << "...";
//    init_gen_rand(options.randseed);
//    fill_array64(randArray, blockSize);

	dsfmt_init_gen_rand(&dsfmt, (int)options.randseed);
	dsfmt_fill_array_close_open(&dsfmt, randArray, blockSize);
	std::cout << "DONE." << endl;
	
	
	//dsfmt_init_gen_rand(&dsfmt, options.randseed);

}


void Numerics::precomputeLogTable(double *LUT) { 
	min = 0.0000000001;
	max = 5.000000000;
	incr = (max-min)/n;
	
	for(int i=0;i<n;i++) { 
		LUT[i] = log((min + (i*incr))); //save the log value 
	//	std::cout << "Log precomputed for " << (min + (i*incr)) << " = " << logLookupTable[i] << endl;
	}
	
//	std::cout << "TEST" << endl;
//	for (int i=0;i<n;i++) {
//		std::cout << "REAL LOG: " << log( (min + (i*incr) + (incr/2) ) ) << "\tAPPROX: " << logarithm( (min+(i*incr)+(incr/2) )) << endl;
//	}
} 

void Numerics::precomputeExpTable(double *LUT) { 
	expMin = -5.00000;
	expMax = 5.000000;
	expIncr = (expMax-expMin)/n;
	
	for(int i=0;i<n;i++) { 
		LUT[i] = exp((expMin + (i*expIncr))); //save the log value 
	//	std::cout << "Log precomputed for " << (min + (i*incr)) << " = " << logLookupTable[i] << endl;
	}
	
//	std::cout << "TEST" << endl;
//	for (int i=0;i<n;i++) {
//		std::cout << "REAL LOG: " << log( (min + (i*incr) + (incr/2) ) ) << "\tAPPROX: " << logarithm( (min+(i*incr)+(incr/2) )) << endl;
//	}
} 
//float logTable[15];
//fill_icsi_log_table(15, &logTable);


/* Computes an approximation of log(val) quickly. 
val is a IEEE 754 float value, must be >0. 
logLookupTable and n must be the same values as 
provided to fill_icsi_table. 
returns: log(val). No input checking performed. 
*/ 
//inline float Numerics::logarithm(register float val) { 
double Numerics::logarithm(double val) { 
	if (theMode==0) { return log(val); }
	if (val == 1e-70) { return logFloor;}
	if ((val > max) || (val < min)) { return log(val); } //std::cout << "WARNING: LOG LUT EXCEEDED for " << val << std::endl; 
	static int x1,x2;
	static double y1,y2, theX1, theX2;
	x1 = static_cast <int>( (val-min)/incr );
	x2 = x1+1;
	theX1 = min + (x1*incr);
	theX2 = min + (x2*incr);
	
	y1 = logLookupTable[x1]; y2 = logLookupTable[x2];
	// y2 is at (min + (x2*incr))
	// val is between x1 and x2
//	std::cout << "requested fxn for " << val << "; closest is " << theX1 << ", and " << theX2 << endl;
//	std::cout << "fxn values for x1 and x2 are " << y1 << " and " << y2 << endl;
//	exit(1);
	return y1 + (val-theX1)*(y2-y1)/(theX2-theX1);
} 

double Numerics::exponential(double val) { 
	if (theMode==0) { return exp(val); }
	if ((val > expMax) || (val < expMin)) { return exp(val); } //std:: cout << "EXP BOUND " << val << endl;
	static int x1,x2;
	static double y1,y2, theX1, theX2;
	x1 = static_cast <int>( (val-expMin)/expIncr );
	x2 = x1+1;
	theX1 = expMin + (x1*expIncr);
	theX2 = expMin + (x2*expIncr);
	
		y1 = expLookupTable[x1]; y2 = expLookupTable[x2];
	// y2 is at (min + (x2*incr))
	// val is between x1 and x2
//	std::cout << "requested fxn for " << val << "; closest is " << theX1 << ", and " << theX2 << endl;
//	std::cout << "fxn values for x1 and x2 are " << y1 << " and " << y2 << endl;
//	exit(1);
	return y1 + (val-theX1)*(y2-y1)/(theX2-theX1);
} 

int Numerics::PAML_PMatUVRoot (DoubleMatrix *P, double t, int n, double U[], double V[], double Root[])
{
/* P(t) = U * exp{Root*t} * V
*/
   int i,j,k;
    double expt, uexpt; //, *pP;
   double smallp = 0;
   if (t<0) { t = 0.0; }

	for (i=0;i<n;i++) {
		for (j=0;j<n;j++) {
			P->mRows.mDataArray[i].mDataArray[j] = 0.0;
		}
	}
	
//   NPMatUVRoot++;
//   if (t<-0.1) std::cout << "\nt = " << t << "in PMatUVRoot" << std::endl;
   if (t<1e-100) {
		for (i=0;i<n;i++) {
			P->mRows.mDataArray[i].mDataArray[i] = 1.0;
		}
		return(0);
	}
	

	//,pP=P
//	std::cout << "n is " << n << endl;
   for (k=0; k<n; k++) {
      for (i=0,expt=exp(t*Root[k]); i<n; i++) {
//		std::cout << "Root[k] = " << Root[k] << endl;
         for (j=0,uexpt=U[i*n+k]*expt; j<n; j++) {
//			std::cout << "	incr is " << uexpt*V[k*n+j] << endl;
			P->mRows.mDataArray[i].mDataArray[j] += uexpt*V[k*n+j]; //*pP++ += uexpt*V[k*n+j];
		 }
	  }
   }
			
  for(i=0;i<n;i++) {
	for(j=0;j<n;j++) {
		if ( P->mRows.mDataArray[i].mDataArray[j]<smallp) P->mRows.mDataArray[i].mDataArray[j]=0;
//		std::cout << setprecision(2) << P->mRows.mDataArray[i].mDataArray[j] << "\t";
	} //std::cout << endl;
  } //std::cout << endl;
  
//		<smallp) P->mRows.mDataArray[i].mDataArray[j]=0;

   return (0);
}

/*
int Numerics::PAML_PMatQRev(double Q[], double pi[], double t, int n, double *U, double *V, double *Root)
{
// This calculates P(t) = exp(Q*t), where Q is the rate matrix for a 
//   time-reversible Markov process.

//   Q[] or P[] has the rate matrix as input, and P(t) in return.
//   space[n*n*2+n*2]

   PAML_eigenQREV(Q, pi, n, Root, U, V);
   PAML_PMatUVRoot(Q, t, n, U, V, Root);
   return(0);
} */

int Numerics::PAML_eigenQREV (double Q[], double pi[], int n, 
               double Root[], double U[], double V[])
{
/* 
   This finds the eigen solution of the rate matrix Q for a time-reversible 
   Markov process, using the algorithm for a real symmetric matrix.
   Rate matrix Q = S * diag{pi} = U * diag{Root} * V, 
   where S is symmetrical, all elements of pi are positive, and U*V = I.
   space[n] is for storing sqrt(pi).

   [U 0] [Q_0 0] [U^-1 0]    [Root  0]
   [0 I] [0   0] [0    I]  = [0     0]

   Ziheng Yang, 25 December 2001 (ref is CME/eigenQ.pdf)
*/

   int i,j, inew, jnew, nnew, status;
   static double *pi_sqrt;
   double small=1e-100;
   static int allocated = 0;
   
	if (allocated==0) {
		pi_sqrt = new double[n];
		allocated=1;
	}
   for(j=0,nnew=0; j<n; j++)
      if(pi[j]>small)
         pi_sqrt[nnew++]=sqrt(pi[j]);

   /* store in U the symmetrical matrix S = sqrt(D) * Q * sqrt(-D) */
		
   if(nnew==n) {
      for(i=0; i<n; i++)
         for(j=0,U[i*n+i] = Q[i*n+i]; j<i; j++)
            U[i*n+j] = U[j*n+i] = (Q[i*n+j] * pi_sqrt[i]/pi_sqrt[j]);

      status=PAML_eigenRealSym(U, n, Root, V);
      for(i=0;i<n;i++) for(j=0;j<n;j++)  V[i*n+j] = U[j*n+i] * pi_sqrt[j];
      for(i=0;i<n;i++) for(j=0;j<n;j++)  U[i*n+j] /= pi_sqrt[i];
   }
   else {
      for(i=0,inew=0; i<n; i++) {
         if(pi[i]>small) {
            for(j=0,jnew=0; j<i; j++) 
               if(pi[j]>small) {
                  U[inew*nnew+jnew] = U[jnew*nnew+inew] 
                                    = Q[i*n+j] * pi_sqrt[inew]/pi_sqrt[jnew];
                  jnew++;
               }
            U[inew*nnew+inew] = Q[i*n+i];
            inew++;
         }
      }

      status=PAML_eigenRealSym(U, nnew, Root, V);
	  
      for(i=n-1,inew=nnew-1; i>=0; i--)   /* construct Root */
         Root[i] = (pi[i]>small ? Root[inew--] : 0);
      for(i=n-1,inew=nnew-1; i>=0; i--) {  /* construct V */
         if(pi[i]>small) {
            for(j=n-1,jnew=nnew-1; j>=0; j--)
               if(pi[j]>small) {
                  V[i*n+j] = U[jnew*nnew+inew]*pi_sqrt[jnew];
                  jnew--;
               }
               else 
                  V[i*n+j] = (i==j);
            inew--;
         }
         else 
            for(j=0; j<n; j++)  V[i*n+j] = (i==j);
      }
      for(i=n-1,inew=nnew-1; i>=0; i--) {  /* construct U */
         if(pi[i]>small) {
            for(j=n-1,jnew=nnew-1;j>=0;j--)
               if(pi[j]>small) {
                  U[i*n+j] = U[inew*nnew+jnew]/pi_sqrt[inew];
                  jnew--;
               }
               else 
                  U[i*n+j] = (i==j);
            inew--;
         }
         else 
            for(j=0;j<n;j++)
               U[i*n+j] = (i==j);
      }
   }

/*   This routine works on P(t) as well as Q. */
/*
   if(fabs(Root[0])>1e-10 && noisy) printf("Root[0] = %.5e\n",Root[0]);
   Root[0]=0; 
*/
   return(status);
}

/* eigen solution for real symmetric matrix */
int Numerics::PAML_eigenRealSym(double A[], int n, double Root[], double work[])
{
/* This finds the eigen solution of a real symmetrical matrix A[n*n].  In return, 
   A has the right vectors and Root has the eigenvalues. 
   work[n] is the working space.
   The matrix is first reduced to a tridiagonal matrix using HouseholderRealSym(), 
   and then using the QL algorithm with implicit shifts.  

   Adapted from routine tqli in Numerical Recipes in C, with reference to LAPACK
   Ziheng Yang, 23 May 2001
*/
   int status=0;
   PAML_HouseholderRealSym(A, n, Root, work);
   status=PAML_EigenTridagQLImplicit(Root, work, n, A);
   PAML_EigenSort(Root, A, n);

   return(status);
}


void Numerics::PAML_EigenSort(double d[], double U[], int n)
{
/* this sorts the eigen values d[] and rearrange the (right) eigen vectors U[]
*/
   int k,j,i;
   double p;

   for (i=0;i<n-1;i++) {
      p=d[k=i];
      for (j=i+1;j<n;j++)
         if (d[j] >= p) p=d[k=j];
      if (k != i) {
         d[k]=d[i];
         d[i]=p;
         for (j=0;j<n;j++) {
            p=U[j*n+i];
            U[j*n+i]=U[j*n+k];
            U[j*n+k]=p;
         }
      }
   }
}

void Numerics::PAML_HouseholderRealSym(double a[], int n, double d[], double e[])
{
/* This uses HouseholderRealSym transformation to reduce a real symmetrical matrix 
   a[n*n] into a tridiagonal matrix represented by d and e.
   d[] is the diagonal (eigends), and e[] the off-diagonal.
*/
   int m,k,j,i;
   double scale,hh,h,g,f;

   for (i=n-1;i>=1;i--) {
      m=i-1;
      h=scale=0;
      if (m > 0) {
         for (k=0;k<=m;k++)
            scale += fabs(a[i*n+k]);
         if (scale == 0)
            e[i]=a[i*n+m];
         else {
            for (k=0;k<=m;k++) {
               a[i*n+k] /= scale;
               h += a[i*n+k]*a[i*n+k];
            }
            f=a[i*n+m];
            g=(f >= 0 ? -sqrt(h) : sqrt(h));
            e[i]=scale*g;
            h -= f*g;
            a[i*n+m]=f-g;
            f=0;
            for (j=0;j<=m;j++) {
               a[j*n+i]=a[i*n+j]/h;
               g=0;
               for (k=0;k<=j;k++)
                  g += a[j*n+k]*a[i*n+k];
               for (k=j+1;k<=m;k++)
                  g += a[k*n+j]*a[i*n+k];
               e[j]=g/h;
               f += e[j]*a[i*n+j];
            }
            hh=f/(h*2);
            for (j=0;j<=m;j++) {
               f=a[i*n+j];
               e[j]=g=e[j]-hh*f;
               for (k=0;k<=j;k++)
                  a[j*n+k] -= (f*e[k]+g*a[i*n+k]);
            }
         }
      } 
      else
         e[i]=a[i*n+m];
      d[i]=h;
   }
   d[0]=e[0]=0;

   /* Get eigenvectors */
   for (i=0;i<n;i++) {
      m=i-1;
      if (d[i]) {
         for (j=0;j<=m;j++) {
            g=0;
            for (k=0;k<=m;k++)
               g += a[i*n+k]*a[k*n+j];
            for (k=0;k<=m;k++)
               a[k*n+j] -= g*a[k*n+i];
         }
      }
      d[i]=a[i*n+i];
      a[i*n+i]=1;
      for (j=0;j<=m;j++) a[j*n+i]=a[i*n+j]=0;
   }
}

#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))

int Numerics::PAML_EigenTridagQLImplicit(double d[], double e[], int n, double z[])
{
/* This finds the eigen solution of a tridiagonal matrix represented by d and e.  
   d[] is the diagonal (eigenvalues), e[] is the off-diagonal
   z[n*n]: as input should have the identity matrix to get the eigen solution of the 
   tridiagonal matrix, or the output from HouseholderRealSym() to get the 
   eigen solution to the original real symmetric matrix.
   z[n*n]: has the orthogonal matrix as output

   Adapted from routine tqli in Numerical Recipes in C, with reference to
   LAPACK fortran code.
   Ziheng Yang, May 2001
*/
   int m,j,iter,niter=30, status=0, i,k;
   double s,r,p,g,f,dd,c,b, aa,bb;

   for (i=1;i<n;i++) e[i-1]=e[i];  e[n-1]=0;
   for (j=0;j<n;j++) {
      iter=0;
      do {
         for (m=j;m<n-1;m++) {
            dd=fabs(d[m])+fabs(d[m+1]);
            if (fabs(e[m])+dd == dd) break;  /* ??? */
         }
         if (m != j) {
            if (iter++ == niter) {
               status=-1;
               break;
            }
            g=(d[j+1]-d[j])/(2*e[j]);

            /* r=pythag(g,1); */

            if((aa=fabs(g))>1)  r=aa*sqrt(1+1/(g*g));
            else                r=sqrt(1+g*g);

            g=d[m]-d[j]+e[j]/(g+SIGN(r,g));
            s=c=1;
            p=0;
            for (i=m-1;i>=j;i--) {
               f=s*e[i];
               b=c*e[i];

               /*  r=pythag(f,g);  */
               aa=fabs(f); bb=fabs(g);
               if(aa>bb)       { bb/=aa;  r=aa*sqrt(1+bb*bb); }
               else if(bb==0)             r=0;
               else            { aa/=bb;  r=bb*sqrt(1+aa*aa); }

               e[i+1]=r;
               if (r == 0) {
                  d[i+1] -= p;
                  e[m]=0;
                  break;
               }
               s=f/r;
               c=g/r;
               g=d[i+1]-p;
               r=(d[i]-g)*s+2*c*b;
               d[i+1]=g+(p=s*r);
               g=c*r-b;
               for (k=0;k<n;k++) {
                  f=z[k*n+i+1];
                  z[k*n+i+1]=s*z[k*n+i]+c*f;
                  z[k*n+i]=c*z[k*n+i]-s*f;
               }
            }
            if (r == 0 && i >= j) continue;
            d[j]-=p; e[j]=g; e[m]=0;
         }
      } while (m != j);
   }
   return(status);
}

// Minimizer from NR...
#include <math.h>
#include "nrutil.h"
#define ITMAX 400
#define EPS 3.0e-8
#define TOLX (4*EPS)
#define STPMX 100.0
#define LS_TOLX 1.0e-7
#define ALF 1.0e-4

void Numerics::NR_lnsrch(int n, double xold[], double fold, double g[], double p[], double x[],
	double *f, double stpmax, int *check, double (*func)(double [], int n))
{
	int i;
	double a,alam,alam2,alamin,b,disc,f2,fold2,rhs1,rhs2,slope,sum,temp,
		test,tmplam;

	*check=0;
	for (sum=0.0,i=1;i<=n;i++) sum += p[i]*p[i];
	sum=sqrt(sum);
	if (sum > stpmax)
		for (i=0;i<n;i++) p[i] *= stpmax/sum;
	for (slope=0.0,i=0;i<n;i++)
		slope += g[i]*p[i];
	test=0.0;
	for (i=0;i<n;i++) {
		temp=fabs(p[i])/DMAX(fabs(xold[i]),1.0);
		if (temp > test) test=temp;
	}
	alamin=LS_TOLX/test;
	alam=1.0;
	for (;;) {
		for (i=0;i<n;i++) x[i]=xold[i]+alam*p[i];
		*f=(*func)(x,n);
	//	printf("Got fxn eval: %f\n", *f);
		if (alam < alamin) {
			for (i=0;i<n;i++) x[i]=xold[i];
			*check=1;
			return;
		} else if (*f <= fold+ALF*alam*slope) return;
		else {
			if (alam == 1.0)
				tmplam = -slope/(2.0*(*f-fold-slope));
			else {
				rhs1 = *f-fold-alam*slope;
				rhs2=f2-fold2-alam2*slope;
				a=(rhs1/(alam*alam)-rhs2/(alam2*alam2))/(alam-alam2);
				b=(-alam2*rhs1/(alam*alam)+alam*rhs2/(alam2*alam2))/(alam-alam2);
				if (a == 0.0) tmplam = -slope/(2.0*b);
				else {
					disc=b*b-3.0*a*slope;
					//if (disc<0.0) nrerror("Roundoff problem in lnsrch.");
					//else 
					tmplam=(-b+sqrt(disc))/(3.0*a);
				}
				if (tmplam>0.5*alam)
					tmplam=0.5*alam;
			}
		}
		alam2=alam;
		f2 = *f;
		fold2=fold;
		alam=DMAX(tmplam,0.1*alam);
	}
}

int Numerics::PAML_gradient(int n, double x[], double f0, double g[], double (*fun)(double x[],int n), double space[], int Central) {
/*  f0 = fun(x) is always given.
*/
   int i,j;
   double *x0=space, *x1=space+n, eh0=1e-8, eh;  /* 1e-7 */

   if (Central) {
      for(i=0; i<n; i++)  {
         for(j=0; j<n; j++) 
            x0[j]=x1[j]=x[j];
         eh=pow(eh0*(fabs(x[i])+1), 0.67);
         x0[i]-=eh; x1[i]+=eh;
         g[i] = ((*fun)(x1,n) - (*fun)(x0,n))/(eh*2.0);
      }
   }
   else {
      for(i=0; i<n; i++)  {
         for(j=0; j<n; j++)
            x1[j]=x[j];
         eh=eh0*(fabs(x[i])+1);
         x1[i]+=eh;
         g[i] = ((*fun)(x1,n)-f0)/eh;
      }
   }
   return(0);
}

#define FREEALL free_vector(xi,1,n);free_vector(pnew,1,n); free_matrix(hessin,1,n,1,n);free_vector(hdg,1,n);free_vector(g,1,n); free_vector(dg,1,n);
#define FREEALL2 free(xi);free(pnew); free(hdg);free(g); free(dg); free(hessin);

void Numerics::NR_dfpmin(double p[], int n, double gtol, int *iter, double *fret,
	double(*func)(double [], int np), void (*dfunc)(double [], int, double [])) {
	
	//void lnsrch(int n, double xold[], double fold, double g[], double p[], double x[],
	//	 double *f, double stpmax, int *check, double (*func)(double [], int np));
	int check,i,its,j;
	double den,fac,fad,fae,fp,stpmax,sum=0.0,sumdg,sumxi,temp,test;
	static double *dg,*g,*hdg,*hessin,*pnew,*xi;
	static int allocd = 0;
	
	if (allocd == 0) {
		std::cout << "Allocating memory... (n is " << n << ")" << endl;
		
		dg = new double[n]; for (int i=0; i<n; i++) { dg[i] = 1.0; } //memset( dg, 1.0, n) ; //malloc(sizeof(double)*n);  //vector(1,n);
		g =new double[n]; for (int i=0; i<n; i++) { g[i] = 1.0; } // malloc(sizeof(double)*n); memset( g, 1.0, n) ; //vector(1,n);
		hdg = new double[n]; for (int i=0; i<n; i++) { hdg[i] = 1.0; } //=malloc(sizeof(double)*n); memset( hdg, 1.0, n) ; //vector(1,n);
		pnew = new double[n]; for (int i=0; i<n; i++) { pnew[i] = 1.0; } //=malloc(sizeof(double)*n); memset( pnew, 1.0, n) ; //vector(1,n);
		xi = new double[n]; for (int i=0; i<n; i++) { xi[i] = 1.0; } //malloc(sizeof(double)*n); memset( xi, 1.0, n) ; //vector(1,n);
		hessin = new double[n*n]; for (int i=0; i<n*n; i++) { hessin[i] = 1.0; } //(sizeof(double*)*n*n);
		std::cout << "Done." << endl;
		allocd = 1;
	}
	
	printf("	Beginning optimization of loss function...\n");
	fp=(*func)(p,n);
	printf("	[Initial fxn eval: %f]\n", fp);
	
//	printf("Gradient eval...\n");
	(*dfunc)(p,n,g);
//	printf("Returned.\n");
//	for (i=0;i<n;i++) {
//		printf("Grad %d = %f\n", i, g[i]);
//	}
	
	for (i=0;i<n;i++) {
		for (j=0;j<n;j++) hessin[i*n+j]=0.0;
		hessin[i*n+i]=1.0;
		xi[i] = -g[i];
		sum += p[i]*p[i];
	}
	stpmax=STPMX*DMAX(sqrt(sum),(double)n);
	for (its=1;its<=ITMAX;its++) {
		printf("	Iteration %d...\n", its);
		
		*iter=its;
		NR_lnsrch(n,p,fp,g,xi,pnew,fret,stpmax,&check,func);
		fp = *fret;
		for (i=0;i<n;i++) {
			xi[i]=pnew[i]-p[i]; //printf("Coord diff after linesrc %d -> %f\n", i, xi[i]);
			p[i]=pnew[i];
		}
		test=0.0;
		for (i=0;i<n;i++) {
			temp=fabs(xi[i])/DMAX(fabs(p[i]),1.0);
			if (temp > test) test=temp;
		}
		if (test < TOLX) {
		//	FREEALL2
			return;
		}
		for (i=0;i<n;i++) dg[i]=g[i];
		(*dfunc)(p,n,g);
		test=0.0;
		den=DMAX(*fret,1.0);
		for (i=0;i<n;i++) {
			temp=fabs(g[i])*DMAX(fabs(p[i]),1.0)/den;
			if (temp > test) test=temp;
		}
		if (test < gtol) {
		//	FREEALL2
			return;
		}
		for (i=0;i<n;i++) dg[i]=g[i]-dg[i];
		for (i=0;i<n;i++) {
			hdg[i]=0.0;
			for (j=0;j<n;j++) hdg[i] += hessin[i*n+j]*dg[j];
		}
		fac=fae=sumdg=sumxi=0.0;
		for (i=0;i<n;i++) {
			fac += dg[i]*xi[i];
			fae += dg[i]*hdg[i];
			sumdg += SQR(dg[i]);
			sumxi += SQR(xi[i]);
		}
		if (fac*fac > EPS*sumdg*sumxi) {
			fac=1.0/fac;
			fad=1.0/fae;
			for (i=0;i<n;i++) dg[i]=fac*xi[i]-fad*hdg[i];
			for (i=0;i<n;i++) {
				for (j=0;j<n;j++) {
					hessin[i*n+j] += fac*xi[i]*xi[j]
					-fad*hdg[i]*hdg[j]+fae*dg[i]*dg[j];
				}
			}
		}
		for (i=0;i<n;i++) {
			xi[i]=0.0;
			for (j=0;j<n;j++) xi[i] -= hessin[i*n+j]*g[j];
		}
	}
	std::cout << "Error in dfpmin - too many iterations" << endl;
	//nrerror("too many iterations in dfpmin");
//	FREEALL2
}
/* (C) Copr. 1986-92 Numerical Recipes Software ?421.1-9. */

#undef SIGN
