/*
 *  Numerics.h
 *  PLEX
 *
 *  A.P. Jason de Koning, jason.de.koning@gmail.com
 *  http://jasondk.org
 *
 *  David Pollock lab, University of Colorado School of Medicine
 *  http://www.evolutionarygenomics.com
 *  All rights reserved.
 
 */


#ifndef Numerics_h__
#define Numerics_h__

#include <math.h>
#include "CLAMPTypes.h"
#include "RandomGenerator.h"


#define	HAVE_SSE2
#define _XOPEN_SOURCE 600
#include "dSFMT.h"

class Numerics {
	public:
		Numerics(void);
		double Random(void);
		static Numerics& getGlobalNumerics(void);
		void precomputeLogTable(double *logLookupTable);
		void precomputeExpTable(double *expLookupTable);
		double logarithm( double val);
		double exponential(double y);
		
		double calculateB1uprob( const DoubleMatrix &theMatrix, double t, int x, int y);
		double calculateB1uprob( const DoubleMatrix &theMatrix, double t, int x, int y, double mu );
		double calculateB1uprob( const DoubleMatrix &theMatrix, double t, int x, int y, double mu, double rateScaleFactor );

		double calculateB1prob( const DoubleMatrix &theMatrix, double t, int x, int y ) ;
		double calculateB2prob( const DoubleMatrix &theMatrix, double t, int x, int y, int z ) ;
		double box_muller(double m, double s);
		void setMode(int val);
	
		// Stuff for Conjugate gibbs sampling (Feb 2009)...
		double LnGamma (double alpha);
		double factln(double n);
		double gammln(double xx);

		void   simpleSort( double *list, int listSize);
	
		int returnModelIndex( int modelassignment, int rateCatAssignment, int discreteBranchLengthIndex );
	
		double gsl_stats_quantile_from_sorted_data(const double sorted_data[],const size_t stride,const size_t n,const double f);
		double betaVariate( double a, double b );	
		double gammaVariate( double a );
		double expVariate(void);
		double factorial(int x);
		double normalizedLikelihood_B1u_instRate( double lambdaXY, double LambdaXsp, double C1, double CXY, double mu);
        double MLE_B1u_branchlength( double C1, double C0, double mu);
		double MLE_B1u_instRate( double LambdaXsp, double C1, double CXY, double mu);
		double Variance_B1u_instRate( double LambdaXsp, double C1, double CXY, double mu);
		double Variance_B1u_GivenLambda( double lambdaXY, double LambdaXsp, double C1, double CXY, double mu);
		double MLE_B1u_instRate_double( double LambdaXsp1, double LambdaXsp2, double C11, double C12, double CXY1, double CXY2, double mu);

		int evaluateProposal_logL( double logLikRatio );
		int evaluateProposal_logL( double newLik, double lastLik );
		int evaluateProposal_logL( double newLik, double lastLik, double likHeat, double posteriorHeat );
		int evaluateProposal_logL( double newLik, double lastLik, double proposalRatio, double priorRatio, double likHeat, double posteriorHeat );
	
		// end
	
		int NR_ludcmp(double *a, int n, int *indx, double *d);
		void NR_lubksb(double *a, int n, int *indx, double b[]);
		void NR_dfpmin(double p[], int n, double gtol, int *iter, double *fret,double(*func)(double [], int np), void (*dfunc)(double [], int, double []));
		void NR_lnsrch(int n, double xold[], double fold, double g[], double p[], double x[],double *f, double stpmax, int *check, double (*func)(double [], int n));
		int  PAML_gradient(int n, double x[], double f0, double g[], double (*fun)(double x[],int n), double space[], int Central);
		// Functions adapted from PAML for matrix exponentiation...
		void PAML_EigenSort(double d[], double U[], int n);
		int PAML_PMatUVRoot (DoubleMatrix *P, double t, int n, double U[], double V[], double Root[]);
		int PAML_eigenQREV (double Q[], double pi[], int n, double Root[], double U[], double V[]);
		RandomGenerator globalGenerator;
		
		double returnVariateByARSP( int numDists, double *CXY, double *C1, double *Deltas, double mu, double *workingSpace );

	private:
		int PAML_eigenRealSym(double A[], int n, double Root[], double work[]);
		void PAML_HouseholderRealSym(double a[], int n, double d[], double e[]);
		int PAML_EigenTridagQLImplicit(double d[], double e[], int n, double z[]);

		int n;
		double min, max, incr;
		double expMin, expMax, expIncr;
		double *logLookupTable;
		double *expLookupTable;
		int theMode;
		double logFloor;
		void initLookup(int theN);		

		//uint64_t *randArray;
		//const int blockSize = 10000000; // 10 million
		double *randArray;
		static const int blockSize = 10000000;
		dsfmt_t dsfmt;
};

#endif