//
/*  ObservedData.cpp
*  PLEX
*
*  A.P. Jason de Koning, jason.de.koning@gmail.com
*  http://jasondk.org
*
*  David Pollock lab, University of Colorado School of Medicine
*  http://www.evolutionarygenomics.com
*  All rights reserved.
*/


#include "ObservedData.h"

ObservedData::ObservedData() {
    // Defaults
    numSites = numTaxa = 0;
    tupleSize = 1;
    numStates = 0;
    
    _currentState = 0;
    _overlappingTuples = false; // no overlapping tuples by default (good for codon models)
    _getStatesFromData = true;  // by default get the state index from the datafile
    _gapsAsCharacters = false;  // do not model gaps as states by default
}

void ObservedData::setNumStates(int val) {
    numStates = val;
}

void ObservedData::setTupleSize(int val) {
    tupleSize = val;
}

int ObservedData::getAlignmentLength(void) {
    return theStates[0].size();
}

int ObservedData::getNumSequences(void) {
    return taxonNames.size();
}

int ObservedData::getNumStates(void) {
    return numStates;
}

string ObservedData::getState(int which) {
	return reverseStateIndex[ which ];
}

string& ObservedData::getTaxonName_ref(int which) {
    return taxonNames[ which ];
}

std::vector<int>& ObservedData::getSequence_ref(int which) {
    return theStates[ which ];
}

// --- Internal functions -------------------------------------

bool ObservedData::_hasGap( string theTuple ) {
    // Check if the current tuple has a gap...
    if ( ( theTuple.find('-') == string::npos ) && ( theTuple.find('?') == string::npos ) ) {
        return false;
    } else {
        return true;
    }
}

void ObservedData::_removeAllGapColumns( void ) {
    // This function ALSO sets realDatasetColumn and reverseRealDatasetColumn even if gaps are being modelled...
    int i;
    int origCol, newCol;
    
    vector<int> *currentSeq;
    
    for (i=0; i<theStates.size(); i++) {
        currentSeq = &theStates.at(i);
        
        origCol = newCol = 0;
        for (std::vector<int>::iterator j = currentSeq->begin(); j != currentSeq->end(); ++j, origCol++) {
            // If gaps are being deleted...
            if (! _gapsAsCharacters ) {
                if ( _excludedPositions[i] ) {
                    // Remove
                    currentSeq->erase(j);
                } else {
                    if (i==0) {
                        // If position not removed, set
                        reverseRealDatasetColumn[origCol] = newCol;
                        realDatasetColumn[newCol++] = origCol;
                    }
                }
            } else {
                if (i==0) {
                    reverseRealDatasetColumn[origCol] = newCol++;
                }
            }
        }
    }
}

// --- Primary functions -------------------------------------

void ObservedData::usePresetStateIndex(int which) {
    _getStatesFromData = false;
    
    static const char aas[21] = { 'M','T','N','I','R','K','S','H','P','L','V','E','F','D','A','W','G','C','Q','Y' };
    static const char nts[ 5] = { 'T', 'C', 'A', 'G' };
    
    numStates = which;
    const char *whichPreset;
    
    if (numStates == 4) {
        whichPreset = nts;
    } else if (numStates == 20) {
        whichPreset = aas;
    }
    
    stringstream ss;
    for (int i=0; i < numStates; i++) {
        ss.clear();
        ss << whichPreset[i];
        stateIndex[ ss.str() ] = i;
        reverseStateIndex[ i ] = ss.str();
    }
}


void ObservedData::readData( string fileName ) {
    // Read and encode the sequence data from a FASTA formatted file...
    int seqNum = 0;
    
    ifstream infile( fileName.c_str() );
     
    static vector<int> currentSeqVector;
    string line;
    string currentSeq;
    
    currentSeq.clear();
    currentSeqVector.clear();
    
    while (getline( infile, line )) {
        if (! infile.eof() ) {
            line.erase( std::remove( line.begin(), line.end(), ' ' ), line.end() );
            
            if (line.at(0) == '>') {
                line.erase( std::remove( line.begin(), line.end(), '>' ), line.end() );
                
                // New sequence
                taxonNames.push_back( line );
                numTaxa++;
                
                // If we have a seq, add it...
                if (currentSeq.length() != 0) {
                    currentSeqVector.clear();
                    
                    encodeSequence( currentSeq, currentSeqVector );
                    theStates.push_back( currentSeqVector );
                    
                    // Set the number of sites from the first sequence
                    if (seqNum == 0) {
                        numSites = currentSeqVector.size();
                    }
                    
                    currentSeq.clear();
                    seqNum++;
                }
            } else {
                line.erase( std::remove( line.begin(), line.end(), '\n' ), line.end() );
                currentSeq = currentSeq + line;
            }
        }
    }
    
    // Encode the last sequence...
    if (currentSeq.length() != 0) {
        taxonNames.push_back( line );
        numTaxa++;
        
        currentSeqVector.clear();
        encodeSequence( currentSeq, currentSeqVector );
        theStates.push_back( currentSeqVector );
        currentSeq.clear();
        seqNum++;
    }
    
    _removeAllGapColumns();

    infile.close();
}

void ObservedData::encodeSequence( string characters, vector<int> &encodedStates ) {
    // Encode a sequence to a list of states (integers) using the stateIndex
    // - if  _getStatesFromData, the index is built incrementally
    // - if  _overlappingTuples, all tuples of a given size are built and returned
    // - if !_overlappingTuples, states are harvested as expected (use this for codon models)
    // - an index of gap containing states is updated while processing; need to remove gap columns at the end
    
    int i;
    string currentTuple;
    int incr;
    
    // Set the increment (varies when getting tuples, etc.)
    if ( (tupleSize == 1) || (_overlappingTuples) ) {
        incr = 1;
    } else {
        incr = tupleSize;
    }
    
    encodedStates.clear();
    // For each position...
    for (i=0; i<characters.length(); i+=incr) {
        currentTuple = characters.substr( i, tupleSize );
        
        // If the character state is non-gap (or gaps being modelled)...
        if ( !_hasGap( currentTuple ) || (_gapsAsCharacters && !_hasGap( currentTuple ) ) ) {
            
            // Add state to the state index if appropriate...
            if (_getStatesFromData) {
                if ( stateIndex.find( currentTuple ) == stateIndex.end() ) {
                    stateIndex[ currentTuple ] = _currentState;
                    reverseStateIndex[ _currentState++ ] = currentTuple;
                    numStates++;
                }
            }
            
            encodedStates.push_back( stateIndex[ currentTuple ] );
            
        } else if ( !_gapsAsCharacters ) {
            // Mark gapped position for removal
            _excludedPositions[ i ] = true;
            
            // Mark the gapped position in the character string
            encodedStates.push_back( -1 );
        }
    }

    // Debugging code:
    /*
    cout << "Just encoded sequence: " << characters << endl;
    cout << "  as: ";
    for (i=0; i<encodedStates.size(); i++) {
        cout << encodedStates.at(i) << " ";
    } cout << endl;
    */
}
