// ********************************************************************************************************************************
// PLEX.cpp
// PLEX - Phylogenetics, Likelihood, Evolution, and Complexity
// ********************************************************************************************************************************

/*
 *  A.P. Jason de Koning, jason.de.koning@gmail.com
 *  http://jasondk.org
 *
 *  with Wanjun Gu and David Pollock
 *
 *  David Pollock lab, University of Colorado School of Medicine
 *  http://www.evolutionarygenomics.com
 *  All rights reserved.
 */

// ********************************************************************************************************************************
// Current Version: 0.81
// This is an unfinished experimental alpha pre-release release.

// Please note this code has a variety of vestigial features from several earlier versions of related code from our lab.  As a
// result it may appear to be quite a mess.  This is only an illusion: don't panic!

// With the publication of several related manuscripts, new features will be added, and old, experimental code purged. For details
// of the core underlying methods see:

//      de Koning, Gu, and Pollock (2009).  Rapid likelihood analysis on large
//      phylogenies using partial sampling of substitution histories.  MBE (and references therein)

// For any questions please contact Jason (Jason.de.Koning@gmail.com)
// ********************************************************************************************************************************

#include "ObservedData.h"
#include "ReadInputs.h"
#include "PLEXApp.h"
#include "GTRModel.h"
#include "ProteinModel.h"
#include "MCMC.h"
#include "Numerics.h"
#include "Options.h"
#include "NewickTreeParser.h"
#include "NodeDistanceMaker.h"
#include "Tree.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <iostream>
#include <string>
#include <fstream>
#include <iomanip>

#include <omp.h>

#include <sys/param.h>
#include <sys/times.h>
#include <sys/types.h>

using std::cerr;
using std::endl;
using std::ofstream;
using std::setw;
using std::setprecision;
using std::showpos;
using std::noshowpos;
using std::fixed;

Options options;
Numerics *num;
Tree *treeLink;

using namespace std;

//---------------------------------------------------------------------------
bool GetTree(
             ifstream&		inTreeStream,
             Tree&	outTree) {
    
	NewickTreeParser	parser;
	parser.SetNodeMaker(new NodeDistanceMaker);
	
	Tree	tree;
    
	bool success = parser.Read(inTreeStream, tree);
	
	if (success)
	{
		tree.MakeRootless();
		tree.swap(outTree);
	}
	
	return success;
} 

//---------------------------------------------------------------------------
void GetTree(
             const char*		m_strSequencefile,
             Tree&	outTree)
{
    const string cFileOpenErr = "Error opening file: ";
    const string cFileReadErr = "Error reading file: ";
    
    PLEXApp* m_oPLEXApp = new(PLEXApp);
	ifstream treeFile(m_strSequencefile);
    
	if (treeFile.is_open())
	{
		m_oPLEXApp->ReportError( "Reading tree file........", false );
		
		if (!GetTree(treeFile, outTree))
			m_oPLEXApp->ReportError(cFileReadErr + m_strSequencefile);
	}
	else 
		m_oPLEXApp->ReportError(cFileOpenErr + m_strSequencefile);
    
	m_oPLEXApp->ReportError("Done read tree file.", false);
	
	cerr << endl;
	
	delete m_oPLEXApp;
}


int main( int argc, char *argv[]) {
	int	returnCode = 0;
	
	// Set the number of threads
	if (options.numThreads == -1) {
		int nthreads, tid;
		
#pragma omp parallel shared( nthreads ) private( tid )
		{
            tid = omp_get_thread_num();
			
            if (tid == 0) {
                nthreads = omp_get_num_threads();
            }
        }
		
		std::cout << "INITIALIZING NUMBER OF THREADS TO " << nthreads << "." << std::endl;
		options.numThreads = nthreads;
    }
 
	// JdK, Sept 18/07: for execution timing...
	time_t startTime;
	startTime = time(NULL);
	
	try {
		int m_nInitseed,  programTask;
		
		ReadInputs* readInputs = new(ReadInputs);
		
		if ( !readInputs->ReadControlfile(CONTROLFILE) )
			exit(1);
		
		// *********************************************
		// Test the Randomize function
		// *********************************************
		
		PLEXApp* m_oPLEXApp = new(PLEXApp);
		
		// Set seed for random number generator
        m_oPLEXApp->SetSeed(m_nInitseed);
		options.randseed = m_nInitseed;
		
        // Only one task defined so far...
		if ( programTask == 0) {
            
			// ************************************************************************************************************
			//              Read Control file
			// ************************************************************************************************************
			if (!readInputs->ReadControlfile(CONTROLFILE) ) {
				throw(1);
			}
			
			// ************************************************************************************************************
            //            Read Input files                
			// ************************************************************************************************************
            ObservedData sequenceData;
            
            string fname( options.m_strSeqfile );
            sequenceData.readData( fname );
            
            int m_nAlignlength = sequenceData.getAlignmentLength();
			
 			// ************************************************************************************************************
			// Read the tree file to get the tree topology
			// ************************************************************************************************************
			Tree  Tree;
			treeLink = &Tree;
			
			GetTree( options.m_strTreefile,  Tree);
			
			int m_nLeaves =  Tree.leaves();
			int m_nInternodes =  Tree.internalnodes();
			int m_nBranches = options.m_nNumBranches = Tree.branches();
            Tree.numRegularNodes = m_nInternodes + m_nLeaves;
			
			// ************************************************************************************************************
			// Optionally subdivide long branches so a maximum length is not exceeded
			// ************************************************************************************************************
			std::cout << "Before processing tree: leaves = " << m_nLeaves << ", m_nInternalNodes = " << m_nInternodes << ", branches " << m_nBranches << endl;
			double loss = 0.0;
			
			MCMC* m_oMCMC = new(MCMC);
			m_oMCMC->InitializeTree( Tree, sequenceData );
			
            Tree.root()->_mixSequences( Tree.root()->GetData().sequence, Tree.root()->Left()->GetData().sequence, Tree.root()->Right()->GetData().sequence );
			// Tree.root()->GetData().sequence.Mix(& Tree.root()->Left()->GetData().sequence, & Tree.root()->Right()->GetData().sequence);
			
			if (options.m_nMethod != 3 && options.m_nMethod != 4  && options.m_nMethod != 5) {
				// Subdivide all branches so each segment is <= maxBL...
                Tree.subdivideLongBranches( Tree.root(), options.maxBL);
			}
			
			if (options.m_nMethod == 3 || options.m_nMethod == 5) {
				// Use a branchlength discrete approximation...
                Tree.root()->_mixSequences( Tree.root()->GetData().sequence, Tree.root()->Left()->GetData().sequence, Tree.root()->Right()->GetData().sequence );
 				
				loss =  Tree.discreteBranchLengthApproximation( Tree.root());
				std::cout << "Got actual loss of " << sqrt(loss/m_nBranches) << endl;
				
				
			} else {
                Tree.MakeRootless();
			}
			
			if (options.m_nMethod == 3 || (options.m_nMethod == 5)) { std::cout << endl << "TOTAL BRANCHLENGTH TRUNCATION ERROR (RMSE): " << sqrt(loss/m_nBranches) << endl << endl; }
			
			m_nLeaves =  Tree.leaves();
			m_nInternodes =  Tree.internalnodes();
			m_nBranches =  Tree.branches();
			
			std::cout << "After processing tree: leaves = " << m_nLeaves << ", m_nInternalNodes = " << m_nInternodes << ", branches " << m_nBranches << endl;
			
			// ************************************************************************************************************
			
			
			cerr << "Tips counted: " << m_nLeaves << endl
			<< "Sequences counted: " << sequenceData.getNumSequences() << endl;
			cerr << "Number Internal Nodes: " << m_nInternodes << endl
			<< "Number of Branches: " << m_nBranches << endl << endl;
			
			ofstream outTreeStream(options.m_strTreeOutfile);
			
			if (options.Debugging)
                Tree.printtree(outTreeStream, -1, 1);
			else
			{
				if (options.m_bTreeflag)
                    Tree.printtree(outTreeStream, -1, 1);
			}
			
			m_oMCMC->InitializeTreeAlternate( Tree, sequenceData );
			
			m_nLeaves =  Tree.leaves();
			m_nInternodes =  Tree.internalnodes();
			m_nBranches =  Tree.branches();
			
			cerr << "Tips counted: " << m_nLeaves << endl
			<< "Sequences counted: " << sequenceData.getNumSequences() << endl;
			cerr << "Number Internal Nodes: " << m_nInternodes << endl
			<< "Number of Branches: " << m_nBranches << endl << endl;
			
			if (options.Debugging)
                Tree.printtree(outTreeStream, -1, 1);
			else
			{
				if (options.m_bTreeflag)
                    Tree.printtree(outTreeStream, -1, 1);
			}	
			
			ofstream outSequenceStream(options.m_strSeqOutfile);
			
			
            IntMatrix  TreeChanges;
            // ************************************************************************************************************
            // Count changes on tree -> matrix
            // ************************************************************************************************************
            // count state changes of each branch, and sum over to get the whole tree changes, use this to initialize the model
            Tree.CountChangesAlongEveryBranch( sequenceData );			
            Tree.SumChangesOverTree( TreeChanges );
            
            // ************************************************************************************************************
            

			// ************************************************************************************************************
			// Model definition and setup
			// ************************************************************************************************************
			
			MixtureModel *m_gtrModel;
			
			if ( (options.m_nDatatype ==2) ) { //(options.doRAR == 1) {
				// Use new protein-model class...
				m_gtrModel = new(ProteinModel);
				
				// Need to provide a vector of initial values for gamma RAR...		
				double *gammaShapes = new double[options.m_nNumModels];
				for (int i=0; i<options.m_nNumModels;i++) {
					gammaShapes[i] = 1.0; // 0.42; // * (Numerics::getGlobalNumerics().Random() * 5.0);
				}
				
				// Basic options...
				(*(ProteinModel *)m_gtrModel).setReversible(0);
				(*(ProteinModel *)m_gtrModel).setFixedFreqs(1);
				(*(ProteinModel *)m_gtrModel).SetNumModels( options.m_nNumModels );
				
				if (options.fixedRates != -1) {
					// For predefined-matrix stuff...
					//	std::cout << "NON-GAMMA MODEL DEFINED..." << endl; exit(1);
					(*(ProteinModel *)m_gtrModel).setRARdistributed(0);
					(*(ProteinModel *)m_gtrModel).setPredefinedMatrix(options.fixedRates);
					(*(ProteinModel *)m_gtrModel).setScaleMatrix(1);
					(*(ProteinModel *)m_gtrModel).setReversible(1);
					(*(ProteinModel *)m_gtrModel).initRateScale(0.03); //0.03  //0.034493);
				} else if (options.doRAR) {
					//	std::cout << "GAMMA MODEL DEFINED..." << endl; exit(1);
					(*(ProteinModel *)m_gtrModel).setRARdistributed(1);
					(*(ProteinModel *)m_gtrModel).setScaleMatrix(0);
				} 
				
				if (options.freeFreqs == 1) {
					(*(ProteinModel *)m_gtrModel).setFixedFreqs(0);					
				}
				
				// For gamma-distributed RAR models...
				
				(*(ProteinModel *)m_gtrModel).setIndependentFractionPrior(options.DRAprior);
				(*(ProteinModel *)m_gtrModel).setDependentRates(options.DRA);
				//				(*(ProteinModel *)m_gtrModel).randomInitRateAssignments();


				if ( (*(ProteinModel *)m_gtrModel).RARdistributedRates ) (*(ProteinModel *)m_gtrModel).initGamma3D(options.m_nNumModels, options.numCat, gammaShapes);
				m_oMCMC->InitializeMixtureModel( sequenceData, (*m_gtrModel), TreeChanges );
				
				//	if ( (*(ProteinModel *)m_gtrModel).RARdistributedRates ) (*(ProteinModel *)m_gtrModel).setDistributionParmsFromCounts(  Tree );
				
			} else {	// GTR model
				m_gtrModel = new(GTRModel);

				m_oMCMC->InitializeMixtureModel( sequenceData, (*m_gtrModel), TreeChanges );
			}
			
			// ************************************************************************************
			// Initialize the ASRV distribution if necessary...
			// ************************************************************************************
			//if ( ( Tree.root()->GetData().siteCats == NULL) ) {
			std::cout << "CREATING SITE CATEGORY DATA AT ROOT NODE..." << endl;
            Tree.root()->GetData().siteCats = new SiteCategory( m_nAlignlength );
			//}
			// ************************************************************************************		
			
			ofstream outCountStream(options.m_strCountOutfile);
			
			if (options.Debugging) {
				if ( (options.m_nDatatype != 2)) {
					(*(GTRModel *)m_gtrModel).PrintCountMatrix(outCountStream, sequenceData);
				} else {
					(*(ProteinModel *)m_gtrModel).PrintCountMatrix(outCountStream, sequenceData);
				}
			}
			ofstream outRatematrixStream(options.m_strMatrixOutfile);
			
			// Jason
			ofstream outSiteAssignmentStream("siteAssignments");
			ofstream outMixtureDependencyStream("mixtureDependencies");
			ofstream outRatesAcrossRatesStream("RARparameters");
			ofstream posteriorPredictiveStream("posteriorPredictive");
			ofstream freqStream("stateFrequencies");
			
			
			if (options.Debugging)
			{
				(*m_gtrModel).PrintRateMatrix3D(outRatematrixStream, sequenceData);
				if ((options.m_nDatatype != 2)) {
					(*(GTRModel *)m_gtrModel).PrintFrequencyMatrix(freqStream, sequenceData);
				} else {
					(*(ProteinModel *)m_gtrModel).PrintFrequencyMatrix(freqStream, sequenceData);
				}
				//m_gtrModel.PrintAlphaMatrix3D(outRatematrixStream);
			}
			else
			{
				outRatematrixStream << setw(6) << "Count ";
				m_oPLEXApp->PrintSequenceStatesLine(outRatematrixStream, sequenceData );	//print the changes type of different states in a line
			}
			
			
			// ************************************************************************************************************
			// Run mixture model MCMC 
			// ************************************************************************************************************
			// Force PLEX to use the standard mode
			int m_nMixtureScheme = 1;
            
             int tempCalcProbDist = Tree.root()->getCalculateFullProbabilityDistribution();
            Tree.root()->setCalculateFullProbabilityDistribution(1); //override
            Tree.setprobmatrixalongbranch( (*m_gtrModel), (*m_gtrModel).GetFreqModels() );
            Tree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
			
			if (m_nMixtureScheme == 1) {
				// ************************************************************************************************************
				// Run the Markov chain and save/output state data
				// ************************************************************************************************************
				IntVector m_vecSiteAssignment;
				
				// Doing this for the temporal model, temporarily...
				m_vecSiteAssignment.resize( m_nAlignlength, 0 );
				
				m_oMCMC->setAllAncestralAndTransientStates_full( Tree, (*m_gtrModel), sequenceData, m_vecSiteAssignment );
				
				m_oMCMC->countChangesAlongTree( Tree, (*m_gtrModel), sequenceData, m_vecSiteAssignment );
                
				m_oMCMC->InitializeAssignment( Tree, m_vecSiteAssignment, (*m_gtrModel), sequenceData );
				
                
				bool m_bModel, m_bState, m_bAssignment;
				double logLikelihood, m_fMaxLnLikelihood;
				
				m_fMaxLnLikelihood = logLikelihood = 0.0;
				
				ofstream outLikelihoodStream(options.m_strLikelihoodOutfile);
				outLikelihoodStream << "Count\tUpdates\tLnLikelihood\tIntegratedLnLikelihood\tMaxLnLikelihood" << setw(0) << endl;
				
				int m_nUnacceptedModel = 0;
				int m_nUpdateNode = 0;
				int m_nAcceptModel = 0;
				int m_nUpdateAssignment = 0;
				int SampleSequence = 0;
				int maxSampleSequence = 10000;
				
				StringMatrix InternalNodeSeq(m_nInternodes, maxSampleSequence);
				StringVector AncestralNodeSeq(m_nInternodes);
				IntVector NodeID(m_nInternodes);
				
				ofstream ASRVoutfile("siteRateParameters");
				ofstream siteRatesOutfile("siteRates");
				
				siteRatesOutfile << "Gen\t";
				for (int jj=0; jj< Tree.root()->GetData().siteCats->siteAssignments.size(); jj++) {
					if (jj != ( Tree.root()->GetData().siteCats->siteAssignments.size() - 1) ) {
						siteRatesOutfile << "Site " << (jj+1) << "\t";
					} else {
						siteRatesOutfile << "Site " << (jj+1) << endl;
					}
				}
				
				long int totalUpdates = 0;
				for (unsigned int i = 0; i < options.m_nGenerations; i++) {
					if (options.suppressFileOutput == 0) if ( (i % 1000) == 0 ) { std::cout << "	Generation " << i << endl; }
					
					if (options.m_nDatatype != 2) {
                        // Refactor this for new simpler calling routine; what about m_vecSiteAssignment?
                        
						m_oMCMC->RunMarkovChain( Tree, (*(GTRModel *)m_gtrModel), m_vecSiteAssignment, sequenceData, m_bModel, m_bState, m_bAssignment); //, posteriorPredictiveStream);
					} else {
						totalUpdates += m_oMCMC->RunMarkovChain( Tree, (*(ProteinModel *)m_gtrModel), m_vecSiteAssignment, sequenceData, m_bModel, m_bState, m_bAssignment, posteriorPredictiveStream);
					}
					if ((options.suppressFileOutput == 1)  && (i % 1000 == 0)) std::cout << "	REAL GEN " << i << "..." << endl;
					if (options.suppressFileOutput == 1) continue;
					
					if (m_bState) {
						m_nUpdateNode++;
					} else if (m_bModel) {
						m_nAcceptModel++;
					} else if (m_bAssignment) {
						m_nUpdateAssignment++;
					} else
						m_nUnacceptedModel++;
					
					if ( i % options.outputFrequency == 0 ) { 			
						//output the count matrix of total changes in different states
                        TreeChanges.fill(0);
                        Tree.SumChangesOverTree( TreeChanges);
						m_oPLEXApp->PrintMatrix(outCountStream, TreeChanges, sequenceData );
						
						outRatematrixStream <<endl << setw(2) << i << endl;
						
						if (options.m_nDatatype == 2) {
							(*(ProteinModel *)m_gtrModel).PrintRateParameters(outRatesAcrossRatesStream);	// only applies to protein models so far
						}
						(*m_gtrModel).PrintRateMatrix3D(outRatematrixStream, sequenceData);

						if (options.m_nDatatype != 2) {
							(*(GTRModel *)m_gtrModel).PrintFrequencyMatrix(freqStream, sequenceData);
						} else {
							(*(ProteinModel *)m_gtrModel).PrintFrequencyMatrix(freqStream, sequenceData);
						}
						if (options.m_nNumModels > 1) (*m_gtrModel).PrintModelAssignments(outSiteAssignmentStream, m_vecSiteAssignment);
						
						if (options.doRAR == 1) {
							if ((*(ProteinModel *)m_gtrModel).getDependentRates() == 1) (*(ProteinModel *)m_gtrModel).PrintDependentRates(outMixtureDependencyStream);
						}
						
						// If the uniformization constant isn't capped, enlarge if necessary...
						if ( (i==0) && (options.m_nMethod == 6) ) {
							(*m_gtrModel).enlargeUniformization();
						}
						
						// Output the likelihood with current rate matrix and count changes:				
						// ************************************************************************************************************
						// CHECK THIS: SHOULDN'T NEED TO RECALCULATE THE COMPLETE_DATA LOG-L HERE
						// ************************************************************************************************************
						// logLikelihood =  Tree.completeDataLnL;
						logLikelihood = m_oMCMC->CalculateCompleteDataLogLikelihood( Tree, (*m_gtrModel), m_vecSiteAssignment );
						
						if (m_nAcceptModel <= 1)
							m_fMaxLnLikelihood = logLikelihood;
						else {
							if (m_fMaxLnLikelihood < logLikelihood)
								m_fMaxLnLikelihood = logLikelihood;
						}
						
						outLikelihoodStream << i << "\t" << setw(11) << totalUpdates << "\t" << setprecision(4) << fixed << showpos << logLikelihood << "\t" <<  Tree.incompleteDataLnL << "\t" << m_fMaxLnLikelihood << setprecision(0) << setw(0) << noshowpos << endl;
						
						
						// Output tree with branchlengths...
						if (options.updatebls)  Tree.printtreechanges(outTreeStream);
						
						// Any type of update...
						if (options.allowASRV) {
							if ( i % options.outputFrequency == 0 ) {
								ASRVoutfile << "Gamma shape:\t" <<  Tree.root()->GetData().siteCats->getShapeParm() << "\tRates:\t";
								for (int ii=0; ii<options.numCatASRV; ii++) {
									ASRVoutfile <<  Tree.root()->GetData().siteCats->getRate(ii) << "\t";
								}
								
								if (options.outputSiteRates) {
									ASRVoutfile << "Num sites per class:\t";
									static int *tempCounts;
									static int countAlloc = 0;
									if (countAlloc == 0) {
										tempCounts = new int[options.numCatASRV];
										countAlloc = 1;
									}
									for (int ii=0; ii<options.numCatASRV; ii++) {
										tempCounts[ii] = 0;
									}
									
									siteRatesOutfile << i << "\t";
									for (int jj=0; jj< Tree.root()->GetData().siteCats->siteAssignments.size(); jj++) {
										tempCounts[  Tree.root()->GetData().siteCats->getAssignment(jj) ]++;
										siteRatesOutfile <<  Tree.root()->GetData().siteCats->getRate(  Tree.root()->GetData().siteCats->getAssignment(jj) ) << "\t";
									} siteRatesOutfile << endl;
									
									for (int ii=0; ii<options.numCatASRV; ii++) {
										ASRVoutfile << tempCounts[ii] << "\t";
									} ASRVoutfile << endl;
								} else {
									ASRVoutfile << endl;
								}
								
							}
						}
					}			
				}
				
				std::cout << "DONE SAMPLING..." << endl;

				if (options.Debugging)
				{
					outLikelihoodStream << flush;
					
					std::cout << endl << endl << setw(40) << "Number of Updated Node" << setw(40) << "Number of Accepted Proposal Model" << setw(40) << "Number of Unaccepted Proposal Model" << setw(0) << endl;
					std::cout << setw(40) << m_nUpdateNode << setw(40) << m_nAcceptModel << setw(40) << m_nUnacceptedModel << setw(0) << endl;
					std::cout << flush;
				}
			}
			
		}
    }
    
	
	// ************************************************************************************************************
	// Error handling and reporting
	// ************************************************************************************************************
    
	catch(int newreturnCode) {
		returnCode = newreturnCode;
		
		cerr << endl << endl
		<< "**********************************" <<endl
		<< " ------------- ERROR ------------ " << endl
		<< endl <<endl;
		
		if (newreturnCode == 1)
			cerr << "error " << returnCode << ": " 
			<< "Open controlfile error" << endl;
		else if (newreturnCode == 2)
			cerr << "error " << returnCode << ": " 
			<< "Read sequence file error" << endl;
		else if (newreturnCode == 3)
			cerr << "error " << returnCode << ": " 
			<< "Read position file error" << endl;
		else if (newreturnCode == 4)
			cerr << "error " << returnCode << ": " 
			<< "Read OL file error" << endl;
		else if (newreturnCode == 5)
			cerr << "error " << returnCode << ": " 
			<< "Read genome size file error" << endl;
		else if (newreturnCode == 6)
			cerr << "error " << returnCode << ": " 
			<< "Not all sequence position information inputted" << endl;
		else if (newreturnCode == 7)
			cerr << "error " << returnCode << ": " 
			<< "Not all OL or size information inputted" << endl;
		else if (newreturnCode == 8)
			cerr << "error " << returnCode << ": " 
			<< "Read tree file error" << endl;
		else if (newreturnCode == 9)
			cerr << "error " << returnCode << ": " 
			<< "Method to calculate probability is error" << endl;
		
		cerr << endl << endl
		<< "**********************************" <<endl
		<< endl <<endl <<endl;
		
		exit(returnCode);
	}
	
	catch(...) {
		cerr << "**********************" << endl;
		cerr << "Unknown error" << endl;
		returnCode = -1;
		exit(returnCode);
	}
	
	// ************************************************************************************************************
	// Calculation of timetaken
	// ************************************************************************************************************
	
	time_t theT;
	theT = time(NULL)-startTime;
	int h,m,s;
	h=theT/3600; m=(theT%3600)/60; s=theT-(theT/60)*60;
	printf("\n\nTime elapsed: ");
	if(h)  printf("%d:%02d:%02d", h,m,s);
	else   printf("%2d:%02d", m,s); 
	ofstream outTimeStream("timetakenfile");
	if (h) outTimeStream << h << ":" << m << ":" << s << endl;
	else outTimeStream << m << ":" << s << endl;
    
	// ************************************************************************************************************
	
	return returnCode;
}
