/*
 *  ProteinModel.cpp
 *  PLEX
 *
 *  A.P. Jason de Koning, jason.de.koning@gmail.com
 *  http://jasondk.org
 *
 *  David Pollock lab, University of Colorado School of Medicine
 *  http://www.evolutionarygenomics.com
 *  All rights reserved.
 
 */

#include "ProteinModel.h"
#include "PLEXApp.h"
#include "MixtureModel.h"

#pragma mark -
//---------------------------------------------------------------------------
ProteinModel::ProteinModel(
		const DoubleMatrix3D& inAlphaMatrix3D,
		const DoubleMatrix& inFrequencyMatrix,
		const IntMatrix& inCountMatrix)
		:	MixtureModel()
{
	num = &Numerics::getGlobalNumerics();
	mAlphaMatrix3D = inAlphaMatrix3D;
	mFrequencyMatrix = inFrequencyMatrix;
	mCountMatrix = inCountMatrix;
	RARdistributedRates = false;
	setScaleMatrix(0);

	setReversible(0);
	setFixedFreqs(0);

	//initRateScale(1.0);
	setDependentRates(0);
	setPredefinedMatrix(-1);
	setDependentRates(0);
}		

ProteinModel::ProteinModel(
		const DoubleMatrix3D& inAlphaMatrix3D,
		const DoubleMatrix& inFrequencyMatrix,
		const IntMatrix& inCountMatrix, int theNumRates, double *gammaParm)
		:	MixtureModel()
{
	// Version 1 constructor: include gamma distribution init...
	num = &Numerics::getGlobalNumerics();
	
	mAlphaMatrix3D = inAlphaMatrix3D;
	mFrequencyMatrix = inFrequencyMatrix;
	mCountMatrix = inCountMatrix;
	setScaleMatrix(0);

	setReversible(0);
	setFixedFreqs(0);
	
	//initRateScale(1.0);
	setPredefinedMatrix(-1);

	// Gamma distribution stuff..
	setDependentRates(0);
	RARdistributedRates = false;
	randomInitRateAssignments();
	initGamma3D( GetNumModels(), theNumRates, gammaParm );
}		

//---------------------------------------------------------------------------
ProteinModel::ProteinModel(
		const ProteinModel&	inModel)
		:	MixtureModel(inModel)
{
	num = &Numerics::getGlobalNumerics();
	mAlphaMatrix3D = inModel.mAlphaMatrix3D;
//	mRateMatrix3D = inModel.mRateMatrix3D;
	mFrequencyMatrix = inModel.mFrequencyMatrix;
	mCountMatrix = inModel.mCountMatrix;
	
	RARdistributedRates = inModel.RARdistributedRates;
	setReversible(inModel.reversibleModel);
	setFixedFreqs(inModel.fixedFreqs);

	setDependentRates( inModel.getDependentRates() );
	if (inModel.getDependentRates() == 1) {
		setIndependentFractionPrior( inModel.getIndependentFractionPrior() );
	}
	
	if (getDependentRates() == 1) {
		if (dependencyIndicators.empty()) {
			dependencyIndicators.resize(GetNumModels(), mNumStates, mNumStates, 0);
		}
		
		for (int i=0; i<GetNumModels(); i++) {
			for ( int j=0; j<mNumStates; j++) {
				for (int k=0; k<mNumStates; k++) {
					dependencyIndicators[i][j][k] = inModel.dependencyIndicators[i][j][k];
				}
			}
		}
	}
	
//	std::cout << "HERE HERE!  COPY CONSTRUCTOR INIT FOR RATESCALES... (numModels = " << inModel.GetNumModels() << ")" << endl;
	setScaleMatrix( inModel.scaleMatrix);
	if (inModel.scaleMatrix == 1) {
		//	std::cout << "Apparently, inmodel has getscaleMatrix: " << inModel.getScaleMatrix() << endl;
			initRateScale(inModel.getRateScale(0));
			for (int i=0; i<inModel.mNumModels; i++) { setRateScale(i, inModel.rateScaleFactor[i]); } //std::cout << "TEST, in rate is " << inModel.rateScaleFactor[i] << endl; 
			setScaleMatrix( inModel.scaleMatrix);
	}
	setPredefinedMatrix( inModel.predefinedMatrix );

	// Gamma distribution stuff..
	if (inModel.RARdistributedRates) {
		numDiscreteRates = inModel.numDiscreteRates;

		discreteRates.resize(numDiscreteRates*GetNumModels());
		gammaShapeParameter.resize(GetNumModels());
		gammaScaleParameter.resize(GetNumModels());
		rateAssignments.resize(mNumStates*mNumStates*mNumModels);
		
		for (int i=0; i<(inModel.GetNumModels()*numDiscreteRates); i++) {
			discreteRates[i] = inModel.discreteRates[i];
		}
		
		for (int i=0; i<inModel.GetNumModels(); i++) {
			gammaShapeParameter[i] = inModel.gammaShapeParameter[i];
			gammaScaleParameter[i] = inModel.gammaScaleParameter[i];
			for (int j=0; j<mNumStates*mNumStates;j++) {
				rateAssignments[(i*mNumStates*mNumStates)+j] = inModel.rateAssignments[(i*mNumStates*mNumStates)+j];
			}
		}
//		initAndCopyGamma3D(GetNumModels(), numDiscreteRates, inModel.gammaShapeParameter, inModel.rateAssignments);
	}
}


void ProteinModel::copy( const ProteinModel& inModel)
{
//	m_oPLEXApp = new(PLEXApp);
	mRateMatrix3D = inModel.mRateMatrix3D;
	mNumStates = inModel.mNumStates;
	mStateIndex = inModel.mStateIndex;
	mStates = inModel.mStates;
	mMinState = inModel.mMinState;
	mNumModels = inModel.mNumModels;
	mFreqModels = inModel.mFreqModels;
	mFixedFreq = inModel.mFixedFreq;
	mWindowsize = inModel.mWindowsize;
	mNucleotides = inModel.mNucleotides;

	num = &Numerics::getGlobalNumerics();
	mAlphaMatrix3D = inModel.mAlphaMatrix3D;
//	mRateMatrix3D = inModel.mRateMatrix3D;
	mFrequencyMatrix = inModel.mFrequencyMatrix;
	mCountMatrix = inModel.mCountMatrix;
	
	RARdistributedRates = inModel.RARdistributedRates;
	setReversible(inModel.reversibleModel);
	setFixedFreqs(inModel.fixedFreqs);

	setDependentRates( inModel.getDependentRates() );
	if (inModel.getDependentRates() == 1) {
		setIndependentFractionPrior( inModel.getIndependentFractionPrior() );
	}
	
	if (getDependentRates() == 1) {
		if (dependencyIndicators.empty()) {
			dependencyIndicators.resize(GetNumModels(), mNumStates, mNumStates, 0);
		}
		
		for (int i=0; i<GetNumModels(); i++) {
			for ( int j=0; j<mNumStates; j++) {
				for (int k=0; k<mNumStates; k++) {
					dependencyIndicators.mRows.mDataArray[i].mRows.mDataArray[j].mDataArray[k] = inModel.dependencyIndicators.mRows.mDataArray[i].mRows.mDataArray[j].mDataArray[k];
				}
			}
		}
	}
	
	//std::cout << "HERE HERE!  COPY CONSTRUCTOR INIT FOR RATESCALES... (numModels = " << inModel.GetNumModels() << ")" << endl;
	if (inModel.getScaleMatrix() == 1) {
		//	std::cout << "Apparently, inmodel has getscaleMatrix: " << inModel.getScaleMatrix() << endl;
			initRateScale(inModel.GetNumModels(), 1.0);
			for (int i=0; i<inModel.GetNumModels(); i++) { setRateScale(i, inModel.rateScaleFactor[i]); } // std::cout << "TEST, in rate is " << inModel.rateScaleFactor[i] << endl; }
			setScaleMatrix( inModel.scaleMatrix);
	}
	setPredefinedMatrix( inModel.predefinedMatrix );

	// Gamma distribution stuff..
	if (inModel.RARdistributedRates) {
		numDiscreteRates = inModel.numDiscreteRates;

	//	std::cout << "numdiscrates is " << numDiscreteRates << ", inmodels is " << inModel.GetNumModels() << endl;
		
		discreteRates.resize(numDiscreteRates*inModel.GetNumModels());
		gammaShapeParameter.resize(inModel.GetNumModels());
		gammaScaleParameter.resize(inModel.GetNumModels());
		rateAssignments.resize(inModel.mNumStates*inModel.mNumStates*inModel.mNumModels);
		
	//	std::cout << "About to copy over the discreteRates...there are " << (inModel.GetNumModels() * numDiscreteRates) << endl;
		
		for (int i=0; i<(inModel.GetNumModels()*numDiscreteRates); i++) {
	//		std::cout << "i is " << i << endl;
			discreteRates[i] = inModel.discreteRates[i];
		}
		
		for (int i=0; i<inModel.GetNumModels(); i++) {
			gammaShapeParameter[i] = inModel.gammaShapeParameter[i];
			gammaScaleParameter[i] = inModel.gammaScaleParameter[i];
			for (int j=0; j<mNumStates*mNumStates;j++) {
				rateAssignments[(i*mNumStates*mNumStates)+j] = inModel.rateAssignments[(i*mNumStates*mNumStates)+j];
			}
		}
//		initAndCopyGamma3D(GetNumModels(), numDiscreteRates, inModel.gammaShapeParameter, inModel.rateAssignments);
	}
}


//---------------------------------------------------------------------------
ProteinModel::~ProteinModel(void)
{
//delete [] rateScaleFactor;
}

//---------------------------------------------------------------------------
void ProteinModel::setFixedFreqs(int val) {
	fixedFreqs = val;
}

int ProteinModel::getFixedFreqs(void) {
	return fixedFreqs;
}

//---------------------------------------------------------------------------
void ProteinModel::randomInitRateAssignments(void) {
	// Assuming non-symmetric rates: assign rate category to each transition...
	if (mNumStates == 0) { return; } // BIG ERROR...

	int u;
	std::cout << "Doing random init of rate assignments..." << endl;
	//double normalized[numDiscreteRates], sum;
	
//	sum = 0.0;
//	for (int i=0; i<numDiscreteRates; i++) {
//		sum += i * (1.0 / numDiscreteRates); //discreteRates[i];
//	}
//	for (int i=0; i<numDiscreteRates; i++) {
//		normalized[i] = discreteRates[i]/sum;
///	}
	
	std::cout << "numstates is " << mNumStates << endl;
	std::cout << "discreteRate 0 is " << discreteRates[0] << ", and 4 " << discreteRates[4] << endl;


/*	double initMatrix[20][20] = {
	{0, 4, 2, 4, 0, 1, 3, 1, 2, 4, 4, 0, 2, 1, 4, 1, 2, 0, 0, 2},
	{4, 0, 3, 4, 1, 2, 4, 2, 2, 4, 3, 1, 2, 2, 4, 1, 0, 0, 1, 0},
	{0, 3, 0, 3, 2, 3, 4, 3, 0, 0, 2, 1, 2, 4, 0, 1, 1, 0, 1, 2},
	{4, 4, 2, 0, 0, 0, 1, 1, 2, 4, 4, 1, 3, 0, 4, 0, 2, 0, 0, 0},
	{0, 1, 1, 0, 0, 0, 1, 2, 1, 1, 0, 0, 2, 0, 2, 0, 2, 0, 3, 0},
	{2, 3, 4, 1, 0, 0, 1, 3, 0, 2, 0, 2, 0, 0, 2, 1, 1, 2, 0, 3},
	{3, 4, 3, 0, 0, 1, 0, 0, 3, 3, 2, 1, 2, 2, 4, 0, 3, 2, 2, 2},
	{1, 1, 3, 0, 2, 2, 2, 0, 2, 1, 0, 2, 1, 3, 0, 0, 1, 1, 3, 4},
	{0, 2, 2, 2, 0, 1, 3, 1, 0, 2, 1, 1, 1, 1, 0, 0, 0, 1, 1, 0},
	{4, 3, 1, 4, 0, 2, 3, 1, 1, 0, 3, 0, 4, 0, 3, 2, 1, 1, 2, 2},
	{4, 4, 1, 4, 2, 0, 0, 0, 2, 4, 0, 0, 3, 0, 4, 0, 1, 0, 0, 1},
	{1, 2, 3, 1, 1, 3, 1, 0, 1, 1, 1, 0, 1, 3, 1, 0, 2, 1, 3, 1},
	{2, 2, 2, 3, 1, 1, 2, 0, 1, 4, 2, 0, 0, 0, 1, 0, 1, 1, 0, 4},
	{0, 2, 4, 1, 0, 1, 2, 1, 1, 1, 2, 3, 0, 0, 1, 1, 2, 2, 1, 0},
	{3, 4, 1, 3, 1, 0, 3, 0, 2, 2, 4, 1, 2, 2, 0, 1, 2, 1, 0, 0},
	{0, 1, 0, 1, 1, 0, 0, 1, 0, 0, 1, 1, 0, 1, 1, 0, 0, 1, 0, 0},
	{2, 2, 2, 1, 1, 0, 3, 1, 0, 2, 0, 1, 0, 1, 2, 0, 0, 0, 1, 0},
	{3, 4, 4, 4, 2, 0, 4, 1, 3, 3, 3, 1, 4, 0, 0, 2, 1, 0, 0, 2},
	{2, 2, 2, 1, 4, 3, 1, 3, 0, 3, 0, 3, 0, 0, 0, 0, 0, 1, 0, 0},
	{0, 0, 2, 1, 1, 0, 2, 4, 2, 2, 1, 0, 4, 2, 1, 1, 1, 0, 2, 0}};
*/

	std::cout << "Test: getDeprates() is " << getDependentRates() << endl;
	std::cout << "Test2: getNumModels() is " << GetNumModels() << endl;
	
	double rowsum;
	for (int i=0; i<GetNumModels(); i++) {
		if (getDependentRates()) {
			if (dependencyIndicators.empty()) {
				dependencyIndicators.resize(GetNumModels(), mNumStates, mNumStates, 0);
			}
		}
		
		double u2;
		for (int j1=0; j1<mNumStates; j1++) {
			rowsum = 0.0;
			for (int j2=0; j2<mNumStates; j2++) {
				std::cout << "j2 = " << j2 << ", j1 = " << j1 << endl;
				
				if (getDependentRates() && (i > 0) ) {
					
					if (j2 != j1) {
						u2 = Numerics::getGlobalNumerics().Random();
					//	std::cout << " rand() was " << u2 << "; prior was " << getIndependentFractionPrior() << endl;
						if (u2 > getIndependentFractionPrior()) {
							// Make dependent
							dependencyIndicators[i][j1][j2] = 1;
					//		std::cout << "mixture " << i << " state change " << j1 << " -> " << j2 << " set to " << dependencyIndicators[i][j1][j2] << endl;
							rateAssignments[(i*mNumStates*mNumStates)+(j1*mNumStates) + j2] = rateAssignments[(0)+(j1*mNumStates) + j2];
							mAlphaMatrix3D[i][j1][j2] = mAlphaMatrix3D[0][j1][j2]; //initMatrix[j1][j2]]; //initMatrix[j1][j2]
							mRateMatrix3D[i][j1][j2] = mRateMatrix3D[0][j1][j2];
							if (getReversible())  mRateMatrix3D[i][j1][j2] *= mFrequencyMatrix[i][j2]; //initMatrix[j1][j2]// [BLOCKING USE OF FREQS HERE - IMPLICIT IN NON-REV MODELS]
							rowsum += mRateMatrix3D[i][j1][j2];
						} else {
							dependencyIndicators[i][j1][j2] = 0;
					//		std::cout << "mixture " << i << " state change " << j1 << " -> " << j2 << " set to " << dependencyIndicators[i][j1][j2] << endl;
							u = static_cast<int>(Numerics::getGlobalNumerics().Random() * numDiscreteRates);
						
							rateAssignments[(i*mNumStates*mNumStates)+(j1*mNumStates) + j2] = u; //initMatrix[j1][j2]; //initMatrix[j1][j2]
							mAlphaMatrix3D[i][j1][j2] = discreteRates[(i*numDiscreteRates) + u]; //initMatrix[j1][j2]]; //initMatrix[j1][j2]
							if (options.scaleByFrequencies) {
								mRateMatrix3D[i][j1][j2] = discreteRates[(i*numDiscreteRates) + u] * mFrequencyMatrix[i][j2]; //initMatrix[j1][j2]// [BLOCKING USE OF FREQS HERE - IMPLICIT IN NON-REV MODELS]
							} else {
								mRateMatrix3D[i][j1][j2] = discreteRates[(i*numDiscreteRates) + u]; // * (1.0/mNumStates) ; //initMatrix[j1][j2]// [BLOCKING USE OF FREQS HERE - IMPLICIT IN NON-REV MODELS]
							}
							rowsum += mRateMatrix3D[i][j1][j2];
						}
					}
				} else {
					std::cout << "Here we go..." << endl;
					if (j2 != j1) {
						u = (int)(Numerics::getGlobalNumerics().Random() * numDiscreteRates);
						//u = static_cast<int>(Numerics::getGlobalNumerics().Random() * numDiscreteRates);
						std::cout << u << " ";
						rateAssignments[(i*mNumStates*mNumStates)+(j1*mNumStates) + j2] = u; //initMatrix[j1][j2]; //initMatrix[j1][j2]
						mAlphaMatrix3D[i][j1][j2] = discreteRates[(i*numDiscreteRates) + u]; //initMatrix[j1][j2]]; //initMatrix[j1][j2]
						mRateMatrix3D[i][j1][j2] = discreteRates[(i*numDiscreteRates) + u];
						if (getReversible())  mRateMatrix3D[i][j1][j2] *= mFrequencyMatrix[i][j2]; //initMatrix[j1][j2]// [BLOCKING USE OF FREQS HERE - IMPLICIT IN NON-REV MODELS]
						rowsum += mRateMatrix3D[i][j1][j2];
					}
				}
				//numDiscreteRates-1;
				//for (int k=0; k<numDiscreteRates; k++) {
				//	if (u < normalized[ k ]) { rateAssignments[(i*mNumStates*mNumStates)+(j1*mNumStates) + j2] = rateAssignments[(i*mNumStates*mNumStates)+(j2*mNumStates) + j1] = k; break; }
				//}
			}
			mAlphaMatrix3D[i][j1][j1] = 0.0;
			mRateMatrix3D[i][j1][j1] = -1 * rowsum; //std::cout << "row " << j1 << ": " << mRateMatrix3D[i][j1][j1] << endl;
		}
	}
	
//	std::cout << "Done!" << endl;
}

//---------------------------------------------------------------------------
void ProteinModel::copyRateAssignments(int *inRates) {
	// Not assuming symmetric rates: assign rate category to each transition...
	
	for (int i=0; i<GetNumModels(); i++) {
		for (int j=0; j<mNumStates*mNumStates; j++) {
			rateAssignments[ (i*mNumStates*mNumStates) + j] = inRates[ (i*mNumStates*mNumStates) + j ];
		}
	}
}

void ProteinModel::backupRateAssignments(int *inRates) {
	// Not assuming symmetric rates: assign rate category to each transition...
	
	for (int i=0; i<GetNumModels(); i++) {
		for (int j=0; j<mNumStates*mNumStates; j++) {
			inRates[ (i*mNumStates*mNumStates) + j ] = rateAssignments[ (i*mNumStates*mNumStates) + j];
		}
	}
}


//---------------------------------------------------------------------------
void ProteinModel::initAndCopyGamma3D(int numModels, int numCat, double *value, int *inRates) {
	initGamma3D(numModels, numCat, value, noInitAssignments);
	copyRateAssignments(inRates);
}

//---------------------------------------------------------------------------
void ProteinModel::initGamma3D(int numModels, int numCat, double *value) {
	// Wrapper to initGamma3D (default - do random initialization at the end)...
	initGamma3D(numModels, numCat, value, randomInitAssignments);
}

//---------------------------------------------------------------------------
int ProteinModel::round(double a) {
	return int(a + 0.5);
}

void ProteinModel::initGamma3D(int numModels, int numCat, double *value, int initRatesMode) {
	// Set basic parameters...
	numDiscreteRates = numCat;

	if (RARdistributedRates) { 
		// Allocate memory if needed...
		gammaShapeParameter.resize(numModels); 
		gammaScaleParameter.resize(numModels); 
		discreteRates.resize(numModels * numDiscreteRates);
		rateAssignments.resize(numModels*mNumStates*mNumStates);
	}

	for (int i=0; i<numModels; i++) {
		gammaShapeParameter[i] = value[i]; //round((value[i] / 0.01))*0.01;
		gammaScaleParameter[i] = 2.0; //4.67 * (Numerics::getGlobalNumerics().Random() * 2.0); //round((value[i] / 0.01)) * 0.046875;
		
		// Setup rates vector...
//		setGammaDistribution( i );
		setAllGammaDistributions();
	}

	// Activate gamma model
	RARdistributedRates = true;
	
	// Random init if requested...
	//if (initRatesMode == randomInitAssignments) 
		randomInitRateAssignments();
}

/*
//---------------------------------------------------------------------------
void ProteinModel::setDistributionParmsFromCounts(void) {
	double a,b;

	double sum[mNumStates];
	double sampleMean = 0.0;
	double sampleVar = 0.0;
	
	for (unsigned int i=0; i<mNumStates; i++) {
		sum[i] = 0;
		for (unsigned int j=0; j < mNumStates; j++) {
			if (j != i) sum[i] += mCountMatrix[i][j];
		}
	}
	
	for (unsigned int i=0; i<mNumStates; i++) {
		for (unsigned int j=0; j<mNumStates; j++) {
			if (j != i) sampleMean += (mCountMatrix[i][j] / sum[i]);
		}
	}
	sampleMean /= (mNumStates-1)*mNumStates;

	for (unsigned int i=0; i<mNumStates; i++) {
		for (unsigned int j=0; j<mNumStates; j++) {
			if (j != i) sampleVar += ((mCountMatrix[i][j] / sum[i]) - sampleMean) * ((mCountMatrix[i][j] / sum[i]) - sampleMean);
		}
	}
	sampleVar /= (mNumStates-1)*mNumStates;
	
	// Beta distribution parameter estimates:
//	for (int i=0; i<mNumModels; i++) {
//		gammaShapeParameter[i] = sampleMean * ( ( ( sampleMean * (1-sampleMean) ) / sampleVar ) - 1.0 );
//		gammaScaleParameter[i] = (1.0 - sampleMean) * ( ( ( sampleMean * (1-sampleMean) ) / sampleVar ) - 1.0 );
//	} 
	
	// Gamma distribution parameter estimates:
	for (int i=0; i<mNumModels; i++) {
		gammaShapeParameter[i] = ( sampleMean / sqrt(sampleVar) ) * ( sampleMean / sqrt(sampleVar) );
		gammaScaleParameter[i] = ( sampleMean / sampleVar);
	}
	
//	std::cout << "			[Setting dist. param. 1 from counts:\t" << gammaShapeParameter[0] << "]" << endl;
//	std::cout << "			[Setting dist. param. 1 from counts:\t" << gammaScaleParameter[0] << "]" << endl;
	
	setAllGammaDistributions();
}
*/

void ProteinModel::setMaximumUniformization_RAR(void) {
	int i;
	double maxRate, maxLambda;
	
	maxRate = 0;
	for (i=0; i<numDiscreteRates; i++) {
		if ( (discreteRates[i] > maxRate) ) {
			maxRate = discreteRates[i];
		}
	}
	
	double maxRowSum = 0.0;
	for (int j=0; j < GetNumModels(); j++) {
		for (int k=0; k < mNumStates; k++) {
			if ( -mRateMatrix3D[j][k][k] > maxRowSum ) {
				maxRowSum = -mRateMatrix3D[j][k][k];
			}
		}
	}
	
				
	maxLambda = maxRowSum + ( maxRate * 3 ); //maxRate * mNumStates / 10.0;

	std::cout << "		[ Setting uniformization constants to maximum possible LambdaX " << maxLambda << " from the RAR distribution" << endl;
	for (i=0; i< mNumModels; i++) {
		setUniformization( i, maxLambda );
	}
}

//---------------------------------------------------------------------------
void ProteinModel::setDistributionParmsFromCounts( Tree& mTree ) {
	// Use this version to set parms based on rate distribution...
	//double a,b;
	static double *tempRates;
	static int tempalloc = 0;
	static double *array;
	static int alloc = 0;
	
	if (alloc == 0) {
		array = new double[ (mNumStates * (mNumStates) ) ];
		alloc = 1;
	}
	
	if (tempalloc == 0) { 
		// Wasn't allocating this properly before!  Need numDiscRates for each NumModel...
		tempRates = new double[(numDiscreteRates*mNumModels)+1]; 
		tempalloc = 1; 
	}
	
	
//	std::cout << "OVERRIDING setDistributionFromCounts()...has a problem with > 1 site class..." << endl;
//	return;
	
	static int firstTime = 0;
//	std::cout << "Got realchangemat as " << &realChangeCount3D << endl;
	
	for (unsigned int k = 0; k < mNumModels; k++)
	{
		
		double sum[mNumStates];
		double sampleMean = 0.0;
		double sampleVar = 0.0;
		
		// Force gamma updater...
		// firstTime = 1;
		
		if (firstTime == 1) {
			
			// mCountMatrix[i][j]
			
			// Method 1: using counts and/or rates...
			for (unsigned int i=0; i<mNumStates; i++) {
				sum[i] = 0;
				for (unsigned int j=0; j < mNumStates; j++) {
					if (j != i) sum[i] += mCountMatrix[i][j];
				}
			}
			
			for (unsigned int i=0; i<mNumStates; i++) {
				for (unsigned int j=0; j<mNumStates; j++) {
					if (j != i) sampleMean += (mCountMatrix[i][j] / sum[i]);
				}
			}
			sampleMean /= (mNumStates-1)*mNumStates;
			
			for (unsigned int i=0; i<mNumStates; i++) {
				for (unsigned int j=0; j<mNumStates; j++) {
					if (j != i) sampleVar += ((mCountMatrix[i][j] / sum[i]) - sampleMean) * ((mCountMatrix[i][j] / sum[i]) - sampleMean);
				}
			}
			sampleVar /= (mNumStates-1)*mNumStates;
			
			std::cout << "sample mean = " << sampleMean << ", sampleVar = " << sampleVar << endl;
			// Gamma distribution parameter estimates:
			gammaShapeParameter[k] = ( sampleMean / sqrt(sampleVar) ) * ( sampleMean / sqrt(sampleVar) );
			gammaScaleParameter[k] = ( sampleMean / sampleVar);
			
			std::cout << "			[Setting dist. param. 1 from rate distribution:\t" << gammaShapeParameter[k] << "]" << endl;
			std::cout << "			[Setting dist. param. 1 from rate distribution:\t" << gammaScaleParameter[k] << "]" << endl;
	
			// Needed for method 1...
			setAllGammaDistributions();

		} else {
		  /*
		// mCountMatrix[i][j]
			sampleMean = sampleVar = 0.0;
		// Method 1: using counts and/or rates...
			 for (unsigned int i=0; i<mNumStates; i++) {
				 sum[i] = 0;
				 for (unsigned int j=0; j < mNumStates; j++) {
					 if (j != i) sum[i] += (*realChangeCount3D).mRows.mDataArray[k].mRows.mDataArray[i].mDataArray[j];
				 }
				 std::cout << "row sum for model " << k << ", row " << i << " is " << sum[i] << endl;
			 }
			 
			 for (unsigned int i=0; i<mNumStates; i++) {
				 for (unsigned int j=0; j<mNumStates; j++) {
				 if (j != i) sampleMean += ((*realChangeCount3D).mRows.mDataArray[k].mRows.mDataArray[i].mDataArray[j] / sum[i]);
				 }
			 }
			 sampleMean /= (mNumStates-1)*mNumStates;
			
			 for (unsigned int i=0; i<mNumStates; i++) {
				 for (unsigned int j=0; j<mNumStates; j++) {
				 if (j != i) sampleVar += (((*realChangeCount3D).mRows.mDataArray[k].mRows.mDataArray[i].mDataArray[j] / sum[i]) - sampleMean) * (((*realChangeCount3D).mRows.mDataArray[k].mRows.mDataArray[i].mDataArray[j] / sum[i]) - sampleMean);
				 }
			 }
			 sampleVar /= (mNumStates-1)*mNumStates;
			 
			 std::cout << "		Sample mean = " << sampleMean << ", sampleVar = " << sampleVar << endl;
			 
			// Gamma distribution parameter estimates:
			 gammaShapeParameter[k] = ( sampleMean / sqrt(sampleVar) ) * ( sampleMean / sqrt(sampleVar) );
			 gammaScaleParameter[k] = ( sampleMean / sampleVar);

			 std::cout << "			[Setting dist. param. 1 from rate distribution:\t" << gammaShapeParameter[k] << "]" << endl;
			 std::cout << "			[Setting dist. param. 1 from rate distribution:\t" << gammaScaleParameter[k] << "]" << endl;
		}
		*/
		// Method 2: using empirical distribution's quantiles...
			

			for (unsigned int i=0; i<mNumStates; i++) {
				sum[i] = 0;
				for (unsigned int j=0; j < mNumStates; j++) {
					if (j != i) sum[i] += mCountMatrix[i][j];
				}
			}
			
			int count=0;
			for (unsigned int i=0; i<mNumStates; i++) {
				for (unsigned int j=0; j<mNumStates; j++) {
					if (j != i) array[count++] = mAlphaMatrix3D[k][i][j]; // * mFrequencyMatrix[k][j];
				}
			}
			num->simpleSort( array, (mNumStates*(mNumStates-1)) );
			
	//		std::cout << "ACTUAL (ORIGINAL) DIST:" << endl;
	//			for (int i=0; i<count; i++) {
	//				std::cout << i << ": " << array[i] << endl;
	//			}
			
			double min,max;
			min = num->gsl_stats_quantile_from_sorted_data(array, 1, (mNumStates * (mNumStates-1)), 0);
			max = num->gsl_stats_quantile_from_sorted_data(array, 1, (mNumStates * (mNumStates-1)), 1);
			
			double incr = (max - min) / (numDiscreteRates-1);

			std::cout << "	Model " << k << "..." << endl;
			for (int i=0; i<numDiscreteRates; i++) {
				if (i==0) tempRates[(k * numDiscreteRates) + i] = min;
				else if (i==(numDiscreteRates-1)) tempRates[(k * numDiscreteRates) + i] = max;
				else {
					tempRates[(k * numDiscreteRates) + i] = num->gsl_stats_quantile_from_sorted_data(array, 1, (mNumStates * (mNumStates-1)), (min + incr*i)/max);	
				}
		//		std::cout << "comp " << i << ": (incr is " << (min + incr*i)/max << ") " << discreteRates[(k*numDiscreteRates)+i] << endl;
			}
			
			for (int i=0; i<numDiscreteRates; i++) {
				if (i==0) discreteRates[(k * numDiscreteRates) + i] = min;
				else if (i==(numDiscreteRates-1)) discreteRates[(k * numDiscreteRates) + i] = max;
				else {
					discreteRates[(k * numDiscreteRates) + i] = tempRates[(k * numDiscreteRates) + i]; //-1] + (tempRates[(k * numDiscreteRates) + i] - tempRates[(k * numDiscreteRates) + i-1]) / 2;	
				}
				std::cout << "		Discrete category " << i << ": (incr is " << (min + incr*i)/max << ") " << discreteRates[(k*numDiscreteRates)+i] << endl;
			}
		}
		
		// Now need to set the rateAssignments to the closest rate...
		double temp;
		int m;
		for (int i=0; i<mNumStates; i++) {
			for (int j=0; j<mNumStates; j++) {
				if (j != i) {
					int closestID = -1;
					double closest = std::numeric_limits<double>::max( );
					for ( m=0; m<numDiscreteRates; m++) {
						temp =  mAlphaMatrix3D[k][i][j]; // * mFrequencyMatrix[k][j];
						if ( pow( ( temp - discreteRates[ (k*numDiscreteRates) + m]), 2.0) < closest) {
							closest = pow( ( temp - discreteRates[ (k*numDiscreteRates) + m]), 2.0);
							closestID = m;
						}
					}
				//	std::cout << "exchange " << i << ", " << j << ": closest is " << closestID << " - " << temp << " for " << discreteRates[ (k*numDiscreteRates) + closestID] << endl;
					if (closestID < 0) { std::cout << "		ERROR!  closestID is -1" << endl; }
					rateAssignments[ (k*mNumStates*mNumStates) + (i*mNumStates) + j] = closestID;
				}
			}
		}
		
	} // for k .. all models...
	rewriteAlphaMatrix3D();
	ProteinModelParameterstoRates();
	if (!options.capUniformizationRate) forceUpdateUniformization();
	
	//std::cout << "		Done initialization of RAR distribution." << endl;

	// Needed for method 1...
	// setAllGammaDistributions();
	
	firstTime=0;
}

//---------------------------------------------------------------------------
void ProteinModel::setGammaDistribution(int theModel) {
	// Code adapted from PIDAMEBL, A.P. Jason de Koning
	int which, status;
	double p,q,x,a,b, bound;

//	double sum = 0.0;
//	double theMin = 0.0001;
//	double increment;
//	double y;
	
	for (int k=0; k<numDiscreteRates; k++) {
		a = gammaShapeParameter[theModel];
		b = gammaScaleParameter[theModel]; 
	//	a = gammaShapeParameter[theModel] * (1.0 / gammaScaleParameter[theModel]);
	//	b = gammaShapeParameter[theModel] * (gammaScaleParameter[theModel]);
		
		which = 2;											// (calculate P and Q)
		p = (2.0*k+1.0) / (2.0 * numDiscreteRates);	// set region to integrate over
		q = (1-p);
		
		// Do the inverseCDF search for gamma distributions...
		cdfgam( &which, &p, &q, &x, &a, &b, &status, &bound);
		//cdfbet( &which, &p, &q, &x, &y, &a, &b, &status, &bound);
        //cdfnor( &which, &p, &q, &x, &mean, &sd, &status, &bound );
        
        //		sum += ( x ); // / options.numCategories );
//		discreteRates[(theModel * numDiscreteRates) + k] = x; // =a; // SET TO a TO SAMPLE A SHARED OVERALL RATE!

	//	if ( ( k==0 ) && ( x <= theMin ) ) { increment = (x - theMin ); } else { increment = 0.0; }
		discreteRates[(theModel * numDiscreteRates) + k] = x; //+ increment; // =a; // SET TO a TO SAMPLE A SHARED OVERALL RATE!
//
//		std::cout << "Just set rate " << k << " to " << x << endl;
	}

//	for (int k=0; k<numDiscreteRates; k++) {
//		discreteRates[(theModel * numDiscreteRates) + k] /= sum;
//		discreteRates[(theModel * numDiscreteRates) + k] *= gammaScaleParameter[theModel];
//	}

}
		
//---------------------------------------------------------------------------
void ProteinModel::setAllGammaDistributions(void) {
	// Code adapted from PIDAMEBL, A.P. Jason de Koning
	int which, status;
	double p,q,x,a,b, bound; //, y;
	// SET ALL THE MIXTURE COMPONENTS TO USE THE SAME GAMMA DIST (#0)...
//	double theMin = 0.0001;
//	double increment;
	
//	double sum = 0.0;
	for (int i=0; i<mNumModels; i++) {
	//	sum = 0.0;
		for (int k=0; k<numDiscreteRates; k++) {
			//a = b = gammaShapeParameter[0]; //theModel];
			a = gammaShapeParameter[0];
			b = gammaScaleParameter[0];
		//	a = gammaShapeParameter[0] * (1.0 / gammaScaleParameter[0]);
		//	b = gammaShapeParameter[0] * (gammaScaleParameter[0]); //gammaScaleParameter[theModel];

			which = 2;											// (calculate P and Q)
			p = (2.0*k+1.0) / (2.0 * numDiscreteRates);	// set region to integrate over
			q = (1-p);
			
			// Do the inverseCDF search for gamma distributions...
			cdfgam( &which, &p, &q, &x, &a, &b, &status, &bound);
			//cdfbet( &which, &p, &q, &x, &y, &a, &b, &status, &bound);

//			sum += ( x ); // / options.numCategories );
	//		discreteRates[(theModel * numDiscreteRates) + k] = x; // =a; // SET TO a TO SAMPLE A SHARED OVERALL RATE!
//			if ( x <= theMin ) { increment = theMin - x; } else { increment = 0.0; }
			discreteRates[(i * numDiscreteRates) + k] = x; // + increment; // =a; // SET TO a TO SAMPLE A SHARED OVERALL RATE!
	//
	//		std::cout << "Just set rate " << k << " to " << x << endl;
		}
		
//		for (int k=0; k<numDiscreteRates; k++) {
//			discreteRates[(i * numDiscreteRates) + k] /= sum; //*= gammaScaleParameter[0]/
//		}
	}
}
	
//---------------------------------------------------------------------------
void ProteinModel::updateDiscreteRates(void) {
	// This may not work very well.  testing...
	
//	double rannum = Numerics::getGlobalNumerics().Random(); //[jdk] Numerics::getGlobalNumerics().Random();
	double changeamount = num->box_muller( 0.0, 0.05 ); // (rannum - 0.5) * mWindowsize;
	double proposal;
	
	for (int i=0; i<mNumModels; i++) {
		for (int k=0; k<numDiscreteRates; k++) {
			if (Numerics::getGlobalNumerics().Random() > 0.2) {
				proposal = discreteRates[(i * numDiscreteRates) + k] + changeamount;
								
				if ((proposal <= 0.0 )) {
					proposal = discreteRates[(i * numDiscreteRates) + k] - changeamount;
				}

				discreteRates[(i * numDiscreteRates) + k] = proposal;
			}
		}
		
	}
}
//---------------------------------------------------------------------------

void ProteinModel::setUniformization(DoubleVector* vals) {
	if (uniformizationConstants.empty()) { uniformizationConstants.resize( mNumModels, 0.0 ); }
	
	for (int i=0; i<mNumModels; i++) {
		uniformizationConstants.mDataArray[i] = vals->mDataArray[i];
		//	std::cout << "		SetUniformization() -> setting to " << vals->mDataArray[i] << endl;
	}
}

void ProteinModel::setUniformization(int which, double val) {
	if (uniformizationConstants.empty()) { uniformizationConstants.resize( mNumModels, 0.0 ); }

	uniformizationConstants.mDataArray[which] = val;
}

const ProteinModel& ProteinModel::operator=(
				const ProteinModel&	inModel)
{
	copy(inModel);
    return *this;
} 

//---------------------------------------------------------------------------
void ProteinModel::swap(
			ProteinModel&	inModel)
{
	MixtureModel::swap(inModel);
	mAlphaMatrix3D.swap(inModel.mAlphaMatrix3D);
	std::swap(mFrequencyMatrix, inModel.mFrequencyMatrix);
	mCountMatrix.swap(inModel.mCountMatrix);
}

//---------------------------------------------------------------------------
void ProteinModel::SetCountMatrix(const IntMatrix& inCountMatrix)
{
	mCountMatrix = inCountMatrix;
}


//---------------------------------------------------------------------------
void ProteinModel::InitializeFrequencyMatrix(
		const IntMatrix& inCountMatrix)
{
	if (mNumStates == 0)
		mNumStates = inCountMatrix.rows();
	
	//	std::cout << "INITIALIZNG FREQ MATRIX" << endl;
			
	if (mFrequencyMatrix.empty())
		mFrequencyMatrix.resize(mNumModels, mNumStates, 0);
	
	double	total=0.0;
	
	DoubleVector sum(mNumStates, 0);

	for (unsigned int i=0; i<mNumStates; i++)
	{
		for (unsigned int j=0; j < mNumStates; j++)
			sum[i]+= inCountMatrix[i][j];

		total += sum[i];
	}
	
	//  Freqs initialized here...
	for (unsigned int k = 0; k < mNumModels; k++)
		for (unsigned int i=0; i < mNumStates; i++) {
			if (getFixedFreqs()) {
				mFrequencyMatrix[k][i] = (1.0/mNumStates); //sum[i]/total; //
			} else {
				mFrequencyMatrix[k][i] = sum[i]/total;
			}
		}
}
			
						
void ProteinModel::setPredefinedRatesMatrix(int which, DoubleMatrix& theRates) {
	for (int i=0; i<GetNumStates(); i++) {
		for (int j=0; j<GetNumStates(); j++) {
			if (which == POISSON) {
				theRates[i][j] = 0.05;
			} else {
				theRates[i][j] = predefinedRates(which, i, j);
			}
		}
	}
}

void ProteinModel::setPredefinedFreqs(int which, DoubleMatrix& theFreqs) {
	if (getFixedFreqs()) {
		for (int i=0; i<GetNumModels(); i++) {
			for (int j=0; j<GetNumStates(); j++) {
				if (which == POISSON) {
					theFreqs[i][j] = 0.05;
				} else {
					theFreqs[i][j] = predefinedFreqs(which, j);
				}
			}
		}
	}
}



//---------------------------------------------------------------------------
double ProteinModel::predefinedFreqs(int which, int i) {
	if (which == JTT) {
		return freqs_JTT[i];
	} else if (which == MTMAM) {
		return freqs_MTMAM[i];
	} else {
		std::cout << "REQUESTED AN UNDEFINED RATE MATRIX." << endl; exit(-1);
	}
}

//---------------------------------------------------------------------------
double ProteinModel::predefinedRates(int which, int i, int j) {
	double returnVal;
	if (which == JTT) {
		returnVal = rates_JTT[i][j] / freqs_JTT[j];
	} else if (which == MTMAM) {
		returnVal = rates_MTMAM[i][j] / freqs_MTMAM[j];
	} else {
		std::cout << "REQUESTED AN UNDEFINED RATE MATRIX." << endl; exit(-1);
	}

	if ( (j!=i) && (returnVal < 0.00001) ) { returnVal = 0.00001; }
	return returnVal;
}

//---------------------------------------------------------------------------
void ProteinModel::InitializeAlphaMatrix3D(
		const IntMatrix& inCountMatrix)
{
	if (mNumStates == 0)
		mNumStates = inCountMatrix.rows();

	if (mFrequencyMatrix.empty())
		InitializeFrequencyMatrix(inCountMatrix);

	if (mAlphaMatrix3D.empty())
		mAlphaMatrix3D.resize(mNumModels, mNumStates, mNumStates, 0.0);
	
	DoubleVector sum(mNumStates, 0);
	
	DoubleMatrix Rates(mNumStates, mNumStates, 0);

	for (unsigned int i=0; i<mNumStates; i++)
		for (unsigned int j=0; j < mNumStates; j++)
			sum[i]+= inCountMatrix[i][j];

	std::cout << endl;
	// Rates initialized here...

	if (getPredefinedMatrix() != -1) {
		std::cout << "*** Setting rate matrix to pre-defined version..." << endl;
		setPredefinedRatesMatrix(getPredefinedMatrix(), Rates); // [override] inCountMatrix[i][j] / sum[i]; //1.0;
		setPredefinedFreqs(getPredefinedMatrix(), mFrequencyMatrix);
	} else {
		std::cout << "*** Setting rate matrix normally..." << endl;
		for (unsigned int i=0; i<mNumStates; i++) {
			for (unsigned int j=0; j<mNumStates; j++) {
			//	std::cout << "Setting matrix by counts..." << endl;
				double tempVal;
				if (sum[i] != 0)  {
					tempVal = inCountMatrix[i][j] / sum[i];
				} else {
					tempVal = 0.00001;
				}
				
				if (tempVal < 0.00001) tempVal = 0.00001;
				Rates[i][j] = tempVal;

         //      std::cout << "(" << i << "," << j << ") = " << Rates[i][j] << " (" << inCountMatrix[i][j] << ")" << endl; 
			} //std::cout << endl;
		}
	}
	
//	std::cout << "part2..." << endl;
	
	IntMatrix found(mNumStates, mNumStates, 1);
			
	for (unsigned int k = 0; k < mNumModels; k++)
	{
		for (unsigned int i=0; i < mNumStates; i++)
			for (unsigned int j=0; j < mNumStates; j++)
				if (i == j)
					found[i][j] = 0;
				else
					found[i][j] = 1;

		for (unsigned int i=0; i < mNumStates; i++) {
			for (unsigned int j=0; j < mNumStates; j++) {
				if (found[i][j]) {
		//			std::cout << "DIAG: alphamatrix3d is " << mAlphaMatrix3D.empty() << ", i is " << i<< ", j is " << j << ", k (model) is " << k << endl;
		//			std::cout << "	test of rates[i,j]: " << Rates[i][j] << endl;
		//			std::cout << "	test of rates[j,i]: " << Rates[j][i] << endl;
		//			std::cout << "	test of mFrequencyMatrix[k,j]: " << mFrequencyMatrix[k][j] << endl;
		//			std::cout << "	test of mFrequencyMatrix[k,i]: " << mFrequencyMatrix[k][i] << endl;
		//			std::cout << "	test of mAlphaMatrix3D[k,i,j]: " << mAlphaMatrix3D[k][i][j] << endl;
		//			std::cout << "  test of calc: " << ((Rates[i][j] / mFrequencyMatrix[k][j] + Rates[j][i] / mFrequencyMatrix[k][i]) / 2) << endl;
		//			std::cout << "  confirming k,i,j = " << k << ", " << i << ", " << j << endl;
// [blocking the init routine for conjugate gibbs approach ]

// [block]					if (!RARdistributedRates) {
						if (getReversible()) {
						// [Mar 31 - Jason changing this; why on earth would we divide by the frequency?? should be multiplying]
							if (options.scaleByFrequencies) {
								mAlphaMatrix3D[k][i][j] = (Rates[i][j] / mFrequencyMatrix[k][j] + Rates[j][i] / mFrequencyMatrix[k][i]) / 2;
							} else {
								mAlphaMatrix3D[k][i][j] = (Rates[i][j]  + Rates[j][i] ) / 2;
//								mAlphaMatrix3D[k][i][j] = (Rates[i][j] / (1.0/mNumStates) + Rates[j][i]/ (1.0/mNumStates) ) / 2;
							}// mAlphaMatrix3D[k][i][j] = (Rates[i][j] * mFrequencyMatrix[k][j] + Rates[j][i] * mFrequencyMatrix[k][i]) / 2;
						} else {
							mAlphaMatrix3D[k][i][j] = Rates[i][j]; // / mFrequencyMatrix[k][j];
//							mAlphaMatrix3D[k][i][j] = Rates[i][j] / mFrequencyMatrix[k][j];
						}
					
//					std::cout << "	test of mAlphaMatrix3D[k,i,j]: " << mAlphaMatrix3D[k][i][j] << endl;

// [block]
/*						} else {
			//			std::cout << "doing rate assignments from discreterates distribution" << endl;
			//			std::cout << " there are " << numDiscreteRates << " disc rates" << endl;
						rateAssignments.resize(mNumStates*mNumStates*mNumModels);

			//			std::cout << " rate assign is: " <<  rateAssignments[ (k*mNumStates*mNumStates) + (i*mNumStates) + j] << endl;
						
						if (getReversible()) {
							mAlphaMatrix3D[k][i][j] = discreteRates[ (k*numDiscreteRates) + rateAssignments[ (k*mNumStates*mNumStates) + (i*mNumStates) + j] ];
							mAlphaMatrix3D[k][j][i] = mAlphaMatrix3D[k][i][j];
						} else {
							mAlphaMatrix3D[k][i][j] = discreteRates[ (k*numDiscreteRates) + rateAssignments[ (k*mNumStates*mNumStates) + (i*mNumStates) + j] ];
						}
					}
 */
					if (RARdistributedRates) {
						rateAssignments.resize(mNumStates*mNumStates*mNumModels);
						//setDistributionParmsFromCounts(mTree);
					}
					
					//		std::cout << "Done..." << endl;
					if (getScaleMatrix()) {
						mAlphaMatrix3D[k][i][j] *= getRateScale(k);
					}
				}
			}
		}

//		std::cout << "done" << endl;
	}		
}
	
//---------------------------------------------------------------------------
void ProteinModel::ProteinModelParameterstoRates(void) {
	ProteinModelParameterstoRates( &mFrequencyMatrix );
}
	
void ProteinModel::ProteinModelParameterstoRates(DoubleMatrix* freqs) {
	// Modified by Jason to use a pointer to the rates matrix...
	double sum;
	
	if (mNumStates == 0)
		mNumStates = mCountMatrix.rows();

	if (mFrequencyMatrix.empty()) {
//		std::cout << "doing init on freqs!" << endl;
	//	std::cout << "this call uses mCountMatrix!" << endl;
		InitializeFrequencyMatrix(mCountMatrix);
	}
	
	if (mAlphaMatrix3D.empty()) {
//		std::cout << "doing init on alphamatrix3d" << endl;
	//	std::cout << "this call uses mCountMatrix!" << endl;
		InitializeAlphaMatrix3D(mCountMatrix);
	}
	
	IntMatrix found(mNumStates, mNumStates, 1);
			
	DoubleMatrix3D *Rates3D = returnRateMatrix3D();

	if (Rates3D->empty()) {
//		std::cout << "doing resize on Rates3D matrix -it was empty!" << endl;
		Rates3D->resize(mNumModels, mNumStates, mNumStates, 0.0);
	} 
	
	for (unsigned int k = 0; k < mNumModels; k++)
	{
		for (unsigned int i=0; i < mNumStates; i++){
			for (unsigned int j=0; j < mNumStates; j++) {
				if (i == j) {
					found[i][j] = 0;
				} else {
					found[i][j] = 1;
				}
			}
		}

		for (unsigned int i=0; i<mNumStates; i++)
		{
			for (unsigned int j=0; j<mNumStates; j++) {
				if (found[i][j]) {
			//		if ( (*Rates3D)[k][i][j] != mAlphaMatrix3D[k][i][j] * mFrequencyMatrix[k][j] ) {
//[bigdebug]			//			std::cout << "parms->rates error " << k << ", " << i << ", " << j << ": orig " << (*Rates3D)[k][i][j] << "; new: " << (mAlphaMatrix3D[k][i][j] * mFrequencyMatrix[k][j]) << ", freq part " << mFrequencyMatrix[k][j] << endl;
			//		}
					if (options.scaleByFrequencies) {
						(*Rates3D)[k][i][j] = mAlphaMatrix3D[k][i][j] * (*freqs)[k][j];// [BLOCKING USE OF FREQS HERE - IMPLICIT IN NON-REV MODELS]
					} else {
						(*Rates3D)[k][i][j] = mAlphaMatrix3D[k][i][j]; // / (1.0/mNumStates) ;// [BLOCKING USE OF FREQS HERE - IMPLICIT IN NON-REV MODELS]
					}
						// [??]			if (getReversible()) (*Rates3D)[k][i][j] *= mFrequencyMatrix[k][j];
				//	std::cout << i << ","<<j<<": " << setprecision(3) << (*Rates3D)[k][i][j] << "\t" << setprecision(7);
				//	std::cout << "model " << k << ", i si " << i << ", j " << j << ", malphamat " << mAlphaMatrix3D[k][i][j] << ", mfreq " << mFrequencyMatrix[k][j] << endl;
				} //else {
			//		std::cout << "found says no " << i << ", " << j << endl;
		//		}
			}//std::cout << endl;
		}//std::cout << endl << endl;

		for (unsigned int i=0; i<mNumStates; i++)
		{
			sum = 0.0;
		
			for (unsigned int j=0; j<mNumStates; j++)
			{
				if (i != j) {
					sum += (*Rates3D)[k][i][j];
				}
			}
		
//[bigdebug]			if ( (*Rates3D)[k][i][i] != (-1.0 * sum)) {
//				std::cout << "row sum is incorrect for " << k << ", " << i << " x " << i << ": orig " << (*Rates3D)[k][i][i] << ", new: " << (-1.0 * sum) << endl;
//			}
			(*Rates3D)[k][i][i] = -1.0 * sum;
	//		std::cout << "sum for row i " << i << " is " << sum << endl;
		}
	}
	
// [jdk]	SetRateMatrix3D(Rates3D);
}
			
//---------------------------------------------------------------------------
void ProteinModel::InitializeParameter(
		const IntMatrix& inCountMatrix)
{
//	std::cout << "countmatrix" << endl;
	SetCountMatrix(inCountMatrix);
//	std::cout << "InitializeFrequencyMatrix" << endl;
	InitializeFrequencyMatrix(inCountMatrix);
//	std::cout << "InitializeAlphaMatrix3D" << endl;
	InitializeAlphaMatrix3D(inCountMatrix);
//	std::cout << "randominitrateassign..." << endl;
//	randomInitRateAssignments();
//	std::cout << "ProteinModelParameterstoRates" << endl;
	ProteinModelParameterstoRates();
//	std::cout << "InitializeFreqModels" << endl;
	InitializeFreqModels();
//	std::cout << "SetModelNumNucleotides" << endl;
	SetModelNumNucleotides(1);
	
	if (options.freeFreqs) updateStationaryFreqs(1);
}
			
			
//---------------------------------------------------------------------------
void ProteinModel::PrintRateParameters(ostream&	outStream) {
//[jdk]	PLEXApp* m_oPLEXApp = new(PLEXApp);

	if (RARdistributedRates) {
		// First print gamma scale parameters for each model...
		for (int j=0; j<mNumModels; j++) {
			outStream << "RAR Distribution parms:\t" << setprecision(5) <<  gammaShapeParameter[j] << "\t" << setprecision(5) <<  gammaScaleParameter[j] << "\t";
			// shape parameter:\t";
			//m_oPLEXApp->PrintDoubleArray(outStream, mNumModels, &gammaShapeParameter);
			//outStream << "Gamma scale parameter:\t";
			//m_oPLEXApp->PrintDoubleArray(outStream, mNumModels, &gammaScaleParameter);

			// Second print the median rates in each category...
//			for (int i=0; i<mNumModels; i++) {
				m_oPLEXApp->PrintDoubleArray(outStream, numDiscreteRates, &discreteRates, (j*numDiscreteRates));
//			}
		}
	}
//[jdk]	delete(m_oPLEXApp);
}

//---------------------------------------------------------------------------
void ProteinModel::PrintDependentRates(ostream& outStream) {
//[jdk]	PLEXApp* m_oPLEXApp = new(PLEXApp);
	if (outStream) {		
		for (int i=1; i<mNumModels; i++) {
			outStream << "DEPENDENCY ASSIGNMENTS:\t";	
			outStream << "MODEL " << i << endl;
			for (int j=0; j<GetNumStates(); j++) {
				m_oPLEXApp->PrintIntVector(outStream, dependencyIndicators[i][j]);
			}
			outStream << endl;
		}
	}
//[jdk]	delete(m_oPLEXApp);
}



//---------------------------------------------------------------------------
void ProteinModel::PrintAlphaMatrix3D(
			ostream& outStream, ObservedData& theData )
{
//[jdk]	PLEXApp* m_oPLEXApp = new(PLEXApp);

	m_oPLEXApp->PrintMatrix3D(outStream, mAlphaMatrix3D, theData);
	
//[jdk]	delete(m_oPLEXApp);
}

//---------------------------------------------------------------------------
void ProteinModel::PrintCountMatrix(
			ostream& outStream, ObservedData& theData)
{
//[jdk]	PLEXApp* m_oPLEXApp = new(PLEXApp);

	m_oPLEXApp->PrintMatrix(outStream, mCountMatrix, theData);
	
//[jdk]	delete(m_oPLEXApp);
}

//---------------------------------------------------------------------------
void ProteinModel::PrintFrequencyMatrix(
			ostream& outStream, ObservedData& theData )
{
//[jdk]	PLEXApp* m_oPLEXApp = new(PLEXApp);

	//DoubleVector tempVector(mNumStates, 0);
	
	for (unsigned int k = 0; k< mNumModels; k++)
	{
	//	tempVector = mFrequencyMatrix[k];
		
		m_oPLEXApp->PrintVector(outStream, theData, mFrequencyMatrix[k]);
	}
	
//[jdk]	delete(m_oPLEXApp);
}


//---------------------------------------------------------------------------
void ProteinModel::UpdateFrequencyParameter(
		double& inParameter)
{
//PLEXApp* m_oPLEXApp = new(PLEXApp);

	double rannum = Numerics::getGlobalNumerics().Random(); //[jdk] Numerics::getGlobalNumerics().Random();
	
//[bad]
	double changeamount = (rannum - 0.5) * mWindowsize;
//	double changeamount = num->box_muller(0.0, mWindowsize);
//	std::cout << "using variance " << inParameter*mWindowsize << " - got variate " << changeamount << " for PROPOSAL: " << (inParameter + changeamount) << endl;
	double proposal = inParameter + changeamount;
							
	if ((proposal <= 0.0 ) || (proposal >= 1.0))
		proposal = inParameter - changeamount;
		
//	std::cout << "Proposal for freq parameter; FROM " << inParameter << " to " << proposal << endl;

	inParameter = proposal;
	//[jdk] delete(m_oPLEXApp);
}

//---------------------------------------------------------------------------
void ProteinModel::UpdateAlphaParameter( double& inParameter)
{
//	PLEXApp* m_oPLEXApp = new(PLEXApp);

	if (options.doRAR == 1) { std::cout << "ERROR! Called the wrong parameter update for the gamma model" << endl; exit(-1); }

//	double rannum = Numerics::getGlobalNumerics().Random(); //[jdk] Numerics::getGlobalNumerics().Random();
	double changeamount = num->box_muller(0.0, inParameter*0.01);
//	double changeamount = (rannum - 0.5) * mWindowsize;
	double proposal = inParameter + changeamount;
							
	if (proposal <= 0.0 )
		proposal = inParameter - changeamount;
		
	inParameter = proposal;
	
							
//[jdk]	delete(m_oPLEXApp);
}

void ProteinModel::UpdateAlphaParameter(int& theRate) {
//	PLEXApp* m_oPLEXApp = new(PLEXApp);
	// Adjust rate category membership by 1...
	
	if (!RARdistributedRates) { std::cout << "ERROR! Called the wrong parameter update for a non-gamma model" << endl; exit(-1); }

	int proposal = Numerics::getGlobalNumerics().Random() < 0.5 ? -1 : +1;
	if ( ( (proposal+theRate) < 0) || ( (proposal+theRate) >= numDiscreteRates) ) { proposal *= -1; } 
	//std::cout << "Going from " << theRate << "to " << proposal << endl;
	theRate += proposal;

//[jdk]	delete(m_oPLEXApp);
}


//---------------------------------------------------------------------------
/*void ProteinModel::UpdateFrequency(DoubleVector& mFrequency)
{
	bool m_bUpdate = false;
	
	DoubleVector tempFrequency;

	while (!m_bUpdate)
	{
		double sum = 0;
	
		tempFrequency = mFrequency;
	
		IntVector found(mNumStates, 0);
	
		for (int i = 0; i < mNumStates; i++)
		{
			if ( i == 0)
				found[i] = 0;
			else
				found[i] = 1;
		}
	
		for (int i = 0; i < mNumStates; i++)
		{
			if (found[i])
			{
				UpdateFrequencyParameter(tempFrequency[i]);
		
				sum += tempFrequency[i];
			}
		}
	
		tempFrequency[0] = 1 - sum;
		
		if ((tempFrequency[0] >= 1.0) || (tempFrequency[0] <= 0.0))
			m_bUpdate = false;
		else
			m_bUpdate = true;
	}
	
	if (m_bUpdate)
		mFrequency = tempFrequency;
}
*/

void ProteinModel::UpdateFrequency(DoubleVector& mFrequency)
{
	int i;
	int whichFreq[mNumStates];
	// Called for each model component...
	bool m_bUpdate = false;
	
	DoubleVector tempFrequency;

	while (!m_bUpdate) {
		for (i = 0; i < mNumStates; i++) {
			whichFreq[i] = 0;
		}
		
		double sum = 0;
	
		tempFrequency = mFrequency;
	
		IntVector found(mNumStates, 0);
	
		// Update one frequency...
		double updateFraction =  num->Random() / 2.0; //Numerics::getGlobalNumerics().Random() / 2.0; //(2.0 / mNumStates); // / (mNumStates * mNumStates); // Numerics::getGlobalNumerics().Random(); //[jdk] Numerics::getGlobalNumerics().Random();

		if (updateFraction*mNumStates < 1) { updateFraction = 1/mNumStates; }
		
		int updatedCount = 0;
		
		while (updatedCount < (updateFraction * mNumStates) ) {
			int choice = (int) ( Numerics::getGlobalNumerics().Random() * (mNumStates) );
		//	std::cout << "Choice is " << choice << endl;
			if (whichFreq[choice] != 1) {
				whichFreq[choice] = 1;
				updatedCount++;
			}
		}
		
//		for (i=0; i<mNumStates; i++) {
//			whichFreq[i] = Numerics::getGlobalNumerics().Random() <= (updateFraction)? 1 : 0; //[jdk] Numerics::getGlobalNumerics().Random();
		//	std::cout <<  "Whichfreq for " << i << " (prob of update = " << (updateFraction) << "); update? " << whichFreq[i] << endl;
//		}
		
		for (i = 0; i < mNumStates; i++)
		{
			if ( i == 0)
				found[i] = 0; /// was 0
			else
				found[i] = 1;
		}
	
		sum = 0.0;
		for (i = 0; i < mNumStates; i++)
		{
			if (found[i])
			{
		//		UpdateFrequencyParameter(tempFrequency[i]);
				if (whichFreq[i]) UpdateFrequencyParameter(tempFrequency[i]);
		 
				sum += tempFrequency[i];
			}
		}
	
	//	for (i = 0; i < mNumStates; i++) {
	//		tempFrequency[i] /= sum;
	//	}

	  tempFrequency[0] = 1 - sum;
		
		if ((tempFrequency[0] >= 1.0) || (tempFrequency[0] <= 0.0))
			m_bUpdate = false;
		else
			m_bUpdate = true;
	}
	
	if (m_bUpdate)
		mFrequency = tempFrequency;
}
		
//---------------------------------------------------------------------------
void ProteinModel::UpdateFrequencyMatrix(void)
{
	//std::cout << "UPDATING FREQS..." << endl;
	
	// Update the frequencies over all model components...
	for (unsigned int k = 0; k< mNumModels; k++)
		UpdateFrequency(mFrequencyMatrix[k]);
}

//---------------------------------------------------------------------------
/*void ProteinModel::UpdateAlphaMatrix(DoubleMatrix& mAlphaMatrix, const DoubleVector& mFrequency, const int& m_nMethod)
{
	int i,j;
	bool m_bUpdate = false;
//	std::cout << "UpdateAlphaMatrix begins..." << endl;
	static DoubleMatrix tempAlphaMatrix;
	
	if (options.m_nMethod == 0)
	{
		while (!m_bUpdate)
		{
			tempAlphaMatrix = mAlphaMatrix;
		
			IntMatrix found(mNumStates, mNumStates, 0);
	
			for (i = 0; i < mNumStates; i++)
			{
				for (j = 0; j < mNumStates; j++)
				{
					if ( i == j)
						found[i][j] = 0;
					else
						found[i][j] = 1;
				}
			}
	
			for (i = 0; i < mNumStates; i++)
			{
				for (j = 0; j < mNumStates; j++)
				{
					if (found[i][j])
					{
						UpdateAlphaParameter(tempAlphaMatrix[i][j]);
						tempAlphaMatrix[j][i] = tempAlphaMatrix[i][j];
						found[i][j] = 0;
						found[j][i] = 0;
					}
				}
			
				double sum = 0;
	
				double tempRate;
			
				for (j = 0; j < mNumStates; j++)
				{
					if (i != j)
					{
						tempRate = tempAlphaMatrix[i][j] * mFrequency[j];
						sum += tempRate;
					}
				}

				tempRate = 1 - sum;
			
				if ((tempRate <= 0.0) || (tempRate >= 1.0))
				{
					m_bUpdate = false;
					break;
				}
				else
				{
					m_bUpdate = true;
					tempAlphaMatrix[i][i] = 0;
				}
			}
		}
	}
	else
	{
		tempAlphaMatrix = mAlphaMatrix;
		
		IntMatrix found(mNumStates, mNumStates, 0);
	
		for (i = 0; i < mNumStates; i++)
		{
			for (j = 0; j < mNumStates; j++)
			{
				if ( i == j)
					found[i][j] = 0;
				else
					found[i][j] = 1;
			}
		}
	
		for (i = 0; i < mNumStates; i++)
		{
			for (j = 0; j < mNumStates; j++)
			{
				if (found[i][j])
				{
					UpdateAlphaParameter(tempAlphaMatrix[i][j]);
					tempAlphaMatrix[j][i] = tempAlphaMatrix[i][j];
					found[i][j] = 0;
					found[j][i] = 0;
				}
			}
		
			m_bUpdate = true;
			tempAlphaMatrix[i][i] = 0;
		}
	}

	if (m_bUpdate)
		mAlphaMatrix = tempAlphaMatrix;
		
//	std::cout << "updateAlphaMatrix done" << endl;
}
*/
//::updatePmatB1_singleParameter(int x, int y) {

// for gamma - pass in UpdateAlphaParameter( theAssignments[(i*mNumStates)+j] );

void ProteinModel::UpdateSingleParameterB1(int whichModel, int x, int y) { 
	// Updates a specified parameter only; this function ONLY changes what needs changing - and updates the ratematrix, alphamatrix, probmatrix entries
	if (RARdistributedRates) { std::cout << "Wrong updatePmatB1_singleParameter called" << endl; exit(-1); }

	DoubleMatrix3D *Rates3D = returnRateMatrix3D();
	DoubleMatrix& mRateMatrix = (*Rates3D)[whichModel];
	DoubleMatrix& mAlphaMatrix = mAlphaMatrix3D[whichModel];
	DoubleVector& mFrequency = mFrequencyMatrix[whichModel];

//	int i,j;
//	bool m_bUpdate = false;
	double oldRate, newRate;
	
//	if (options.m_nMethod == 0) {
		oldRate = mRateMatrix[x][y];
		newRate = oldRate;
		UpdateAlphaParameter( newRate );
		
		// Update rate and alpha matrix with newRate...
		mAlphaMatrix[x][y] = newRate;
		if (options.scaleByFrequencies) {
			mRateMatrix[x][y] = newRate * mFrequency[y];
		} else {
			mRateMatrix[x][y] = newRate; // * (1.0/mNumStates);
		}
		// Do the same for y->x if reversible model...
		if (getReversible()) {
			mAlphaMatrix[y][x] = newRate;
			if (options.scaleByFrequencies) {
				mRateMatrix[y][x] = newRate * mFrequency[x];
			} else {
				mRateMatrix[y][x] = newRate; // * (1.0/mNumStates);
			}
			// Update y->y too...
			if (x != y) {
				if (options.scaleByFrequencies) {
					mRateMatrix[y][y] -= (oldRate * mFrequency[x]);
					mRateMatrix[y][y] += mAlphaMatrix[y][x];
				} else {
					mRateMatrix[y][y] -= oldRate; // * (1.0/mNumStates);
					mRateMatrix[y][y] += mAlphaMatrix[y][x];		
				}
			}
		}
		
		// Update x->x...
	if (options.scaleByFrequencies) {
		mRateMatrix[x][x] -= (oldRate * mFrequency[y]);
		mRateMatrix[x][x] += mAlphaMatrix[x][y];
	} else {
		mRateMatrix[x][x] -= oldRate; // * (1.0/mNumStates) ;
		mRateMatrix[x][x] += mAlphaMatrix[x][y];		
	}
		// NOTE: THE ROW AND COL FOR X,Y of the Pmat must be updated too!
			
//	}
}

void ProteinModel::scaleDiscreteRates( double factor ) {
	// Rescale the RAR rates distribution...
	for (int m=0; m < GetNumModels(); m++ ) {
		for (int i=0; i < numDiscreteRates; i++) {
			discreteRates[(m*numDiscreteRates) + i] /= factor;
		}
	}
}

void ProteinModel::UpdateRateAssignmentB1(int whichModel, int x, int y, int newCategory) { 
	// Updates a specified parameter only; this function ONLY changes what needs changing - and updates the ratematrix, alphamatrix, probmatrix entries
	if (!RARdistributedRates) { std::cout << "Wrong updatePmatB1_singleParameter called" << endl; exit(-1); }

	// JUST ADDED IN A BUNCH OF FREQUENCY STUFF HERE...
	
//	int i,j;
//	bool m_bUpdate = false;
	int oldAssign, newAssign;
	double oldRate, newRate, oldRate2;
	//double oldFreq, oldFreq2;
	
	DoubleMatrix3D *Rates3D = returnRateMatrix3D();
	DoubleMatrix& mRateMatrix = (*Rates3D)[whichModel];
	DoubleMatrix& mAlphaMatrix = mAlphaMatrix3D[whichModel];
	DoubleVector& mFrequency = mFrequencyMatrix[whichModel];
	if (x==y) { std::cout << "ERROR!  UpdateRateAssignmentB1 called for same state!  Don't do that." << endl; exit(-1); }
	
		oldAssign = rateAssignments[(whichModel * mNumStates * mNumStates) + (x * mNumStates) + y];
		newAssign = newCategory;
		
		rateAssignments[(whichModel * mNumStates * mNumStates) + (x * mNumStates) + y] = newAssign;
		if (getReversible()) {
			rateAssignments[(whichModel * mNumStates * mNumStates) + (y * mNumStates) + x] = newAssign;
		}
		
		oldRate  = mRateMatrix[x][y]; //discreteRates[(whichModel*numDiscreteRates) + oldAssign]; //mAlphaMatrix[x][y];
		oldRate2 = mRateMatrix[y][x]; //discreteRates[(whichModel*numDiscreteRates) + oldAssign]; //mAlphaMatrix[x][y];
		
	//	oldFreq   = mRateMatrix[x][y] / discreteRates[(whichModel*numDiscreteRates) + oldAssign];
	//	oldFreq2  = mRateMatrix[y][x] / discreteRates[(whichModel*numDiscreteRates) + oldAssign];
		//newRate = oldRate;
		newRate = discreteRates[(whichModel*numDiscreteRates) + newAssign];
		
		// Update rate and alpha matrix with newRate...
		mAlphaMatrix[x][y] = newRate;
		if (options.scaleByFrequencies) {
			mRateMatrix[x][y] = newRate * mFrequency[y];  // [BLOCKING USE OF FREQS HERE - IMPLICIT IN NON-REV MODELS] * mFrequency[y];// [BLOCKING USE OF FREQS HERE - IMPLICIT IN NON-REV MODELS]
		} else {
			mRateMatrix[x][y] = newRate; // * (1.0/mNumStates);  // [BLOCKING USE OF FREQS HERE - IMPLICIT IN NON-REV MODELS] * mFrequency[y];// [BLOCKING USE OF FREQS HERE - IMPLICIT IN NON-REV MODELS]
		}
		// Do the same for y->x if reversible model...
		if (getReversible()) {
			if (options.scaleByFrequencies) {
				mAlphaMatrix[y][x] = newRate;
				mRateMatrix[y][x] = newRate * mFrequency[x];// [BLOCKING USE OF FREQS HERE - IMPLICIT IN NON-REV MODELS]

				// Update y->y too...
				if (x != y) {
					mRateMatrix[y][y] += ( oldRate2 );
					mRateMatrix[y][y] -= ( mRateMatrix[y][x] );
				}
			} else {
				mAlphaMatrix[y][x] = newRate;
				mRateMatrix[y][x] = newRate; // * (1.0/mNumStates);// [BLOCKING USE OF FREQS HERE - IMPLICIT IN NON-REV MODELS]
				
				// Update y->y too...
				if (x != y) {
					mRateMatrix[y][y] += ( oldRate2 );
					mRateMatrix[y][y] -= ( mRateMatrix[y][x] );
				}
				
			}
		}
		
	//	std::cout << "CHANGED " << oldRate << " to " << newRate << endl;
		// Update x->x...
		mRateMatrix[x][x] += (oldRate ); // * mFrequency[y]);
		mRateMatrix[x][x] -= ( mRateMatrix[x][y] );
	//	std::cout << "After our little update for x,x = " << x << " the sum is " << mRateMatrix[x][x] << endl;
		// NOTE: THE ROW AND COL FOR X,Y of the Pmat must be updated too!
//	}
}

void ProteinModel::UpdateAlphaMatrix(DoubleMatrix& mAlphaMatrix, const DoubleVector& mFrequency, const int& m_nMethod) {
	double dummy = -1;
//	std::cout << "UpdateAlphaMat-0 called with arg." << endl;
	UpdateAlphaMatrix(mAlphaMatrix, mFrequency, options.m_nMethod, dummy);
}

void ProteinModel::UpdateAlphaMatrix(DoubleMatrix& mAlphaMatrix, const DoubleVector& mFrequency, const int& m_nMethod, double& rateScale)
{
	int i,j;
	bool m_bUpdate = false;
//	std::cout << "UpdateAlphaMatrix-1 begins..." << endl;
//	static DoubleMatrix tempAlphaMatrix;
// WTF???	int whichP[ mNumStates * mNumStates ];
	
//	std::cout << "THE inRATESCALE IS " << rateScale << endl;
	
/*	if (options.m_nMethod == 0)
	{
		while (!m_bUpdate)
		{
			//tempAlphaMatrix = mAlphaMatrix;
		
			IntMatrix found(mNumStates, mNumStates, 0);
	
			// Update one frequency...
			//(5 / (mNumStates*mNumStates)); //
			double updateFraction = (5 / (mNumStates*mNumStates)); //Numerics::getGlobalNumerics().Random(); //[jdk] Numerics::getGlobalNumerics().Random();
	
			
			std::cout << "OBSOLETE??" << endl;
			
			for (i = 0; i < mNumStates; i++)
			{
				for (j = 0; j < mNumStates; j++)
				{
					if ( i == j)
						found[i][j] = 0;
					else {
						found[i][j] = 1;
						if (!getScaleMatrix()) {
							if (getReversible()) {
								whichP[(i*mNumStates)+j] = whichP[(j*mNumStates)+i] = Numerics::getGlobalNumerics().Random() < (updateFraction)? 1 : 0; //[jdk] Numerics::getGlobalNumerics().Random();
							} else {
								whichP[(i*mNumStates)+j] = Numerics::getGlobalNumerics().Random() < (updateFraction)? 1 : 0; //[jdk] Numerics::getGlobalNumerics().Random();
							}
						} else { whichP[(i*mNumStates)+j] = whichP[(j*mNumStates)+i]  = 1; }
					}
				}
			}
	
			double newRateScale, oldRateScale = rateScale;
			if (getScaleMatrix()) {
				newRateScale = oldRateScale;
				UpdateAlphaParameter(newRateScale);
				if (newRateScale <= 0.0) { newRateScale += ((oldRateScale - newRateScale)*2.0); }
				rateScale = newRateScale;
				std::cout << "Proposed changing ratescale from " << oldRateScale << " to " << newRateScale << endl;
			}
			
			for (i = 0; i < mNumStates; i++)
			{
				for (j = 0; j < mNumStates; j++)
				{
					if (found[i][j])
					{
						if ( whichP[ (i*mNumStates)+j] == 1 ) {
							if (!getScaleMatrix()) {
								UpdateAlphaParameter(mAlphaMatrix[i][j]);
								found[i][j] = 0;
								if (getReversible()) { mAlphaMatrix[j][i] = mAlphaMatrix[i][j]; found[j][i] = 0; }
							} else {
								if (rateScale == NULL) { std::cout << "ERROR!  Update parameter called without a rateScaling factor" << endl; exit(-1); }
								mAlphaMatrix[i][j] /= oldRateScale;
								mAlphaMatrix[i][j] *= newRateScale;
								
								if (getReversible()) { mAlphaMatrix[j][i] = mAlphaMatrix[i][j]; found[j][i] = 0; }
								found[i][j] = 0;
							}
						}
					}
				}
			
				double sum = 0;
	
				double tempRate;
			
				for (j = 0; j < mNumStates; j++)
				{
					if (i != j)
					{
						tempRate = mAlphaMatrix[i][j];// [BLOCKING USE OF FREQS HERE - IMPLICIT IN NON-REV MODELS] * mFrequency[j];// [BLOCKING USE OF FREQS HERE - IMPLICIT IN NON-REV MODELS]
						sum += tempRate;
					}
				}

				tempRate = 1 - sum;
		//		std::cout << "Row " << i << ", sum is " << sum << ", 1-sum = " << tempRate << endl;
				
			// WHY DO THIS?  THE Q-RATES CAN BE WHATEVER WE WANT...ARE WE FORCING THEM TO BE BETWEEN 0 and 1?
				if ((tempRate <= 0.0) || (tempRate >= 1.0))
				{
					m_bUpdate = false;
					break;
				}
				else
				{
					m_bUpdate = true;
					mAlphaMatrix[i][i] = 0;
				}
			}
		}
	}
	else
	{ */
//	std::cout << "DOING OLD UPDATE" << endl;
	//	tempAlphaMatrix = mAlphaMatrix;
		
/*		IntMatrix found(mNumStates, mNumStates, 0);
	
		for (i = 0; i < mNumStates; i++)
		{
			for (j = 0; j < mNumStates; j++)
			{
				if ( i == j)
					found[i][j] = 0;
				else
					found[i][j] = 1;
			}
		}
*/		
//		double updateFraction = (5 / (mNumStates*mNumStates)); //Numerics::getGlobalNumerics().Random(); //[jdk] Numerics::getGlobalNumerics().Random();

/*		for (i = 0; i < mNumStates; i++)
		{
			for (j = 0; j < mNumStates; j++)
			{
				if ( i == j)
					found[i][j] = 0;
				else {
					found[i][j] = 1;
					//if (found[j][i] != 1) {
					if (!getScaleMatrix()) {
						if (getReversible()) whichP[(i*mNumStates)+j] = whichP[(j*mNumStates)+i] = Numerics::getGlobalNumerics().Random() < (updateFraction)? 1 : 0; //[jdk] Numerics::getGlobalNumerics().Random();
						else whichP[(i*mNumStates)+j] = Numerics::getGlobalNumerics().Random() < (updateFraction)? 1 : 0;
					} else {
						// Override selective resampling...
						whichP[(i*mNumStates)+j] = 1;
					}
					//if (whichP[(i*mNumStates)+j] == 1) std::cout << "Updating parm " << i << ", " << j << endl; 
					//}
				}
			}
		}

	*/
		double newRateScale, oldRateScale = rateScale;
	//	std::cout << "Going to update rate scale..." << endl;
		if (getScaleMatrix()) {
			newRateScale = oldRateScale;
			UpdateAlphaParameter(newRateScale);
			if (newRateScale <= 0.0) { newRateScale += ((oldRateScale - newRateScale)*2.0); }
			rateScale = newRateScale;

	//		std::cout << "Proposed changing ratescale from " << oldRateScale << " to " << newRateScale << endl;
		}

	for (i = 0; i < mNumStates; i++)
	{
		for (j = 0; j < mNumStates; j++)
		{
			if (j != i)
			{
				//					if ( whichP[ (i*mNumStates)+j] == 1 ) {
				if (!getScaleMatrix()) {
				//	std::cout << "Before update, parm " << i << ", " << j << " is " << mAlphaMatrix[i][j] << endl;
					UpdateAlphaParameter(mAlphaMatrix[i][j]);
				//	std::cout << "After update, parm " << i << ", " << j << " is " << mAlphaMatrix[i][j] << endl;
					if (getReversible()) { mAlphaMatrix[j][i] = mAlphaMatrix[i][j]; }// found[j][i] = 0; }
					//							found[i][j] = 0;
				} else {
					if (rateScale == NULL) { std::cout << "ERROR!  Update parameter called without a rateScaling factor" << endl; exit(-1); }
					//		std::cout << "TEST: ratescale is " << rateScale << endl;
					mAlphaMatrix[i][j] /= oldRateScale;
					//	if ((i==0) && (j==1)) std::cout << "CHECK: MAT[0,1] = " << mAlphaMatrix[0][1] << endl;
					mAlphaMatrix[i][j] *= newRateScale;
					//		std::cout << "new alpha " << i << ", " << j << " is " << mAlphaMatrix[i][j] << endl;
					
					if (getReversible()) { mAlphaMatrix[j][i] = mAlphaMatrix[i][j]; }// found[j][i] = 0; }
					//							found[i][j] = 0;
				}
			}
		}
		mAlphaMatrix[i][i] = 0;
	}
	
	m_bUpdate = true;
	
	//	}
//	}

//	if (m_bUpdate) mAlphaMatrix = tempAlphaMatrix;

//	std::cout << "updateAlphaMatrix done" << endl;
}


// THIS IS THE VERSION THAT SHOULD BE DEPRECATED...
// Rate matrix update when using discrete rate variation...
void ProteinModel::UpdateAlphaMatrix(DoubleMatrix& mAlphaMatrix, const DoubleVector& mFrequency, const int& m_nMethod, int *theAssignments, double *discRates)
{
	//	std::cout << "Called updateAlphaMatrix-2!" << endl;
	
	if (options.doRAR == 0) { std::cout << "ERROR! UpdateAlphaMatrix for gamma called inappropriately!" << endl; exit(-1); }
	
	int i,j;
	bool m_bUpdate = false;
	
	static DoubleMatrix tempAlphaMatrix;
	int whichP[ mNumStates * mNumStates ];
	
	if (options.m_nMethod == 0)
	{
		//		std::cout << "m_nMethod = 0" << endl;
		while (!m_bUpdate)
		{
			tempAlphaMatrix = mAlphaMatrix;
			
			IntMatrix found(mNumStates, mNumStates, 0);
			
			// Update one frequency...
			//(5 / (mNumStates*mNumStates)); //
			double updateFraction = (1 / (mNumStates*mNumStates)); //Numerics::getGlobalNumerics().Random(); //[jdk] Numerics::getGlobalNumerics().Random();
			
			for (i = 0; i < mNumStates; i++)
			{
				for (j = 0; j < mNumStates; j++)
				{
					if ( i == j)
						found[i][j] = 0;
					else {
						found[i][j] = 1;
						//	if (found[j][i] != 1) {
						if (getReversible()) whichP[(i*mNumStates)+j] = whichP[(j*mNumStates)+i] = Numerics::getGlobalNumerics().Random() < (updateFraction)? 1 : 0; //[jdk] Numerics::getGlobalNumerics().Random();
						else whichP[(i*mNumStates)+j] = Numerics::getGlobalNumerics().Random() < (updateFraction)? 1 : 0; 
						//if (whichP[(i*mNumStates)+j] == 1) std::cout << "Updating parm " << i << ", " << j << endl; 
						//	}
					}
				}
			}
			
			for (i = 0; i < mNumStates; i++)
			{
				for (j = 0; j < mNumStates; j++)
				{
					if (found[i][j])
					{
						std::cout << "("<<i<<","<<j<<") " << theAssignments[(i*mNumStates)+j] << " ";
						if ( whichP[ (i*mNumStates)+j] == 1 ) {
							//				UpdateAlphaParameter( theAssignments[(i*mNumStates)+j] );
						}
						//						if ( tempAlphaMatrix[i][j] != discRates[ theAssignments[ (i*mNumStates) + j]]) std::cout << "ERROR! current and old assignments should be the same! " << i << ", " << j << "; orig " << tempAlphaMatrix[i][j] << ", new "<< discRates[ theAssignments[ (i*mNumStates) + j]] << endl;
						tempAlphaMatrix[i][j] = discRates[ theAssignments[(i*mNumStates)+j] ];
						if (getReversible()) { tempAlphaMatrix[j][i] = tempAlphaMatrix[i][j]; found[j][i] = 0; }
						found[i][j] = 0;
					}
				} std::cout << endl;
				
				double sum = 0;
				
				double tempRate;
				
				for (j = 0; j < mNumStates; j++)
				{
					if (i != j)
					{
						tempRate = tempAlphaMatrix[i][j];// [BLOCKING USE OF FREQS HERE - IMPLICIT IN NON-REV MODELS]  * mFrequency[j];// [BLOCKING USE OF FREQS HERE - IMPLICIT IN NON-REV MODELS]
						sum += tempRate;
					}
				}
				
				tempRate = 1 - sum;
				//		std::cout << "Row " << i << ", sum is " << sum << ", 1-sum = " << tempRate << endl;
				
				// WHY DO THIS?  THE Q-RATES CAN BE WHATEVER WE WANT...ARE WE FORCING THEM TO BE BETWEEN 0 and 1?
				if ((tempRate <= 0.0) || (tempRate >= 1.0))
				{
					m_bUpdate = false;
					break;
				}
				else
				{
					m_bUpdate = true;
					tempAlphaMatrix[i][i] = 0.0; //-sum;// 0
				}
			}
		}
	}
	else
	{
		//	std::cout << "DOING OLD UPDATE" << endl;
		tempAlphaMatrix = mAlphaMatrix;
		
		IntMatrix found(mNumStates, mNumStates, 0);
		
		for (i = 0; i < mNumStates; i++)
		{
			for (j = 0; j < mNumStates; j++)
			{
				if ( i == j)
					found[i][j] = 0;
				else
					found[i][j] = 1;
			}
		}
		
		double updateFraction = 0.10 + (Numerics::getGlobalNumerics().Random() / 3.0);
		//(1 / (mNumStates*mNumStates)); //Numerics::getGlobalNumerics().Random(); //[jdk] Numerics::getGlobalNumerics().Random();
		
		for (i = 0; i < mNumStates; i++)
		{
			for (j = 0; j < mNumStates; j++)
			{
				if ( i == j)
					found[i][j] = 0;
				else {
					found[i][j] = 1;
					//if (found[j][i] != 1) {
					if (getReversible()) whichP[(i*mNumStates)+j] = whichP[(j*mNumStates)+i] = Numerics::getGlobalNumerics().Random() < (updateFraction)? 1 : 0; //[jdk] Numerics::getGlobalNumerics().Random();
					else whichP[(i*mNumStates)+j] = Numerics::getGlobalNumerics().Random() < (updateFraction)? 1 : 0;
					//if (whichP[(i*mNumStates)+j] == 1) std::cout << "Updating parm " << i << ", " << j << endl; 
					//}
				}
			}
		}
		
		
		for (i = 0; i < mNumStates; i++)
		{
			for (j = 0; j < mNumStates; j++)
			{
				if (found[i][j])
				{
					if ( whichP[ (i*mNumStates)+j] == 1 ) {
						//		 UpdateAlphaParameter( theAssignments[(i*mNumStates)+j] );
					}
					///					if ( tempAlphaMatrix[i][j] != discRates[ theAssignments[ (i*mNumStates) + j]])  std::cout << "ERROR! current and old assignments should be the same! " << i << ", " << j << "; orig " << tempAlphaMatrix[i][j] << ", new "<< discRates[ theAssignments[ (i*mNumStates) + j]] << endl;
					
					tempAlphaMatrix[i][j] = discRates[ theAssignments[(i*mNumStates)+j] ];
					
					//UpdateAlphaParameter(tempAlphaMatrix[i][j]);
					if (getReversible()) { tempAlphaMatrix[j][i] = tempAlphaMatrix[i][j]; found[j][i] = 0; }
					found[i][j] = 0;
				}
			}
			
			m_bUpdate = true;
			tempAlphaMatrix[i][i] = 0;
		}
	}
	
	if (m_bUpdate) mAlphaMatrix = tempAlphaMatrix;
}

// THIS IS THE SENSIBLE VERSION...
// Rate matrix update when using discrete rate variation...
void ProteinModel::rewriteAlphaMatrix(DoubleMatrix& mAlphaMatrix, const DoubleVector& mFrequency, int *theAssignments, double *discRates)
{
	// The purpose of this should just be to recalc the alpha matrix and calc the diagonal...
	
	if (!RARdistributedRates) { std::cout << "ERROR! UpdateAlphaMatrix for gamma called inappropriately!" << endl; exit(-1); }
	
	int i,j;

	DoubleMatrix *tempAlphaMatrix;
	
	tempAlphaMatrix = &mAlphaMatrix;
	
	for (i = 0; i < mNumStates; i++)
	{
		for (j = 0; j < mNumStates; j++)
		{
			if (i != j) {
				(*tempAlphaMatrix)[i][j] = discRates[ theAssignments[(i*mNumStates)+j] ];
			//	std::cout << i << ", " << j << " = " << discRates[ theAssignments[(i*mNumStates)+j] ] << " (with assignment " << theAssignments[(i*mNumStates)+j] << ")" << endl;
				if (getReversible()) { (*tempAlphaMatrix)[j][i] = (*tempAlphaMatrix)[i][j]; }// found[j][i] = 0;
			}
		}
		
		double sum = 0;
		for (j = 0; j < mNumStates; j++)
		{
			if (i != j)
			{
				sum += (*tempAlphaMatrix)[i][j];// [BLOCKING USE OF FREQS HERE - IMPLICIT IN NON-REV MODELS]  * mFrequency[j];// [BLOCKING USE OF FREQS HERE - IMPLICIT IN NON-REV MODELS]
			}
		}
		
		(*tempAlphaMatrix)[i][i] = 0.0-sum;				
	} // for i
	
}

			
//---------------------------------------------------------------------------
void ProteinModel::UpdateAlphaMatrix3D(const int& m_nMethod)
{
	DoubleVector tempFrequency(mNumStates, 0);
	
	for (unsigned int k = 0; k< mNumModels; k++)
	{
		tempFrequency = mFrequencyMatrix[k];
		
		if (options.doRAR == 0) {
			if (getScaleMatrix()) {			
				UpdateAlphaMatrix(mAlphaMatrix3D[k], tempFrequency, options.m_nMethod, returnRateScale(k));
				//UpdateAlphaMatrix(mAlphaMatrix3D[k], tempFrequency, options.m_nMethod);
			} else {
				// Need to be careful with uniformization cap
				UpdateAlphaMatrix(k, options.m_nMethod);
				//UpdateAlphaMatrix(mAlphaMatrix3D[k], tempFrequency, options.m_nMethod);
			}
		} else {
			UpdateAlphaMatrix(mAlphaMatrix3D[k], tempFrequency, options.m_nMethod, &(rateAssignments[(k * mNumStates * mNumStates)]), &discreteRates[(k*numDiscreteRates)]);
		}
	}
}

void ProteinModel::rewriteAlphaMatrix3D(void)
{
	DoubleVector tempFrequency(mNumStates, 0);
	
	for (unsigned int k = 0; k< mNumModels; k++)
	{
		tempFrequency = mFrequencyMatrix[k];
		
		if (RARdistributedRates) {
			rewriteAlphaMatrix(mAlphaMatrix3D[k], tempFrequency, &(rateAssignments[(k * mNumStates * mNumStates)]), &discreteRates[(k*numDiscreteRates)]);
		}
	}
}

//---------------------------------------------------------------------------
void ProteinModel::UpdateAlphaMatrix3D(const int& m_nMethod, int k)
{
	DoubleVector tempFrequency(mNumStates, 0);
	
	tempFrequency = mFrequencyMatrix[k];
	
	if (options.doRAR == 0) {
		if (getScaleMatrix()) {			
			UpdateAlphaMatrix(mAlphaMatrix3D[k], tempFrequency, options.m_nMethod, returnRateScale(k));
		} else {
			UpdateAlphaMatrix(k, options.m_nMethod);
//			UpdateAlphaMatrix(mAlphaMatrix3D[k], tempFrequency, options.m_nMethod);
		}
	} else UpdateAlphaMatrix(mAlphaMatrix3D[k], tempFrequency, options.m_nMethod, &(rateAssignments[(k * mNumStates * mNumStates)]), &discreteRates[(k*numDiscreteRates)]);
}

//---------------------------------------------------------------------------
void ProteinModel::UpdateGammaShape3D( double *var ) {
//	updateDiscreteRates(); return;
	
	double theVar;
	for (int k = 0; k< GetNumModels(); k++) {
		if (var == NULL) { theVar = 0.01; }
		else { theVar = var[k]; }
		UpdateGammaShape( k, theVar );
	}
}

void ProteinModel::UpdateRateDistributionRAR_orderedDiscrete_MC(int which, double var) {
	// Update the discrete rate distribution
	
	// NOTE: this uses a very crude proposal mechanism that keeps the components ordered
	//       (if a component is proposed to get bigger or smaller than its right or left most neighbours,
	//		  we simply resample)
	
	double proposal, delta;
	double last, next;
	int i, k; // valid;
	int numModels, startmodel;
//	double small = 0.0001;
//	var = 0.00001;

	if (which != -1) { startmodel = which; numModels = startmodel+1; }
	else { startmodel = 0; numModels = GetNumModels(); }
	
	double startVal, scaledVariance;
	
	double updateFraction = num->Random() / 2.0;
	
	for (k = startmodel; k< numModels; k++) {
		// i = (int)( num->Random() * numDiscreteRates);
			for (i=0; i<numDiscreteRates; i++) {
				if ( num->Random() <= updateFraction ) { // (5 / numDiscreteRates) ) {

					startVal = discreteRates[(k * numDiscreteRates) + i];
					scaledVariance = startVal * var;
					
					// Enforce minimum variance for small rate parameters
					if (scaledVariance < 0.001) scaledVariance = 0.001;
					
					proposal = discreteRates[(k * numDiscreteRates) + i];
					delta = num->box_muller(0.0, scaledVariance);
					proposal += delta;
					
					// setting rate min,max to 0..500
					last = 0.0;
					next = 500.0;
					if ( i>0 ) { last = discreteRates[(k * numDiscreteRates) + (i-1)]; }
					if ( i<(numDiscreteRates-1) ) { next = discreteRates[(k * numDiscreteRates) + (i+1)]; }
					
					// This is a really crude method of restricting our search space...
					while ( (proposal <= last) || (proposal >= next ) ) {
						delta = num->box_muller(0.0, scaledVariance);
						proposal = discreteRates[(k * numDiscreteRates) + i] + delta;
					}
					
				//	std::cout << "Proposing update of rate " << i << " (" << discreteRates[(k*numDiscreteRates)+i] << ") by " << proposal << "..." << endl;
/*
					valid = 0;
					while (!valid) {
						if ( proposal < 0 ) {
							proposal *= -1;
							//num->box_muller(0.0, var);
							//proposal += discreteRates[(k * numDiscreteRates) + i];
						} else {
							valid = 1;
						}
					}
*/	
 //	std::cout << " (final prop = " << proposal << ")" << endl;
					discreteRates[(k * numDiscreteRates) + i] = proposal;
					//break;
					
				} // if we choose to update this one
		 } // for each rate (i)
	} // for each model (k)
	
/*	std::cout << "Outputting modified discrete rates:" << endl;
	
	for (k = startmodel; k< numModels; k++) {
		for (i=0; i<numDiscreteRates; i++) {
			std::cout << "Rate " << i << " (model " << k << "): " << discreteRates[(k * numDiscreteRates) + i] << endl;
		}
	}
*/	
}

//---------------------------------------------------------------------------
void ProteinModel::UpdateGammaShape(int which, double var)
{
//	updateDiscreteRates(); return;
	//var = 0.1;
	if (!RARdistributedRates) return;
//	std::cout << "		UpdateGammaShape()	Called for mixture component " << which << ", with sigma = " << var << endl;
//	std::cout << "		STARTING GAMMA PARM = " << gammaShapeParameter[which] << endl;
	
	/*
	double tot;
	int counts[numDiscreteRates];
	for (int i=0; i<(mNumModels); i++) {
		std::cout << "			[Distrib. parm. 1: " << gammaShapeParameter[i] << "; Distrib. parm. 2: " << gammaScaleParameter[i] << "; Assignment counts:	";
		for (int ii=0; ii<numDiscreteRates; ii++) { counts[ii] = 0; }
		tot = 0;
		for (int j=0; j<(mNumStates*mNumStates);j++) {
			counts[ rateAssignments[ (i*mNumStates*mNumStates) + j] ]++; tot++;
		}
		for (int ii=0; ii<numDiscreteRates; ii++) { std::cout << counts[ii]; if (ii < (numDiscreteRates-1)) { std::cout << "\t"; } } std::cout << "]" << endl;
	}
	*/
	
//	double proposal = (rannum - 0.5) * mWindowsize;
//	gammaShapeParameter[which] += proposal;
//	if ((which > 0)) return;
	
/*	if ( Numerics::getGlobalNumerics().Random() < 0.1) {
		double proposal = gammaShapeParameter[which];
	//	std::cout << "			Trying inversion proposal (gamma shape parameter)..." << endl;
		proposal = 1.0/proposal;
	////[old]	proposal = round( (proposal/0.01) ) * 0.01;
		gammaShapeParameter[which] = proposal;
	} else if ( Numerics::getGlobalNumerics().Random() < 0.1) {
		double proposal = gammaScaleParameter[which];
	//	std::cout << "			Trying inversion proposal (gamma scale parameter)..." << endl;
		proposal = 1.0/proposal;
	////[old]	proposal = round( (proposal/0.01) ) * 0.01;
		gammaScaleParameter[which] = proposal;
	} else {
*/
//		if ( Numerics::getGlobalNumerics().Random() <= 0.5) {

		// First...
		double proposal = gammaShapeParameter[which];
	if ( Numerics::getGlobalNumerics().Random() <= 0.9) { //[Dec]  0.2) {
		var = 0.01;
			double rannum = num->box_muller(1.0, var); //001); //Numerics::getGlobalNumerics().Random();

			// [Dec] double rannum = num->box_muller(0.0, var); //001); //Numerics::getGlobalNumerics().Random();
			//	double rannum = (int)(Numerics::getGlobalNumerics().Random() * (100.0 * var) - (50.0 * var)) * 0.01; //001); //Numerics::getGlobalNumerics().Random();
				if (rannum == 0.0) { 
					if ( Numerics::getGlobalNumerics().Random() <= 0.5) {
						rannum = 0.0001;
					} else {
						rannum = -0.0001;
					}
				}

				if (proposal < 0.0) { std::cout << "ERROR!  Gamma shape shouldn't be initialized with value < 0." << endl; exit(-1); }
				
				if ( (proposal + rannum) < 0.0) {
					if (rannum == 0.0) { rannum = -0.0001; }
					rannum *= -1;
					proposal += rannum;
				} else {
					proposal += rannum;
				}
				
//[Dec]				gammaShapeParameter[which] *= proposal;

		// current variance is gammaShapeParameter[which]/pow(gammaScaleParameter[which],2)
		if ( Numerics::getGlobalNumerics().Random() <= 0.5 ) {
				gammaShapeParameter[which] *= rannum;
				gammaScaleParameter[which] /= rannum;
		} else {
				gammaShapeParameter[which] /= rannum;
				gammaScaleParameter[which] *= rannum;
		}

	} else if (Numerics::getGlobalNumerics().Random() < 0.2) {
			// Second...
			double rannum = num->box_muller(0.0, var); //001); //Numerics::getGlobalNumerics().Random();
		//	double rannum = (int)(Numerics::getGlobalNumerics().Random() * (100.0 * var) - (50.0 * var)) * 0.01; //001); //Numerics::getGlobalNumerics().Random();
			if (rannum == 0.0) { 
				if ( Numerics::getGlobalNumerics().Random() <= 0.5) {
					rannum = 0.0001;
				} else {
					rannum = -0.0001;
				}
			}

			// rannum = num->box_muller(0.0, var);
			double proposal = gammaScaleParameter[which];
			if (proposal < 0.0) { std::cout << "ERROR!  Gamma scale shouldn't be initialized with value < 0." << endl; exit(-1); }
			
			if ( (proposal + rannum) < 0.0) {
		////[old]		while ( (proposal + rannum) < 0.0) {
		////[old]			rannum = ((rannum/0.01)/2.0)*0.1;
					if (rannum == 0.0) { rannum = -0.0001; }
					if (( (proposal + rannum) < 0.0) ) {
						rannum *= -1;
					}
			////[old]	}
				proposal += rannum;
			//	proposal -= rannum;
			////[old]}////[old] else if ( (which > 0) ) {
			////[old]	if ( (proposal+rannum) < gammaScaleParameter[(which-1)] ) {
			////[old]		proposal -= rannum;
			////[old]	} else {
			////[old]		proposal += rannum;
			////[old]	}
			} else {
				proposal += rannum;
			}
			
			gammaScaleParameter[which] = proposal;
	} else {
				double rannum = num->box_muller(0.0, var); //0001); //001); //Numerics::getGlobalNumerics().Random();
			//	double rannum = (int)(Numerics::getGlobalNumerics().Random() * (100.0 * var) - (50.0 * var)) * 0.01; //001); //Numerics::getGlobalNumerics().Random();
				if (rannum == 0.0) { 
					if ( Numerics::getGlobalNumerics().Random() <= 0.5) {
						rannum = 0.0001;
					} else {
						rannum = -0.0001;
					}
				}

				double proposal = gammaShapeParameter[which];

				if (proposal < 0.0) { std::cout << "ERROR!  Gamma shape shouldn't be initialized with value < 0." << endl; exit(-1); }
				
				if ( (proposal + rannum) < 0.0) {
					if (rannum == 0.0) { rannum = -0.0001; }
					rannum *= -1;
					proposal += rannum;
				} else {
					proposal += rannum;
				}
				
				gammaShapeParameter[which] = proposal;

			
				proposal = gammaScaleParameter[which];

				if (proposal < 0.0) { std::cout << "ERROR!  Gamma shape shouldn't be initialized with value < 0." << endl; exit(-1); }

				double rannum2 = num->box_muller(0.0, var); //001); //Numerics::getGlobalNumerics().Random();
				if (rannum2 == 0.0) { 
					if ( Numerics::getGlobalNumerics().Random() <= 0.5) {
						rannum2 = 0.0001;
					} else {
						rannum2 = -0.0001;
					}
				}
				
				if ( (proposal + rannum2) < 0.0) {
					if (rannum == 0.0) { rannum2 = -0.01; }
					rannum2 *= -1;
					proposal += rannum2;
				} else {
					proposal += rannum2;
				}
				
				gammaScaleParameter[which] = proposal;

	}
	
//	}

	// Update the values...
	setGammaDistribution(which);
//	setAllGammaDistributions();
}

void ProteinModel::UpdateAlphaMatrix(int theModel, const int& m_nMethod) {
	int i,j;
	bool m_bUpdate = false;
	
	
	int numUpdated = 0;
	int updateLimit =(int)(Numerics::getGlobalNumerics().Random()*4) + 1;
	
	int tester;
	i=0;j=0;
	while (numUpdated <= updateLimit ) {
		i = (int)(Numerics::getGlobalNumerics().Random() * mNumStates);
		j = (int)(Numerics::getGlobalNumerics().Random() * mNumStates);
		while (j == i) j = (int)(Numerics::getGlobalNumerics().Random() * mNumStates);
		
		// Non-reversible evaluation
		tester = UpdateAlphaParameter( theModel, i, j, 0);
		if (tester == 0) continue;
		
		numUpdated++;
	}
	
	m_bUpdate = true;
}


//---------------------------------------------------------------------------
int ProteinModel::UpdateAlphaParameter( int theModel, int i, int j, int reversibleEvaluation) {
	int rejectNum = 0;
	double changeamount, proposal, var;
	
	// THIS IS NOT VERY EFFICIENT, BUT WILL GUARANTEE VALID PROPOSALS UNDER OPTIONS.CAPUNIFORMIZATION...
	// std::cout << "NEW PARAMETER UPDATE FUNCTION - SHOULD RESPECT UNI CAP" << endl;
	
	double newWaitingTime1, sumCorrection;
	
	double inParameter = mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[i].mDataArray[j];
	double oldRate1     = inParameter * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[j];
	double oldRate2     = inParameter * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[i];
	
	/* This implementation assumes matrix scaling 
	double rateSum = 0.0;
	for (int ii=0; ii < mNumStates; ii++) {
		for (int jj=0; jj < mNumStates; jj++) {
			if (reversibleEvaluation) {
				if (jj != ii) rateSum += mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[ii].mDataArray[jj] * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[jj] * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[ii];
			} else {
				if (jj != ii) rateSum += mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[ii].mDataArray[jj] * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[jj];
			}
		}
	}
	*/
	
	double rateSum = 0.0;
	for (int jj=0; jj < mNumStates; jj++) {
		if (jj != i) rateSum += mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[i].mDataArray[jj] * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[jj];
	}
	
	int notDone = 1;
	while (notDone) {
		
		// 100% Gaussians with variance = 0.05 works well, but slower parameters mix more slowly...
		if ( Numerics::getGlobalNumerics().Random() < 0.05 ) {
			changeamount = ( Numerics::getGlobalNumerics().Random() - 0.5) * mWindowsize * inParameter;
			//trying to scale change amount by parm size to equalize mixing...
			proposal = inParameter + changeamount;
			
			if (proposal <= 0.0 )
				proposal = inParameter - changeamount;
		} else {
			var = 0.1;
			//var *= 0.1;
			
			changeamount = num->box_muller( 0.0, var ); // (rannum - 0.5) * mWindowsize;
			
			while ( (inParameter + changeamount) <= 0.0 ) {
				changeamount *= -1;
				if ( (inParameter + changeamount) < 0.0 ) {
					changeamount = num->box_muller( 0.0, var ); // (rannum - 0.5) * mWindowsize;
				}					
			}
			proposal = inParameter + changeamount;
		}
		
		if (proposal < 0) {
			std::cout << "Error, our proposal is < 0" << endl;
		}
		
		if (proposal > 10.0) {
			std::cout << "WHOA, proposal is " << proposal << ", in parm was " << inParameter << endl;
			notDone = 1;
			rejectNum++;
		} else {
			if (!options.capUniformizationRate) notDone = 0;
		}
		
		if (options.capUniformizationRate) {
			if (rejectNum > 3) {
				if (num->Random() > 0.5) {
					proposal = inParameter - changeamount;
					if (proposal < 0.0) {
						proposal = inParameter + changeamount;
					}
				}
			}
			
			if (rejectNum > 4) {
				return 0;
			}
			
			double rowSum;
			if ( reversibleEvaluation ) {
				std::cout << "REV EVAL" << endl;
				sumCorrection =   (rateSum - oldRate1* mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[i] - oldRate2* mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[j] + (2.0 * proposal * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[j]*mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[i]));
				notDone = 0;
				for (int kk=0; kk<mNumStates; kk++) {
					rowSum = 0.0;
					for (int k=0; k < mNumStates; k++) {
						if (k != kk) rowSum += mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[kk].mDataArray[k] * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[k];
					}
					double term2;
					if (kk == i) {
						term2 = -oldRate1 + (proposal * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[j]);
					} else if (kk == j) {
						term2 = -oldRate2 + (proposal * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[i]);
					} else {
						term2 = 0.0;
					}
					
					newWaitingTime1 = (rowSum + term2) / sumCorrection;
					//newWaitingTime2 = (rowSum2 - oldRate2 + (proposal * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[i])) / sumCorrection;
					
					
					if ( ( (newWaitingTime1 * options.__largestRate) > uniformizationConstants.mDataArray[theModel] ) ) {
						notDone = 1;
					}
				}
				
				if (notDone == 1) {
					rejectNum++;
				} 			
			} else {
			//	sumCorrection =  rateSum / (rateSum - oldRate1  + (proposal * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[j]) );
				newWaitingTime1 = rateSum - oldRate1 + (proposal * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[j]); // * sumCorrection;
				
				// Just to be safe...
				newWaitingTime1 += 0.1;
				
			//	double newWaitingTime1a = rateSum - oldRate1 + ( proposal * mFrequencyMatrix.mRows.mDataArray[theModel].mDataArray[j] );

			//	std::cout << "proposal from " << inParameter << " to " << proposal << "; newWT1 = " << newWaitingTime1 << endl;
			//		std::cout << "  uni is " << uniformizationConstants.mDataArray[theModel] << ", largest rate is " << options.__largestRate << endl;
				
				if ( (newWaitingTime1 * options.__largestRate) > uniformizationConstants.mDataArray[theModel] ) {
					notDone = 1;
					rejectNum++;
				} else {
					notDone = 0;
				}
			}
			
			
		}
	}
	
	mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[i].mDataArray[j] = proposal;
	if (reversibleEvaluation) mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[j].mDataArray[i] = proposal;
	
	
	return 1;
}

//---------------------------------------------------------------------------
void ProteinModel::UpdateParameter(const int& m_nMethod)
{
		
/*	if (RARdistributedRates) {
		dontUpdateRateAssignments = 0;
		std::cout << "PRIOR TO PROPOSAL:" << endl;
		int tot = 0;
		int counts[numDiscreteRates];
		for (int i=0; i<(mNumModels); i++) {
			std::cout << "		gammaShape = " << gammaShapeParameter[i] << ", counts of rate membership: "<< endl;
			for (int ii=0; ii<numDiscreteRates; ii++) { counts[ii] = 0; }
			tot = 0;
			for (int j=0; j<(mNumStates*mNumStates);j++) {
				counts[ rateAssignments[ (i*mNumStates*mNumStates) + j] ]++; tot++;
			}
			for (int ii=0; ii<numDiscreteRates; ii++) { std::cout << "		model " << i << "; rate " << ii << " has count " << counts[ii] << "/" << tot << endl; }
		}
	}
*/

//	std::cout << "UpdateParameter() called for protein model" << endl;
	
	// This function is called when constructing a model update proposal...
	if (!getFixedFreqs()) UpdateFrequencyMatrix();	

	// Update the gamma shape parameter...
	// if (RARdistributedRates) UpdateGammaShape3D(NULL);
	UpdateAlphaMatrix3D(options.m_nMethod); // block re-assignment of rates update
	ProteinModelParameterstoRates();

	// Update mixture model component weights if appropriate...
	UpdateFreqModels();
}

		
