/*
 *  ProteinModel.h
 *  PLEX
 *
 *  A.P. Jason de Koning, jason.de.koning@gmail.com
 *  http://jasondk.org
 *
 *  David Pollock lab, University of Colorado School of Medicine
 *  http://www.evolutionarygenomics.com
 *  All rights reserved.
 
 */


#ifndef ProteinModel_h__
#define ProteinModel_h__


#include "MixtureModel.h"
#include "dcdflib.h"
#include "Numerics.h"
#include <vector>
#include <iostream>
#include <iomanip>

class Tree;

const int	noInitAssignments = 0;
const int	randomInitAssignments = 1;

//---------------------------------------------------------------------------
class ProteinModel : public MixtureModel
{
	public:
		ProteinModel(
			const DoubleMatrix3D& inAlphaMatrix3D = DoubleMatrix3D(),
			const DoubleMatrix& inFrequencyMatrix = DoubleMatrix(),
			const IntMatrix& inCountMatrix = IntMatrix());
			
		ProteinModel(
			const ProteinModel&	inModel);
			
		ProteinModel(
				const DoubleMatrix3D& inAlphaMatrix3D,
				const DoubleMatrix& inFrequencyMatrix,
				const IntMatrix& inCountMatrix, int theNumRates, double *gammaParm);
				
				
		virtual	~ProteinModel(
			void);
			
	IntMatrix3D dependencyIndicators;			// relevant for mixture model components > 0, indicates if rate is the same as mixture #0
	double		independentFractionPrior;		// the fraction of rates allowed to be different from mix #0
	
	void setPredefinedRatesMatrix(int which, DoubleMatrix& theRates);
	void setPredefinedFreqs(int which, DoubleMatrix& theFreqs);
	double predefinedRates(int which, int i, int j);
	double predefinedFreqs(int which, int i);
	
	// 'Dependent rates' means that rate assignments are made based on a sampled indicator...
	int getDependentRates(void) { return dependentRates; }
	void setDependentRates(int val) { dependentRates = val; }

	void setIndependentFractionPrior(double val) { independentFractionPrior = val; setDependentRates(1); }
	double getIndependentFractionPrior(void) { return independentFractionPrior; }
	
	void scaleDiscreteRates( double factor );
	
	void copy(const ProteinModel& inModel);

		void setUniformization(DoubleVector* vals);	
		void setUniformization(int which, double val);
		double getUniformization(int which) { return uniformizationConstants.mDataArray[which]; }
		
		const ProteinModel& operator=(		//copy a GTR model
			const ProteinModel&	inModel);
					
		void swap(				//swap a GTR model
			ProteinModel& inModel);
			
		void PrintAlphaMatrix3D(		//print the alpha matrix for each model to outStream
			ostream& outStream, ObservedData& theData);
			
		void PrintDependentRates(ostream& outStream);

		void PrintCountMatrix(			//print the count matrix to outStream
			ostream& outStream, ObservedData& theData);

		void PrintFrequencyMatrix(		//print the state frequency vector of each model to outStream
			ostream& outStream, ObservedData& theData);
			
//		void PrintModelAssignments(ostream& outStream, IntVector& m_vecSiteAssignment);

		void updateDiscreteRates(void);

		virtual void InitializeParameter(	//initialize the parameters of GTR model based on state changes count matrix
			const IntMatrix& inCountMatrix);
			
		virtual void UpdateParameter(		//update all the parameters for each model in mixture model,
			const int& m_nMethod);		//given the conditional pathway calculation method
		
		// Jason's functions:
		void initGamma3D(int numModels, int numCat, double *value);
		void setGammaDistribution(int which);
		void setAllGammaDistributions(void);
		
		void setFixedFreqs(int val);	// flag for whether to set state freqs equal or not
		int getFixedFreqs(void);
		
		virtual void PrintRateParameters(ostream&	outStream);

		void UpdateSingleParameterB1(int whichModel, int x, int y);
		void UpdateRateAssignmentB1(int whichModel, int x, int y, int newCategory);
		
		int returnNumRates(void) { return numDiscreteRates; }
		std::vector<int>* returnRateAssignments(void) { return &rateAssignments; }
		std::vector<double>* returnDiscreteRates(void) { return &discreteRates; }

		std::vector<double> gammaShapeParameter;	// the scale parameter values for each model component
		std::vector<double> gammaScaleParameter;	// the scale parameter values for each model component

		bool getRARdistributed(void) { return RARdistributedRates; }
	void setRARdistributed(int value) { if (value == 1) { RARdistributedRates = true; } else { RARdistributedRates = false; } }
	
		void UpdateRateDistributionRAR_orderedDiscrete_MC(int which, double var);
		void UpdateGammaShape3D( double *var );
		void UpdateAlphaMatrix3D(		//update the alpha matrix for all models
			const int& m_nMethod);
			
		void ProteinModelParameterstoRates(		//transform the alpha matrix and frequency parameters for each model to rate matrix
			void);

		void ProteinModelParameterstoRates(DoubleMatrix* freqs);

		
		void UpdateGammaShape(int which, double val);		// update the gamma shape parameter += U(-0.5, +0.5)		
		void UpdateAlphaMatrix3D(const int& m_nMethod, int k);


//		void adjustProposalVariance( int which, double val ) { proposalVariances[which] += val; }
//		double getProposalVariance( int which ) { return proposalVariances[which]; }
//		void setProposalVariance(int which, double val) { proposalVariances[which] = val; }
		double getGammaShape( int which ) { return gammaShapeParameter[which]; }
		double getGammaScale( int which ) { return gammaScaleParameter[which]; }
		void   setGammaShape( int which, double val ) { gammaShapeParameter[which] = val; }
		void   setGammaScale( int which, double val ) { gammaScaleParameter[which] = val; }

		// New stuff...
		void rewriteAlphaMatrix3D(void);

//		DoubleMatrix& returnStateFrequencies(void) { return mFrequencyMatrix; }
		void randomInitRateAssignments(void);
		void setDistributionParmsFromCounts(Tree& mTree);
		
		// Made this public for convenience...
		IntMatrix	mCountMatrix;		//count matrix of state changes in alignment
		void copyRateAssignments(int *inRates);
		void backupRateAssignments(int *inRates);
		bool RARdistributedRates;		// set to true for gamma distributed rates
	
		void setMaximumUniformization_RAR(void);

	protected:
		Numerics *num;			// [JdK] - an interface to accelerated numeric functions

	private:
		// New stuff...
		int dependentRates; // for mixture models where the rate assignments are done 'dependently'
		
	int UpdateAlphaParameter( int theModel, int i, int j, int reversibleEvaluation);
	void UpdateAlphaMatrix(int theModel, const int& m_nMethod);

		// Old stuff...
		void InitializeFrequencyMatrix(		//initialize the frequency vector for each model
			const IntMatrix& inCountMatrix);
			
		void InitializeAlphaMatrix3D(		//initialize the rate matrix for each model
			const IntMatrix& inCountMatrix);	
	
		void UpdateFrequency(			//update the frequency vector for a single model
			DoubleVector& mFrequency);

		void UpdateFrequencyMatrix(		//update the frequency vectors for all models
			void);
			
		void UpdateAlphaMatrix(			//update the alpha matrix for a single model
			DoubleMatrix& mAlphaMatrix, 
			const DoubleVector& mFrequency, 
			const int& m_nMethod);
			
		void UpdateAlphaMatrix(DoubleMatrix& mAlphaMatrix, const DoubleVector& mFrequency, const int& m_nMethod, double& rateScale);

		void SetCountMatrix(			//set the count matrix for this model
			const IntMatrix& inCountMatrix);

		void UpdateFrequencyParameter(		//update a single frequency parameter
			double& inParameter);

		void UpdateAlphaParameter(		//update a single alpha parameter
			double& inParameter);

		void UpdateAlphaParameter(int& theRate);	// update a single parameter when using (ordered) discrete rate classes

//		DoubleMatrix	mFrequencyMatrix;	//frequency vector for each model
		
		// Control options...
		int fixedFreqs;					// set equal state frequencies
		int numDiscreteRates;			// number of rate categories to consider
		
		// Stuff for gamma distributed rates...
		std::vector<double> discreteRates;			// a vector of the actual rates for each model component
//		double *gammaShapeParameter;	// the scale parameter values for each model component
		std::vector<int> rateAssignments;			// BIG vector of what rate category each transition belongs to
		
		// Jason's functions for gamma / discrete-rate variation...
		void UpdateAlphaMatrix(DoubleMatrix& mAlphaMatrix, const DoubleVector& mFrequency, const int& m_nMethod, int* theAssignments, double *discRates);
		void rewriteAlphaMatrix(DoubleMatrix& mAlphaMatrix, const DoubleVector& mFrequency, int *theAssignments, double *discRates);
	
		void initGamma3D(int numModels, int numCat, double *value, int initRatesMode);
		void initAndCopyGamma3D(int numModels, int numCat, double *value, int *inRates);

		int round(double a);
//		void allocateProposalVariances(int num) { proposalVariances = new double[num]; }
//		double *proposalVariances;
};

#endif // ProteinModel_h__
