/*
 *  Tokenizer.cpp
 *  CLAMP
 *
 *  Created by Chip Jarred.
 *  Copyright (c) 2005 Dr. David Pollock. All rights reserved.
 *
 */

#include "Tokenizer.h"

#include <algorithm>
#include <istream>
#include <stack>

using std::istream;
using std::stack;
using std::string;
using std::swap;

//---------------------------------------------------------------------------
class Tokenizer_Impl
{
		static	const string						cDefaultDelimiters;
		static	const string						cDefaultWhitespaceChars;
		static	const string						cDefaultQuoteChars;
		static	const bool							cDefaultSkipWhitespace = false;
		static	const bool							cDefaultRemoveQuotes = false;
		
	public:
													Tokenizer_Impl(void);
													Tokenizer_Impl(
														const Tokenizer_Impl&	inSrc);
													~Tokenizer_Impl(void);
													
				void								SetStream(
														istream&				inStream);
				void								ClearStream(void);
				bool								HasStream(void) const;
				void								Reset(void);
		
				bool								ReadToken(
														string&					outToken);
				void								PutBackToken(
														const string&			inToken);
				void								EatWhitespace(void);
															
				void								SetDelimiters(
														const string&			inDelimiters);
				const string&						GetDelimiters(void) const;
				bool								IsDelimiter(
														const string&			inToken) const;
				
				void								SetWhitespaceChars(
														const string&			inWhitespaceChars);
				const string&						GetWhitespaceChars(void) const;
				
				void								SetQuoteChars(
														const string&			inQuoteChars);
				const string&						GetQuoteChars(void) const;
				
				void								SkipWhitespace(
														bool					inSkipWhitespace = true);
				bool								SkippingWhitespace(void) const;
		
				void								RemoveQuotes(
														bool					inRemoveQuotes = true);
				bool								RemovingQuotes(void) const;
				
	protected:
	private:				
		typedef	stack<string>						StringStack;
		
				istream*							mStream;
				StringStack							mTokenStack;

				bool								mSkipWhitespace;
				bool								mRemoveQuotes;
				string								mDelimiters;
				string								mWhitespaceChars;
				string								mQuoteChars;
};

//---------------------------------------------------------------------------
const string	Tokenizer_Impl::cDefaultDelimiters = "";
const string	Tokenizer_Impl::cDefaultWhitespaceChars = " ";
const string	Tokenizer_Impl::cDefaultQuoteChars = "\"'";

#pragma mark -
//---------------------------------------------------------------------------
Tokenizer_Impl::Tokenizer_Impl(void)
	:	mStream(0),
		mTokenStack(),
		mSkipWhitespace(cDefaultSkipWhitespace),
		mRemoveQuotes(cDefaultRemoveQuotes),
		mDelimiters(cDefaultDelimiters),
		mWhitespaceChars(cDefaultWhitespaceChars),
		mQuoteChars(cDefaultQuoteChars)
{
}

//---------------------------------------------------------------------------
Tokenizer_Impl::Tokenizer_Impl(
	const Tokenizer_Impl& inSrc)
{
	mStream		= inSrc.mStream;
	mTokenStack	= inSrc.mTokenStack;
}

//---------------------------------------------------------------------------
Tokenizer_Impl::~Tokenizer_Impl(void)
{
	ClearStream();
}

//---------------------------------------------------------------------------
void
Tokenizer_Impl::SetStream(
	istream&	inStream)
{
	if (&inStream != mStream)
	{
		mStream = &inStream;
		mTokenStack = StringStack();
	}
}

//---------------------------------------------------------------------------
void
Tokenizer_Impl::ClearStream(void)
{
	mStream = 0;
}

//---------------------------------------------------------------------------
bool
Tokenizer_Impl::HasStream(void) const
{
	return mStream != 0;
}

//---------------------------------------------------------------------------
void
Tokenizer_Impl::Reset(void)
{
	while (!mTokenStack.empty())
		mTokenStack.pop();
}

//---------------------------------------------------------------------------
inline
void
Tokenizer_Impl::SetDelimiters(
	const string&	inDelimiters)
{
	mDelimiters = inDelimiters;
}

//---------------------------------------------------------------------------
inline
const string&
Tokenizer_Impl::GetDelimiters(void) const
{
	return mDelimiters;
}

//---------------------------------------------------------------------------
inline
bool
Tokenizer_Impl::IsDelimiter(
	const string&	inToken) const
{
	bool isDelim = false;
	
	if (inToken.size() == 1)
		isDelim = mDelimiters.find(inToken[0]) != string::npos;
	return isDelim;
}

//---------------------------------------------------------------------------
inline
void
Tokenizer_Impl::SetWhitespaceChars(
	const string&	inWhitespaceChars)
{
	mWhitespaceChars = inWhitespaceChars;
}

//---------------------------------------------------------------------------
inline
const string&
Tokenizer_Impl::GetWhitespaceChars(void) const
{
	return mWhitespaceChars;
}

//---------------------------------------------------------------------------
inline
void
Tokenizer_Impl::SetQuoteChars(
	const string&	inQuoteChars)
{
	mQuoteChars = inQuoteChars;
}

//---------------------------------------------------------------------------
inline
const string&
Tokenizer_Impl::GetQuoteChars(void) const
{
	return mQuoteChars;
}

//---------------------------------------------------------------------------
inline
void
Tokenizer_Impl::SkipWhitespace(
	bool	inSkipWhitespace)
{
	mSkipWhitespace = inSkipWhitespace;
}

//---------------------------------------------------------------------------
inline
bool
Tokenizer_Impl::SkippingWhitespace(void) const
{
	return mSkipWhitespace;
}

//---------------------------------------------------------------------------
inline
void
Tokenizer_Impl::RemoveQuotes(
	bool	inRemoveQuotes)
{
	mRemoveQuotes = inRemoveQuotes;
}

//---------------------------------------------------------------------------
inline
bool
Tokenizer_Impl::RemovingQuotes(void) const
{
	return mRemoveQuotes;
}

//---------------------------------------------------------------------------
void
Tokenizer_Impl::EatWhitespace(void)
{
	if (mSkipWhitespace && mStream)
	{
		while(!mStream->eof())
		{
			char c = mStream->get();
			if (!mStream->eof())
			{
				if (mWhitespaceChars.find(c) == string::npos)
				{
					mStream->putback(c);
					break;
				}
			}
		}
	}
}

//---------------------------------------------------------------------------
bool
Tokenizer_Impl::ReadToken(
	string&	outToken)
{
	string	token;
	
	if (mTokenStack.empty())
	{
		EatWhitespace();
		
		char	curQuoteChar = 0;
		
		while(!mStream->eof())
		{
			char c = mStream->get();
			if (!mStream->eof())
			{
				if (curQuoteChar != 0)
				{
					if (curQuoteChar == c)	// terminate quote
					{
						curQuoteChar = 0;
						if (!mRemoveQuotes)
							token.push_back(c);
					}
					else  token.push_back(c);	// inside quotes, just put everything in token
				}
				else
				{
					if (mQuoteChars.find(c) != string::npos)	// begin quote
					{
						curQuoteChar = c;
						if (!mRemoveQuotes)
							token.push_back(c);
					}
					else if (mDelimiters.find(c) != string::npos)
					{
						if (token.empty())
							token.push_back(c);
						else mStream->putback(c);
						
						break;
					}
					else if (mSkipWhitespace && mWhitespaceChars.find(c) != string::npos)
						break;
					else token.push_back(c);
				}
			}
		}
	}
	else
	{
		token = mTokenStack.top();
		mTokenStack.pop();
	}
	
	outToken = token;
	return token.size();
}

//---------------------------------------------------------------------------
void
Tokenizer_Impl::PutBackToken(
	const std::string&	inToken)
{
	mTokenStack.push(inToken);
}

#pragma mark -
//---------------------------------------------------------------------------
Tokenizer::Tokenizer(void)
	:   mImpl(new Tokenizer_Impl)
{
}

//---------------------------------------------------------------------------
Tokenizer::Tokenizer(
	const Tokenizer& inSrc)
	:   mImpl(new Tokenizer_Impl(*(inSrc.mImpl.get())))
{
}

//---------------------------------------------------------------------------
Tokenizer::~Tokenizer(void)
{
}

//---------------------------------------------------------------------------
const Tokenizer&
Tokenizer::operator=(
	const Tokenizer& inSrc)
{
	Tokenizer	temp(inSrc);
	Swap(temp);
	return *this;
}

//---------------------------------------------------------------------------
void
Tokenizer::SetStream(
	istream&	inStream)
{
	mImpl->SetStream(inStream);
}

//---------------------------------------------------------------------------
bool
Tokenizer::HasStream(void) const
{
	return mImpl->HasStream();
}

//---------------------------------------------------------------------------
bool
Tokenizer::ReadToken(
	string&	outToken)
{
	return mImpl->ReadToken(outToken);
}

//---------------------------------------------------------------------------
void
Tokenizer::PutBackToken(
	const string&	inToken)
{
	mImpl->PutBackToken(inToken);
}

//---------------------------------------------------------------------------
void
Tokenizer::Reset(void)
{
	mImpl->Reset();
}

//---------------------------------------------------------------------------
void
Tokenizer::EatWhitespace(void)
{
	mImpl->EatWhitespace();
}

//---------------------------------------------------------------------------
void
Tokenizer::SetDelimiters(
	const string&	inDelimiters)
{
	mImpl->SetDelimiters(inDelimiters);
}

//---------------------------------------------------------------------------
const string&
Tokenizer::GetDelimiters(void) const
{
	return mImpl->GetDelimiters();
}

//---------------------------------------------------------------------------
bool
Tokenizer::IsDelimiter(
	const string&	inToken) const
{
	return mImpl->IsDelimiter(inToken);
}

//---------------------------------------------------------------------------
void
Tokenizer::SetWhitespaceChars(
	const string&	inWhitespaceChars)
{
	mImpl->SetWhitespaceChars(inWhitespaceChars);
}

//---------------------------------------------------------------------------
const string&
Tokenizer::GetWhitespaceChars(void) const
{
	return mImpl->GetWhitespaceChars();
}

//---------------------------------------------------------------------------
void
Tokenizer::SetQuoteChars(
	const string&	inQuoteChars)
{
	mImpl->SetQuoteChars(inQuoteChars);
}

//---------------------------------------------------------------------------
const string&
Tokenizer::GetQuoteChars(void) const
{
	return mImpl->GetQuoteChars();
}

//---------------------------------------------------------------------------
void
Tokenizer::SkipWhitespace(
	bool	inSkipWhitespace)
{
	mImpl->SkipWhitespace(inSkipWhitespace);
}

//---------------------------------------------------------------------------
bool
Tokenizer::SkippingWhitespace(void) const
{
	return mImpl->SkippingWhitespace();
}

//---------------------------------------------------------------------------
void
Tokenizer::RemoveQuotes(
	bool	inRemoveQuotes)
{
	mImpl->RemoveQuotes(inRemoveQuotes);
}

//---------------------------------------------------------------------------
bool
Tokenizer::RemovingQuotes(void) const
{
	return mImpl->RemovingQuotes();
}

//---------------------------------------------------------------------------
void
Tokenizer::Swap(
	Tokenizer&		ioSwapee)
{
	swap(mImpl, ioSwapee.mImpl);
}
