/*
 *  Tree.h
 *  CLAMP
 *
 *  Created by Chip Jarred.
 *  Revised by Wanjun Gu, Jason de Koning.
 
 *  Copyright (c) 2004 Dr. David Pollock. All rights reserved.
 *
 */

#ifndef Tree_h__
#define Tree_h__

#include "TreeNode.h"
#include "TreeNodeData.h"
#include "RandomGenerator.h"
#include "Numerics.h"
#include "Options.h"

#include <algorithm>
#include <deque>
#include <istream>
#include <string>
#include <map>

extern Options options;
class MixtureModel;

//---------------------------------------------------------------------------

class Tree
{
	
public:
    Numerics *num;			//  an interface to accelerated numeric functions
    
    typedef	unsigned int		size_type;	
    typedef	TreeNode			node_type;
    
    Tree(void);
    ~Tree(void);
    
    node_type*			root(void);		//get the root node of the tree
    
    const node_type*			root(void) const;	//get the root node of the tree
    
    size_type					size(void) const;	//get the node number of the tree
    
    size_type					leaves(void) const;	//get the leaf number of the tree
    
    size_type					branches(void) const;	//get the branch number of the tree
    
    size_type					internalnodes(void) const;//get the internal node number of the tree	 
    
    bool						empty(void) const;	//if the tree is empty
    
    void						printtree(ofstream& outStream, const int& site, const double& scale) const;	//print the tree topology or site information along the tree
    
    void						printtreechanges(ofstream& outStream) const;	//print tree with count changes along branch as the branch length
    
    void						printtreesequence(ofstream& outStream) const;	//print all the sequences along the tree
    
    void 						printtreesequencecontent(		//print character contents in sequences along the tree
                                                         ofstream& outStream,
                                                         const IntVector& inStateindex,
                                                         const int& minstate,
                                                         const int& numStates) const;
    
    void 						printnodemeancontent(	//print the state content of the whole tree
                                                     ofstream& outStream,
                                                     const IntVector& inStateindex,
                                                     const int& minstate,
                                                     const int& numStates) const;
    
    void						printnodesequence(ofstream& outStream, const int& node) const;	//print the sequence of a specific node(nodeID)
    
    void						printnodesequence(ofstream& outStream, const char* nodename) const; //print the sequence of a specific node(node name)
    
    void						printinternalnodesequence(ofstream& outStream) const; //print the internal node sequences
    
    void						printleafsequence(ofstream& outStream) const;	//print leaf sequences
    
    void						printnodesequence(ofstream& outStream, const int& node, const int& count) const;	//print node sequence on a specific node with sampling number
    
    void						printnodesequence(ofstream& outStream, const char* nodename, const int& count) const;	//print node sequences on a specific node with sampling number
    
    void						printinternalnodesequence(ofstream& outStream, const int& count) const;	//print internal node sequences with sampling number
    
    void						printinternalnodesequence(StringMatrix& InternalnodeSeq, const int& m_nSampleSequence, IntVector& NodeID) const;	//print internal sequences into a string matrix
    
    void						setinternalnodes(void) const;	//mapping initial internal nodes based on leaf sequences
    
    void						CountChangesAlongEveryBranch(	//count the changes of each branch
                                                             ObservedData& theData );
	
    void						printbranchlength(	//print branch lengths
                                                  ofstream& outStream);
    
    void						printbranchlabel(	//print the label of each branch
                                                 ofstream& outStream);
    
    // Many different versions of this.  Some are probably redundant
    void						setprobmatrixalongbranch(	//set the probability matrix of each branch based on conditional pathway method
                                                         MixtureModel& mModel,
                                                         DoubleVector& inModelFreqVector);
    
    void						setprobmatrixalongbranch(MixtureModel& mModel);
    
    void						setprobmatrixalongbranch(MixtureModel& mModel,
                                                         DoubleVector& inModelFreqVector, int dummyGibbs);
    
    void						setprobmatrixalongbranch(
                                                         //set the probability matrix of each branch based on conditional pathway method
                                                         MixtureModel& mModel,
                                                         int dummyGibbs);
    
    void						setprobmatrixalongbranch(MixtureModel& mModel,
                                                         DoubleVector& inModelFreqVector, int theModel, int rowz);
    
    void						setprobmatrixalongbranch(MixtureModel& mModel,
                                                         int theModel, int row);
    
    void						setprobmatrixalongbranch(MixtureModel& mModel,
                                                         DoubleVector& inModelFreqVector, int dummyGibbs, int theModel, int row);
    
    void						setprobmatrixalongbranch(MixtureModel& mModel,
                                                         int dummyGibbs, int theModel, int row);
    
    void						calculatetreelnlikelihood(	//calculate the log likelihood of the tree
                                                    //     MixtureModel& mModel,
                                                          double& m_Lnlikelihood );
    
    void	calculatetreelnlikelihooddeltab1( double& m_Lnlikelihood, int theModel, int row); //MixtureModel& mModel, 
    
    void	SumChangesOverTree(IntMatrix& m_matCountChange); //count the state changes all over the tree
    
    
    void AssignFromPosteriorPredictive(
                                       const DoubleMatrix& stateFreqs,
                                       const IntVector& m_vecSiteAssignment,
                                       const bool&	m_bSite,
                                       const IntVector& inStateindex,
                                       const std::vector<char>& inStateVector,
                                       const int& minstate,
                                       const int& numStates, std::vector<int>& tipStates);
    
    double calculateObservedSequenceTestStatistic(
                                                  const DoubleMatrix& stateFreqs,
                                                  const IntVector& m_vecSiteAssignment,
                                                  const bool&	m_bSite,
                                                  const IntVector& inStateindex,
                                                  const std::vector<char>& inStateVector,
                                                  const int& minstate,
                                                  const int& numStates);
    
    int	AssignAncestralStatesByMarginalProbability(MixtureModel& theModel,
												   ObservedData& theData,
												   IntVector& m_vecSiteAssignment,
												   DoubleMatrix& stateFreqMatrix);
    
    void MarkNodesBelow(
                        const int& theNode,
                        const IntVector& m_vecSiteAssignment,
                        const bool&	m_bSite,
                        const IntVector& inStateindex,
                        const std::vector<char>& inStateVector,
                        const int& minstate,
                        const int& numStates);
    
    int	AssignAncestralStatesByMarginalProbability(MixtureModel& theModel,
												   ObservedData& theData,
												   IntVector& m_vecSiteAssignment,
												   DoubleMatrix& stateFreqMatrix,
												   int assignMode);
    
    double	getIncompleteDataLikelihood(
                                        const DoubleMatrix& stateFreqs,
                                        const IntVector& m_vecSiteAssignment,
                                        const bool&	m_bSite,
                                        const IntVector& inStateindex,
                                        const std::vector<char>& inStateVector,
                                        const int& minstate,
                                        const int& numStates);
    
    void subdivideLongBranches(TreeNode* theNode, double maxBL);
    void subdivideLongBranches(TreeNode* theNode, double maxBL, int recurseTree );
    void subdivideLongBranches(TreeNode* theNode, double maxBL, DoubleVector* uniformizationConstants, int recurseTree);
    
    double discreteBranchLengthApproximation(TreeNode* theNode);
    //double discreteBranchLengthApproximation(TreeNode* theNode, double fixedBL);
    
    double returnLossFromBLApproximation(TreeNode* theNode, int numSizes, double *theSizes);
    
    void initSiteLikelihoods(
                             IntVector& m_vecSiteAssignment,
                             int numStates);
    
    void initSiteLikelihoods(
                             IntVector& m_vecSiteAssignment,
                             int numStates,
                             int startSite,
                             int endSite);
    
    void						clear(void);
    
    void						swap(
                                     Tree&					inSrc);
    
    void						DetachRoot(void);
    
    void						SetRoot(
                                        node_type*				inNewRoot);
    
    void						MakeRootless(void);
    
    template <typename Functor>
    Functor						RemoveFreeLeaves(
                                                 const Functor&			ioFunctor);
    
    template <typename Functor>
    Functor						for_each(
                                         const Functor&			ioFunctor);
    
    Tree(
         const Tree&	inSrc);
    
    const Tree&					operator=(
                                          const Tree&	inSrc);
    
    void 						setrootsequence(	//set root sequence of the tree
                                                const vector<int>& m_strAncSeq);
    
    void						evolvetree(	//mapping mutations on branches without output
                                           const DoubleMatrix& inRates,
                                           const IntVector& inStateindex,
                                           const std::vector<char>& inStateVector,
                                           const int& minstate,
                                           const int& numStates);
    
    void						evolvetree(	//mapping mutations on branches with output
                                           ofstream& outStream,
                                           const DoubleMatrix& inRates,
                                           const IntVector& inStateindex,
                                           const std::vector<char>& inStateVector,
                                           const int& minstate,
                                           const int& numStates);
    
    void						updateBranchLength(	//update branch lengths
                                                   double& m_fWindowsize);
    
    int						KEEP_updateBranchLength(
                                               //int startNode, const double& m_fWindowsize);
                                               int startNode,
                                               double& windowSize,
                                               MixtureModel& theModel,
                                               ObservedData& theData,
                                               IntVector& m_vecSiteAssignment,
                                               int version);
    
    void 						treecopy(	//copy a tree
                                         Tree& inTarget);
    
    void 						treedatacopy(	//copy tree data to another tree
                                             Tree& inTarget);
    
    void 						printtreechangescount(	//print state changes along the whole tree
                                                      ofstream& outStream);
    
    void 						printtreesitechangescount(	//print the state changes for each site all over the tree
                                                          ofstream& outStream);
	
    int                     ResampleSiteAssignments_MixtureModel(MixtureModel& mModel ,
                                                                 ObservedData& theData,
                                                                 IntVector& m_vecSiteAssignment);
    

    void KEEP_CountFullChangesOverTree(
                                                  MixtureModel& theModel,
                                                  ObservedData& theData,
                                                   IntVector& m_vecSiteAssignment);

        void KEEP_CountFullChangesOverTree(
                                                      MixtureModel& theModel,
                                                      ObservedData& theData,
                                                       IntVector& m_vecSiteAssignment,
                                                      int version);

        
	int						KEEP_ResampleSiteAssignments_ASRV_rates(
                                                               MixtureModel& theModel,
                                                               ObservedData& theData,
                                                               IntVector& m_vecSiteAssignment);	
	
	
    void calculateJointDwellTimeOverBranches( int s1, int s2,
                                             double siteRate1, double siteRate2,
											 MixtureModel& theModel,
											 ObservedData& theData,
											 IntVector& m_vecSiteAssignment);
	
    int calculateDwellTimeOverBranches( int s,
                                       double siteRate,
									   MixtureModel& theModel,
									   ObservedData& theData,
									   IntVector& m_vecSiteAssignment);
	
    
    void						calculateCompleteDataLogL(	//calculate the tree log likelihood when we use gibbs sampling for mixture
                                                          MixtureModel& theModel,
                                                          IntVector& m_vecSiteAssignment,
                                                          double& m_Lnlikelihood, int version);
    
    void						calculateCompleteDataLogLdeltab1(
                                                                 MixtureModel& mModel,
                                                                 IntVector& m_vecSiteAssignment,
                                                                 double& m_Lnlikelihood, int theModel, int x);
    
    double						calculateCompleteDataLogLdeltab1uni(
                                                                    MixtureModel& mModel,
//                                                                    DoubleMatrix3D *rateMat,
//                                                                    DoubleVector *uniformizationConstants,							
                                                                    IntVector& m_vecSiteAssignment,
                                                                    double& m_Lnlikelihood, int theModel, int x, int y);
    
    int 						ResampleAncestralAndTransientStates_local(	//update the internal states when we use gibbs sampling for mixture
 																		  int& rannode,
																		  MixtureModel& theModel,
																		  ObservedData& theData,
																		  IntVector& m_vecSiteAssignment,
																		  DoubleMatrix& stateFreqMatrix);
    
    double						simulatePosteriorPredictiveTestStatistic_multinomialLikelihood(
                                                                                               const DoubleMatrix& stateFreqs,
                                                                                               const IntVector& m_vecSiteAssignment,
                                                                                               const bool&	m_bSite,
                                                                                               const IntVector& inStateindex,
                                                                                               const std::vector<char>& inStateVector,
                                                                                               const int& minstate,
                                                                                               const int& numStates);
    
    void insertNodeTest(void);
    void assignSequencesToLeaves( ObservedData &theData );
    
    int							numRegularNodes;
    
    double completeDataLnL, incompleteDataLnL;
    
    // Need to maintain a pool of 'unused' nodes that we can re-assign during branchlength sampling...
    int addNodesToList(int howMany) { for (int i=0; i<howMany; i++) { nodePool.push_back( new node_type ); freeNodes++; } std::cout << "		[ Just added " << howMany << " nodes to free node pool ]" << endl; return freeNodes; }
    void recycleNode( node_type* theNode ) { nodePool.push_back( theNode ); }
    int freeNodes; // how many of the nodes in the list are free?
    
    node_type* getFreeNode(void) { if (freeNodes <= 1) { int temp = addNodesToList(50); } freeNodes--; return nodePool.at( nodePool.size() - freeNodes ); } // std::cout << "Allocating 10 new nodes" << endl; 
    
    TreeNode* getRoot(void);
    DoubleVector&					GetModelFrequencyVector(void);
    

    // Need to keep track of which nodes are used, and which ones aren't...
private:
    typedef	std::deque<node_type*>			NodeList;
    NodeList								nodePool;
    
    friend	class UTTree;
//    static	void					AssignChildrenToParent(
//                                                           node_type&				inParent,
//                                                           NodeList&				ioChildren);

	// Why was this static?
	void					AssignChildrenToParent(
                                                           node_type&				inParent,
                                                           NodeList&				ioChildren);
	
    
    void						SetModelFrequencyVector
    (const DoubleVector&	inModelFreqVector);
    
    double						rndu(void);
    
    node_type*					mRoot;
    
    DoubleVector					mModelFreqVector;
};


//---------------------------------------------------------------------------
#pragma mark -
//---------------------------------------------------------------------------


inline
Tree::Tree(void)
:	mRoot(0)
{
	freeNodes = 0;
	num = &Numerics::getGlobalNumerics();
    
	// Allocate 10 free nodes to start (if not using B1xM, this should be a really small amount of mem)...
	addNodesToList(10);
}


//---------------------------------------------------------------------------

inline
Tree::~Tree(void)
{
	clear();
}


//---------------------------------------------------------------------------

inline void
Tree::calculatetreelnlikelihood(
                                //       MixtureModel& mModel,
                                double& m_Lnlikelihood)
{
	DoubleVector tempFreqModels = GetModelFrequencyVector();
    
	mRoot->CalTreeLnLikelihood(tempFreqModels, m_Lnlikelihood);
}


//---------------------------------------------------------------------------

inline void
Tree::calculatetreelnlikelihooddeltab1(
                                       //             MixtureModel& mModel,
                                       double& m_Lnlikelihood, int theModel, int row)
{
	DoubleVector tempFreqModels = GetModelFrequencyVector();
    
	mRoot->CalTreeLnLikelihoodDeltaB1(  tempFreqModels, m_Lnlikelihood, theModel, row);
}


//---------------------------------------------------------------------------
inline int Tree::KEEP_updateBranchLength(int startNode,
                                  double& windowSize,
                                  MixtureModel& theModel,
                                  ObservedData& theData,
                                  IntVector& m_vecSiteAssignment,
                                  int version) {
    
    int found=0;
	mRoot->GetData().treeLength = 0;
	
	int dirtyTree = 0;
	mRoot->KEEP_UpdateBranchLength_fromStartNode(startNode, *this, found, 0, windowSize, theModel, theData, m_vecSiteAssignment, version, dirtyTree);
    
	if ( (dirtyTree > 0) && (options.m_nMethod == 6) && (!options.capNumTreeNodes ) && (num->Random() < 0.2) ) {
		std::cout << "		[ Re-parsing the tree structure to keep branch-lengths short enough for B1u; there were " << dirtyTree << " segments that exceeded " << options.maxBL << " ]" << endl;
		subdivideLongBranches( mRoot, options.maxBL, theModel.returnUniformizationConstants(), 1 ); // recursive version
		
		// count
		
		// should now recount changes too probably...
	}
    
    return dirtyTree;
}
//---------------------------------------------------------------------------

inline void
Tree::printbranchlength(
                        ofstream& outStream)
{
	mRoot->PrintBranchLength(outStream);
}

//---------------------------------------------------------------------------

inline void
Tree::printbranchlabel(
                       ofstream& outStream)
{
	mRoot->PrintBranchLabel(outStream);
}

//---------------------------------------------------------------------------

inline void
Tree::treecopy(
               Tree& inTarget)
{
	node_type* temp = new node_type;
	
	inTarget.mRoot = temp; 
	
	mRoot->CopyTree(inTarget.mRoot);
}

//---------------------------------------------------------------------------

inline void
Tree::treedatacopy(
                   Tree& inTarget)
{
	mRoot->CopyTreeData(inTarget.mRoot);
}

//---------------------------------------------------------------------------

inline
void
Tree::printtreechanges(
                       ofstream& outStream) const
{
	mRoot->PrintTreeChanges(outStream);
}

//---------------------------------------------------------------------------

inline
void
Tree::printtreechangescount(
                            ofstream& outStream)
{
	mRoot->PrintBranchChangecount(outStream);
}

//---------------------------------------------------------------------------

inline
void
Tree::printtreesitechangescount(
                                ofstream& outStream)
{
	mRoot->PrintSiteChangecount(outStream);
}

//---------------------------------------------------------------------------

inline
double
Tree::rndu(void)
{
	return Numerics::getGlobalNumerics().Random();
}


//---------------------------------------------------------------------------

inline
int
Tree::calculateDwellTimeOverBranches(int s,
                                     double siteRate,
                                     MixtureModel& theModel,
                                     ObservedData& theData,
                                     IntVector& m_vecSiteAssignment)
{
	if ( mRoot->GetData().tempTreeDepth.empty() ) {
		mRoot->GetData().tempTreeDepth.resize( theModel.GetNumModels(), 0.0 );
	} else {
		mRoot->GetData().tempTreeDepth.fill( 0.0 );
	}
	
	double returnVal = mRoot->calculateDwellTimeOverBranches( s, siteRate, theModel, theData, m_vecSiteAssignment );
	
	DoubleMatrix* marginalDwell = mRoot->getMarginalDwell( mRoot );
	DoubleVector* treeDepth = &mRoot->GetData().tempTreeDepth;
	
    //	for (int k=0; k<numModels; k++) {
    for (int i=0; i<theData.getNumStates(); i++) {
        marginalDwell->mRows.mDataArray[ 0 ].mDataArray[ i ] /= (*treeDepth)[0];
        //			std::cout << "Model " << k << ", site " << s << ", state " << i << ", marginal dwell time (depth normalized): " << ( marginalDwell->mRows.mDataArray[ k ].mDataArray[ i ] / treeDepth) << endl;
    }
    //	}
	
	return returnVal;
}

//---------------------------------------------------------------------------

inline
void
Tree::calculateJointDwellTimeOverBranches( int s1, int s2,
                                          double siteRate1, double siteRate2,
                                          MixtureModel& theModel,
                                          ObservedData& theData,
                                          IntVector& m_vecSiteAssignment)
{
	mRoot->calculateJointDwellTimeOverBranches( s1, s2, siteRate1, siteRate2, theModel, theData, m_vecSiteAssignment );
	
    //	DoubleMatrix* jointDwell = mRoot->getJointDwell( mRoot );
    //	DoubleVector* treeDepth = &mRoot->GetData().tempTreeDepth;
	
    //	for (int k=0; k<numModels; k++) {
    //	std::cout << "DIAG OUTPUT: TREE DEPTH IS " << (*treeDepth)[0] << endl;
	
    //		for (int i=0; i<numStates; i++) {
    //			std::cout << i << ": before scaled " << jointDwell->mRows.mDataArray[ 0 ].mDataArray[ i ]  << " and after ";
    //			jointDwell->mRows.mDataArray[ 0 ].mDataArray[ i ] /= ( (*treeDepth)[0] * 2.0 );
    //			std::cout << jointDwell->mRows.mDataArray[ 0 ].mDataArray[ i ] << endl;
    
    // BLOCKING RESCALE BY TREE DEPTH FOR NOW...
    
    //			std::cout << "Model " << k << ", site " << s << ", state " << i << ", marginal dwell time (depth normalized): " << ( marginalDwell->mRows.mDataArray[ k ].mDataArray[ i ] / treeDepth) << endl;
    //		}
    //	}
}

// --- Uniformized method here... ------

//---------------------------------------------------------------------------

inline
void
Tree::KEEP_CountFullChangesOverTree(
                                    MixtureModel& theModel,
                                    ObservedData& theData,
                                    IntVector& m_vecSiteAssignment)
{
	KEEP_CountFullChangesOverTree(theModel, theData, m_vecSiteAssignment, 0);
}

//---------------------------------------------------------------------------

inline
void
Tree::KEEP_CountFullChangesOverTree(
                                    MixtureModel& theModel,
                                    ObservedData& theData,
                                    IntVector& m_vecSiteAssignment,
                                    int version)
{
	// version 0: do all the counting
	// version 1: resample / calculate the virtual subs
	int numStates = theData.getNumStates();
	int numModels = theModel.GetNumModels();
	
	if (options.m_nMethod != 6) std::cout << "ERROR: CountFullChangesOverTreeUni() called for non-uniformized method" << endl;
	mRoot->countChangesAlongBranchUni( theModel, theData, m_vecSiteAssignment, version);
    
	if ( ( options.doB1uStar == 0 ) && (options.m_nMethod == 6) && (version == 0) ) { 
		DoubleMatrix   *augmentedLogSum;
		DoubleMatrix3D *ChangeCount3D;
		DoubleMatrix3D *ChangeCount3D_scaling;
		DoubleMatrix *augmentedLogSum_scaling;
        
		augmentedLogSum			= root()->getAugmentedLogSum( mRoot );
		ChangeCount3D			= root()->getChangeMatrix3Duni( mRoot ); 
		augmentedLogSum_scaling	= root()->getAugmentedLogSum_scaling( mRoot );
		ChangeCount3D_scaling	= root()->getChangeMatrix3D_scaling( mRoot ); 
        
		// Take logarithms and subtract off the scaling factors...
		for (int k=0; k<numModels; k++) {
			for (int i=0; i<numStates; i++) {
				for (int j=0; j<numStates; j++) {
					if ((*ChangeCount3D).mRows.mDataArray[k].mRows.mDataArray[i].mDataArray[j] != 0) {
						(*ChangeCount3D).mRows.mDataArray[k].mRows.mDataArray[i].mDataArray[j] = num->logarithm( (*ChangeCount3D).mRows.mDataArray[k].mRows.mDataArray[i].mDataArray[j] ) - (*ChangeCount3D_scaling).mRows.mDataArray[k].mRows.mDataArray[i].mDataArray[j];		
					} else {
						(*ChangeCount3D).mRows.mDataArray[k].mRows.mDataArray[i].mDataArray[j] = num->logarithm( 1e-70 ) - (*ChangeCount3D_scaling).mRows.mDataArray[k].mRows.mDataArray[i].mDataArray[j];		
					}
				}
				if ((*augmentedLogSum).mRows.mDataArray[k].mDataArray[i] != 0) {
					(*augmentedLogSum).mRows.mDataArray[k].mDataArray[i] = num->logarithm( (*augmentedLogSum).mRows.mDataArray[k].mDataArray[i] ) - (*augmentedLogSum_scaling).mRows.mDataArray[k].mDataArray[i];		
				} else {
					(*augmentedLogSum).mRows.mDataArray[k].mDataArray[i] = num->logarithm( 1e-70 ) - (*augmentedLogSum_scaling).mRows.mDataArray[k].mDataArray[i];		
				}
			}
		}
	}
    
}
// ----


//---------------------------------------------------------------------------

inline
int
Tree::KEEP_ResampleSiteAssignments_ASRV_rates(
                                              MixtureModel& theModel,
                                              ObservedData& theData,
                                              IntVector& m_vecSiteAssignment)
{
	int numUpdated = 0;
	int doFullCount = 0;
	//std::cout << "doing gibbs assignment" << endl;
	
    //	setprobmatrixalongbranch(options.m_nMethod, numStates, RateMatrix3D, uniformizationConstants, stateFreqs, 1);
	
	int numStates = theData.getNumStates();
	int numModels = theModel.GetNumModels();
	int numRateModels = options.numCatASRV;
	
	int i, m;
	
	static DoubleVector LnLikelihoodVector(numRateModels, 1.0);
	static DoubleVector LnLikelihoodVector2(numRateModels, 1.0);
	static DoubleVector m_fLikelihoodRatio(numRateModels, 0);
	static DoubleVector m_fCummulativeRatio(numRateModels, 0);
	
	IntVector* siteAssignments = &root()->GetData().siteCats->siteAssignments;
	
	LnLikelihoodVector.fill(1.0);
	LnLikelihoodVector2.fill(1.0);
	m_fLikelihoodRatio.fill(0);
	m_fCummulativeRatio.fill(0);
	
	if (options.m_bSite)	//assign different models for different sites
	{
		if (siteAssignments->empty()) {
			siteAssignments->resize(theData.getAlignmentLength(), 0);
			KEEP_CountFullChangesOverTree( theModel, theData, m_vecSiteAssignment );
            
		} else {
			// siteAssignments->fill(-1);
		}
		
		if (options.numCatASRV <= 1) {
			return 0;
		}
		
		// Set the P matrix for each model on each branch
		if (options.m_nMethod != 6) {
			int tempCalcProbDist = mRoot->getCalculateFullProbabilityDistribution();
			mRoot->setCalculateFullProbabilityDistribution(1); //override
			setprobmatrixalongbranch( theModel, 1);
			mRoot->setCalculateFullProbabilityDistribution(tempCalcProbDist);				
		} else {
			// We don't need the branch-wise substitution prob matrices for B1u...
			int tempCalcProbDist = mRoot->getCalculateFullProbabilityDistribution();
			mRoot->setCalculateFullProbabilityDistribution(0); // '1' should not be necessary here for B1u
			setprobmatrixalongbranch( theModel, GetModelFrequencyVector(), 1 );
			mRoot->setCalculateFullProbabilityDistribution(tempCalcProbDist);				
		}
		
		// THIS STEP IS QUESTIONABLE: BOTH HERE AND IN STANDARD ResampleSiteAssignments_MixtureModel()...
        /*		if (options.m_nMethod == 6) {
         // mModel.updateUniformization();
         // Need to recalculate all the prob matricies...
         int tempCalcProbDist = mRoot->getCalculateFullProbabilityDistribution();
         mRoot->setCalculateFullProbabilityDistribution(1);
         setprobmatrixalongbranch(options.m_nMethod, numStates, RateMatrix3D, uniformizationConstants, stateFreqs, 1);
         mRoot->setCalculateFullProbabilityDistribution(tempCalcProbDist);
         } 
         */
		
        //	std::cout << "assignments: ";
        
		// IS THIS NEEDED?
        //		if (options.m_nMethod == 6) KEEP_CountFullChangesOverTree(options.m_nMethod, m_vecSiteAssignment, numModels, uniformizationConstants, inStateindex, minstate, numStates, RateMatrix3D);	
        //	else KEEP_CountFullChangesOverTree(options.m_nMethod, m_vecSiteAssignment, numModels, inStateindex, minstate, numStates);
        
        
		for (i = 0; i < theData.getAlignmentLength(); i++)
		{
			for (int j=0; j<numRateModels; j++) { LnLikelihoodVector[j] = 1.0; }
			
			// calculate the site specific likelihood (rather than log-likelihood) for each model
			
			// NOTE: THIS IS OVERRIDING THE USE OF THE B1U SPECIFIC FUNCTION FOR CalculateSiteSpecificLnLikelihoodUni_ASRV(), WHICH DOESN'T SEEM TO BE WORKING...
			if (options.m_nMethod != 6) {
				mRoot->CalculateSiteSpecificLnLikelihood_ASRV( i, theModel, theData, m_vecSiteAssignment, LnLikelihoodVector );
			} else {
                // *******************************************************************************************************************************************************
                // The first method uses a site-specific calc that is INTEGRATED OVER VIRTUAL SUBSTITUTIONS...
                // *******************************************************************************************************************************************************
                /*	int tempCalcProbDist = mRoot->getCalculateFullProbabilityDistribution();
                 mRoot->setCalculateFullProbabilityDistribution(1);
                 setprobmatrixalongbranch(options.m_nMethod, numStates, RateMatrix3D, uniformizationConstants, stateFreqs, 1);
                 mRoot->setCalculateFullProbabilityDistribution(tempCalcProbDist);				
                 mRoot->CalculateSiteSpecificLnLikelihood_ASRV(i, theModel, theData, m_vecSiteAssignment, LnLikelihoodVector );
                 double lik1, lik2;
                 lik1 = 0.0;
                 std::cout << "SITE " << i << "..." << endl;
                 for (int ii=0; ii<LnLikelihoodVector.size(); ii++) {
                 lik1 += LnLikelihoodVector[ mRoot->GetData().siteCats->getAssignment(i) ];
                 std::cout << "First method, site lik for " << ii << " is: " << LnLikelihoodVector[ii] << ", vs 0: " << exp( LnLikelihoodVector[ii] - LnLikelihoodVector[0]) << endl;
                 }
                 mRoot->setCalculateFullProbabilityDistribution(0);
                 
                 for (int kk=0; kk<10; kk++) {	
                 setprobmatrixalongbranch(options.m_nMethod, numStates, RateMatrix3D, uniformizationConstants, stateFreqs, 1);
                 // *******************************************************************************************************************************************************
                 
                 
                 KEEP_CountFullChangesOverTree(options.m_nMethod, m_vecSiteAssignment, numModels, uniformizationConstants, inStateindex, minstate, numStates, RateMatrix3D, 1);	
                 */	
				// *******************************************************************************************************************************************************
				// This method uses a delta-logL that is NOT integrated over virtual substitutions: we are keeping them fixed...
                mRoot->CalculateSiteSpecificLnLikelihoodUni_ASRV( i, theModel, theData, m_vecSiteAssignment, LnLikelihoodVector );
                /*		lik2 = 0.0;
                 for (int ii=0; ii<LnLikelihoodVector.size(); ii++) {
                 std::cout << "Second method, site lik for " << ii << " is: " << LnLikelihoodVector[ii]  << ", vs 0: " << exp( LnLikelihoodVector[ii] - LnLikelihoodVector[0]) << endl;
                 lik2+= LnLikelihoodVector[ii];
                 }
                 std::cout << "site lik 1: " << lik1 << endl;
                 std::cout << "site lik 2: " << lik2 << endl;
                 
                 // *******************************************************************************************************************************************************
                 } //exit(-1);
                 */
			}
			
			// Gibbs sampling using the likelihood space of each model
			double logLikelihoodModelMax;
			
			for (m = 0; m < numRateModels; m++)
			{
				if (m== 0)
					logLikelihoodModelMax = LnLikelihoodVector[m];
				else if (LnLikelihoodVector[m] > logLikelihoodModelMax)
					logLikelihoodModelMax = LnLikelihoodVector[m];
			}
			
			m_fLikelihoodRatio.fill(0);
			
			double sum = 0.0;
			
			for (m = 0; m < numRateModels; m++)
			{
				m_fLikelihoodRatio[m] = exp( LnLikelihoodVector[m] - logLikelihoodModelMax);
				
				sum += m_fLikelihoodRatio[m];
			}
			
			for (m = 0; m < numRateModels; m++) m_fCummulativeRatio[m] = 0.0;
			
			for (m = 0; m < numRateModels; m++)
			{
				if (m == 0)
					m_fCummulativeRatio[m] = m_fLikelihoodRatio[m] / sum;
				else
					m_fCummulativeRatio[m] = m_fCummulativeRatio[m-1] + m_fLikelihoodRatio[m] / sum;
				}
			
			// assign each site with the model to the assignment vector
			double rannum = rndu();
			
			int lastAssignment = siteAssignments->mDataArray[i]; //mRoot->GetData().siteCats->getAssignment( i ); // 
			int itWasSet=0;
			for (m = 0; m < numRateModels; m++)
			{
				if (rannum <= m_fCummulativeRatio[m])
				{
					siteAssignments->mDataArray[i] = m;
					//std::cout << m << " ";
					itWasSet = 1;
					break;
				}
			}
			numUpdated++;
           
			if (itWasSet == 0) { std::cout << "ERROR! ResampleSiteAssignments_MixtureModel not mode!!!  rannum was " << rannum << "; cum[numRateModels] = " << m_fCummulativeRatio[(numRateModels-1)] << ", numRateModels = " << numRateModels << "; sum = " << sum << endl; }
			
			if ( ( options.m_nMethod == 6) && ( siteAssignments->mDataArray[i] != lastAssignment ) ) {
				// Re-allocate the counts from one category to another...
				if ( lastAssignment < 0 ) {
					doFullCount = 1;
				} else {
					doFullCount = 0;
					mRoot->reassignSiteCountsBetweenRateCategories_Uni( i, lastAssignment, siteAssignments->mDataArray[i], theData, m_vecSiteAssignment );
				}
			}
   		}
		// calculate the changes for each model in each branches
		if (options.m_nMethod != 6) {
			KEEP_CountFullChangesOverTree(theModel, theData, m_vecSiteAssignment );
		}
		
		// Handling counts separately...
		//else {
		if ( (options.m_nMethod == 6) && (doFullCount == 1) ) {
            // [block]	std::cout << "	[ Rebuilding full substitution count index, since last counts were invalid or unset. ]" << endl;
			KEEP_CountFullChangesOverTree(theModel, theData, m_vecSiteAssignment );
		}
		
	}
	else			//assign different models for different lineages(sequences)
	{
		std::cout << "	[ERROR: lineage-based mixtures not implemented yet for ResampleSiteAssignments_ASRV_rates]" << endl;
		exit(-1);
		// calculate the likelihood of each model in each branch, sampling from the likelihood of each branch
		// and assign each branch to each model
		mRoot->GibbsSamplingLineageAssignment();
	}
	
    
	return numUpdated;
}


//---------------------------------------------------------------------------

inline void
Tree::calculateCompleteDataLogL(
                                MixtureModel& theModel,
                                IntVector& m_vecSiteAssignment,
                                double& m_Lnlikelihood, int version)
{
	
	mRoot->CalcLogLikelihood_branch( theModel, m_vecSiteAssignment, m_Lnlikelihood, version);
}

//---------------------------------------------------------------------------

inline void
Tree::calculateCompleteDataLogLdeltab1(
                                       MixtureModel& mModel,
                                       IntVector& m_vecSiteAssignment,
                                       double& m_Lnlikelihood, int theModel, int x)
{
	mRoot->CalcLogLikelihood_branchDeltaB1( mModel, m_vecSiteAssignment, m_Lnlikelihood, theModel, x);
}


//---------------------------------------------------------------------------

inline double
Tree::calculateCompleteDataLogLdeltab1uni(          MixtureModel& mModel,
                                          //         DoubleMatrix3D *rateMat,
                                          //       DoubleVector *uniformizationConstants,
                                          IntVector& m_vecSiteAssignment,
                                          double& m_Lnlikelihood, int theModel, int x, int y)
{
	return mRoot->CalcLogLikelihood_branchDeltaB1uni( mModel, m_vecSiteAssignment, m_Lnlikelihood, theModel, x, y);
}


//---------------------------------------------------------------------------

inline int
Tree::ResampleAncestralAndTransientStates_local(
                                                int& rannode,
                                                MixtureModel& theModel,
                                                ObservedData& theData,
                                                IntVector& m_vecSiteAssignment,
                                                DoubleMatrix& stateFreqMatrix) {
	int found=0;
 
	
	//---------------------------------------------------------------------------		
	// Looking up the random internal node which serves as the start of updation
	//---------------------------------------------------------------------------		
	return mRoot->ResampleAncestralAndTransientStates_local_fromNode( rannode, found, theModel, theData, m_vecSiteAssignment, stateFreqMatrix ); 
}

/*
 //---------------------------------------------------------------------------
 
 void
 Tree::MarkNodesBelow(
 const int& theNode,
 const IntVector& m_vecSiteAssignment,
 const bool&	m_bSite,
 const IntVector& inStateindex,
 const std::vector<char>& inStateVector,
 const int& minstate,
 const int& numStates)
 {
 int found=0;
 //	std::cout << "	call to ResampleAncestralAndTransientStates_local (tree.h); calling mRoot->ResampleAncestralAndTransientStates_local_fromNode..." << endl;
 
 //---------------------------------------------------------------------------		
 // Looking up the random internal node which serves as the start of updation
 //---------------------------------------------------------------------------		
 mRoot->markNodesBelow(theNode, found, m_vecSiteAssignment, options.m_bSite, inStateindex, inStateVector, minstate, numStates); 
 }
 */


//---------------------------------------------------------------------------

inline int
Tree::AssignAncestralStatesByMarginalProbability(MixtureModel& theModel,
                                                 ObservedData& theData,
                                                 IntVector& m_vecSiteAssignment,
                                                 DoubleMatrix& stateFreqMatrix)
{
	// Default to sampling from posterior (mode 1)...
	return AssignAncestralStatesByMarginalProbability( theModel, theData, m_vecSiteAssignment, stateFreqMatrix, 1);
}

inline int Tree::AssignAncestralStatesByMarginalProbability(MixtureModel& mModel,
                                                     ObservedData& theData,
                                                     IntVector& m_vecSiteAssignment,
                                                     DoubleMatrix& stateFreqMatrix,
                                                     int assignMode) {
	
	// When assignMode = 0, only do the complete conditional posterior calculation...
    int numStates = theData.getNumStates();
	int updateCount = 0;
	
	if (!options.m_bSite) {
		std::cout << "ERROR.  Ancestral state reconstruction only checked/implemented for integration by site, for now." << endl;
		exit(-1);
	}
	
	// Initialize the site likelihood data structure
	initSiteLikelihoods( m_vecSiteAssignment, numStates);
    
	// Do the post-order traversal
    int mode = 0;
    
	mRoot->setSiteLikelihoods_downPass( mRoot, mModel, theData, m_vecSiteAssignment, GetModelFrequencyVector(), mode );
	updateCount = mRoot->setSiteLikelihoods_doSimpleAssignment(mRoot, mModel, theData, m_vecSiteAssignment, GetModelFrequencyVector(), assignMode);

    DoubleVector *scalingFactors  = & mRoot->GetData().scalingFactor;
    
    double lnL = 0.0;
	double temp;
	int theModel;
	for (int i=0; i<m_vecSiteAssignment.size();i++) {
		theModel = m_vecSiteAssignment.mDataArray[i];
		temp = 0.0;
		for (int j=0; j<numStates; j++) {
			// [?? Using stateFreqs here instead of stationaryFreqs... ??]
			temp +=  exp((*scalingFactors)[i] ) * mRoot->GetData().siteLikelihoodBelow[(i*numStates) + j]* mModel.returnStateFrequencies().mRows.mDataArray[theModel].mDataArray[j]; 
		}
		
		if (temp <= 0) temp = 1e-70;

        // This led to quite different likelihoods...
        // lnL += num->logarithm( temp );
		
        // We want full precision here...
        lnL += log(temp);
	}
    
	incompleteDataLnL = lnL;
	return updateCount;
}


/*
 
 // Needs to be converted over
 
 double
 Tree::simulatePosteriorPredictiveTestStatistic_multinomialLikelihood(
 const DoubleMatrix& stateFreqs,
 const IntVector& m_vecSiteAssignment,
 const bool&	m_bSite,
 const IntVector& inStateindex,
 const std::vector<char>& inStateVector,
 const int& minstate,
 const int& numStates)
 {
 std::map<std::vector<int>,int> sitePatternCounts;
 
 std::vector<int> tipStates;
 tipStates.resize( leaves() );
 
 for (int i=0; i< m_vecSiteAssignment.size(); i++) {
 mRoot->posteriorPredictive_doSimpleAssignment(mRoot, stateFreqs, GetModelFrequencyVector(), m_vecSiteAssignment, inStateindex, inStateVector, minstate, numStates, tipStates);
 sitePatternCounts[tipStates]++;
 }
 
 // Now do the test statistic...
 std::map<std::vector<int>,int>::iterator iter;   
 double testStat = 0.0;
 for( iter = sitePatternCounts.begin(); iter != sitePatternCounts.end(); iter++ ) {
 testStat += iter->second * log( iter->second );
 }
 
 testStat -= m_vecSiteAssignment.size() * log( m_vecSiteAssignment.size() );
 //testStat *= m_vecSiteAssignment.size() /5000.0;
 
 std::cout << "Posterior predictive T(X) = " << testStat << " (over " << m_vecSiteAssignment.size() << " sites, " << sitePatternCounts.size() << " unique site patterns)" << endl;
 return testStat;
 }
 
 
 
 double
 Tree::calculateObservedSequenceTestStatistic(
 const DoubleMatrix& stateFreqs,
 const IntVector& m_vecSiteAssignment,
 const bool&	m_bSite,
 const IntVector& inStateindex,
 const std::vector<char>& inStateVector,
 const int& minstate,
 const int& numStates)
 {
 std::map<std::vector<int>,int> sitePatternCounts;
 
 std::vector<int> tipStates;
 tipStates.resize( leaves() );
 
 for (int i=0; i< m_vecSiteAssignment.size(); i++) {
 mRoot->posteriorPredictive_getSite(i, mRoot, stateFreqs, GetModelFrequencyVector(), m_vecSiteAssignment, inStateindex, inStateVector, minstate, numStates, tipStates);
 sitePatternCounts[tipStates]++;
 }
 
 // Now do the test statistic...
 std::map<std::vector<int>,int>::iterator iter;   
 double testStat = 0.0;
 for( iter = sitePatternCounts.begin(); iter != sitePatternCounts.end(); iter++ ) {
 testStat += iter->second * log( iter->second );
 }
 
 testStat -= m_vecSiteAssignment.size() * log( m_vecSiteAssignment.size() );
 
 //std::cout << "Observed T(X) = " << testStat << " (over " << m_vecSiteAssignment.size() << " sites, " << sitePatternCounts.size() << " unique site patterns)" << endl;
 
 return testStat;
 }
 
 
 void
 Tree::AssignFromPosteriorPredictive(
 const DoubleMatrix& stateFreqs,
 const IntVector& m_vecSiteAssignment,
 const bool&	m_bSite,
 const IntVector& inStateindex,
 const std::vector<char>& inStateVector,
 const int& minstate,
 const int& numStates, std::vector<int>& tipStates)
 {
 if (!m_bSite) {
 std::cout << "ERROR.  Posterior predictive sampling only implemented for integration over sites, for now." << endl;
 exit(-1);
 }
 
 return mRoot->posteriorPredictive_doSimpleAssignment(mRoot, stateFreqs, GetModelFrequencyVector(), m_vecSiteAssignment, inStateindex, inStateVector, minstate, numStates, tipStates);
 
 }
 
 
 
 //---------------------------------------------------------------------------
 
 double 
 Tree::getIncompleteDataLikelihood(
 const DoubleMatrix& stateFreqs,
 const IntVector& m_vecSiteAssignment,
 const bool&	m_bSite,
 const IntVector& inStateindex,
 const std::vector<char>& inStateVector,
 const int& minstate,
 const int& numStates)
 {
 // Likelihood integrated over all ancestral states...
 if (!m_bSite) {
 std::cout << "ERROR.  Ancestral state reconstruction only checked/implemented for integration by site, for now." << endl;
 exit(-1);
 }
 
 return mRoot->incompleteDataLikelihood(mRoot, stateFreqs, GetModelFrequencyVector(), m_vecSiteAssignment ); 
 }
 
 */
//---------------------------------------------------------------------------

inline void
Tree::initSiteLikelihoods(
                          IntVector& m_vecSiteAssignment,
                          int numStates)
{
	mRoot->initSiteLikelihoods(mRoot, m_vecSiteAssignment, numStates, 0, m_vecSiteAssignment.size() );
	
	if (mRoot->GetData().scalingFactor.size() == 0) {
		mRoot->GetData().scalingFactor.resize( m_vecSiteAssignment.size(), 0.0 );
	} else {
		for (int i=0; i< m_vecSiteAssignment.size(); i++) { mRoot->GetData().scalingFactor[i] = 0.0; }
	}
    
	// After the first init, we have set node scaling.  Need to turn off resetting it.
	mRoot->GetData().doScaling = 1; // this means don't reassign scaling to nodes
	
}

//---------------------------------------------------------------------------

inline void
Tree::initSiteLikelihoods(
                          IntVector& m_vecSiteAssignment,
                          int numStates,
                          int startSite,
                          int endSite)
{
	mRoot->initSiteLikelihoods(mRoot, m_vecSiteAssignment, numStates, startSite, endSite); //m_vecSiteAssignment, inStateindex, inStateVector, minstate,
	if (mRoot->GetData().scalingFactor.size() == 0) {
		std::cout << "WARNING!  SCALINGFACTOR VECTOR AT ROOT MUST BE ALLOCATED BEFORE CALLING THREADED initSiteLikelihoods()" << endl;
	} else {
		for (int i=startSite; i< endSite; i++) { mRoot->GetData().scalingFactor[i] = 0.0; }
	}
	mRoot->GetData().doScaling = 1; // this means don't reassign scaling to nodes
}




inline void
Tree::insertNodeTest(void) {
	TreeNode* theNewNode[10];
	int numNodes = leaves() + internalnodes() + 1;
	double bl = mRoot->Left()->Left()->GetData().distance;
    
	bl /= 10;
	
	mRoot->Left()->Left()->GetData().distance = bl;
	
	for (int i=0; i<9; i++) {
		theNewNode[i] = new TreeNode();
		mRoot->Left()->Left()->insertNodeAbove( theNewNode[i] );
 		theNewNode[i]->GetData().id = numNodes++;
		theNewNode[i]->GetData().distance = bl; // / 10 * rndu();
	}
}


inline double
Tree::returnLossFromBLApproximation(TreeNode* theNode, int numSizes, double *theSizes) {
	// This has to be called BEFORE doing the subdivision...
	for (int i=0; i<numSizes; i++) { if (theSizes[i] <= 0.005) { theSizes[i] = 0.005; } }
	for (int i=0; i<numSizes; i++) { for (int j=i+1; j<numSizes; j++) { if (theSizes[j] > theSizes[i]) { theSizes[j] = theSizes[i] - 0.0001; } } }
	
	int numNodes;
	static double loss = 0.0;
	static int segCount = 0;
	if (theNode == mRoot) { segCount = 0; loss = 0.0; }
	
	double bl = theNode->GetData().distance;
    
	if ( (theNode != NULL) ) {
		if (theNode->Left() != NULL) {
			returnLossFromBLApproximation(theNode->Left(), numSizes, theSizes);
		}
		
		if (theNode->Right() != NULL) {
			returnLossFromBLApproximation(theNode->Right(), numSizes, theSizes);
		}
        
		numNodes = leaves() + internalnodes() + 1;
		if ( (theNode->GetData().id != mRoot->GetData().id) && ( theNode->Up() != theNode->Up()->Up() ) ) {
			int totalNodeCount = 1;
			double newsize = theNode->GetData().distance;
            
			totalNodeCount = floor( newsize / theSizes[0] );
			
			if (numSizes <= 1) loss += (newsize -  (totalNodeCount * theSizes[0])) * (newsize -  (totalNodeCount * theSizes[0]));
			segCount++;
			
			int lastNodeCount = 0;
			// Now need to add extra branches of different sizes...
			double remainder = (newsize -  (totalNodeCount * theSizes[0]));
			for (int j=1; j<numSizes; j++) {
				segCount++;
				lastNodeCount = totalNodeCount;
				newsize = remainder;
				totalNodeCount = floor( newsize / theSizes[j] );
				remainder = (remainder -  (totalNodeCount * theSizes[j]));
                
				if (j == (numSizes-1)) { 
					loss += (remainder * remainder);
				}
				
				if (lastNodeCount == 0) { lastNodeCount = 1; }
			}
		} // not root
		else {
			// Root
			// do Nothing!
		}
	} // node not null
    
    //	if ( (theNode->Left() == NULL) && (theNode->Right() == NULL) ) std::cout << "Loss: " << loss << "; segCount: " << segCount << endl;
	return (loss*segCount);
}




inline
int
Tree::ResampleSiteAssignments_MixtureModel(MixtureModel& mModel,
                                           ObservedData& theData,
                                           IntVector& m_vecSiteAssignment
                                           ) {
    // This function resamples site assignments (m_vecSiteAssignment) using an unfortunately tedious process.
	// 1) Calculate all pmats while integrating over virtual substitutions, O(N^2 b)
	// 2) Calculate site-specific likelihoods for each model class
	// 3) Sample assignments
	// 4) When complete, recount substitutions (this should only be done for virtual subs)
	
	int numModels = mModel.GetNumModels();
	
	static DoubleVector LnLikelihoodVector(numModels, 1.0);
	static DoubleVector LnLikelihoodVector2(numModels, 1.0);
	static DoubleVector m_fLikelihoodRatio(numModels, 0);
	static DoubleVector m_fCummulativeRatio(numModels, 0);
    
  	int m_nAlignLength = theData.getAlignmentLength();
	
	int i, m;
	int numUpdated = 0;
	
	LnLikelihoodVector.fill(1.0);
	LnLikelihoodVector2.fill(1.0);
	m_fLikelihoodRatio.fill(0);
	m_fCummulativeRatio.fill(0);
	
	// There seems to be a memory problem with the allocation of this vector.  Forcing resize:
	if (m_vecSiteAssignment.empty() ) {
		m_vecSiteAssignment.resize(m_nAlignLength, 0);
		KEEP_CountFullChangesOverTree( mModel, theData, m_vecSiteAssignment );
	}	
	
	if (numModels <= 1) {
		return -1;
	}
	
	// set the P matrix for each model on each branch
	if (options.m_nMethod != 6) {
		setprobmatrixalongbranch(mModel, GetModelFrequencyVector(), 1);
	} else if ( (options.m_nMethod == 6) ) {
		// mModel.updateUniformization();
		
		// Need to recalculate all the prob matricies...
		int tempCalcProbDist = mRoot->getCalculateFullProbabilityDistribution();
		mRoot->setCalculateFullProbabilityDistribution(1);
		setprobmatrixalongbranch(mModel, GetModelFrequencyVector(), 1);
		mRoot->setCalculateFullProbabilityDistribution(tempCalcProbDist);
	} 
	
	
	for (i = 0; i < theData.getAlignmentLength(); i++) {
		for (int j=0; j<numModels; j++) { LnLikelihoodVector[j] = 1.0; }
		
		mRoot->CalculateSiteSpecificLnLikelihood( numModels, i, LnLikelihoodVector, theData );
		
		// Gibbs sampling using the likelihood space of each model
		double logLikelihoodModelMax;
		
		for (m = 0; m < numModels; m++)
		{
			if (m== 0)
				logLikelihoodModelMax = LnLikelihoodVector[m];
			else if (LnLikelihoodVector[m] > logLikelihoodModelMax)
				logLikelihoodModelMax = LnLikelihoodVector[m];
		}
		
		m_fLikelihoodRatio.fill(0);
		
		double sum = 0.0;
		
		for (m = 0; m < numModels; m++) {
			m_fLikelihoodRatio[m] = exp( LnLikelihoodVector[m] - logLikelihoodModelMax);
			sum += m_fLikelihoodRatio[m];
		}
		
		for (m = 0; m < numModels; m++) m_fCummulativeRatio[m] = 0.0;
		
		for (m = 0; m < numModels; m++) {
			if (m == 0)
				m_fCummulativeRatio[m] = m_fLikelihoodRatio[m] / sum;
			else
				m_fCummulativeRatio[m] = m_fCummulativeRatio[m-1] + m_fLikelihoodRatio[m] / sum;
		}
		
		// assign each site with the model to the assignment vector
		double rannum = rndu();
		
		int itWasSet=0;
		for (m = 0; m < numModels; m++) {
			if (rannum <= m_fCummulativeRatio[m]) {
				m_vecSiteAssignment[i] = m;
				itWasSet = 1;
				break;
			}
		}
		numUpdated++;
		
		if (itWasSet == 0) { std::cout << "ERROR! ResampleSiteAssignments_MixtureModel not mode!!!  rannum was " << rannum << "; cum[numModels] = " << m_fCummulativeRatio[(numModels-1)] << ", numModels = " << numModels << "; sum = " << sum << endl; }
	}
	
	// Resample the virtual substitutions...
	KEEP_CountFullChangesOverTree(mModel, theData, m_vecSiteAssignment, 1 );
	
	return numUpdated;
}

//---------------------------------------------------------------------------

inline void
Tree::setprobmatrixalongbranch(MixtureModel& mModel,
                               DoubleVector& inModelFreqVector) {
	setprobmatrixalongbranch( mModel, inModelFreqVector, -1, -1);
}


inline void
Tree::setprobmatrixalongbranch(
                               MixtureModel& mModel,
                               DoubleVector& inModelFreqVector, 
                               int theModel, 
                               int row)
{
    int numStates = mModel.GetNumStates();
    
	SetModelFrequencyVector(inModelFreqVector);
    
	if (options.m_nMethod == 0) {
		if (theModel != -1) {
			std::cout << "ERROR!  setprobmatrixalongbranch not implemented for row specific calcs in BL yet." << endl; exit(-1);
		} else {
			mRoot->SetTreeBLProb(numStates, *mModel.returnRateMatrix3D() );
		}
	} else if ( (options.m_nMethod == 1) || (options.m_nMethod == 3)) {
		if (theModel != -1) {
			mRoot->SetTreeB1Prob( numStates, *mModel.returnRateMatrix3D() , theModel, row);
		} else {
			mRoot->SetTreeB1Prob( numStates, *mModel.returnRateMatrix3D() , -1, -1);
		}
	} else if (options.m_nMethod == 6) {
        if (theModel != -1) {
            mRoot->SetTreeB1uProb( numStates, *mModel.returnRateMatrix3D() , mModel.returnUniformizationConstants(), theModel, row);
        } else {
            mRoot->SetTreeB1uProb( numStates, *mModel.returnRateMatrix3D() , mModel.returnUniformizationConstants(), -1, -1);
        }
 	} else if (options.m_nMethod == 2) {
		if (theModel != -1) {
			std::cout << "ERROR!  setprobmatrixalongbranch not implemented for row specific calcs in B2 yet." << endl; exit(-1);
		} else {
			mRoot->SetTreeB2Prob(numStates, *mModel.returnRateMatrix3D());
		}
	} else if ( (options.m_nMethod == 4) || (options.m_nMethod == 5) ) {
		// 4 = Binf, 5 = BinfxM
		if (theModel != -1) {
			mRoot->SetTreeBinfProb( numStates, *mModel.returnRateMatrix3D(), &mModel.returnStateFrequencies(), theModel);
		} else {
			mRoot->SetTreeBinfProb( numStates, *mModel.returnRateMatrix3D(), &mModel.returnStateFrequencies(), -1);
		}	
	}
	else
		throw(9);
}

//---------------------------------------------------------------------------

inline void
Tree::setprobmatrixalongbranch(MixtureModel& mModel,
                               DoubleVector& inModelFreqVector, 
                               int dummyGibbs) {	
    
	setprobmatrixalongbranch( mModel, inModelFreqVector, dummyGibbs, -1, -1 );
    
}


inline void
Tree::setprobmatrixalongbranch(
                               MixtureModel& mModel,
                               DoubleVector& inModelFreqVector, 
                               int dummyGibbs, 
                               int theModel, 
                               int row)
{	
	int numStates = mModel.GetNumStates();
	
	SetModelFrequencyVector(inModelFreqVector);
    
	// set probability calculation method here
	if (options.m_nMethod == 0) {
		if (theModel != -1) {
			std::cout << "ERROR!  setprobmatrixalongbranch not implemented for row specific calcs in BL yet." << endl; exit(-1);
		} else {
			mRoot->SetTreeBLProb(numStates, *mModel.returnRateMatrix3D(), dummyGibbs);
		}
	} else if ( (options.m_nMethod == 1) || (options.m_nMethod == 3)) {
		if (theModel != -1) {
			mRoot->SetTreeB1Prob( numStates, *mModel.returnRateMatrix3D(), dummyGibbs, theModel, row);
		} else {
			mRoot->SetTreeB1Prob( numStates, *mModel.returnRateMatrix3D(), dummyGibbs);
		}
	} else if ( (options.m_nMethod == 6)) {
        //	if ((mRoot->getCalculateFullProbabilityDistribution() == 0) ) {
        if (theModel != -1) {
            mRoot->SetTreeB1uProb( numStates, *mModel.returnRateMatrix3D(), mModel.returnUniformizationConstants(), dummyGibbs, theModel, row);
        } else {
            mRoot->SetTreeB1uProb( numStates, *mModel.returnRateMatrix3D(), mModel.returnUniformizationConstants(), dummyGibbs);
        }
 	} else if (options.m_nMethod == 2) {
		if (theModel != -1) {
			std::cout << "ERROR!  setprobmatrixalongbranch not implemented for row specific calcs in B2 yet." << endl; exit(-1);
		} else {
			mRoot->SetTreeB2Prob(numStates, *mModel.returnRateMatrix3D(), dummyGibbs);
		}
	} else if ( (options.m_nMethod == 4) || (options.m_nMethod == 5) ) {
		// 4 = Binf, 5 = BinfxM
		if (theModel != -1) {
			mRoot->SetTreeBinfProb( numStates, *mModel.returnRateMatrix3D(), &mModel.returnStateFrequencies(), theModel);
		} else {
			mRoot->SetTreeBinfProb(  numStates, *mModel.returnRateMatrix3D(), &mModel.returnStateFrequencies(), -1);
		}	
	}
	else
		throw(9);
}

//---------------------------------------------------------------------------

inline void
Tree::setprobmatrixalongbranch(MixtureModel& theModel) {
	setprobmatrixalongbranch( theModel, -1, -1 );
}


inline void
Tree::setprobmatrixalongbranch(MixtureModel& mModel,
                               int theModel,
                               int row) {
    int numStates = mModel.GetNumStates();
	
	if (options.m_nMethod == 0) {
		if (theModel != -1) {
			std::cout << "ERROR!  setprobmatrixalongbranch not implemented for row specific calcs in BL yet." << endl; exit(-1);
		} else {
			mRoot->SetTreeBLProb(numStates, *mModel.returnRateMatrix3D());
		}
	} else if ( (options.m_nMethod == 1) || (options.m_nMethod == 3) )  {
		if (theModel != -1) {
			mRoot->SetTreeB1Prob(  numStates, *mModel.returnRateMatrix3D(), theModel, row);
		} else {
			mRoot->SetTreeB1Prob(  numStates, *mModel.returnRateMatrix3D());
		}
	} else if ( (options.m_nMethod == 6) )  {
        if (theModel != -1) {
            mRoot->SetTreeB1uProb( numStates, *mModel.returnRateMatrix3D(), mModel.returnUniformizationConstants(), theModel, row);
        } else {
            mRoot->SetTreeB1uProb(  numStates, *mModel.returnRateMatrix3D(), mModel.returnUniformizationConstants());
        }
	} else if (options.m_nMethod == 2) {
		if (theModel != -1) {
			std::cout << "ERROR!  setprobmatrixalongbranch not implemented for row specific calcs in B2 yet." << endl; exit(-1);
		} else {
			mRoot->SetTreeB2Prob(numStates, *mModel.returnRateMatrix3D());
		}
	} else if ( (options.m_nMethod == 4) || (options.m_nMethod == 5) ) {
		// 4 = Binf, 5 = BinfxM
		if (theModel != -1) {
			mRoot->SetTreeBinfProb(numStates, *mModel.returnRateMatrix3D(), &mModel.returnStateFrequencies(), theModel);
		} else {
			mRoot->SetTreeBinfProb( numStates, *mModel.returnRateMatrix3D(), &mModel.returnStateFrequencies(), -1);
		}	
	}
	else
		throw(9);
}

//---------------------------------------------------------------------------

inline void
Tree::setprobmatrixalongbranch(MixtureModel& mModel,
                               int dummyGibbs) {
	setprobmatrixalongbranch( mModel, dummyGibbs, -1, -1);
}


inline void
Tree::setprobmatrixalongbranch(MixtureModel& mModel,
                               int dummyGibbs,
                               int theModel,
                               int row) {
    
	int numStates = mModel.GetNumStates();
	
	if (options.m_nMethod == 0) {
		if (theModel != -1) {
			std::cout << "ERROR!  setprobmatrixalongbranch not implemented for row specific calcs in BL yet." << endl; exit(-1);
		} else {
			mRoot->SetTreeBLProb(numStates, *mModel.returnRateMatrix3D(), dummyGibbs);
		}
	} else if ((options.m_nMethod == 1) || (options.m_nMethod == 3)) {
		if (theModel != -1) {
			mRoot->SetTreeB1Prob(numStates, *mModel.returnRateMatrix3D(), dummyGibbs, theModel, row);
		} else {
			mRoot->SetTreeB1Prob(numStates, *mModel.returnRateMatrix3D(), dummyGibbs);
		}
	} else if ((options.m_nMethod == 6)) {
        if (theModel != -1) {
            mRoot->SetTreeB1uProb(numStates, *mModel.returnRateMatrix3D(), mModel.returnUniformizationConstants(), dummyGibbs, theModel, row);
        } else {
            mRoot->SetTreeB1uProb(numStates, *mModel.returnRateMatrix3D(), mModel.returnUniformizationConstants(), dummyGibbs);
        }
	} else if (options.m_nMethod == 2) {
		if (theModel != -1) {
			std::cout << "ERROR!  setprobmatrixalongbranch not implemented for row specific calcs in B2 yet." << endl; exit(-1);
		} else {
			mRoot->SetTreeB2Prob(numStates, *mModel.returnRateMatrix3D(), dummyGibbs);
		}
	} else if ( (options.m_nMethod == 4) || (options.m_nMethod == 5) ) {
		// 4 = Binf, 5 = BinfxM
		if (theModel != -1) {
			mRoot->SetTreeBinfProb(numStates, *mModel.returnRateMatrix3D(), &mModel.returnStateFrequencies(), theModel);
		} else {
			mRoot->SetTreeBinfProb(numStates, *mModel.returnRateMatrix3D(), &mModel.returnStateFrequencies(), -1);
		}	
	}
	else
		throw(9);
}



inline TreeNode* Tree::getRoot(void) { return mRoot; }

//---------------------------------------------------------------------------
// Note this method doesn't necessarily return a true "root" - any node in 
// the tree can act as a root.  It simply returns the node we're using
// to hold onto the tree structure.

inline Tree::node_type*
Tree::root(void)
{
	return mRoot;
}

//---------------------------------------------------------------------------

inline const Tree::node_type*
Tree::root(void) const
{
	return 0;
}

//---------------------------------------------------------------------------

inline Tree::size_type
Tree::size(void) const
{
	return mRoot 
    ? mRoot->CountNodes() 
    : 0;
}

//---------------------------------------------------------------------------

inline Tree::size_type
Tree::leaves(void) const
{
	return mRoot 
    ? mRoot->CountLeaves() 
    : 0;
}

//---------------------------------------------------------------------------

inline Tree::size_type
Tree::internalnodes(void) const
{
	return mRoot 
    ? mRoot->CountInternalNodes() 
    : 0;
}

//---------------------------------------------------------------------------

inline Tree::size_type
Tree::branches(void) const
{
	return mRoot 
    ? mRoot->CountBranches() 
    : 0;
}

//---------------------------------------------------------------------------

inline
void
Tree::printtree(ofstream& outStream, const int& site, const double& scale) const
{
	// [block]!!
	return;
	
    //	mRoot->PrintTree(outStream, site, scale);
}


 inline
 void
 Tree::printnodesequence(ofstream& outStream, const int& node) const
 {
 mRoot->PrintNodeSequence(outStream, node);
 }
 
 //---------------------------------------------------------------------------
 
 inline
 void
 Tree::printnodesequence(ofstream& outStream, const char* nodename) const
 {
 mRoot->PrintNodeSequence(outStream, nodename);
 }
 
 
inline
void
Tree::setinternalnodes(void) const
{
	mRoot->SetInternalNodes();
}

//---------------------------------------------------------------------------

inline
bool
Tree::empty(void) const
{
	return size() == 0;
}

//---------------------------------------------------------------------------

inline
void
Tree::clear(void)
{
	delete mRoot;
	mRoot = 0;
}

//---------------------------------------------------------------------------

inline
void
Tree::swap(
           Tree&	inSrc)
{
	std::swap(mRoot, inSrc.mRoot);
}

//---------------------------------------------------------------------------

inline Tree::Tree(
           const Tree&	inSrc)
{
	mRoot = inSrc.mRoot;
	mRoot->calculateFullProbabilityDistribution = 0;
}

//---------------------------------------------------------------------------

inline
const Tree&
Tree::operator=(
                const Tree&	inSrc)
{
	mRoot = inSrc.mRoot;
	return(*this);
}

//---------------------------------------------------------------------------

inline void
Tree::SetRoot(
              node_type*	inNewRoot)
{
	mRoot = inNewRoot;
}

//---------------------------------------------------------------------------

inline void
Tree::DetachRoot(void)
{
	mRoot = 0;
}

//---------------------------------------------------------------------------

inline void
Tree::MakeRootless(void)
{
    
	if (mRoot)
	{
		if (!mRoot->Up() || mRoot->Up()->Up() != mRoot)
		{
			node_type*	node = new node_type;
            //			mRoot->ConnectUp(node);
            
			//add here to make the up node of root not a leaf node
			node->ConnectLeft(mRoot);
            
			node->ConnectUp(mRoot);
			
			node->ConnectRoot(mRoot);
            
		}
	}
}

//---------------------------------------------------------------------------

inline void
Tree::AssignChildrenToParent(
                             node_type&	inParent,
                             NodeList&	ioChildren)
{
	node_type*	child = 0;
	
	NodeList::size_type	numChildren = ioChildren.size();
	
	if (numChildren <= 3)
	{
		// If we have 3 or fewer children, just connect them directly.
		if (numChildren > 0)
		{
			// for 1, 2 or 3 nodes, we can connect the left node directly
			child = ioChildren.front();
			ioChildren.pop_front();
			inParent.ConnectLeft(child);
			
			if (numChildren > 1)
			{
				if (numChildren < 3)
				{
					// For 2 nodes, we can connect the right node directly too
					child = ioChildren.front();
					ioChildren.pop_front();
				}
				else
				{
					// for three nodes, we need to recursively call with 
					child = new node_type;
					child->ConnectRoot( root() );
					AssignChildrenToParent(*child, ioChildren);
				}
				inParent.ConnectRight(child);
			}
		}
	}
	else
	{
		// We have more than 3 children, so we need to divide the child list evenly
		NodeList::size_type	midPoint = numChildren / 2;
		{
			NodeList	leftChildren(ioChildren.begin(), ioChildren.begin() + midPoint);
			
			child = new node_type;
			child->ConnectRoot(root());
			AssignChildrenToParent(*child, leftChildren);
			inParent.ConnectLeft(child);
		}
		
		{
			NodeList	rightChildren(ioChildren.begin() + midPoint, ioChildren.end());
			
			child = new node_type;
			child->ConnectRoot(root());
			AssignChildrenToParent(*child, rightChildren);
			inParent.ConnectRight(child);
		}
		
		ioChildren.clear();
	}
}

//---------------------------------------------------------------------------

template <typename Functor>
inline
Functor
Tree::RemoveFreeLeaves(
                       const Functor&	inFunctor)
{
	Functor	localFunctor(inFunctor);
	
	if (mRoot)
		SetRoot(mRoot->RemoveFreeLeaves(localFunctor));
    
	return localFunctor;
}


//---------------------------------------------------------------------------
// NEW function: more simply assign sequences to leaves of the tree

inline void Tree::assignSequencesToLeaves( ObservedData &theData ) {
    mRoot->assignSequencesToLeaves( theData );
}

//---------------------------------------------------------------------------

inline
void
Tree::CountChangesAlongEveryBranch( ObservedData& theData ) {
	mRoot->CountSimpleChangesOnSubtree(theData );
}

//---------------------------------------------------------------------------

inline
void
Tree::SumChangesOverTree(IntMatrix& m_matCountChange) {
	m_matCountChange.fill(0);
	mRoot->SumChangesOverSubtree(m_matCountChange);
}

//---------------------------------------------------------------------------

inline
void
Tree::SetModelFrequencyVector(const DoubleVector& inModelFreqVector)
{
	mModelFreqVector = inModelFreqVector;
}

//---------------------------------------------------------------------------

inline
DoubleVector&
Tree::GetModelFrequencyVector(void)
{
	return(mModelFreqVector);
}


inline void
Tree::subdivideLongBranches(TreeNode* theNode, double maxBL) {
	// By default, assume recursive behavior over the tree...
	subdivideLongBranches(theNode, maxBL, 1);
}


inline void
Tree::subdivideLongBranches(TreeNode* theNode, double maxBL, int recurseTree) {
	// Use this in conjunction with B1 or B2 to limit the maximum length of each branch...
	
	TreeNode* theNewNode[5000];
	
	int numNodes = leaves() + internalnodes() + 1;
	double bl;
	bl = theNode->GetData().distance;
    
	if ( (theNode != NULL) ) {
		if ( (theNode->Left() != NULL) && recurseTree)  {
			subdivideLongBranches(theNode->Left(), maxBL);
		}
		
		if ( (theNode->Right() != NULL) && recurseTree)  {
			subdivideLongBranches(theNode->Right(), maxBL);
		}
        
		if (bl <= maxBL) return;
		if (theNode != mRoot ) {
			int totalNodeCount = 1;
			double newsize = theNode->GetData().distance;
			
			// Check if branch above is a segment; if so, take from its length if possible...
			while (newsize > maxBL) {
				totalNodeCount++;
				newsize = theNode->GetData().distance / totalNodeCount;
			}
            
			theNode->GetData().distance = newsize;
            
            for (int i=0; i<(totalNodeCount-1); i++) {
                theNewNode[i] = getFreeNode(); //new TreeNode();
                theNewNode[i]->GetData().id = ++numNodes;
                theNewNode[i]->GetData().distance = newsize;
                theNewNode[i]->ConnectRoot(mRoot);
                root()->_mixSequences( theNewNode[i]->GetData().sequence, theNode->Up()->GetData().sequence, theNode->GetData().sequence );
                theNode->insertNodeAbove( theNewNode[i] );
            }
		} // not root
		
	} // node not null
}


inline void
Tree::subdivideLongBranches(TreeNode* theNode, double maxBL, DoubleVector* uniformizationConstants, int recurseTree) {
	// Use this in conjunction with B1 or B2 to limit the maximum length of each branch...
    
    //	double mu = (*uniformizationConstants)[0];
	
	TreeNode* theNewNode[5000];
	
	int numNodes = leaves() + internalnodes() + 1;
	double bl;
	bl = theNode->GetData().distance;
	
	if ( (theNode != NULL) ) {
		if ( (theNode->Left() != NULL) && recurseTree)  {
			subdivideLongBranches(theNode->Left(), maxBL, uniformizationConstants, recurseTree);
		}
		
		if ( (theNode->Right() != NULL) && recurseTree)  {
			subdivideLongBranches(theNode->Right(), maxBL, uniformizationConstants, recurseTree);
		}
		
        
		if (theNode != mRoot ) {
			int totalNodeCount = 1;
			double newsize = theNode->GetData().distance;
			
			int notProcessed = 1;
            
			if (notProcessed) {
				if (bl <= maxBL) return;
				
				while (newsize > maxBL) {
					totalNodeCount++;
					newsize = theNode->GetData().distance / totalNodeCount;
				}
				
				theNode->GetData().distance = newsize;
				
				for (int i=0; i<(totalNodeCount-1); i++) {
					theNewNode[i] = getFreeNode();
					theNewNode[i]->GetData().id = ++numNodes;
					theNewNode[i]->GetData().distance = newsize;
					theNewNode[i]->ConnectRoot(mRoot);
					root()->_mixSequences( theNewNode[i]->GetData().sequence, theNode->Up()->GetData().sequence, theNode->GetData().sequence );
					theNode->insertNodeAbove( theNewNode[i] );
				}
			}
		} // not root
		
	} // node not null
}


#include <vector>


inline double
Tree::discreteBranchLengthApproximation(TreeNode* theNode) {
	// Use this in conjunction with B1xm (probmethod=3) to represent each branch as an integral number of short branch segments....
	TreeNode* theNewNode[1000];
    
	int numSegSizes = options.numDiscreteBLs;
	std::vector<double> *theSizes = &options.discreteBLs;
    
	double fixedBL = (*theSizes)[0];
    
	if (theNode == mRoot) {
        //		numSegSizes++;
		// Set the global info for discrete branchlengths...
		mRoot->numDiscreteBLs = numSegSizes;
		mRoot->discreteBLs = (double *)malloc( sizeof(double) * (double)numSegSizes );
		//mRoot->discreteBLs[0] = minBL;
        //		for (int i=1; i<numSegSizes; i++) { mRoot->discreteBLs[i] = theSizes[i-1]; }
		for (int i=0; i<numSegSizes; i++) { mRoot->discreteBLs[i] = (*theSizes)[i]; }
	}
    //	std::cout << "NumNodes at start: leaves = " << leaves() << ", numinternal = " << internalnodes() << endl;
	int numNodes;
	static double loss = 0.0;
	int branchTooSmall = 0;
	
    //	double bl = theNode->GetData().distance;
    
	if ( (theNode != NULL) ) {
		if (theNode->Left() != NULL) {
			discreteBranchLengthApproximation(theNode->Left());
		}
		
		if (theNode->Right() != NULL) {
			discreteBranchLengthApproximation(theNode->Right());
		}
        
		numNodes = leaves() + internalnodes() + 1;
		if ( (theNode->GetData().id != mRoot->GetData().id) && ( theNode->Up() != theNode->Up()->Up() ) ) {
			int totalNodeCount = 1;
			double newsize = theNode->GetData().distance;
			// [want to not go over, for now] totalNodeCount = round( newsize / fixedBL );
			totalNodeCount = floor( newsize / fixedBL );
			// [blocking this - want the minSize if anything] if (totalNodeCount == 0) { totalNodeCount = 1; }
			
			std::cout << "Subdividing branch: " << theNode->Up()->GetData().id << ".." << theNode->GetData().id << " using " << totalNodeCount << " segments; remainder is: " << (newsize -  (totalNodeCount * fixedBL)) << endl;
			if (numSegSizes <= 1) loss += (newsize -  (totalNodeCount * fixedBL)) * (newsize -  (totalNodeCount * fixedBL));
			
			theNode->GetData().distance = fixedBL;
			if (totalNodeCount == 0) branchTooSmall = 1; //theNode->GetData().distance = 0.0; //totalNodeCount=1; // 
			
			if (totalNodeCount > 1000) { std::cout << "WARNING!  Hard-coded limit violated on the number of node pointers!  Fix this." << endl; exit(-1); }
			
			for (int i=0; i<(totalNodeCount-1); i++) {
				theNewNode[i] = new TreeNode();
				theNewNode[i]->GetData().id = numNodes++;
				theNewNode[i]->GetData().distance = fixedBL;
				theNewNode[i]->GetData().branchLengthIndex = 0;
				root()->_mixSequences( theNewNode[i]->GetData().sequence, theNode->Up()->GetData().sequence, theNode->GetData().sequence );
				theNode->insertNodeAbove( theNewNode[i] );
			}
			
			int lastNodeCount = 0;
            
			// Now need to add extra branches of different sizes...
			double remainder = (newsize -  (totalNodeCount * fixedBL));
			for (int j=1; j<numSegSizes; j++) {
				std::cout << "---SEGSIZE " << j << "---" << endl;
				
				lastNodeCount = totalNodeCount;
				
				newsize = remainder;
				totalNodeCount = floor( newsize / mRoot->discreteBLs[j] );
				remainder = (remainder -  (totalNodeCount * mRoot->discreteBLs[j]));
                
				std::cout << "	Further subdividing branch (using " << mRoot->discreteBLs[j] << "): " << theNode->Up()->GetData().id << ".." << theNode->GetData().id << " using " << totalNodeCount << " more segments; new remainder is: " << remainder << endl;
				if (j == (numSegSizes-1)) { 
					loss += (remainder * remainder);
				}
				
				theNode->GetData().distance = mRoot->discreteBLs[j];
                //		if (totalNodeCount == 0) theNode->GetData().distance = 0.0; //totalNodeCount=1; // 
                //	if ((totalNodeCount == 0) && (j==(numSegSizes-1))) {
                //		if ( ( remainder > ((mRoot->discreteBLs[j] - minBL)/2.0) ) ) {
                //			totalNodeCount = 1;
                //		}
                //	}
				
				if (totalNodeCount > 1000) { std::cout << "WARNING!  Hard-coded limit violated on the number of node pointers!  Fix this." << endl; exit(-1); }
				
				if (lastNodeCount == 0) { lastNodeCount = 1; }
				
				for (int i=lastNodeCount-1; i<((lastNodeCount-1)+(totalNodeCount)); i++) {
                    std::cout << "	-- STARTING SUBDIV NODE ADDING AT " << (lastNodeCount-1) << endl;
					theNewNode[i] = new TreeNode();
					theNewNode[i]->GetData().id = numNodes++;
					theNewNode[i]->GetData().distance = mRoot->discreteBLs[j];
					theNewNode[i]->GetData().branchLengthIndex = j;
					root()->_mixSequences( theNewNode[i]->GetData().sequence, theNode->Up()->GetData().sequence, theNode->GetData().sequence );
					theNode->insertNodeAbove( theNewNode[i] );
					//		std::cout << "			adding secondary subdivision of len " << mRoot->discreteBLs[j] << "; id is " << (numNodes-1) << endl;
					branchTooSmall=0;
				}
				
				
			}
			// Debug: output node info...
            
			if (branchTooSmall == 1) {
				theNode->GetData().distance = mRoot->discreteBLs[numSegSizes-1];
				theNode->GetData().branchLengthIndex = numSegSizes-1;
			}
			
		} // not root
		else {
			// Root
			theNode->GetData().distance = 0.0; //fixedBL;
		}
	} // node not null
    
	return loss;
}

#endif // Tree_h__
