/*
 *  TreeParser.h
 *  PLEX
 *
 *  Created by Chip Jarred.
 *  Copyright (c) 2004 Dr. David Pollock. All rights reserved.
 *
 *
 *	This class is an ABSTRACT class representing ways to read/write trees.  It must be subclassed in
 *	to implement particular formats.  Subclasses must override the Read and Write methods.
 */

#ifndef TreeParser_h__
#define TreeParser_h__

#include <iostream>
#include <stack>
#include <string>

#include "Tokenizer.h"
//#include "Tree.h"
class Tree;
#include "TreeNode.h"

//---------------------------------------------------------------------------

class TreeParser
{
	public:
		typedef	TreeNode	node_type;
		typedef Tree		tree_type;
		
		class NodeMaker
		{
			public:
				virtual				~NodeMaker(void);
				
				virtual node_type*	Make(
										const std::string&	inNodeDescription,
										node_type*			inNodeToSetFromDescription = 0) const = 0;
		};
		
									TreeParser(void);
		virtual						~TreeParser(void);
		
		virtual	bool				Read(
										std::istream&		inTreeStream,
										tree_type&			outTree)		= 0;
		virtual	bool				Write(
										std::ostream&		outTreeStream,
										const tree_type&	inTree)			= 0;
										
				void				SetErrorStream(
										std::ostream&		inErrorStream);
									
				void				SetNodeMaker(
										NodeMaker*			inNodeMaker);

	protected:
		friend class UTTreeParser;
				void				SetDelimiters(
										const std::string&	inDelimiters);
				const std::string&	GetDelimiters(void) const;
				bool				IsDelimiter(
										const std::string&	inToken) const;
				
				void				SetWhitespaceChars(
										const std::string&	inWhitespaceChars);
				const std::string&	GetWhitespaceChars(void) const;
				
				void				SetQuoteChars(
										const std::string&	inQuoteChars);
				const std::string&	GetQuoteChars(void) const;
				
				void				SkipWhitespace(
										bool				inSkipWhitespace = true);
				bool				SkippingWhitespace(void) const;
		
				void				RemoveQuotes(
										bool				inRemoveQuotes = true);
				bool				RemovingQuotes(void) const;
		
				void				EatWhitespace(
										std::istream&		inStream);
				bool				ReadToken(
										std::istream&		inTreeStream,
										std::string&		outToken);
				void				PutBackToken(
										const std::string&	inToken);
										
				void				ReportError(
										const std::string&	inError) const;
										
				node_type*			MakeNode(
										const std::string&	inNodeDescription,
										node_type*			inNodeToSetFromDescription = 0) const;
										
		virtual	void				reset(void);
										
	private:
				Tokenizer			mTokenizer;
						
				NodeMaker*			mNodeMaker;
				std::ostream*		mErrorStream;
};

//---------------------------------------------------------------------------

inline TreeParser::NodeMaker::~NodeMaker(void)
{
}

#pragma mark -
//---------------------------------------------------------------------------

inline TreeParser::TreeParser(void)
:	mTokenizer(),
mNodeMaker(0),
mErrorStream(&std::cerr)
{
}

//---------------------------------------------------------------------------

inline TreeParser::~TreeParser(void)
{
	delete mNodeMaker;
}

//---------------------------------------------------------------------------

inline
void
TreeParser::SetDelimiters(
                          const std::string&	inDelimiters)
{
	mTokenizer.SetDelimiters(inDelimiters);
}

//---------------------------------------------------------------------------

inline
const std::string&
TreeParser::GetDelimiters(void) const
{
	return mTokenizer.GetDelimiters();
}

//---------------------------------------------------------------------------

inline
bool
TreeParser::IsDelimiter(
                        const std::string&	inToken) const
{
	return mTokenizer.IsDelimiter(inToken);
}

//---------------------------------------------------------------------------

inline
void
TreeParser::SetWhitespaceChars(
                               const std::string&	inWhitespaceChars)
{
	mTokenizer.SetWhitespaceChars(inWhitespaceChars);
}

//---------------------------------------------------------------------------

inline
const std::string&
TreeParser::GetWhitespaceChars(void) const
{
	return mTokenizer.GetWhitespaceChars();
}

//---------------------------------------------------------------------------

inline
void
TreeParser::SetQuoteChars(
                          const std::string&	inQuoteChars)
{
	mTokenizer.SetQuoteChars(inQuoteChars);
}

//---------------------------------------------------------------------------

inline
const std::string&
TreeParser::GetQuoteChars(void) const
{
	return mTokenizer.GetQuoteChars();
}

//---------------------------------------------------------------------------

inline
void
TreeParser::SkipWhitespace(
                           bool	inSkipWhitespace)
{
	mTokenizer.SkipWhitespace(inSkipWhitespace);
}

//---------------------------------------------------------------------------

inline
bool
TreeParser::SkippingWhitespace(void) const
{
	return mTokenizer.SkippingWhitespace();
}

//---------------------------------------------------------------------------

inline
void
TreeParser::RemoveQuotes(
                         bool	inRemoveQuotes)
{
	mTokenizer.RemoveQuotes(inRemoveQuotes);
}

//---------------------------------------------------------------------------

inline
bool
TreeParser::RemovingQuotes(void) const
{
	return mTokenizer.RemovingQuotes();
}

//---------------------------------------------------------------------------

inline void
TreeParser::EatWhitespace(
                          std::istream&	inStream)
{
	mTokenizer.SetStream(inStream);
	mTokenizer.EatWhitespace();
}

//---------------------------------------------------------------------------

inline bool
TreeParser::ReadToken(
                      std::istream&	inTreeStream,
                      std::string&	outToken)
{
	mTokenizer.SetStream(inTreeStream);
	return mTokenizer.ReadToken(outToken);
}

//---------------------------------------------------------------------------

inline void
TreeParser::PutBackToken(
                         const std::string&	inToken)
{
	mTokenizer.PutBackToken(inToken);
}

//---------------------------------------------------------------------------

inline void
TreeParser::SetErrorStream(
                           std::ostream&	inErrorStream)
{
	mErrorStream = &inErrorStream;
}

//---------------------------------------------------------------------------

inline void
TreeParser::ReportError(
                        const std::string&	inError) const
{
	(*mErrorStream) << inError << std::endl;
}

//---------------------------------------------------------------------------

inline void
TreeParser::SetNodeMaker(
                         NodeMaker*	inNodeMaker)
{
	mNodeMaker = inNodeMaker;
}

//---------------------------------------------------------------------------

inline TreeParser::node_type*
TreeParser::MakeNode(
                     const std::string&	inNodeDescription,
                     node_type*			inNodeToSetFromDescription) const
{
	return mNodeMaker
    ?	mNodeMaker->Make(inNodeDescription, inNodeToSetFromDescription)
    :	0;
}

//---------------------------------------------------------------------------

inline void
TreeParser::reset(void)
{
	mTokenizer.Reset();
}

#endif // TreeParser_h__
